<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * datatype mapping tool
 *
 * This class handles transformation of semantic data types to
 * syntactic data types and vice versa.
 *
 * This class is part of the data access layer.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @ignore
 */
class DbDatatype extends Object
{
    /**#@+
     * @access  private
     * @ignore
     */

    /** @var DbStructure */ var $structure = null;
    /** @var DbTrigger   */ var $exec = null;
    /** @var string      */ var $dbDir = null;

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  DbStructure  &$structure
     */
    function __construct(&$structure)
    {
        assert('is_object($structure) && is_a($structure, "DbStructure"); // Wrong argument type for argument 1. Instance of DbStructure expected');
        $this->structure =& $structure;
        $this->exec = new DbTrigger($structure);
    }

    /**
     * prepare data for input to database
     *
     * @access  public
     * @param   mixed   &$value
     * @param   string  $table
     * @param   string  $column
     * @param   string  $row
     * @param   object  $db
     * @return  bool
     */
    function encode(&$value, $table, $column, $row = null, $db = null)
    {
        assert('is_string($table); // Wrong argument type for argument 2. String expected');
        assert('is_string($column); // Wrong argument type for argument 3. String expected');
        assert('is_null($row) || is_scalar($row); // Wrong argument type for argument 4. Scalar value expected');
        assert('is_null($db) || is_object($db); // Wrong argument type for argument 5. Object expected');

        $table  = (string) $table;
        $column = (string) $column;
        $row    = (string) $row;
        $row    = strtolower($row);

        $type = $this->structure->getType($table, $column);
        if ($type === 'select' && !is_array($value) &&  $this->structure->isForeignKey($table, $column)) {
            $table = $this->structure-> getTableByForeignKey($table, $column);
            $type = $this->structure->getType($table, $column);
        }

        switch ($type)
        {
            case 'text':
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_USERTEXT);
            break;

            case 'string': case 'select': case 'ip': case 'url': case 'mail': case 'profile':
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_LINEBREAK);
            break;

            /* handle file uploads
             *
             * syntax: _handleUpload($table, $row, $column, $index, true)
             */
            case 'image': case 'file':

                if ($value === YANA_DB_DELETE_FILE) {
                    if (!empty($row)) {
                        $this->_removeFile($db, $table, $column, $row);
                    }
                    $value = false;
                } else {
                    $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_LINEBREAK);

                    /* if row already has an associated file, overwrite the old one, rather than creating another */
                    if ($row !== '' && $row !== '*' && $row !== '?' && $db->exists("$table.$row.$column")) {
                        $fileId = DbDatatype::_decodeFileId($db->get(strtolower("$table.$row.$column")));
                    } else {
                        $fileId = DbDatatype::_encodeFileId($table, $column);
                    }

                    $isImage = ( $type == 'image' );

                    $value = $this->_handleUpload($table, $row, $column, $value, $fileId, $isImage);
                }
            break;

            case 'int': case 'integer': case 'float': case 'double':
                if (is_numeric($value)) {
                    $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_NONE);
                    if (is_int($value) && $this->structure->isUnsigned($table, $column) && $value < 0) {
                        trigger_error("Invalid input. The value '$value' is not an unsigned integer on column '$table.$column'.", E_USER_NOTICE);
                        return false;

                    } elseif (is_float($value) && $this->structure->isUnsigned($table, $column) && $value < 0) {
                        trigger_error("Invalid input. The value '$value' is not an unsigned float on column '$table.$column'.", E_USER_NOTICE);
                        return false;

                    }
                } else {
                    $value = null;
                }
            break;

            default:
                $value = $this->structure->untaintInput($table, $column, $value, YANA_ESCAPE_NONE);
            break;
        }
        return true;
    }

    /**
     * prepare a database entry for output
     *
     * @access  public
     * @param   mixed    &$out
     * @param   string   $table
     * @param   string   $column
     * @param   mixed    $value
     * @param   string   $key         array address (applies to columns of type array only)
     * @param   DbQuery  $dbQuery
     * @return  bool
     */
    function decode(&$out, $table, $column, $value, $key = '', $dbQuery = null)
    {
        assert('is_string($table); // Wrong argument type for argument 2. String expected');
        assert('is_string($column); // Wrong argument type for argument 3. String expected');
        assert('is_string($key); // Wrong argument type for argument 5. String expected');
        assert('is_null($dbQuery) || is_object($dbQuery); // Wrong argument type for argument 6. Object expected');

        /* 1) check constraint */
        if ($this->exec->constraint(YANA_DB_SELECT, $table, $column, $value) !== false) {

            /* get list of joined tables */
            if (isset($dbQuery)) {
                if (!$this->structure->isColumn($table, $column)) {
                    $table = null;
                    $tables = array_keys($dbQuery->getJoin());
                    if (!empty($tables)) {
                        assert('!isset($element); // Cannot redeclare var $element');
                        foreach ($tables as $element)
                        {
                            if ($this->structure->isColumn($element, $column)) {
                                $table = $element;
                                break;
                            }
                        } /* end foreach */
                        unset($element); /* clean up garbage */
                    }
                }
                if (is_null($table)) {
                    trigger_error("A database type conversion error occurred.\n\t\t" .
                    "There is no such column '${column}' in the current table.", E_USER_NOTICE);
                }
            }

            switch ($this->structure->getType($table, $column))
            {
                case 'integer': case 'int':
                    if (is_numeric($value)) {
                        $out = (int) $value;
                    } elseif ($value === true) {
                        $out = 1;
                    } elseif ($value === false) {
                        $out = 0;
                    }
                    /* apply zerofill */
                    if ($this->structure->isZerofill($table, $column)) {
                        $length = $this->structure->getLength($table, $column);
                        $out = (string) abs($out);
                        if ($length > 0 && strlen($out) < $length) {
                            $out = str_pad($out, $length, '0', STR_PAD_LEFT);
                        }

                    /* apply unsigned */
                    } elseif ($this->structure->isUnsigned($table, $column)) {
                        $out = abs($out);
                    }
                break;

                case 'float': case 'double':
                    if (is_numeric($value)) {
                        $out = (float) $value;
                    } elseif ($value === true) {
                        $out = 1.0;
                    } elseif ($value === false) {
                        $out = 0.0;
                    }
                    /* apply zerofill */
                    if ($this->structure->isZerofill($table, $column)) {
                        $length = $this->structure->getLength($table, $column);
                        $out = (string) abs($out);
                        if ($length > 0 && strlen($out) < $length) {
                            $out = str_pad($out, $length, '0', STR_PAD_LEFT);
                        }

                    /* apply unsigned */
                    } elseif ($this->structure->isUnsigned($table, $column)) {
                        $out = abs($out);
                    }
                break;

                case 'bool': case 'boolean':
                    if (is_bool($value)) {
                        $out = $value;
                    } elseif ($value == 1) {
                        $out = true;
                    } elseif (empty($value)) {
                        $out = false;
                    }
                break;

                case 'array': case 'set':
                    if (is_array($value)) {
                        $out = $value;
                    } else {
                        $out = SML::getFile(explode("\n", $value), CASE_LOWER);
                    }
                    if ($key !== "") {
                        assert('is_array($out); /* unexpected result: $out should be an array */');
                        $out = Hashtable::get($out, strtolower($key));
                        if (is_null($out)) {
                            $out = false;
                        }
                    }
                break;

                case 'image': case 'file':
                    if (is_null($this->dbDir) && $this->dbDir !== false) {
                        global $YANA;
                        if (isset($YANA)) {
                            $dir = $YANA->getVar('DBBLOB');
                        } else {
                            $dir = 'config/db/.blob/';
                        }
                        if (is_dir($dir)) {
                            $this->dbDir = $dir;
                        } else {
                            $this->dbDir = false;
                        }
                    }
                    if (empty($value)) {
                        $out = null;

                    } elseif ($this->dbDir !== false) {
                        $out = $this->dbDir . $value;
                    }
                break;

                case '':
                    if ($this->structure->isStrict()) {
                        trigger_error("A database type conversion error occurred.\n\t\t" .
                        "Found a value in column '${column}' of table '${table}',\n\t\t" .
                        "but the type of this column is undefined.", E_USER_NOTICE);
                    }
                    if (is_string($value)) {
                        $out = stripslashes($value);
                    } else {
                        $out = $value;
                    }
                break;

                default:
                    if (is_scalar($value)) {
                        $out = stripslashes($value);
                    } elseif ($this->structure->isStrict()) {
                        trigger_error("Invalid database value in table '${table}' column '${column}'.\n\t\t" .
                        "Scalar value expected, found '".gettype($value)."' instead.", E_USER_NOTICE);
                    } else {
                        $out = $value;
                    }
                break;
            }
            return true;

        } else {
            return false;
        }
    }

    /**
     * remove a binary large object from database
     *
     * This removes a files (blob) from the database.
     * It will delete all files associated with key: $table.$row.$column.
     *
     * The type of the column must be "file" or "image".
     *
     * IMPORTANT NOTE: This is a low-level function that DOES NOT
     * disassociate the files with the datasets, that reference
     * them.
     *
     * Returns bool(true) on success and bool(false) on error,
     * e.g. the file and / or dataset does not exist.
     *
     * @access  private
     * @param   DbStream  $db
     * @param   string    $table
     * @param   string    $column
     * @param   scalar    $row
     * @return  bool
     * @name    DbStream::removeFile()
     * @since   2.9.6
     */
    function _removeFile($db, $table, $column, $row)
    {
        assert('is_object($db);     // Wrong argument type argument 1. Object expected.');
        assert('is_string($table);  // Wrong argument type argument 2. String expected.');
        assert('is_string($column); // Wrong argument type argument 3. String expected.');

        if (!is_scalar($row)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'Scalar', gettype($row)), E_USER_WARNING);
            return false;
        }

        $type = $this->structure->getType($table, $column);
        if ($type !== 'file' && $type !== 'image') {
            /* error - not a file */
            return false;
        }

        assert('!isset($fileToDelete); /* Cannot redeclare $fileToDelete */');
        $fileToDelete = "";

        if (!empty($row)) {
            /**
             * get filename
             */
            assert('!isset($file); /* Cannot redeclare $file */');
            assert('!isset($result); // Cannot redeclare var $result');
            $result = $db->query("SELECT $column from ".$db->quoteId(YANA_DATABASE_PREFIX.$table)." where ".$this->structure->getPrimaryKey($table)." = ".$db->quote($row), 0, 1);
            if (!$db->isError($result)) {
                $file = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                assert('is_array($file); /* unexpected result: $file */');
                if (($dbDir = $GLOBALS['YANA']->getVar('DBBLOB')) !== false) {
                    $fileToDelete = $dbDir . array_pop($file);

                /* error - directory not found */
                } else {
                    return false;
                }

            /* error - query failed ($table.$row.$column does not exist) */
            } else {
                return false;
            }
            unset($file, $result);

            /**
             * delete file
             */
            if (is_file($fileToDelete)) {
                unlink($fileToDelete);
                $fileToDelete = str_replace('image.', 'thumb.', $fileToDelete);
                if (is_file($fileToDelete)) {
                    unlink($fileToDelete);
                }
                /* success - file(s) have been removed successfully */
                return true;

            /* error - file does not exist */
            } else {
                return false;

            }

        /* error - argument 3 may not be empty */
        } else {
            return false;
        } /* end if */
    }

    /**
     * handle file uploads
     *
     * This handles all file uploads.
     * Set $isImage to true to upload images.
     *
     * @access  private
     * @param   string  $table
     * @param   string  $row
     * @param   string  $column
     * @param   string  $id
     * @param   string  $fileId
     * @param   bool    $isImage
     * @return  string|null
     * @ignore
     */
    function _handleUpload($table, $row, $column, $id, $fileId, $isImage = false)
    {
        assert('is_string($table);  // Wrong argument type for argument 1. String expected');
        assert('is_string($row);    // Wrong argument type for argument 2. String expected');
        assert('is_string($column); // Wrong argument type for argument 3. String expected');
        assert('is_string($id);     // Wrong argument type for argument 4. String expected');
        assert('is_string($fileId); // Wrong argument type for argument 5. String expected');
        assert('is_bool($isImage);  // Wrong argument type for argument 6. Boolean expected');

        global $YANA;

        /* get directory */
        assert('!isset($dir); // Cannot redeclare var $dir');
        if (isset($YANA)) {
            $dir = $YANA->getVar('DBBLOB');
            if ($dir === false) {
                $YANA->report( new Alert("UPLOAD_FAILED") );
                return null;
            }
        } else {
            $dir = 'config/db/.blob/';;
        }

        /* 1. file uploads */
        if (!$isImage) {

            /* name of output file */
            $file = "file.${fileId}.gz";
            /* name of input file */
            $maxSize = $this->structure->getLength($table, $column);
            $result = untaintInput($id, "upload", $maxSize);

        /* 2. image uploads */
        } else {

            /* name of output files */
            $file = "image.${fileId}";
            $thumb = "thumb.${fileId}";
            /* name of input file */
            $args = $this->structure->getImageSettings($table, $column);
            if (!is_array($args)) {
                return null;
            }
            $result = Image::uploadFile($id, $dir.$file, null, $args['size'], $args['width'], $args['height'], $args['ratio'], $args['background']);

        }

        unset($fileId);


        if (is_string($result) && is_file($result) && is_readable($result)) {

            /* 1. file uploads */
            if (!$isImage) {

                assert('!isset($FILES); // Cannot redeclare var $FILES');
                $FILES = array_change_key_case($_FILES, CASE_LOWER);
                assert('!isset($name); // Cannot redeclare var $name');
                if (!empty($FILES[$id]['name'])) {
                    $name = $FILES[$id]['name'];
                    $name = preg_replace('/[^\d\w\_\-\.]/', '_', $name);
                } else {
                    $name = "";
                }
                assert('!isset($mimetype); // Cannot redeclare var $mimetype');
                if (!empty($FILES[$id]['type'])) {
                    $mimetype = $FILES[$id]['type'];
                    $mimetype = preg_replace('/\s/', ' ', $mimetype);
                } else {
                    $name = "";
                }
                unset($FILES);

                /* create zip file */
                assert('!isset($gz); /* Cannot redeclare var $gz */');
                $gz = gzopen($dir . $file, 'w9');

                /**
                 * insert Header information
                 *
                 * 1) original filename
                 * 2) original size
                 * 3) original mime-type
                 *
                 * Note: line break "\n" is used as a delimiter
                 */
                gzwrite($gz, "$name\n");
                gzwrite($gz, filesize($result)."\n");
                gzwrite($gz, "$mimetype\n");

                /* save file contents */
                gzwrite($gz, file_get_contents($result));
                gzclose($gz);
                unset($gz);

                return $file;

            /* 2. image uploads */
            } else {

                /* create thumbnail */
                assert('!isset($width); // Cannot redeclare $width');
                if (isset($args['width']) && $args['width'] < 75) {
                    $width = null;
                } else {
                    $width = 75;
                }
                assert('!isset($height); // Cannot redeclare $height');
                if (isset($args['height']) && $args['height'] < 75) {
                    $height = null;
                } else {
                    $height = 75;
                }
                Image::uploadFile($id, $dir . $thumb, null, null, $width, $height, true, array(128, 128, 128));
                unset($width, $height);
                $result = preg_replace('/^'.toRegExp($dir).'/', '', $result);
                return $result;

            }

        /* handle errors */
        } else {

            $FILES    = array_change_key_case($_FILES, CASE_UPPER);
            $filename = @$FILES[$id]['name'];

            switch ($result)
            {
                case UPLOAD_ERR_SIZE: case UPLOAD_ERR_INI_SIZE: case UPLOAD_ERR_FORM_SIZE:
                    if (isset($YANA)) {
                        $YANA->report( new Alert("TOO_LARGE_FILE", array('FILE' => $filename, 'MAXIMUM' => $args['size'])) );
                    } else {
                        trigger_error("File '$filename' is too large ({$args['size']}).", E_USER_WARNING);
                    }
                break;
                case UPLOAD_ERR_FILE_TYPE:
                    if ($isImage) {
                        if (isset($YANA)) {
                            $YANA->report( new Alert("INVALID_IMAGE", array('FILE' => $filename)) );
                        } else {
                            trigger_error("File '$filename' is not an image.", E_USER_WARNING);
                        }
                    } else {
                        if (isset($YANA)) {
                            $YANA->report( new Alert("UPLOAD_FAILED", array('FILE' => $filename)) );
                        } else {
                            trigger_error("Upload failed for file '$filename'.", E_USER_WARNING);
                        }
                    }
                break;
                case UPLOAD_ERR_INVALID_TARGET:
                    if (isset($YANA)) {
                        $YANA->report( new Log("Unable to write uploaded file to '${file}'.") );
                        $YANA->report( new ErrorMessage("500") );
                    } else {
                        trigger_error("Unable to write to file '$file'.", E_USER_WARNING);
                    }
                break;
                case UPLOAD_ERR_NO_FILE:
                    if ($isImage) {
                        if (isset($YANA)) {
                            $YANA->report( new Alert("INVALID_IMAGE") );
                        } else {
                            trigger_error("Upload failed, no image uploaded.", E_USER_WARNING);
                        }
                    } else {
                        if (isset($YANA)) {
                            $YANA->report( new Alert("UPLOAD_FAILED") );
                        } else {
                            trigger_error("Upload failed, no file uploaded.", E_USER_WARNING);
                        }
                    }
                break;
                default:
                    if (isset($YANA)) {
                        $YANA->report( new Alert("UPLOAD_FAILED", array('FILE' => $filename)) );
                    } else {
                        trigger_error("Upload failed for file '$filename'.", E_USER_WARNING);
                    }
                break;
            } /* end switch */
            return null;

        } /* end error handling */
    }

    /**
     * create unique id to identify a file
     *
     * @access  private
     * @static
     * @param   string  $table
     * @param   string  $column
     * @return  string
     * @ignore
     */
    function _encodeFileId($table, $column)
    {
        assert('is_string($table);  // Wrong argument type for argument 1. String expected');
        assert('is_string($column); // Wrong argument type for argument 2. String expected');
        return md5(uniqid(strtolower($table).'.'.rand().'.'.$column, true));
    }

    /**
     * extract unique file-id from a database value
     *
     * @access  private
     * @static
     * @param   string  $fileId
     * @return  string
     * @ignore
     */
    function _decodeFileId($fileId)
    {
        assert('is_string($fileId); // Wrong argument type for argument 1. String expected');
        return preg_replace('/^.*(?:image|thumb|file)\.(\w+)\.\w+$/', '${1}', $fileId);
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
