<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * column information
 *
 * exports as (example):
 *
 * Array
 * (
 *     [table] => foo
 *     [name] => foo_id
 *     [type] => int
 *     [nullable] =>
 *     [primarykey] => 1
 *     [unique] => 
 *     [index] => 
 *     [auto] => 1
 *     [default] => 
 *     [unsigned] => 
 *     [zerofill] =>
 *     [comment] => id
 *     [length] => 8
 *     [foreignkey] => 
 *     [references] =>
 *     [update] => 1
 *     [insert] => 1
 *     [select] => 1
 * )
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @ignore
 */
class DbInfoColumn extends Object
{
    /**#@+
     * @access  private
     */

    /**
     * table name
     *
     * This can be set automatically and may be ignored.
     *
     * @var string
     */
    var $table = null;

    /**
     * column name
     *
     * mandatory
     *
     * @var string */
    var $name = null;

    /**
     * type of data
     *
     * mandatory
     *
     * @var string
     */
    var $type = null;

    /**
     * NULL allowed? (true = yes, false = no)
     *
     * mandatory
     *
     * @var bool
     */
    var $nullable = true;

    /**
     * is the primary key? (true = yes, false = no)
     *
     * This can be set automatically and may be ignored.
     *
     * @var bool
     */
    var $primaryKey = false;

    /**
     * has a unique constraint? (true = yes, false = no)
     *
     * mandatory
     *
     * @var bool
     */
    var $unique = false;

    /**
     * use index on this column? (true = yes, false = no)
     *
     * optional
     *
     * @var bool
     */
    var $index = false;

    /**
     * is auto-increment, or auto-generated value? (true = yes, false = no)
     *
     * optional
     *
     * @var bool
     */
    var $auto = false;

    /**
     * default value
     *
     * optional
     *
     * @var mixed
     */
    var $default = null;

    /**
     * is unsigned integer? (true = yes, false = no)
     *
     * optional
     *
     * @var bool
     */
    var $unsigned = false;

    /**
     * use zerfofill on integer? (true = yes, false = no)
     *
     * optional
     *
     * @var bool
     */
    var $zerofill = false;

    /**
     * comment or name of column
     *
     * optional
     *
     * @var string
     */
    var $comment = null;

    /**
     * maximum length (interpretation depends on type)
     *
     * optional
     *
     * @var int
     */
    var $length = null;

    /**
     * is a foreign key? (true = yes, false = no)
     *
     * This can be set automatically and may be ignored.
     *
     * @var bool
     */
    var $foreignKey = false;

    /**
     * is updateable? (true = yes, false = no)
     *
     * optional
     *
     * @var bool
     */
    var $update = true;

    /**
     * allow new values to be created? (true = yes, false = no)
     *
     * optional
     *
     * @var bool
     */
    var $insert = true;

    /**
     * show this column? (true = visible, false = hidden)
     *
     * optional
     *
     * @var bool
     */
    var $select = true;

    /**#@-*/

    /**#@+
     * @access  private
     */

    /**
     *
     * @var array
     */
    var $references = null;

    /**#@-*/

    /**
     * Constructor
     *
     * This function creates a new instance of this class.
     *
     * @param   string  $column  column name
     * @access  public
     */
    function __construct($column = null)
    {
        if (is_null($column)) {
            /* intentionally left blank */
            
        } elseif (!is_string($column)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($column)), E_USER_ERROR);

        } else {
            $this->setName($column);

        }
    }

    /**
     * set type
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $name
     * @return  bool
     */
    function setType($name)
    {
        if (!is_string($name)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($name)), E_USER_WARNING);
            return false;

        } else {
            $this->type = strtolower($name);

        }
    }

    /**
     * get type of data
     *
     * Returns table name on success and bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getType()
    {
        if (!is_string($this->type)) {
            return false;

        } else {
            return $this->type;

        }
    }

    /**
     * set column nullable
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isNullable
     * @return  bool
     */
    function setNullable($isNullable)
    {
        if ($isNullable) {
            $this->nullable = true;

        } else {
            $this->nullable = false;

        }
        return true;
    }

    /**
     * check if column is nullable
     *
     * @access  public
     * @return  bool
     */
    function isNullable()
    {
        if ($this->nullable) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set primary key
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isPrimaryKey
     * @return  bool
     */
    function setPrimaryKey($isPrimaryKey)
    {
        if ($isPrimaryKey) {
            $this->primaryKey = true;

        } else {
            $this->primaryKey = false;

        }
        return true;
    }

    /**
     * check if column is primary key
     *
     * @access  public
     * @return  bool
     */
    function isPrimaryKey()
    {
        if ($this->primaryKey) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set foreign key
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isForeignKey
     * @return  bool
     */
    function setForeignKey($isForeignKey)
    {
        if ($isForeignKey) {
            $this->foreignKey = true;

        } else {
            $this->foreignKey = false;

        }
        return true;
    }

    /**
     * check if column is foreign key
     *
     * @access  public
     * @return  bool
     */
    function isForeignKey()
    {
        if ($this->foreignKey) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set unique constraint
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isUnique
     * @return  bool
     */
    function setUnique($isUnique)
    {
        if ($isUnique) {
            $this->unique = true;

        } else {
            $this->unique = false;

        }
        return true;
    }

    /**
     * check if column is unique
     *
     * @access  public
     * @return  bool
     */
    function isUnique()
    {
        if ($this->unique) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set index on column
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $hasIndex
     * @return  bool
     */
    function setIndex($hasIndex)
    {
        if ($hasIndex) {
            $this->index = true;

        } else {
            $this->unique = false;

        }
        return true;
    }

    /**
     * check if column has index
     *
     * @access  public
     * @return  bool
     */
    function hasIndex()
    {
        if ($this->index) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set auto-increment
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isAuto
     * @return  bool
     */
    function setAuto($isAuto)
    {
        if ($isAuto) {
            $this->auto = true;

        } else {
            $this->auto = false;

        }
        return true;
    }

    /**
     * check if column is auto-increment
     *
     * @access  public
     * @return  bool
     */
    function isAuto()
    {
        if ($this->auto) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set if column is updateable
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isUpdateable
     * @return  bool
     */
    function setUpdate($isUpdateable)
    {
        if ($isUpdateable) {
            $this->update = true;

        } else {
            $this->update = false;

        }
        return true;
    }

    /**
     * check if column is updateable
     *
     * @access  public
     * @return  bool
     */
    function isUpdateable()
    {
        if ($this->update) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set if column is selectable
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isSelectable
     * @return  bool
     */
    function setSelect($isSelectable)
    {
        if ($isSelectable) {
            $this->select = true;

        } else {
            $this->select = false;

        }
        return true;
    }

    /**
     * check if column is selectable
     *
     * @access  public
     * @return  bool
     */
    function isSelectable()
    {
        if ($this->select) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set if column is insertable
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isInsertable
     * @return  bool
     */
    function setInsert($isInsertable)
    {
        if ($isInsertable) {
            $this->insert = true;

        } else {
            $this->insert = false;

        }
        return true;
    }

    /**
     * check if column is insertable
     *
     * @access  public
     * @return  bool
     */
    function isInsertable()
    {
        if ($this->insert) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set to unsigned number
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isUnsigned
     * @return  bool
     */
    function setUnsigned($isUnsigned)
    {
        if ($isUnsigned) {
            $this->unsigned = true;

        } else {
            $this->unsigned = false;

        }
        return true;
    }

    /**
     * check if column is unsigned number
     *
     * @access  public
     * @return  bool
     */
    function isUnsigned()
    {
        if ($this->unsigned) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set to zerofill number
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $isUnsigned
     * @return  bool
     */
    function setZerofill($isZerofill)
    {
        if ($isZerofill) {
            $this->zerofill = true;

        } else {
            $this->zerofill = false;

        }
        return true;
    }

    /**
     * check if column is zerofill number
     *
     * @access  public
     * @return  bool
     */
    function isZerofill()
    {
        if ($this->zerofill) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * set table name
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $name
     * @return  bool
     */
    function setTable($name)
    {
        if (!is_string($name)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($name)), E_USER_WARNING);
            return false;

        } else {
            $this->table = $name;

        }
    }

    /**
     * get table name
     *
     * Returns table name on success and bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getTable()
    {
        if (!is_string($this->table)) {
            return false;

        } else {
            return $this->table;

        }
    }

    /**
     * set column name
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $name
     * @return  bool
     */
    function setName($name)
    {
        if (!is_string($name)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($name)), E_USER_WARNING);
            return false;

        } else {
            $this->name = strtolower($name);

        }
    }

    /**
     * get column name
     *
     * Returns column name on success and bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getName()
    {
        if (!is_string($this->name)) {
            return false;

        } else {
            return strtolower($this->name);

        }
    }

    /**
     * get default value
     *
     * Returns the default value,
     * or NULL (not bool(false)!) on error.
     *
     * @access  public
     * @return  mixed|NULL
     */
    function getDefault()
    {
        if (isset($this->default)) {
            return $this->default;

        } else {
            return null;
        }
    }

    /**
     * set default value
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $default
     * @return  bool
     */
    function setDefault($default)
    {
        $this->default = $default;
        return true;
    }

    /**
     * get comment
     *
     * Returns the comment as a string,
     * or bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     */
    function getComment()
    {
        if (is_string($this->comment)) {
            return $this->comment;

        } else {
            return false;
        }
    }

    /**
     * set comment
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $comment
     * @return  bool
     */
    function setComment($comment)
    {
        if (!is_string($comment)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($comment)), E_USER_NOTICE);
            return false;

        } else {
            $this->comment = $comment;
            return true;

        }
    }

    /**
     * get length
     *
     * Returns the length as an integer,
     * or bool(false) on error.
     *
     * @access  public
     * @return  int|bool(false)
     */
    function getLength()
    {
        if (is_int($this->length)) {
            return $this->length;

        } else {
            return false;
        }
    }

    /**
     * set comment
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   int  $length
     * @return  bool
     */
    function setLength($length)
    {
        if (!is_int($length)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($length)), E_USER_WARNING);
            return false;

        } elseif ($length > 0) {
            $this->length = $length;
            return true;

        } else {
            $this->length = null;
            return false;

        }
    }

    /**
     * add reference to foreign table
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $foreignTable   name of destination table
     * @param   string  $foreignColumn  name of column in destination table
     * @return  bool
     */
    function setReference($foreignTable, $foreignColumn)
    {
        if (!is_string($foreignTable)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($foreignTable)), E_USER_WARNING);
            return false;
        }
        if (!is_string($foreignColumn)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($foreignColumn)), E_USER_WARNING);
            return false;
        }        
        $this->foreignKey = true;
        $this->references = array($foreignTable, strtolower($foreignColumn));
        return true;
    }

    /**
     * export object as associative array
     *
     * @access  public
     * @return  array
     */
    function toArray()
    {
        $array = get_object_vars($this);
        $array = array_change_key_case($array, CASE_LOWER);
        return $array;
    }
}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
