<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * Database query builder with SQL parser
 *
 * This class is an extension to the query builder,
 * which enables it to parse real SQL statements
 * to fill the object's properties.
 *
 * Note that the SQL statements are expected to use
 * the syntax understood by FileDb. This sligthly
 * differs from MySQL. No Limit clause is allowed.
 * Use the method "DbQuery::setLimit()" instead.
 * Ids do not need to be quoted. Values are quoted
 * using double-quotes - not single-quotes.
 *
 * For more details on this topic see the manual
 * and the developer's cookbook.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @since       2.9.3
 */
class DbQueryParser extends DbQuery
{
    /**
     * create a new instance
     *
     * This creates and initializes a new
     * instance of this class.
     *
     * @param  object  $database
     */
    function __construct($database)
    {
        parent::__construct($database);

        /**
         * prepare regular expressions
         */
        $this->_delim   = YANA_DB_DELIMITER;
        $this->_id      = '[\w\d_]+';
        $this->_value   = '(?:' . $this->_delim.'.*?'.$this->_delim . '|\d+|true|false)';
        $this->_value2  = $this->_delim.'(.*?)'.$this->_delim;
        $this->_s       = '/^';
        $this->_select1 = 'select\s+1';
        $this->_count   = 'select\s+count\(\*\)';
        $this->_select  = 'select\s+(\S+)';
        $this->_table   = '\s+from\s+('.$this->_id.')';
        $this->_tables  = '\s+from\s+((?:'.$this->_id.'(?:\s*,\s*)?){1,}|'.$this->_id.'\s+NATURAL\sJOIN\s+'.$this->_id.')';
        $this->_insert  = 'insert\s+into\s+('.$this->_id.')';
        $this->_update  = 'update\s+('.$this->_id.')';
        $this->_delete  = 'delete';
        $this->_values  = '\s*\((.*?)\)\s+values\s*\((.*?)\)';
        $this->_set     = '\s+set\s+((?:\s*'.$this->_id.'\s*\=\s*'.$this->_value.'\s*,?)+)';
        $this->_where   = '(?:\s+where\s+(.*?))?';
        $this->_orderBy = '(?:\s+order\s+by\s+((?:' . $this->_id . '|,\s*)+))?(\s+desc)?';
        $this->_e       = '\s*;?\s*$/si';
    }

    /**
     * parse SQL query into query object
     *
     * Returns bool(true) if the input SQL statement
     * is valid (in the sense of this parser) and
     * bool(false) on failure.
     *
     * @access  public
     * @param   string  $sqlStmt
     * @return  bool
     */
    function parseSQL($sqlStmt)
    {
        $sqlStmt = (string) $sqlStmt;

        /* init var $m */
        $m = array();
        switch (true)
        {
            /**
             * A) SELECT 1 FROM table { WHERE }
             */
            case preg_match($this->_s.$this->_select1.$this->_table.$this->_where.$this->_e, $sqlStmt, $m):

                /**
                 * 0) initialize variables
                 */
                $table = $m[1];
                $where = @$m[2];
                unset($m);

                /**
                 * 1) set table
                 */
                if (!$this->setType(YANA_DB_EXISTS)) {
                    return false;
                } elseif (!$this->setTable($table)) {
                    return false;

                /**
                 * 2) set where clause
                 */
                } elseif (empty($where)) {
                    return true;
                } elseif (!$this->_where($where)) {
                    return false;
                } elseif (!$this->setWhere($where)) {
                    return false;
                } else {
                    return true;
                }
            break;

            /**
             * B) SELECT count(*) FROM table { WHERE }
             */
            case preg_match($this->_s.$this->_count.$this->_table.$this->_where.$this->_e, $sqlStmt, $m):

                /**
                 * 0) initialize variables
                 */
                $table = $m[1];
                $where = @$m[2];
                unset($m);

                /**
                 * 1) set table
                 */
                if (!$this->setType(YANA_DB_LENGTH)) {
                    return false;
                } elseif (!$this->setTable($table)) {
                    return false;

                /**
                 * 2) set where
                 */
                } elseif (empty($where)) {
                    return true;
                } elseif (!$this->_where($where)) {
                    return false;
                } elseif (!$this->setWhere($where)) {
                    return false;
                } else {
                    return true;
                }
            break;

            /**
             * B) SELECT column FROM { table1, table2 | table1 NATURAL JOIN table2 } { WHERE } { ORDER BY } { DESC }
             */
            case preg_match($this->_s.$this->_select.$this->_tables.$this->_where.$this->_orderBy.$this->_e, $sqlStmt, $m):

                /**
                 * 0) initialize variables
                 */
                $column  = (@$m[1]);
                $tables  = strtolower($m[2]);
                $where   = @$m[3];
                $orderBy = $this->_orderBy(@$m[4]);
                $desc    = !@empty($m[5]);
                unset($m);

                /**
                 * 1) set table
                 */

                /**
                 * 1.1) resolve natural join
                 */
                while (preg_match('/('.$this->_id.')\s+NATURAL\s+JOIN\s+('.$this->_id.')/', $tables, $m))
                {
                    if (!$this->db->structure->isTable($m[1])) {
                        return false;
                    } elseif (!$this->db->structure->isTable($m[2])) {
                        return false;
                    } else {
                        $tables = str_replace($m[0], $m[1] . ',' . $m[2], $tables);
                        $listA = $this->db->structure->getColumns($m[1]);
                        $listB = $this->db->structure->getColumns($m[2]);
                        foreach ($listA as $colA)
                        {
                            if (in_array($colA, $listB)) {
                                $where .= (($where != '') ? ' and ' : '') . $m[1] . '.' . $colA . '=' . $m[2] . '.' . $colA;
                            }
                        } /* end foreach */
                        unset($listA, $listB, $colA);
                    }
                } /* end while */
                /**
                 * 1.2) create list of tables
                 */
                $tables = preg_split('/\s*,\s*/', $tables);
                $tables = array_unique($tables);

                /**
                 * 1.3) set table
                 */
                if (empty($tables)) {
                    return false;
                } elseif (!$this->setType(YANA_DB_SELECT)) {
                    return false;
                } elseif (!$this->setTable(array_shift($tables))) {
                    return false;
                } elseif (!$this->_column($column)) {
                    return false;
                } elseif (!$this->setColumn($column)) {
                    return false;
                }

                /**
                 * 2) resolve where clause
                 *
                 * Note: is not yet set!
                 */
                if (empty($where)) {
                    /**
                     * Note: where clause is mandatory for joined tables,
                     * to avoid accidental cross-joins.
                     */
                    if (count($tables) === 0) {
                        $where = null;
                    } else {
                        trigger_error("SQL error: accidental cross-join detected in statement '${sqlStmt}'.\n\t\t" .
                        "The statement has been ignored.", E_USER_NOTICE);
                        return false;
                    }
                } elseif (!$this->_where($where)) {
                    return false;
                }

                /**
                 * 3) set order by and direction
                 */
                if (!$this->setOrderBy($orderBy, $desc)) {
                    return false;
                }

                /**
                 * 4) join tables
                 */
                if (count($tables) > 0) {
                    $tables = array_flip($tables);
                    assert('!isset($i);       // Cannot redeclare variable $i');
                    assert('!isset($clause);  // Cannot redeclare variable $clause');
                    assert('!isset($tableA);  // Cannot redeclare variable $tableA');
                    assert('!isset($columnA); // Cannot redeclare variable $columnA');
                    assert('!isset($tableB);  // Cannot redeclare variable $tableB');
                    assert('!isset($columnB); // Cannot redeclare variable $columnB');
                    foreach ($where as $i => $clause)
                    {
                        /* if is join-clause */
                        if (is_array($clause[0]) && is_array($clause[1])) {
                            $tableA  = $clause[0][0];
                            $columnA = $clause[0][1];
                            $tableB  = $clause[1][0];
                            $columnB = $clause[1][1];

                            /**
                             * flip operands
                             */
                            if (strcasecmp($tableA, $this->table) !== 0 && strcasecmp($tableB, $this->table) === 0) {
                                $tableB  = $clause[0][0];
                                $columnB = $clause[0][1];
                                $tableA  = $clause[1][0];
                                $columnA = $clause[1][1];

                            /**
                             * note - ignore when there is nothing to join
                             */
                            } elseif (strcasecmp($tableA, $this->table) !== 0) {
                                /* intentionally left blank */
                                continue;
                            }

                            /**
                             * error - referenced table not found in select clause
                             */
                            if (!in_array($tableB, $tables)) {
                                return false;

                            /**
                             * error - unable to join tables
                             */
                            } elseif (!$this->setJoin($tableB, $columnA, $columnB)) {
                                return false;
                            } else {
                                unset($tables[$tableB]);
                            }
                            unset($where[$i]);
                        }
                    }
                    unset($tableA, $columnA, $tableB, $columnB, $i, $clause);
                }

                /**
                 * 4).1 error - cross join detected
                 */
                if (count($tables) !== 0) {
                    trigger_error("SQL error: accidental cross-join detected in statement '${sqlStmt}'.\n\t\t" .
                    "The statement has been ignored.", E_USER_NOTICE);
                    return false;
                }

                /**
                 * 5) set where clause
                 */
                if (!$this->setWhere($where)) {
                    return false;
                } else {
                    return true;
                }
            break;

            /**
             * C) INSERT INTO table (keys) VALUES (values)
             */
            case preg_match($this->_s.$this->_insert.$this->_values.$this->_e, $sqlStmt, $m):

                /**
                 * 0) initialize variables
                 */
                $table  = $m[1];
                $keys   = @$m[2];
                $values = @$m[3];
                $set    = array(); /* combined array of $keys and $values */

                unset($m);
                /**
                 * 1) set table
                 */
                if (!$this->setType(YANA_DB_INSERT)) {
                    return false;
                } elseif (!$this->setTable($table)) {
                    return false;

                /**
                 * 2) combine arrays of keys and values
                 */
                } else {
                    $values = $this->_parseValues($values);
                    if (!is_array($values)) {
                        /* parse error */
                        return false;
                    } else {
                        $keys = preg_split('/,\s*/', $keys);
                        $set = $this->_set($keys, $values);
                        if (empty($set)) {
                            trigger_error('SQL syntax error. The statement contains illegal values.', E_USER_WARNING);
                            return false;
                        }
                    }

                }
                unset($keys, $values);

                /**
                 * 3) set values
                 */
                if (!$this->setValues($set)) {
                    return false;

                /**
                 * 4) check security constraint
                 */
                } elseif ($this->getExpectedResult() !== YANA_DB_ROW) {
                    if ($this->db->structure->isAuto($this->table, $this->db->structure->getPrimaryKey($this->table))) {
                        return true;
                    } else {
                        trigger_error("SQL security restriction. Cannot insert a table (only rows).", E_USER_WARNING);
                        return false;
                    }
                } else {
                    return true;
                }
            break;

            /**
             * D) UPDATE table SET { WHERE }
             */
            case preg_match($this->_s.$this->_update.$this->_set.$this->_where.$this->_e, $sqlStmt, $m):

                /**
                 * 0) initialize variables
                 */
                $table  = $m[1];
                $values = $m[2];
                $where  = @$m[3];
                unset($m);

                /**
                 * 1) set table
                 */
                if (!$this->setType(YANA_DB_UPDATE)) {
                    return false;
                } elseif (!$this->setTable($table)) {
                    return false;
                }

                /**
                 * 2) where clause
                 */
                if (empty($where)) {
                    trigger_error("SQL security restriction. Cannot update a table (only rows and cells).", E_USER_WARNING);
                    return false;
                } elseif (!$this->_where($where)) {
                    return false;
                } elseif (!$this->setWhere($where)) {
                    return false;
                }

                /**
                 * 3) resolve values
                 */
                assert('!isset($nrOfHits); // Cannot redeclare variable $nrOfHits');
                assert('!isset($m); // Cannot redeclare variable $m');
                $nrOfHits = preg_match_all('/\s*('.$this->_id.')\s*\=\s*'.$this->_value2.'\s*,?/s', $values, $m);
                if ($nrOfHits === false || $nrOfHits < 1) {

                    trigger_error("SQL syntax error. The string '" . print_r($values, true) . "' " .
                    "is not a valid value in the provided SQL-statement.", E_USER_WARNING);
                    return false;

                } else {
                    $values = $this->_set($m[1], $m[2]);
                    if (empty($values)) {
                        trigger_error('SQL syntax error. The statement contains illegal values.', E_USER_WARNING);
                        return false;
                    }
                } /* end if */
                unset($m, $nrOfHits);

                /**
                 * 4) set values
                 */
                $expectedResult = $this->getExpectedResult();
                if (!$this->setValues($values)) {
                    return false;

                /**
                 * 5) check security constraint
                 */
                } elseif ($expectedResult !== YANA_DB_ROW && $expectedResult !== YANA_DB_CELL) {
                    trigger_error("SQL security restriction. Cannot update a table (only rows and cells).", E_USER_WARNING);
                    return false;
                } else {
                    return true;
                }
            break;

            /**
             * E) DELETE FROM table WHERE { ORDER BY } { DESC }
             */
            case preg_match($this->_s.$this->_delete.$this->_table.$this->_where.$this->_orderBy.$this->_e, $sqlStmt, $m):

                /**
                 * 0) initialize variables
                 */
                $table   = $m[1];
                $where   = @$m[2];
                $orderBy = $this->_orderBy(@$m[3]);
                $desc    = !@empty($m[4]);
                unset($m);

                /**
                 * 1) set table
                 */
                if (!$this->setType(YANA_DB_DELETE)) {
                    return false;
                } elseif (!$this->setTable($table)) {
                    return false;

                /**
                 * 2) set order by + direction
                 */
                } elseif (!$this->setOrderBy($orderBy, $desc)) {
                    return false;

                /**
                 * 3) where clause
                 */
                } elseif (empty($where)) {
                    return true;
                } elseif (!$this->_where($where)) {
                    return false;
                } elseif (!$this->setWhere($where)) {
                    return false;
                } else {
                    return true;
                }
            break;

            /**
             * F) invalid statement
             */
            default:
                /* throw a line if unknown statement encountered */

                trigger_error("Invalid or unknown SQL statement: $sqlStmt.", E_USER_WARNING);
                return false;

            break;
        } /* end switch */
    }

    /**
     * resolve where clause
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note that the argument $where is taken by reference.
     *
     * @access  private
     * @param   string  &$where
     * @return  bool
     * @ignore
     */
    function _where(&$where)
    {
        assert('is_string($where);');
        /* settype to STRING */
        $where = (string) $where;
        $where = trim($where);
        $where = explode(' AND ', $where);

        if (($this->type === YANA_DB_UPDATE || $this->type === YANA_DB_DELETE) && $this->getRow() === '*' && $this->getTable() !== false) {
            $primaryKey = $this->db->structure->getPrimaryKey($this->table);
        } else {
            $primaryKey = null;
        }

        if (empty($where)) {
            $where = array();
            return true;
        } else {
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($val); /* cannot redeclare variable $val */');
            foreach($where as $i => $val)
            {
                if (empty($val)) {
                    /* intentionally left blank */
                    continue;
                } else {
                    $m = array();
                    $where[$i] = array();
                    $leftOperand = null;
                    $rightOperand = null;
                    $operator = null;

                    /**
                     * 1) is test for existence
                     */
                    if (preg_match('/^(\S+)\s+is(\s+not)?\s+null\s*;?$/i', $val, $m)) {

                        $leftOperand = strtoupper($m[1]);
                        $rightOperand = null;
                        if (empty($m[2])) {
                            $operator = '=';
                        } else {
                            $operator = '!=';
                        }

                    /**
                     * 2) is comparison
                     */
                    } elseif (preg_match('/^\s*(\S+)\s*(=|REGEXP|regexp|LIKE|like|<|>|\!=|<=|>=)\s*(.+)\s*;?$/', $val, $m)) {

                        /**
                         * 1) left operand
                         */
                        /* a) flip operands, where necessary */
                        if (preg_match('/^'.$this->_value.'$/', $m[1])) {
                            $m[4] = $m[3];
                            $m[3] = $m[1];
                            $m[1] = $m[4];
                            unset($m[4]);
                        }
                        /* b) is column name */
                        if (preg_match('/^'.$this->_id.'$/', $m[1])) {
                            $leftOperand = strtolower($m[1]);
                        /* c) is table + column name */
                        } elseif (preg_match('/^'.$this->_id.'\.'.$this->_id.'$/', $m[1])) {
                            $leftOperand = explode('.', strtolower($m[1]));
                        /* d) syntax error */
                        } else {
                            return false;
                        }

                        /**
                         * 2) operator
                         */
                        $operator = strtolower($m[2]);

                        /**
                         * 3) right operand
                         */
                        /* a) is column name */
                        if (preg_match('/^'.$this->_id.'$/', $m[3])) {
                            $rightOperand = strtolower($m[3]);
                        /* b) is table + column name */
                        } elseif (preg_match('/^'.$this->_id.'\.'.$this->_id.'$/', $m[3])) {
                            $rightOperand = explode('.', strtolower($m[3]));
                        /* c) is value */
                        } elseif (preg_match('/^'.$this->_value2.'$/', $m[3], $m2)) {
                            $rightOperand = $m2[1];
                        /* d) syntax error */
                        } else {
                            return false;
                        }

                        /**
                         * 4) extract primary key
                         */
                        if (!is_null($primaryKey) && $operator === '=' && is_string($rightOperand)) {
                            switch (true)
                            {
                                case is_array($leftOperand)  && strcasecmp($primaryKey, $leftOperand[1]) === 0:
                                case !is_array($leftOperand) && strcasecmp($primaryKey, $leftOperand) === 0:
                                    if ($this->setRow($rightOperand)) {
                                        $primaryKey = null;
                                        unset($where[$i]);
                                        continue 2;
                                    }
                                break;
                            }
                        }
                    } else {
                        /* syntax error */
                        return false;
                    }

                    /**
                     * add the parsed clause
                     */
                    $where[$i][0] = $leftOperand;
                    $where[$i][1] = $rightOperand;
                    $where[$i][2] = $operator;
                } /* end if */
            } /* end foreach */
            unset($i, $val);
            return true;
        }
    }

    /**
     * resolve "column" clause
     *
     * Returns name of column.
     *
     * @access  private
     * @param   string  &$column
     * @param   bool    $checkTable
     * @return  string
     * @ignore
     */
    function _column(&$column, $checkTable = true)
    {
        if ($checkTable) {
            $table = $this->table;
        }
        if (is_string($column)) {
            $column = trim(strtolower($column));
            if (strpos($column, '.') !== false) {
                $column = explode('.', $column);
                if ($checkTable && strcasecmp($column[0], $this->table) === 0) {
                    $column = $column[1];
                    return true;
                } elseif (!$checkTable && $this->db->structure->isTable($column[0])) {
                    $column = $column[1];
                    return true;
                } else {
                    $column = null;
                    return false;
                }
            } else {
                return true;
            }
        } else {
            $column = null;
            return false;
        }
    }

    /**
     * resolve "order by" clause
     *
     * Returns a list of columns.
     *
     * @access  private
     * @param   string  $orderBy
     * @return  array
     * @ignore
     */
    function _orderBy($orderBy)
    {
        assert('is_string($orderBy) || is_null($orderBy);');
        if (empty($orderBy)) {
            return array();
        } else {
            /* settype to STRING */
            $orderBy = (string) $orderBy;

            $orderBy = explode(',', strtoupper(trim($orderBy)));
            $result = array();
            if (empty($orderBy)) {
                return array();
            } else {
                assert('!isset($m); /* cannot redeclare variable $m */');
                assert('!isset($val); /* cannot redeclare variable $val */');
                foreach($orderBy as $val)
                {
                    /* unquote id */
                    if (!empty($val)) {
                        $result[] = trim($val);
                    } else {
                        /* intentionally left blank */
                        continue;
                    } /* end if */
                } /* end foreach */
                unset($val);
                return $result;
            } /* end if */
        } /* end if */
    }

    /**
     * combine a list of keys and values
     *
     * Returns the row-array on success.
     * On failure an empty array is returned.
     *
     * @access  private
     * @param   array  $keys
     * @param   array  $values
     * @return  array
     * @ignore
     */
    function _set($keys, $values)
    {
        assert('is_array($keys);');
        assert('is_array($values);');
        if (!is_array($keys) || !is_array($values)) {
            return array();
        }

        assert('count($keys) == count($values);');
        /* array_combine() became available in PHP 5, here's a code to simulate this behaviour for PHP 4 */
        $insert = array();
        assert('!isset($isStrict); /* Cannot redeclare var $isStrict */');
        $isStrict = $this->db->structure->isStrict();
        assert('!isset($i); /* Cannot redeclare var $i */');
        for ($i = 0; $i < count($keys); $i++)
        {
            if ($isStrict === true && !$this->db->structure->isColumn($this->table, $keys[$i])) {
                trigger_error("SQL syntax error. The column '".$keys[$i]."' does not exist in table '".$this->table."'.", E_USER_WARNING);
                return array();
            } elseif ($this->db->structure->getType($this->table, $keys[$i]) === 'array') {
                $insert[strtoupper($keys[$i])] = SML::decode($values[$i], CASE_UPPER);
            } else {
                $insert[strtoupper($keys[$i])] = $values[$i];
            }
        } /* end foreach */
        unset($i);

        assert('is_array($insert)');
        return $insert;
    }
    
    /**
     * parse value string into array
     *
     * @access  private
     * @param   string  $values
     * @return  array
     * @ignore
     */
    function _parseValues($values)
    {
        $isString = false;
        $result = array();
        $buffer = "";
        $j = 0;
        for ($i = 0; $i < strlen($values); $i++)
        {
            switch ($values[$i])
            {
                case $this->_delim:
                    if ($isString) {
                        if (isset($result[$j])) {
                            /* error - parse error */
                            $result = false;
                            break;
                        } else {
                            $result[$j] = $buffer;
                        }
                        $buffer = "";
                        $isString = false;
                    } else {
                        $buffer = "";
                        $isString = true;
                    }
                break;
                case ",":
                    if ($isString) {
                        $buffer .= $values[$i];
                    } else {
                        if (is_numeric($buffer)) {
                            if (isset($result[$j])) {
                                /* error - parse error */
                                $result = false;
                                break;
                            } else {
                                $result[$j] = $buffer;
                            }
                            $buffer = "";
                        }
                        $j++;
                    }
                break;
                case " ":
                    if ($isString) {
                        $buffer .= $values[$i];
                    } else {
                        /* ignore */
                    }
                break;
                default:
                    if ($isString) {
                        $buffer .= $values[$i];
                    } elseif (preg_match('/[\w\d.,]/', $values[$i])) {
                        $buffer .= $values[$i];
                    } else {
                        /* error - parse error */
                        $result = false;
                        break;
                    }
                break;
            }
        } /* end for */

        /* check: missing value */
        if (!isset($result[$j])) {
            if ($buffer !== "" && is_numeric($buffer)) {
                $result[$j] = $buffer;
            } else {
                /* error - parse error */
                $result = false;
            }
        }

        /* check: unclosed string literal */
        if ($isString) {
            /* error - parse error */
            $result = false;
        }
        
        if (is_array($result)) {
            return $result;

        } else {
            trigger_error("SQL syntax error. The string '${values}' is not a valid value in the provided SQL-statement.\n\t\t" .
            "Error occured at character index '$i', near '" . substr($values, $i-strlen($buffer), 5) . "'.", E_USER_WARNING);
            return false;            
        }
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool
     * @ignore
     */
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                global $YANA;
                $db = new FileDb('check');
                $parser = new DbQueryParser($db);
                $profileId = $YANA->session->getId();

                // exists
                $sql1 = 'SELECT 1 FROM ft WHERE ft.ftid = "2"';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // alternative writing
                $parser = new DbQueryParser($db);
                $sql3 = 'SELECT 1 FROM ft WHERE ftid = "2"';
                if (!$parser->parseSQL($sql3)) {
                    return "Parser error: $sql3";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // length
                $parser = new DbQueryParser($db);
                $sql1 = 'SELECT count(*) FROM ft WHERE ft.ftid = "2"';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // select table
                $parser = new DbQueryParser($db);
                $sql1 = 'SELECT * FROM ft';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // select row
                $parser = new DbQueryParser($db);
                $sql1 = 'SELECT * FROM ft WHERE ft.ftid = "2"';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // select column
                $parser = new DbQueryParser($db);
                $sql1 = 'SELECT ftid FROM ft WHERE ftid = "2" ;';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ('SELECT ft.ftid FROM ft WHERE ft.ftid = "2"' !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // select cross join
                $parser = new DbQueryParser($db);
                $sql1 = 'SELECT ft.ftid FROM ft,t WHERE ft.ftid = "2"';
                if (@$parser->parseSQL($sql1) !== false) { // supposed to fail (due to cross join)
                    return "Parser error: $sql1";
                }

                // missing foreign key
                $parser = new DbQueryParser($db);
                $sql1 = 'INSERT INTO t (tid, tvalue) VALUES ("1", "2")';
                if (@$parser->parseSQL($sql1) !== false) { // supposed to fail (due to missing foreign key)
                    return "Parser error: $sql1";
                }

                // insert into (with profile constraint)
                $parser = new DbQueryParser($db);
                $sql1 = 'INSERT INTO t (tid, tvalue, ftid) VALUES ("1", "2", "1");';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ('INSERT INTO t (tid, tvalue, ftid, profile_id) VALUES ("1", "2", "1", "' . $profileId . '")' !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // missing primary key
                $parser = new DbQueryParser($db);
                $sql1 = 'UPDATE ft SET ftvalue="2" WHERE ftvalue = "2"';
                if (@$parser->parseSQL($sql1) !== false) { // supposed to fail (due to missing primary key)
                    return "Parser error: $sql1";
                }

                // update
                $parser = new DbQueryParser($db);
                $sql1 = 'UPDATE t SET tvalue="2" WHERE tid = "2";';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ('UPDATE t SET tvalue = "2" WHERE profile_id = "' . $profileId . '" AND t.tid = "2"' !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // delete
                $parser = new DbQueryParser($db);
                $sql1 = 'DELETE FROM ft WHERE ft.ftid = "1"';
                if (!$parser->parseSQL($sql1)) {
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                // delete table
                $parser = new DbQueryParser($db);
                $sql1 = 'DELETE FROM ft';
                if (!$parser->parseSQL($sql1)) { // supposed to fail (due to missing primary key)
                    return "Parser error: $sql1";
                }
                $sql2 = $parser->toString();
                if ($sql1 !== $sql2) {
                    return "Statement not resolved: $sql2";
                }

                return true;
            break;
        }
    } /* end of function */
}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
