<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**#@+
 * Load only if database connection is activated.
 *
 * @ignore
 */
if (YANA_DATABASE_ACTIVE) {
    @include_once("DB.php");
}
if (!defined('DB_FETCHMODE_ORDERED')) {
    define('DB_FETCHMODE_ORDERED', 1);
}
if (!defined('DB_FETCHMODE_ASSOC')) {
    define('DB_FETCHMODE_ASSOC', 2);
}
if (!defined('DB_TABLEINFO_ORDER')) {
    define('DB_TABLEINFO_ORDER', 1);
}
/**#@-*/

/**
 * database API
 *
 * this class is a database abstraction api, that uses pear db
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 */
class DbStream extends Object
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var  int       */   var $exists              = YANA_DB_FAILURE;
    /** @var  Object    */   var $database            = null;
    /** @var  DbTrigger */   var $exec                = null;
    /** @var  string    */   var $lastSelectedTable   = "";
    /** @var  string    */   var $path                = "";
    /** @var  array     */   var $queue               = array();
    /** @var  array     */   var $cache               = array();
    /** @var  array     */   var $joins               = array();
    /** @var  array     */   var $reservedSqlKeywords = null;
    /** @var  SML       */   var $lastModified        = null;
    /** @var  array     */   var $dsn                 = array();
    /** @var  string    */   var $errorMessage        = false;

    /**#@-*/

    /**
     * structure of current database
     *
     * This gives you access to the db-schema that is used in the
     * current session.
     *
     * Please note that you should not change this schema unless
     * you REALLY know what you are doing.
     *
     * @access  public
     * @var     DbStructure
     */
    var $structure = null;

    /**
     * @var     DbDatatype
     * @access  public
     * @ignore
     */
    var $datatype  = null;

    /**
     * create a new instance
     *
     * Each database connection depends on a structure file
     * describing the database. Therefore you have to provide
     * a $filename of such structure file (without path or extension)
     *
     * These files are to be found in config/db/*.config
     *
     * Since version 2.8.7 the argument $file can also be an instance
     * of class DbStructure.
     *
     * @param  string|DbStructure  $file
     * @param  dbServer            $server
     */
    function __construct($file = null, $server = null)
    {
        /* 1 CHECK INPUT
         */

        /* 1.1 FIRST ARGUMENT
        */
        if (is_null($file)) {
            $this->path  = "";
            $this->structure = null;
        } elseif (is_object($file)) {

            /* 1.1 a) ERROR CASE: invalid base class
            */
            if (!is_a($file, 'DbStructure')) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbStructure', get_class($file)), E_USER_WARNING);
                $file = "";

            /* 1.1 b) SETTING CLASS ATTRIBUTES
            */
            } else {
                $this->path  = $file->getPath();
                $this->structure = $file;
            }

        } else {

            /* 1.1 a) ERROR CASE: invalid data type
            */
            if (!empty($file) && !is_string($file)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($file)), E_USER_WARNING);
                $this->path  = "";
                $this->structure = null;

            /* 1.1 b) SETTING CLASS ATTRIBUTES
            */
            } else {
                /* settype to STRING */
                $file = (string) $file;
                $this->path = $file;
                $this->structure = new DbStructure($this->getPath());
            }

        }

        /* 1.2 SECOND ARGUMENT
        */

        /* 1.2 a) ERROR CASE: invalid data type
        */
        if (is_null($server)) {
            $server = new DbServer();
        } elseif (is_object($server)) {
            if (!is_a($server, 'DbServer')) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Instance of DbServer', get_class($server)), E_USER_WARNING);
                $server = new dbServer();
            }
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Object', gettype($server)), E_USER_WARNING);
            $server = new DbServer();
        }

        /* 1.2 b) SETTING CLASS ATTRIBUTES
        */
        $this->database = $server->get();
        $this->dsn = $server->getDsn();

        /* 2 LIST OF GLOBALS
         */

        global $YANA;

        /* 3 INIT DATABASE CONNECTION
         */

        if (empty($this->database) || $this->isError($this->database)) {
            /* connection failed - abort */
            $this->exists = YANA_DB_FAILURE;
            $this->handleError($this->database);
        } else {

            /* deactivate autocommit by default */
            $result = $this->database->autoCommit(false);
            if ($this->isError($result)) {
                $this->handleError($result);
            }

            if (is_null($this->structure)) {

                /**
                 * If no structure file is available,
                 * this framework has the ability to estimate the structure
                 * of the database "on the fly" at runtime, via the
                 * "DbStream::buildStructure()" method.
                 */

                $this->buildStructure();
                $this->exists = YANA_DB_NO_FILE;

            } else {

                $test = $this->structure->read();
                if ($test === false) {
                    /* given file does not exist */
                    unset($this->structure);
                    $this->exists = YANA_DB_NOT_AVAILABLE;
                    trigger_error("Unable to initialize database. Structure file '".$this->getPath()."' not found.", E_USER_ERROR);
                    exit(1);
                } else {
                    /* all fine, proceed */
                    $this->exists = YANA_DB_OK;
                }

            } /* end if */
        } /* end if */

        /* 4 INIT IO/FILTER
         */
        if (is_object($this->structure)) {
            $this->datatype = new DbDatatype($this->structure);
        }

        /* 5 INIT TRIGGER
         */
        if (is_object($this->structure)) {
            $this->exec = new DbTrigger($this->structure);
        }
    }

    /**
     * get the DSN
     *
     * This function returns an associative array containing
     * information on the current connection or bool(false) on error.
     *
     * See {@link DbServer::getDsn()} for more details.
     *
     * @access  public
     * @name    DbStream::getDsn()
     * @return  array
     * @since   2.9.8
     */
    function getDsn()
    {
        if (is_array($this->dsn)) {
            return $this->dsn;

        } else {
            return false;
        }
    }

    /**
     * Alias of DbStream::write()
     *
     * @access  public
     * @return  bool
     * @see     DbStream::write()
     */
    function commit()
    {
        return $this->write();
    }

    /**
     * Commit current transaction
     *
     * This writes all changes to the database
     *
     * @access  public
     * @return  bool
     * @name    DbStream::write()
     */
    function write()
    {
        if (!$this->_isWriteable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        if (count($this->queue) == 0) {
            /* Buffer empty */
            return true;
        } else {
            assert('!isset($dbDir); /* Cannot redeclare $dbDir */');
            $dbDir = $GLOBALS['YANA']->getVar('DBBLOB');
            assert('!isset($i); /* Cannot redeclare $i */');
            for ($i = 0; $i < count($this->queue); $i++)
            {
                /**
                 * 1) get query object
                 */
                if (is_array($this->queue[$i]) && isset($this->queue[$i][0])) {
                    $dbQuery =& $this->queue[$i][0];
                } else {
                    $dbQuery =& $this->queue[$i];
                }

                /**
                 * 2) get arguments for trigger
                 */
                if (is_array($this->queue[$i]) && isset($this->queue[$i][1]) && is_array($this->queue[$i][1])) {
                    $args = $this->queue[$i][1];
                } else {
                    $args = null;
                }

                if (YANA_ERROR_REPORTING === YANA_ERROR_LOG)
                {
                    trigger_error("$dbQuery\n\n", E_USER_NOTICE);
                }

                /**
                 * 3.a) handle query object
                 */
                if (is_object($dbQuery)) {
                    switch ($dbQuery->getType())
                    {
                        /**
                         * 3.1) delete a row
                         *
                         * For security reasons all delete queries will automatically
                         * be limited to 1 row at a time.
                         */
                        case YANA_DB_DELETE:
                            /**
                             * 3.1.1) get list of files
                             *
                             * When a row is deleted, blobs associated with it need to be
                             * removed as well.
                             *
                             * A list of these files is created before the row is deleted.
                             * If the row has successfully been removed, the files will
                             * get deleted afterwards.
                             */
                            assert('!isset($table); /* Cannot redeclare $table */');
                            $table = $dbQuery->getTable();
                            assert('!isset($filesToDelete); /* Cannot redeclare $filesToDelete */');
                            assert('!isset($column); /* Cannot redeclare $column */');
                            assert('!isset($file); /* Cannot redeclare $file */');
                            $filesToDelete = array();
                            foreach ($this->structure->getFiles($table) as $column)
                            {
                                $result = $this->database->limitQuery("SELECT $column from ".$this->quoteId(YANA_DATABASE_PREFIX.$table)." where ".$this->structure->getPrimaryKey($table)." = ".$this->quote($dbQuery->getRow()), 0, 1);
                                if (!$this->isError($result)) {
                                    $file = $result->fetchRow(DB_FETCHMODE_ASSOC, 0);
                                    assert('is_array($file); /* unexpected result: $file */');
                                    $filesToDelete[] = $dbDir . array_pop($file);
                                }
                            } /* end foreach */
                            unset($column, $file);
                            $result = $dbQuery->sendQuery();
                        break;
                        /**
                         * 3.2) update a row
                         */
                        case YANA_DB_UPDATE:
                            /* intentionally left blank */
                            $result = $dbQuery->sendQuery();
                        break;
                        /**
                         * 3.3) insert a row
                         */
                        case YANA_DB_INSERT:
                            if (@$this->dsn['DBMS'] === 'mssql' && $dbQuery->getRow() !== '*') {
                                /**
                                 * MSSQL compatibility
                                 *
                                 * {@internal
                                 * MSSQL does not allow entries of identity
                                 * tables to be inserted with a certain primary
                                 * key unless explicitely told to do so.
                                 * This exotic behavior does not apply to any
                                 * other DBMS.
                                 * }}
                                 */
                                assert('!isset($table); /* Cannot redeclare $table */');
                                $table = $dbQuery->getTable();
                                /* test if is identity table */
                                if ($this->structure->isAutonumber($table, $this->structure->getPrimaryKey($table))) {
                                    /* set identity restriction off */
                                    $this->query('SET IDENTITY_INSERT [' . YANA_DATABASE_PREFIX.$table . '] ON;');
                                    $result = $dbQuery->sendQuery();
                                    /* reset identity restriction */
                                    $this->query('SET IDENTITY_INSERT [' . YANA_DATABASE_PREFIX.$table . '] OFF;');
                                } else {
                                    $result = $dbQuery->sendQuery();
                                }
                                unset($table);
                            } else {
                                /* send request to database */
                                $result = $dbQuery->sendQuery();
                            }
                        break;
                        default:
                            /* send request to database */
                            $result = $dbQuery->sendQuery();
                        break;
                    } /* end switch (type) */


                /**
                 * 3.b) handle query string
                 */
                } elseif (is_string($dbQuery)) {
                    $result = $this->query($dbQuery);

                /**
                 * 3.c) error - invalid argument type
                 */
                } else {
                    trigger_error("The value '" . print_r($dbQuery) . "' is not a valid query.", E_USER_WARNING);
                    return false;
                }

                /**
                 * 4.1) error - query failed
                 */
                assert('!isset($success); // cannot redeclare variable $success');
                if ($this->isError($result)) {
                    /**
                     * 4.1.1) fire trigger
                     */
                    if (!is_null($args)) {
                        $success = false;
                        $this->exec->trigger(YANA_TRIGGER_AFTER, $args[0], $args[1], $args[2], $success);
                    }
                    /**
                     * 4.1.2) rollback on error
                     */
                    if (is_string($dbQuery)) {
                        if (YANA_ERROR_REPORTING === YANA_ERROR_LOG)
                        {
                            trigger_error("Failed: $dbQuery\n" . $result->getMessage() . "\n", E_USER_WARNING);
                        }
                        $this->handleError($result, "Statement '${dbQuery}' on database failed.");
                    } else {
                        if (YANA_ERROR_REPORTING === YANA_ERROR_LOG)
                        {
                            trigger_error("Failed: $dbQuery\n" . $result->getMessage() . "\n", E_USER_WARNING);
                        }
                        $this->handleError($result, "Statement '" . $dbQuery->toString() . "' on database failed.");
                    }
                    $result = $this->database->rollback();
                    /**
                     * 4.1.3) when rollback failed, create entry in logs
                     */
                    if ($this->isError($result)) {
                          $this->handleError(null, "Unable to rollback changes. Database might contain corrupt data. Please inform webmaster about time of event and anything you might have done to cause this failure.");
                    }
                    return false;

                /**
                 * 4.2) query was successfull
                 */
                } else {
                    /**
                     * 4.2.1) Handling of associated files and images
                     *
                     * Remove files associated with deleted entry.
                     * The list of files has been created earlier
                     * and stored in array $filesToDelete.
                     */
                    if (is_object($dbQuery)) {
                        if ($dbQuery->getType() === YANA_DB_DELETE && !empty($filesToDelete)) {
                            assert('!isset($file); /* Cannot redeclare $file */');
                            foreach ($filesToDelete as $file)
                            {
                                if (is_file($file)) {
                                    unlink($file);
                                    $file = str_replace('image.', 'thumb.', $file);
                                    if (is_file($file)) {
                                        unlink($file);
                                    }
                                }
                            } /* end foreach */
                            unset($file);
                        }
                    } /* end if */

                    /**
                     * 4.2.2) fire trigger
                     */
                    if (!is_null($args)) {
                        $success = true;
                        $this->exec->trigger(YANA_TRIGGER_AFTER, $args[0], $args[1], $args[2], $success);
                    }
                }
                unset($success, $table, $row, $column, $filesToDelete);
            } /* end for */
            unset($i);

            /**
             * 5) commit changes
             *
             * The time when the database was last modified
             * is updated, to provide protection from race
             * conditions where two transaction try to modify
             * the same data.
             */
            $result = $this->database->commit();
            /**
             * 5.1) commit failed
             */
            if ($this->isError($result)) {
                if (YANA_ERROR_REPORTING === YANA_ERROR_LOG)
                {
                    trigger_error("Failed: $dbQuery\n" . $result->getMessage() . "\n", E_USER_WARNING);
                }
                if (!is_null($this->lastModified)) {
                    $this->lastModified->reset();
                }
                $this->handleError($result);
                return false;

            /**
             * 5.2) commit successfull
             */
            } else {
                if (!is_null($this->lastModified)) {
                    $this->lastModified->write();
                }
                $this->queue = array();
                return true;
            }
            unset($dbDir);
        } /* end if */
    }

    /**
     * get values from the database
     *
     * This returns the values at adress $key
     * starting from $offset and limited to
     * $limit results.
     *
     * The $key parameter has three synopsis.
     * <ul>
     *     <li>
     *         If $key is a string, this parameter is interpreted
     *         as the address of the values you want to retrieve.
     *     </li>
     *     <li>
     *         If $key is an associative array, it overwrites any of the
     *         paramters you provide, without the need to enter default
     *         parameters - this might come in handy in certain situations.
     *     </li>
     *     <li>
     *         Since version 2.9.3 the argument $key may also be an object
     *         of type $dbQuery. If so, no additional parameters need to be
     *         present. This is a shortcut, which e.g. allows you to prepare
     *         a query as an object and reuse it with multiple arguments.
     *     </li>
     * </ul>
     *
     * The array $key can have the following values
     * <ul>
     *     <li>  $key      = the address of the value(s) to retrieve (mandatory) </li>
     *     <li>  $where    = a search string (optional) </li>
     *     <li>  $search   = alias of $where (deprecated) </li>
     *     <li>  $order_by = array of columns to order the resultset by (optional) </li>
     *     <li>  $desc     = boolean, if true results will be ordered in descending, otherwise in ascending order (optional) </li>
     *     <li>  $offset   = integer, the number of the first result to be returned (optional) </li>
     *     <li>  $limit    = integer, maximum number of results to return (optional) </li>
     * </ul>
     *
     * Since version 2.8.5 the parameter $orderBy has two synopsis.
     * <ul>
     *     <li>
     *         $orderBy is the name of the column in the current table to order the resultset by.
     *     </li>
     *     <li>
     *         $orderBy is a numeric array of strings, where each element
     *         is the name of a column in the current table.
     *         The resultset will get ordered by the values of these columns
     *         in the direction in which they are provided.
     *         This feature became available in version 2.8.5
     *     </li>
     * </ul>
     *
     * Since version 2.8.5 the parameter $where has two synopsis.
     * <ul>
     *     <li>
     *         $where is a comma seperated list of column-value pairs using the following syntax:
     *         <pre>column1=value1,column2=value2,...,columnN=valueN</pre>
     *         Note that comparison takes place using the SQL operator 'LIKE', except for
     *         columns that are keys (either a primary key or a foreign key), which are compared
     *         using the SQL operator '='.
     *     </li>
     *     <li>
     *         $where is a two-dimensional numeric array and its entries follow this syntax:
     *         <pre>array(array(0=>column1,1=>value1,2=>operator1),...)</pre>
     *         The "column" name is mandatory and needs to be a string.
     *         The "value" is mandatory and needs to be a scalar value.
     *         The "operator" is optional and defaults to 'LIKE'.
     *         It needs to be a string and can be one of the following:
     *         <pre> '=', 'LIKE', '<', '>', '!=', '<=', '>=', '==' (alias of '=') </pre>
     *     </li>
     * </ul>
     * The second alternative offers more flexibility and should be prefered over the first one.
     *
     * Note: The $desc parameter became available in version 2.8.5
     *
     * @access  public
     * @param   array|string|DbQuery  $key      the address of the value(s) to retrieve
     * @param   array|string          $where    a search string
     * @param   array                 $orderBy  a list of columns to order the resultset by
     * @param   int                   $offset   the number of the first result to be returned
     * @param   int                   $limit    maximum number of results to return
     * @param   bool                  $desc     if true results will be ordered in descending, otherwise in ascending order
     * @return  mixed
     */
    function get($key, $where = array(), $orderBy = array(), $offset = 0, $limit = 0, $desc = false)
    {
        /* 1) check availability of database connection */
        if (!$this->_isAvailable()) {
            return false;
        }

        /* 1.1) handle $key array */
        if (is_array($key)) {
            if (isset($key['where'])) {
                $where = $key['where'];
            } elseif (isset($key['search'])) {
                $where = $key['search'];
            }
            if (isset($key['order_by'])) {
                $orderBy = $key['order_by'];
            }
            if (isset($key['desc'])) {
                $desc = $key['desc'];
            }
            if (isset($key['offset'])) {
                $offset = $key['offset'];
            }
            if (isset($key['limit'])) {
                $limit = $key['limit'];
            }
            if (isset($key['key'])) {
                $key = $key['key'];
            } else {
                return false;
            }
        } elseif (is_object($key) && (is_a($key, 'DbQuery') || is_subclass_of($key, 'DbQuery'))) {
            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            if ($dbQuery->getType() !== YANA_DB_SELECT) {
                trigger_error("Unable to select row. Invalid query.", E_USER_WARNING);
                return false;
            }
        } /* end if */

        /**
         * 2) build query
         */
        if (!isset($dbQuery)) {
            $dbQuery = new DbQuery($this);
            $dbQuery->setType(YANA_DB_SELECT);
    
            if (!is_string($key) || !$dbQuery->setKey($key)) {
                trigger_error("Key argument in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
            } else {
                $table = $dbQuery->getTable();
            }

            if (!empty($where) && !$dbQuery->setWhere($where)) {
                trigger_error("Where clause in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
            } elseif ((!empty($orderBy) || $desc === true) && !$dbQuery->setOrderBy($orderBy, $desc)) {
                trigger_error("Order by clause in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
            }
    
            /**
             * 2.1) resolve joined tables
             */
            if (isset($this->joins[$table]) && is_array($this->joins[$table])) {
                assert('!isset($table2);  /* Cannot redeclare $table2 */');
                assert('!isset($columns); /* Cannot redeclare $columns */');
                foreach ($this->joins[$table] as $table2 => $columns)
                {
                    if (!$dbQuery->setJoin($table2, $columns[0], $columns[1])) {
                        trigger_error("Unable to join tables '${table}' and '${table2}' " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                        return false;
                    }
                }
                unset($table2, $columns);
            }

            /**
             * 2.2) set limit and offset
             */
            if ($offset > 0) {
                $dbQuery->setOffset($offset);
            }    
            if ($limit > 0) {
                $dbQuery->setLimit($limit);
            }

        } /* end if */

        /**
         * 3) send query
         */
        $result = $dbQuery->sendQuery();

        if ($this->isError($result)) {
            $this->handleError($result, "Statement '" . $dbQuery->toString() ."' on database failed");
        } else {
            $returnedType = $dbQuery->getExpectedResult();
            $table = $dbQuery->getTable();
            $column = $dbQuery->getColumn();
            $this->lastSelectedTable = $table;
            $REF = array();
            $id = $this->structure->getPrimaryKey($this->getTable());

            for ($i = 0; $i < $result->numRows(); $i++)
            {
                $row = $result->fetchRow(DB_FETCHMODE_ASSOC, $i);
                if (!is_array($row)) {
                    $this->handleError("DB NOTICE: The returned data for statement '" . $dbQuery->toString() ."' must be of type array. Instead database returned the following value '${row}'. The result was considered to be an error.");
                    break;
                } elseif ($returnedType === YANA_DB_TABLE) {
                    assert('!isset($rowId); // Cannot redeclare var $rowId');
                    $rowId = strtoupper($row[$id]);
                    assert('!isset($refKey); /* Cannot redeclare variable $refKey */');
                    if (!empty($REF[$rowId])) {
                        $REF[$rowId] = array($REF[$rowId]);
                        $REF[$rowId][1] = false;
                        $refKey =& $REF[$rowId][1];
                    } else {
                        $refKey =& $REF[$rowId];
                    }
                    unset($rowId);
                }
                assert('!isset($a); /* Cannot redeclare variable $a */');
                assert('!isset($value); /* Cannot redeclare variable $value */');
                assert('!isset($currentTable); /* Cannot redeclare variable $currentTable */');
                foreach ($row as $a => $value)
                {
                    if (!is_null($value)) {
                        switch ($returnedType)
                        {
                            case YANA_DB_TABLE:
                                if (($currentTable = $dbQuery->getParentByColumn($a)) === false) {
                                    $currentTable = $table;
                                }
                                if (!$this->datatype->decode($refKey[strtoupper($a)], $currentTable, $a, $value, '', $dbQuery)) {
                                    unset($REF[strtoupper($row[$id])]);
                                    break;
                                }
                            break;

                            case YANA_DB_ROW:
                                if (($currentTable = $dbQuery->getParentByColumn($a)) === false) {
                                    $currentTable = $table;
                                }
                                if (!$this->datatype->decode($REF[strtoupper($a)], $currentTable, $a, $value, '', $dbQuery)) {
                                    return array();
                                }
                            break;

                            case YANA_DB_CELL:
                                assert('!isset($arrayAddress); /* Cannot redeclare variable $arrayAddress */');
                                $arrayAddress = $dbQuery->getArrayAddress();
                                if (empty($arrayAddress)) {
                                    if (!$this->datatype->decode($REF, $table, $column, $value, '', $dbQuery)) {
                                        return array();
                                    }
                                } else {
                                    if (!$this->datatype->decode($REF, $table, $column, $value, $arrayAddress, $dbQuery)) {
                                        return array();
                                    }

                                }
                                unset($arrayAddress);
                            break;

                            case YANA_DB_COLUMN:
                                $this->datatype->decode($REF[$i], $table, $column, $value, '', $dbQuery);
                            break;

                            default:
                                trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                                return false;
                            break;
                        } /* end switch */
                    } /* end if */
                } /* end foreach */
                unset($a, $value, $refKey, $currentTable);
            } /* end for */
            return $REF;
        } /* end if */
    }

    /**
     * update a row or cell
     *
     * update $value at position $key
     *
     * If $key does not exist, bool(false) is returned.
     *
     * This function returns bool(true) on success
     * and bool(false) on error.
     *
     * Note, that this function does not auto-commit.
     * This means, changes to the database will NOT be saved
     * until you call $DbStream->write().
     *
     * The argument $key may also be an object of type $dbQuery.
     * If so, no additional parameters need to be present.
     * This is a shortcut, which e.g. allows you to prepare
     * a query as an object and reuse it with multiple arguments.
     *
     * @access  public
     * @param   string|DbQuery  $key
     * @param   mixed           $value
     * @return  bool
     * @name    DbStream::update()
     * @see     DbStream::insertOrUpdate()
     * @see     DbStream::insert()
     * @since   2.9.5
     */
    function update($key, $value = array())
    {
        if (!$this->_isWriteable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        /**
         * 1) input is query object
         */
        if (is_object($key) && (is_a($key, 'DbQuery') || is_subclass_of($key, 'DbQuery'))) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            /* init arguments */
            $table = $dbQuery->getTable();
            $row = $dbQuery->getRow();
            $column = $dbQuery->getColumn();
            $value =& $dbQuery->getValues();
            if ($dbQuery->getType() !== YANA_DB_UPDATE) {
                trigger_error("Unable to update row. Invalid query.", E_USER_WARNING);
                return false;
            }

        /**
         * 2) input is key address
         */
        } else {

            /**
             * 2.1) check input
             */
            /* pessimistic scanning */
            /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

            /* optimistic type casting */
            /* argument 1 */ $key = (string) $key;

            if ($key == '') {
                return false;
            } else {
                $key = strtolower($key);
            }

            /**
             * 2.2) build query
             */
            $dbQuery = new DbQuery($this);

            /**
             * 2.2.1) error - key is not valid
             */
            if (!$dbQuery->setKey($key)) {
                trigger_error("Key argument '$key' in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;

            } elseif (!$dbQuery->setType(YANA_DB_UPDATE)) {
                trigger_error("Key argument '$key' in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;

            /**
             * 2.2.2) error - argument is invalid
             */
            } elseif (!$dbQuery->setValues($value)) {
                $this->handleError("SQL ERROR: Update operation on database failed for key '${key}'. Invalid values.", $value);
                return false;

            } else {
                assert('!isset($table); /* Cannot redeclare var $table */');
                assert('!isset($column); /* Cannot redeclare var $column */');
                $table = $dbQuery->getTable();
                $row = $dbQuery->getRow();
                $column = $dbQuery->getColumn();
            }

        } /* end if */

        /**
         * 2) check whether the row has been modified since last access
         */
        if ($this->structure->isStrict()) {
            if (isset($_SESSION['transaction_isolation_created'])) {
                if ($this->_lastModified($table, $row) > $_SESSION['transaction_isolation_created']) {
                    global $YANA;
                    if (isset($YANA)) {
                        $YANA->report( new Log("SQL ERROR: The user was trying to save form data, which has been changed " .
                        "since he accessed the corrsponding form. The operation has been aborted, as updating this row would " .
                        "have overwritten changes made by another user.") );
                        $YANA->report( new Alert("FORM_EXPIRED") );
                    } else {
                        $this->handleError("SQL ERROR: Input data has been changed since last access. " .
                        "Updating this row would overwrite existing changes. Update aborted for security reasons.");
                    }
                    return false;
                }
            }
        }

        /**
         * 3.2.3) check if element is inside an array
         *
         * If true, get the previous array and merge both.
         */
        if ($dbQuery->getExpectedResult() === YANA_DB_CELL) {
            assert('!isset($arrayAddress); // Cannot redeclare variable $arrayAddress');
            if ($arrayAddress = $dbQuery->getArrayAddress()) {
                assert('!isset($_value); // Cannot redeclare variable $_value');
                assert('!isset($_col); // Cannot redeclare var $_col');
                $_col = strtoupper($column);
                if (isset($this->cache[$table][$row][$_col]) && ($_value = Hashtable::get($this->cache[$table][$row][$_col], $arrayAddress))) {
                    /* intentionally left blank */

                } elseif ($_value = $this->get("$table.$row.$column")) {
                    /* intentionally left blank */

                } else {
                    $_value = array();

                }
                if (!is_array($_value)) {
                    $_value = array();
                }
                Hashtable::set($_value, $arrayAddress, $value);
                $value =& $_value;
                $dbQuery->setValues($value);
                unset($_value);
            }
            unset($arrayAddress);
        }

        /**
         * 3.4) error - updating table / column is illegal
         */
        assert('!isset($expectedResult); /* Cannot redeclare variable $expectedResult */');
        $expectedResult = $dbQuery->getExpectedResult();
        if ($expectedResult !== YANA_DB_ROW && $expectedResult !== YANA_DB_CELL) {
            $this->handleError("Updating a table or column is illegal. Operation aborted.");
            trigger_error('Illegal operation, cannot update a table or column.', E_USER_NOTICE);
            return false;
        }
        unset($expectedResult);

        /**
         * 3.4) before update: check constraints and triggers
         */
        if ($this->exec->constraint(YANA_DB_UPDATE, $table, $column, $value) === false) {
            trigger_error('Cannot set values. Constraint check failed.', E_USER_NOTICE);
            return false;

        } else {
            $this->exec->trigger(YANA_TRIGGER_BEFORE, YANA_DB_UPDATE, $table, $column, $value);
            $triggerArgs = array(YANA_DB_UPDATE, $table, $column);

        }

        /**
         * 4) check input
         */
        switch ($dbQuery->getExpectedResult())
        {
            case YANA_DB_ROW:
                /**
                 * 4.1.1) untaint input
                 */
                assert('!isset($a); /* Cannot redeclare var $a */');
                assert('!isset($b); /* Cannot redeclare var $b */');
                foreach ($value as $a => $b)
                {
                    if ($this->structure->getType($table, $a) !== 'array') {
                        if (!$this->datatype->encode($value[$a], $table, $a, @$row, $this)) {
                            $this->handleError("SQL Error: Update operation on '${table}.${row}.${a}' failed. The value '" . $value[$a] . "' is invalid.");
                            return false;
                        }
                    }
                } /* end foreach */

                /**
                 * 4.1.2) check foreign keys constraints
                 *
                 * Note: this provides a certain level of referential integrity
                 * even with DBMS who do not support this feature at all, like
                 * MySQL on MyISAM tables.
                 */
                assert('!isset($fkey); /* Cannot redeclare var $fkey */');
                assert('!isset($ufkey); /* Cannot redeclare var $fkey */');
                assert('!isset($ftable); /* Cannot redeclare var $ftable */');
                foreach ($this->structure->getForeignKeys($dbQuery->getTable()) as $fkey => $ftable)
                {
                    $fkey = strtolower($fkey);
                    $ufkey = strtoupper($fkey);
                    $ftable = strtolower($ftable);
                    /**
                     * If the referenced row does not exist,
                     * check if there is one recently inserted in cache
                     */
                    if (isset($value[$ufkey]) && !isset($this->cache[$ftable][strtolower($value[$ufkey])])) {
                        if (!$this->exists($ftable . "." . $value[$ufkey])) {
                            $this->handleError("SQL Error: The value '" . $value[$ufkey] . "' for attribute '${fkey}' refers to a non-existing " .
                            "entry in table '${ftable}'. Update operation aborted.", $value);
                            trigger_error("Update on '${key}' failed. Foreign key constraint mismatch.", E_USER_NOTICE);
                            return false;
                        } /* end if */
                    } /* end if */
                } /* end foreach */
                unset($ftable, $fkey, $ufkey);
            break;
            case YANA_DB_CELL:
                /**
                 * 4.2.1) untaint input
                 */
                if ($this->structure->getType($table, $column) !== 'array') {
                    if (!$this->datatype->encode($value, $table, $column, $row, $this)) {
                        $this->handleError("SQL Error: Update operation on '${table}.${row}.${column}' failed. The value '$value' is invalid.");
                        return false;
                    }
                }

                /**
                 * 4.2.2) check foreign keys constraints
                 */
                if ($this->structure->isForeignKey($table, $column)) {
                    assert('!isset($ftable); /* cannot redeclare variable $ftable */');
                    $ftable = $this->structure->getTableByForeignKey($table, $column);
                    /**
                     * If the referenced row does not exist,
                     * check if there is one recently inserted in cache
                     */
                    if (!isset($this->cache[$ftable][strtolower($value)])) {
                        if (!$this->exists("${ftable}.${value}")) {
                            $this->handleError("SQL Error: The value '${value}' for attribute '${column}' refers to a non-existing entry " .
                            "in table '${ftable}'. Update operation aborted.", $value);
                            trigger_error("Update on '${table}.${row}.${column}' failed. Foreign key constraint mismatch.", E_USER_NOTICE);
                            return false;
                        } /* end if */
                    } /* end if */
                    unset($ftable);
                } /* end if */
            break;
            default:
                /* error - may only update rows and cells */
                return false;
            break;
        }

        /**
         * 5) move values to cache
         */
        if ($column === '*') {
            $this->cache[$table][$row] = $value;
        } else {
            $this->cache[$table][$row][$column] = $value;
        }
        $this->lastSelectedTable = $table;
        unset($arrayAddress);

        /**
         * 6) add SQL statement to queue
         */
        if (is_null($triggerArgs)) {
            $this->queue[] = $dbQuery;
        } else {
            $this->queue[] = array($dbQuery, $triggerArgs);
        }

        return true;
    }

    /**
     * alias of DbStream::insertOrUpdate()
     *
     * @access  public
     * @param   string|DbQuery  $key
     * @param   mixed           $value
     * @return  bool
     * @name    DbStream::updateOrInsert()
     * @see     DbStream::insertOrUpdate()
     * @see     DbStream::insert()
     * @see     DbStream::update()
     * @since   2.9.5
     * @ignore
     */
    function updateOrInsert($key, $value = array())
    {
        return $this->insertOrUpdate($key, $value);
    }

    /**
     * update or insert row
     *
     * insert $value at position $key
     *
     * If $key already exists, the previous value
     * gets updated, else the value is created.
     * If you do not like this behaviour, take a look
     * at the functions {@link DbStream::update() update()}
     * and {@link DbStream::insert() insert()} instead,
     * which let you set the operation you want.
     *
     * Note that, as this function has to determine which
     * of both operations to take, it is somewhat slower
     * (approx. 5%) then calling the appropriate function
     * explicitly.
     *
     * This function returns bool(true) on success
     * and bool(false) on error. Note, that this
     * function does not auto-commit. This means,
     * changes to the database will NOT be saved
     * unless you call $DbStream->write().
     *
     * Since version 2.9.3 the argument $key may also be an object
     * of type $dbQuery. If so, no additional parameters need to be
     * present. This is a shortcut, which e.g. allows you to prepare
     * a query as an object and reuse it with multiple arguments.
     *
     * @access  public
     * @param   string|DbQuery  $key
     * @param   mixed           $value
     * @return  bool
     * @name    DbStream::insertOrUpdate()
     * @see     DbStream::insert()
     * @see     DbStream::update()
     * @since   2.9.5
     */
    function insertOrUpdate($key, $value = array())
    {
        $isInsert = null;

        /**
         * 1) input is query object
         */
        if (is_object($key)) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            if ($dbQuery->getType() === YANA_DB_INSERT) {
                $isInsert = true;

            } elseif ($dbQuery->getType() === YANA_DB_UPDATE) {
                $isInsert = false;

            } else {
                trigger_error("Unable to insert or update row. Invalid query.", E_USER_WARNING);
                return false;
            }

        /**
         * 2) input is key address
         */
        } else {

            /**
             * 2.1) check input
             */
            /* pessimistic scanning */
            /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

            /* optimistic type casting */
            /* argument 1 */ $key = (string) $key;

            if ($key == '') {
                return false;

            } else {
                $key = strtolower($key);
            }

            /**
             * 2.2) build query
             */
            $dbQuery = new DbQuery($this);
    
            /**
             * 2.2.1) error - key is not valid
             */
            if (!$dbQuery->setKey($key)) {
                trigger_error("Key argument in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
    
            } else {
                assert('!isset($table); /* Cannot redeclare var $table */');
                assert('!isset($column); /* Cannot redeclare var $column */');
                $table = $dbQuery->getTable();
                $row = $dbQuery->getRow();
                $column = $dbQuery->getColumn();
            }
    
            /**
             * 2.2.2) updating a row or cell
             */
            if ($row !== '*' && (isset($this->cache[$table][$row]) || $this->exists("${table}.${row}"))) {
                $dbQuery->setType(YANA_DB_UPDATE);
                $isInsert = false;

            /**
             * 2.2.3) inserting a row
             */
            } else {
                $dbQuery->setType(YANA_DB_INSERT);
                $isInsert = true;
            }

            /**
             * 2.2.4) set values
             */
            if (!$dbQuery->setValues($value)) {
                $this->handleError("SQL ERROR: Update operation on database failed for key '${key}'. Invalid values.", $value);
                return false;

            }
            
        } /* end if */

        /**
         * 3) call function
         */
        if ($isInsert === true) {
            return $this->insert($dbQuery);

        } elseif ($isInsert === false) {
            return $this->update($dbQuery);

        } else {
            return false;

        }
    }

    /**
     * insert row
     *
     * Insert $value at position $key.
     *
     * This function returns bool(true) on success
     * and bool(false) on error. If $key already exists,
     * the function will return bool(false).
     *
     * Note, that this function does not auto-commit.
     * This means, changes to the database will NOT be saved
     * unless you call $DbStream->write().
     *
     * Since version 2.9.3 the argument $key may also be an object
     * of type $dbQuery. If so, no additional parameters need to be
     * present. This is a shortcut, which e.g. allows you to prepare
     * a query as an object and reuse it with multiple arguments.
     *
     * Compatibility note:
     * As of version 2.9.5, the behaviour of this function has changed.
     * In version 2.9.4 and earlier, when a row already existed, it was
     * updated automatically and no error was thrown.
     * If you need to simulate this behaviour, to get an older plugin
     * working, please set the global constant YANA_DB_COMPAT_INSERT
     * to the value true (Default is false) in file "library.php".
     *
     * @access  public
     * @param   string|DbQuery  $key
     * @param   mixed           $value
     * @return  bool
     * @name    DbStream::insert()
     * @see     DbStream::insertOrUpdate()
     * @see     DbStream::update()
     * @since   2.9.5
     */
    function insert($key, $value = array())
    {
        /**
         * 0) trigger version 2.9.4 compatibility mode
         */
        if (YANA_DB_COMPAT_INSERT) {
            return $this->_insertCompat($key, $value);

        }

        if (!$this->_isWriteable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        /**
         * 1) input is query object
         */
        if (is_object($key) && (is_a($key, 'DbQuery') || is_subclass_of($key, 'DbQuery'))) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            /* init arguments */
            $table = $dbQuery->getTable();
            $row = $dbQuery->getRow();
            $column = $dbQuery->getColumn();
            $value =& $dbQuery->getValues();
            if ($dbQuery->getType() !== YANA_DB_INSERT) {
                trigger_error("Unable to insert row. Invalid query.", E_USER_WARNING);
                return false;
            }

        /**
         * 2) input is key address
         */
        } else {

            /**
             * 2.1) check input
             */
            /* pessimistic scanning */
            /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

            /* optimistic type casting */
            /* argument 1 */ $key = (string) $key;

            if ($key == '') {
                return false;
            } else {
                $key = strtolower($key);
            }

            /**
             * 2.2) build query
             */
            $dbQuery = new DbQuery($this);

            /**
             * 2.2.1) error - key is not valid
             */
            if (!$dbQuery->setKey($key)) {
                trigger_error("Key argument in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;

            /**
             * 2.2.2) error - unable to set type of operation
             */
            } elseif (!$dbQuery->setType(YANA_DB_INSERT)) {
                trigger_error("Key argument '$key' in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;

            /**
             * 2.2.3) error - argument value is invalid
             */
            } elseif (!$dbQuery->setValues($value)) {
                trigger_error("Value argument '$value' in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                $this->handleError("SQL ERROR: Insert on database failed for key '${key}'. Reason: values are invalid.", $value);
                return false;

            } else {
                assert('!isset($table); /* Cannot redeclare var $table */');
                assert('!isset($column); /* Cannot redeclare var $column */');
                $table = $dbQuery->getTable();
                $row = $dbQuery->getRow();
                $column = $dbQuery->getColumn();

            }

        } /* end if */

        /**
         * 3.1) error - inserting updating table / column is illegal
         */
        if ($dbQuery->getExpectedResult() !== YANA_DB_ROW) {
            $this->handleError("Can only use the operation 'insert' with a row - not a table, cell or column. Operation aborted.");
            trigger_error('Illegal operation, cannot insert a table, cell or column.', E_USER_NOTICE);
            return false;

        /**
         * 3.2) error - constraint check failed
         */
        } elseif (@$this->exec->constraint(YANA_DB_INSERT, $table, $column, $value) === false) {
            trigger_error('Cannot set values. Constraint check failed.', E_USER_NOTICE);
            return false;

        /**
         * 3.3) fire trigger
         */
        } else {
            @$this->exec->trigger(YANA_TRIGGER_BEFORE, YANA_DB_INSERT, $table, $column, $value);
            $triggerArgs = array(YANA_DB_INSERT, $table, $column);

        }

        /**
         * 4) check input
         */

        /**
         * 4.1) untaint input
         */
        assert('!isset($a); /* Cannot redeclare var $a */');
        assert('!isset($b); /* Cannot redeclare var $b */');
        foreach ($value as $a => $b)
        {
            if ($this->structure->getType($table, $a) !== 'array') {
                if (!$this->datatype->encode($value[$a], $table, $a, @$row, $this)) {
                    $this->handleError("SQL Error: Insert operation on '${table}.${row}.${a}' failed. The value '" . $value[$a] . "' is invalid.");
                    return false;
                }
            }
        } /* end foreach */

        /**
         * 4.2) check foreign keys constraints
         *
         * Note: this provides a certain level of referential integrity
         * even with DBMS who do not support this feature at all, like
         * MySQL on MyISAM tables.
         */
        assert('!isset($fkey); /* Cannot redeclare var $fkey */');
        assert('!isset($ufkey); /* Cannot redeclare var $fkey */');
        assert('!isset($ftable); /* Cannot redeclare var $ftable */');
        foreach ($this->structure->getForeignKeys($dbQuery->getTable()) as $fkey => $ftable)
        {
            $fkey = strtolower($fkey);
            $ufkey = strtoupper($fkey);
            $ftable = strtolower($ftable);
            /**
             * If the referenced row does not exist,
             * check if there is one recently inserted in cache
             */
            if (isset($value[$ufkey]) && !isset($this->cache[$ftable][strtolower($value[$ufkey])])) {
                if (!$this->exists("$ftable." . $value[$ufkey])) {
                    $this->handleError("SQL Error: The value '" . $value[$ufkey] . "' for attribute '${fkey}' refers to a non-existing " .
                    "entry in table '${ftable}'. Insert operation aborted.", $value);
                    trigger_error("Insert on '${key}' failed. Foreign key constraint mismatch.", E_USER_NOTICE);
                    return false;
                } /* end if */
            } /* end if */
        } /* end foreach */
        unset($ftable, $fkey, $ufkey);

        /**
         * 5) move values to cache
         */
        $this->cache[$table][$row] = $value;
        $this->lastSelectedTable = $table;
        unset($arrayAddress);

        /**
         * 6) add SQL statement to queue
         */
        if (is_null($triggerArgs)) {
            $this->queue[] = $dbQuery;
        } else {
            $this->queue[] = array($dbQuery, $triggerArgs);
        }

        return true;
    }

    /**
     * backwards compatible version of DbStream::insert()
     *
     * This function retains backwards compatibility with
     * versions 2.9.4 and earlier.
     *
     * It is triggered via the global constant YANA_DB_COMPAT_INSERT.
     *
     * @access  private
     * @param   string|DbQuery  $key
     * @param   mixed           $value
     * @return  bool
     * @name    DbStream::insert()
     * @see     DbStream::insertOrUpdate()
     * @see     DbStream::update()
     * @since   2.9.5
     * @ignore
     */
    function _insertCompat($key, $value = array())         
    {
        if (!$this->_isWriteable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        /**
         * 1) input is query object
         */
        if (is_object($key) && (is_a($key, 'DbQuery') || is_subclass_of($key, 'DbQuery'))) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            /* init arguments */
            $table = $dbQuery->getTable();
            $row = $dbQuery->getRow();
            $column = $dbQuery->getColumn();
            $value =& $dbQuery->getValues();
            if ($dbQuery->getType() === YANA_DB_INSERT) {
                $isInsert = true;
            } elseif ($dbQuery->getType() === YANA_DB_UPDATE) {
                $isInsert = false;
            } else {
                trigger_error("Unable to insert or update row. Invalid query.", E_USER_WARNING);
                return false;
            }

        /**
         * 2) input is key address
         */
        } else {

            /**
             * 2.1) check input
             */
            /* pessimistic scanning */
            /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

            /* optimistic type casting */
            /* argument 1 */ $key = (string) $key;

            if ($key == '') {
                return false;
            } else {
                $key = strtolower($key);
            }

            /**
             * 2.2) build query
             */
            $dbQuery = new DbQuery($this);
    
            /**
             * 2.2.1) error - key is not valid
             */
            if (!$dbQuery->setKey($key)) {
                trigger_error("Key argument in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
    
            } else {
                assert('!isset($table); /* Cannot redeclare var $table */');
                assert('!isset($column); /* Cannot redeclare var $column */');
                $table = $dbQuery->getTable();
                $row = $dbQuery->getRow();
                $column = $dbQuery->getColumn();
            }
    
            /**
             * 2.2.2) updating a row or cell
             */
            if ($row !== '*' && (isset($this->cache[$table][$row]) || $this->exists("${table}.${row}"))) {
                $dbQuery->setType(YANA_DB_UPDATE);
                $isInsert = false;

            /**
             * 2.2.2) inserting a row
             */
            } else {
                $dbQuery->setType(YANA_DB_INSERT);
                $isInsert = true;
            }

        } /* end if */

        /**
         * 2) check whether the row has been modified since last access
         */
        if ($this->structure->isStrict()) {
            if (isset($_SESSION['transaction_isolation_created'])) {
                if ($this->_lastModified($table, $row) > $_SESSION['transaction_isolation_created']) {
                    global $YANA;
                    if (isset($YANA)) {
                        $YANA->report( new Log("SQL ERROR: The user was trying to save form data, which has been changed " .
                        "since he accessed the corrsponding form. The operation has been aborted, as updating this row would " .
                        "have overwritten changes made by another user.") );
                        $YANA->report( new Alert("FORM_EXPIRED") );
                    } else {
                        $this->handleError("SQL ERROR: Input data has been changed since last access. " .
                        "Updating this row would overwrite existing changes. Insert/Update aborted for security reasons.");
                    }
                    return false;
                }
            }
        }

        /**
         * 3.2.3) check if element is inside an array
         *
         * If true, get the previous array and merge both.
         */
        if ($dbQuery->getExpectedResult() === YANA_DB_CELL) {
            assert('!isset($arrayAddress); // Cannot redeclare variable $arrayAddress');
            if ($arrayAddress = $dbQuery->getArrayAddress()) {
                assert('!isset($_value); // Cannot redeclare variable $_value');
                if ($_value = Hashtable::get($this->cache[$table][$row], "${column}.${arrayAddress}")) {
                    /* intentionally left blank */
                } elseif ($_value = $this->get("$table.$row.$column")) {
                    /* intentionally left blank */
                } else {
                    $_value = array();
                }
                if (!is_array($_value)) {
                    $_value = array();
                }
                Hashtable::set($_value, $arrayAddress, $value);
                $value =& $_value;
                unset($_value);
            }
            unset($arrayAddress);
        }
    
        /**
         * A) input is string
         */
        if (is_string($key)) {
            /**
             * 3.3) error - argument is invalid
             */
            if (!$dbQuery->setValues($value)) {
                trigger_error("Value argument '$value' in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                $this->handleError("SQL ERROR: Insert on database failed for key '${key}'. Reason: values are invalid.", $value);
                return false;
            }

            /* row may have been changed - need an update */
            $row = $dbQuery->getRow();
        }

        /**
         * 3.4) error - inserting updating table / column is illegal
         */
        assert('!isset($expectedResult); /* Cannot redeclare variable $expectedResult */');
        $expectedResult = $dbQuery->getExpectedResult();
        if ($expectedResult !== YANA_DB_ROW) {
            if ($isInsert) {
                $this->handleError("Can only use the operation 'insert' with a row - not a table, cell or column. Operation aborted.");
                trigger_error('Illegal operation, cannot insert a table, cell or column.', E_USER_NOTICE);
                return false;
            } elseif ($expectedResult !== YANA_DB_CELL) {
                $this->handleError("Updating a table or column is illegal. Operation aborted.");
                trigger_error('Illegal operation, cannot update a table or column.', E_USER_NOTICE);
                return false;
            }
        }
        unset($expectedResult);

        /**
         * 3.4) before insert: check constraints and triggers
         */
        if ($isInsert) {
            if (@$this->exec->constraint(YANA_DB_INSERT, $table, $column, $value) === false) {
                trigger_error('Cannot set values. Constraint check failed.', E_USER_NOTICE);
                return false;
            } else {
                @$this->exec->trigger(YANA_TRIGGER_BEFORE, YANA_DB_INSERT, $table, $column, $value);
            }
            $triggerArgs = array(YANA_DB_INSERT, $table, $column);

        /**
         * 3.5) before update: check constraints and triggers
         */
        } else {
            if ($this->exec->constraint(YANA_DB_UPDATE, $table, $column, $value) === false) {
                trigger_error('Cannot set values. Constraint check failed.', E_USER_NOTICE);
                return false;
            } else {
                $this->exec->trigger(YANA_TRIGGER_BEFORE, YANA_DB_UPDATE, $table, $column, $value);
            }
            $triggerArgs = array(YANA_DB_UPDATE, $table, $column);

        }

        /**
         * 4) check input
         */
        switch ($dbQuery->getExpectedResult())
        {
            case YANA_DB_ROW:
                /**
                 * 4.1.1) untaint input
                 */
                assert('!isset($a); /* Cannot redeclare var $a */');
                assert('!isset($b); /* Cannot redeclare var $b */');
                foreach ($value as $a => $b)
                {
                    if ($this->structure->getType($table, $a) !== 'array') {
                        if (!$this->datatype->encode($value[$a], $table, $a, @$row, $this)) {
                            $this->handleError("SQL Error: Insert/Update operation on '${table}.${row}.${a}' failed. The value '" . $value[$a] . "' is invalid.");
                            return false;
                        }
                    }
                } /* end foreach */

                /**
                 * 4.1.2) check foreign keys constraints
                 */
                assert('!isset($fkey); /* Cannot redeclare var $fkey */');
                assert('!isset($ufkey); /* Cannot redeclare var $fkey */');
                assert('!isset($ftable); /* Cannot redeclare var $ftable */');
                foreach ($this->structure->getForeignKeys($dbQuery->getTable()) as $fkey => $ftable)
                {
                    $fkey = strtolower($fkey);
                    $ufkey = strtoupper($fkey);
                    $ftable = strtolower($ftable);
                    /**
                     * If the referenced row does not exist,
                     * check if there is one recently inserted in cache
                     */
                    if (isset($value[$ufkey]) && !isset($this->cache[$ftable][strtolower($value[$ufkey])])) {
                        if (!$this->exists("$ftable." . $value[$ufkey])) {
                            $this->handleError("SQL Error: The value '" . $value[$ufkey] . "' for attribute '${fkey}' refers to a non-existing " .
                            "entry in table '${ftable}'. Insert/Update operation aborted.", $value);
                            trigger_error("Insert/update on '${key}' failed. Foreign key constraint mismatch.", E_USER_NOTICE);
                            return false;
                        } /* end if */
                    } /* end if */
                } /* end foreach */
                unset($ftable, $fkey, $ufkey);
            break;
            case YANA_DB_CELL:
                /**
                 * 4.2.1) untaint input
                 */
                if ($this->structure->getType($table, $column) !== 'array') {
                    if (!$this->datatype->encode($value, $table, $column, $row, $this)) {
                        $this->handleError("SQL Error: Insert/Update operation on '${table}.${row}.${column}' failed. The value '$value' is invalid.");
                        return false;
                    }
                }

                /**
                 * 4.2.2) check foreign keys constraints
                 */
                if ($this->structure->isForeignKey($table, $column)) {
                    assert('!isset($ftable); /* cannot redeclare variable $ftable */');
                    $ftable = $this->structure->getTableByForeignKey($table, $column);
                    /**
                     * If the referenced row does not exist,
                     * check if there is one recently inserted in cache
                     */
                    if (!isset($this->cache[$ftable][strtolower($value)])) {
                        if (!$this->exists("${ftable}.${value}")) {
                            $this->handleError("SQL Error: The value '${value}' for attribute '${column}' refers to a non-existing entry " .
                            "in table '${ftable}'. Insert/Update operation aborted.", $value);
                            trigger_error("Insert/update on '${table}.${row}.${column}' failed. Foreign key constraint mismatch.", E_USER_NOTICE);
                            return false;
                        } /* end if */
                    } /* end if */
                    unset($ftable);
                } /* end if */
            break;
            default:
                /* error - may only update rows and cells */
                return false;
            break;
        }

        /**
         * 5) move values to cache
         */
        if ($isInsert || $column === '*') {
            $this->cache[$table][$row] = $value;
        } else {
            $this->cache[$table][$row][$column] = $value;
        }
        $this->lastSelectedTable = $table;
        unset($arrayAddress);

        /**
         * 6) add SQL statement to queue
         */
        if (is_null($triggerArgs)) {
            $this->queue[] = $dbQuery;
        } else {
            $this->queue[] = array($dbQuery, $triggerArgs);
        }

        return true;
    }

    /**
     * remove one row
     *
     * For security reasons all delete queries will automatically be limited to 1 row at a time.
     * While this might be seen as a limitation the far more valueable advantage is,
     * no user is able to destroy a whole table - wether by intention or by accident -
     * in only one query. (At least not via this API.)
     *
     * The function returns bool(true) on success and bool(false) on error.
     *
     * Since version 2.8.5 the parameter $where has two synopsis.
     * <ul>
     *     <li>
     *         $where is a comma seperated list of column-value pairs using the following syntax:
     *         <pre>column1=value1,column2=value2,...,columnN=valueN</pre>
     *         Note that comparison takes place using the SQL operator 'LIKE', except for
     *         columns that are keys (either a primary key or a foreign key), which are compared
     *         using the SQL operator '='.
     *     </li>
     *     <li>
     *         $where is a two-dimensional numeric array and its entries follow this syntax:
     *         <pre>array(array(0=>column1,1=>value1,2=>operator1),...)</pre>
     *         The "column" name is mandatory and needs to be a string.
     *         The "value" is mandatory and needs to be a scalar value.
     *         The "operator" is optional and defaults to 'LIKE'.
     *         It needs to be a string and can be one of the following:
     *         <pre> '=', 'LIKE', '<', '>', '!=', '<=', '>=', '==' (alias of '=') </pre>
     *     </li>
     * </ul>
     * The second alternative offers more flexibility and should be prefered over the first one.
     *
     * Since version 2.9.3 the argument $key may also be an object
     * of type $dbQuery. If so, no additional parameters need to be
     * present. This is a shortcut, which e.g. allows you to prepare
     * a query as an object and reuse it with multiple arguments.
     *
     * @access  public
     * @param   string|DbQuery  $key    the address of the row that should be removed
     * @param   string|array    $where  a search string
     * @return  bool
     */
    function remove($key, $where = array())
    {
        if (!$this->_isWriteable()) {
            trigger_error('Operation aborted, not writeable.', E_USER_NOTICE);
            return false;
        }

        /**
         * 1) input is query object
         */
        if (is_object($key) && (is_a($key, 'DbQuery') || is_subclass_of($key, 'DbQuery'))) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            $table = $dbQuery->getTable();
            if ($dbQuery->getType() !== YANA_DB_DELETE) {
                trigger_error("Unable to remove row. Invalid query.", E_USER_WARNING);
                return false;
            }

        /**
         * 2) input is key address
         */
        } else {

            /**
             * 2.1) check input arguments
             */
            if (!is_string($key)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($key)), E_USER_WARNING);
                return false;
            }
            if (!is_string($where) && !is_array($where)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String or array', gettype($where)), E_USER_WARNING);
                return false;
            }

            /**
             * 2.2) Build query
             */
            if (!isset($dbQuery)) {
                $dbQuery = new DbQuery($this);
                $dbQuery->setType(YANA_DB_DELETE);
                if (!$dbQuery->setKey($key)) {
                    trigger_error("Key argument in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                    return false;
                } else {
                    $table = $dbQuery->getTable();
                }
                if (!$dbQuery->setWhere($where)) {
                    trigger_error("Where clause in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                    return false;
                }
            }

        } /* end if */

        /**
         * 3) Security checks
         *
         * 3.1) check constraint
         */
        if ($this->exec->constraint(YANA_DB_DELETE, $table) === false) {
            return false;

        /**
         * 3.2) fire trigger
         */
        } else {
            assert('!isset($value); // cannot redeclare variable $value');
            $value = null;
            $this->exec->trigger(YANA_TRIGGER_BEFORE, YANA_DB_DELETE, $table, null, $value);
            unset($value);
        }

        $this->lastSelectedTable = $table;

        /**
         * 4) add query to queue
         */
        $this->queue[] = $dbQuery;

        /**
         * 5) Logging
         *
         * backup entry before deleting it
         */
        global $YANA;
        if (isset($YANA)) {
            assert('!isset($result); /* cannot redeclare variable $result */');
            $result = $this->get($key, $where, '', 0, 1);
            if (!empty($result)) {
                if (is_string($result)) {
                    $YANA->report(new Log("Deleting entry from ${key}:", preg_replace('/\s/', ' ', $result)));
                } else {
                    $YANA->report(new Log("Deleting entry from ${key}:", $result));
                }
            }
        }

        /* return true to indicate the request was successfull */
        return true;
    }

    /**
     * join the resultsets for two tables
     *
     * Results in an INNER JOIN $table1, $table2 WHERE $table1.$key1 = $table2.$key2 .
     *
     * Note that if you ommit the parameters $key1 and $key2, the API will try to determine
     * the foreign key and target key itself by looking up the foreign key in the database's
     * structure file. The first foreign key association that matches will be used.
     *
     * Also note that two tables may only be joined via one pair of columns - not two or more.
     * Instead if you may add additional rules to the where clause as you see fit.
     *
     * Note that joins are permanent. So in opposition to what you might have learned from
     * common SQL statements and other APIs, you do not need to repeat joins for each query.
     * Instead, this API "remembers" what it was told and once set your joins will automatically
     * be used each time you query the table until you explicitly remove it.
     *
     * To remove all perviously set joins from a table, use the following function call:
     * <code> $dbStream->join('myTable'); </code>
     * As you can see above, if the second argument ($table2) is ommited, all joins bound
     * to 'myTable' are released.
     *
     * Also note, that the wildcard '*' may be used to refer to the "least recently used" table.
     * This is a shortcut that you may use in your scripts.
     *
     * @access  public
     * @param   string $table1  name of the table to join another one with
     * @param   string $table2  name of another table to join table1 with (when omitted will remove all previously set joins from table1)
     * @param   string $key1    name of the foreign key in table1 that references table2 (when omitted the API will look up the key in the structure file)
     * @param   string $key2    name of the key in table2 that is referenced from table1 (may be omitted if it is the primary key)
     * @return  bool
     */
    function join($table1, $table2 = null, $key1 = null, $key2 = null)
    {
        if (!$this->_isAvailable()) {
            return false;
        }

        /**
         * 1) check input
         */

        /* 1.1) argument 1 */
        if (!is_string($table1)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table1)), E_USER_WARNING);
            return false;
        } elseif ($table1 === '' || $table1 === "*") {
            if (empty($this->lastSelectedTable)) {
                trigger_error("Wrong parameter count for function " . __CLASS__ . "->" . __FUNCTION__ . "(). Unable to join tables, since no table has been selected for function join.", E_USER_WARNING);
                return false;
            } else {
                $table1 = $this->lastSelectedTable;
            }
        } else {
            /* lower-case input */
            $table1 = strtolower($table1);
        }

        /* 1.2) argument 2 */
        if (is_null($table2)) {
            /* if second argument is omitted, release association */
            if (isset($this->joins[$table1])) {
                $this->joins[$table1] = array();
            }
            return true;
        } elseif (!is_string($table2)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($table2)), E_USER_WARNING);
            return false;
        } else {
            /* lower-case input */
            $table2 = strtolower($table2);
        }

        /* 1.3) argument 3 */
        if (empty($key1)) {
            assert('!isset($foreignKey);   /* Cannot redeclare variable $foreignKey */');
            assert('!isset($foreignTable); /* Cannot redeclare variable $foreignTable */');
            foreach ($this->structure->getForeignKeys($table1) as $foreignKey => $foreignTable)
            {
                if (strcasecmp($table2, $foreignTable) === 0) {
                    $key1 = strtolower($foreignKey);
                    break;
                }
            } /* end foreach */
            unset($foreignKey, $foreignTable);
            if (empty($key1)) {
                $this->handleError("Cannot auto-join tables '${table1}' and '${table2}'. No foreign key found.");
                return false;
            }
        } elseif (!is_string($key1)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($key1)), E_USER_WARNING);
            return false;
        } elseif (!$this->structure->isColumn($table1, $key1)) {
            trigger_error("Argument 3 is invalid for function " . __CLASS__ . "->" . __FUNCTION__ . "(). Must be a valid column name.", E_USER_WARNING);
            return false;
        } else {
            /* lower-case input */
            $key1 = strtolower($key1);
        }

        /* 1.4) argument 4 */
        if (empty($key2)) {
            $key2 = $this->structure->getPrimaryKey($table2);
        } elseif (!is_string($key2)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'String', gettype($key2)), E_USER_WARNING);
            return false;
        } elseif (!is_null($key2)) {
            /* lower-case input */
            $key2 = strtolower($key2);
        }

        /**
         * 2) strict input checking
         */
        if ($this->structure->isStrict()) {
            if (!$this->structure->isTable($table1) || (!is_null($table2) && !$this->structure->isTable($table2))) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Table does not exist in structure file.");
                return false;
            }
            if (!is_null($key1) && !$this->structure->isColumn($table1, $key1)) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Field '${key1}' does not exist in table '${table1}'.");
                return false;
            }
            if (!is_null($key2) && !$this->structure->isColumn($table2, $key2)) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Field '${key2}' does not exist in table '${table2}'.");
                return false;
            }
        }

        /**
         * 3) reset old association
         */
        if (isset($this->joins[$table1][$table2])) {
            unset($this->joins[$table1][$table2]);
        }

        /**
         * 4) create new association
         */
        $this->joins[$table1][$table2] = array($key1, $key2);
        return true;
    }

    /**
     * optional API bypass
     *
     * Send a sql-statement directly to the PEAR database API, bypassing this API.
     *
     * Note: for security reasons this only sends one single SQL statement at a time.
     * This is done by checking the input for a semicolon followed by anything but
     * whitespace. If such input is found, an E_USER_WARNING is issued and the
     * function will return bool(false).
     *
     * While bypassing the API leaves nearly all of the input checking to you, this
     * is meant to prevent at least a minimum of the common SQL injection attacks.
     * A known attack is to try to terminate a current statement with ';' and afterwards
     * "inject" their own stuff as a second statement. The common attack vector usually
     * is unchecked form data.
     *
     * If you want to send a sequence of statements, call this function multiple times.
     *
     * The function will return bool(false) if the database connection or the
     * PEAR API is not available and otherwise will whatever PEAR sends back as the
     * result of your statement.
     *
     * Note: when database usage is disabled via the administrator's menu,
     * the PEAR-DB API can not be used and this function will return bool(false).
     *
     * The $offset and $limit arguments became available in version 2.8.8
     *
     * Since version 2.9.3 this function has a second synopsis:
     * You may provide a DbQuery object instead of the SQL statement.
     *
     * <code>
     * $dbStream->query($sqlStmt, $offset, $limit);
     * // 2nd synopsis
     * $dbStream->query($dbQuery);
     * </code>
     *
     * Note that when providing the DbQuery object, the $limit and $offset arguments are
     * ignored.
     *
     * @access  public
     * @param   string|DbQuery  $sqlStmt  one SQL statement (or a query object) to execute
     * @param   int             $offset   the row to start from
     * @param   int             $limit    the maximum numbers of rows in the resultset
     * @return  mixed
     */
    function query($sqlStmt, $offset = 0, $limit = 0)
    {
        /**
         * 1) check sql statement
         */
        if (is_object($sqlStmt)) {
            if (!is_a($sqlStmt, 'DbQuery') && !is_subclass_of($sqlStmt, 'DbQuery')) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbQuery', gettype($sqlStmt)), E_USER_WARNING);
                return false;
            } else {
                $offset = $sqlStmt->getOffset();
                $limit = $sqlStmt->getLimit();
                $sqlStmt = $sqlStmt->toString();
                /**
                 * Add this line for debugging purposes
                 *
                 * error_log($sqlStmt . " LIMIT $offset, $limit\n", 3, 'test.log');
                 */
            }
        } elseif (is_string($sqlStmt)) {
            if (is_int(strpos($sqlStmt, ';')) && preg_match("/;.*(?:select|insert|delete|update|create|alter|grant|revoke)\s.*$/is", $sqlStmt)) {
                trigger_error("A semicolon has been found in the current input '${sqlStmt}', indicating multiple queries.\n\t\t" .
                "As this might be the result of a hacking attempt it is prohibited for security reasons and the queries won't be executed.", E_USER_WARNING);
                return false;
            }
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($sqlStmt)), E_USER_WARNING);
            return false;
        }

        /**
         * 2) check other arguments
         */
        if (!is_int($offset) || $offset < 0) {
            trigger_error("Invalid argument 2. Must be a positive integer.", E_USER_WARNING);
            return false;
        } elseif (!is_int($limit) || $limit < 0) {
            trigger_error("Invalid argument 3. Must be a positive integer.", E_USER_WARNING);
            return false;

        /**
         * 3) check if database connection is available
         */
        } elseif (!$this->_isAvailable()) {
            return false;

        /**
         * 4) send query to database
         */
        } else {
            if ($offset > 0 || $limit > 0) {
                return $this->database->limitQuery($sqlStmt, $offset, $limit);
            } else {
                return $this->database->query($sqlStmt);
            }
        }
    }

    /**
     * get the most recently queried table
     *
     * @access  public
     * @return  string
     */
    function getTable()
    {
        assert('is_string($this->lastSelectedTable);');
        return $this->lastSelectedTable;
    }

    /**
     * get the number of entries inside a table
     *
     * Counts and returns the rows of $table.
     * If no table is provided, the least recently used
     * table will be selected instead.
     *
     * Since version 2.9.3 you may also provide the parameter
     * $table as an object of type DbQuery. This offers you
     * more flexibility.
     * In this case the second argument should not be provided.
     * Instead you may add a where clause to your query object.
     *
     * @access  public
     * @param   string|DbQuery  $table   name of a table
     * @param   string|array    $search  optional where clause
     * @return  int
     */
    function length($table = "", $search = array())
    {
        if (!$this->_isAvailable()) {
            return false;
        }

        /**
         * 1) input is query object
         */
        if (is_object($table) && (is_a($table, 'DbQuery') || is_subclass_of($table, 'DbQuery'))) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $table;
            /**
             * 1.1) check type
             */
            switch ($dbQuery->getType())
            {
                case YANA_DB_LENGTH:
                case YANA_DB_EXISTS:
                case YANA_DB_SELECT:
                    /* intentionally left blank */
                break;
                default:
                    trigger_error("Unable to count elements. Invalid query.", E_USER_WARNING);
                    return false;
                break;
            }

        /**
         * 2) input is table name
         */
        } else {

            /**
             * 2.1) check input
             */
            if (!is_string($table)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
                return false;
            }
            if (!is_string($search) && !is_array($search)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String or array', gettype($search)), E_USER_WARNING);
                return false;
            }

            /**
             * 2.2) check table parameter
             */
            if ($table == "" || $table == "*") {
                if (empty($this->lastSelectedTable)) {
                    return false;
                } else {
                    $table = $this->lastSelectedTable;
                }
            } elseif (!$this->structure->isTable($table)) {
                return 0;
            } else {
                $table = strtolower($table);
            }

            /**
             * 2.3) build query
             */
            $dbQuery = new DbQuery($this);
            $dbQuery->setType(YANA_DB_LENGTH);
            if (!$dbQuery->setTable($table)) {
                trigger_error("Table not found in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
            }
            if (!$dbQuery->setWhere($search)) {
                trigger_error("Where clause in query is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                return false;
            }            
        }

        /**
         * 3) get result
         */
        $result = $dbQuery->sendQuery();
        if ($this->isError($result)) {
            return false;
        } else {
            $i = $result->fetchRow(DB_FETCHMODE_ORDERED, 0);
            if (is_null($i) || !isset($i[0])) {
                return false;
            } else {
                return (int) $i[0];
            }
        }
    }

    /**
     * check wether a certain table has no entries
     *
     * Note: if no table is provided, the most recently used
     * table will be tested instead.
     *
     * @access  public
     * @param   string   $table  name of a table
     * @return  bool
     */
    function isEmpty($table = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $table = (string) $table;

        $table = strtolower($table);

        if ($table == "" || $table == "*") {
            if (!$this->lastSelectedTable) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }

        $length = (int) $this->length($table);
        return ($length == 0);
    }

    /**
     * Check wether a certain key exists
     *
     * Returns bool(true), if the adress $key (table, row, column) is defined, and otherwise bool(false).
     * If no argument is provided, the function returns bool(true) if a database connection exists and bool(false) if not.
     *
     * Since version 2.9.3 you may also provide the parameter
     * $key as an object of type DbQuery.
     *
     * @uses    $DbStream->exists('table.5')
     *
     * @access  public
     * @param   string|DbQuery  $key  adress to check
     * @return  bool
     */
    function exists($key = "")
    {
        /**
         * 1) error - database connection failure
         */
        if ($this->exists === YANA_DB_FAILURE) {
            return false;
        }

        /**
         * 2) input is query object
         */
        if (is_object($key) && (is_a($key, 'DbQuery') || is_subclass_of($key, 'DbQuery'))) {

            assert('func_num_args() === 1; // Too many arguments in function ' . __CLASS__ . '->' . __FUNCTION__ . '(). Only 1 argument expected.');
            $dbQuery =& $key;
            /**
             * 2.1) check type
             */
            switch ($dbQuery->getType())
            {
                case YANA_DB_LENGTH:
                case YANA_DB_EXISTS:
                case YANA_DB_SELECT:
                    /* intentionally left blank */
                break;
                default:
                    trigger_error("Unable to test for existence. Invalid query.", E_USER_WARNING);
                    return false;
                break;
            }


        /**
         * 3) input is key address
         */
        } else {
            /* pessimistic scanning */
            /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');
    
            /* optimistic type casting */
            /* argument 1 */ $key = (string) $key;
    
    
            /**
             * 3.1) check file
             */
            if ($key === "") {
                return $this->available();
    
            /**
             * 3.2) check table
             */
            } elseif (strpos($key, '.') === false) {
                return $this->structure->isTable($key);

            /**
             * 3.3) build query to check key
             */
            } else {
                $dbQuery = new DbQuery($this);
                $dbQuery->setType(YANA_DB_EXISTS);
                if (!$dbQuery->setKey($key)) {
                    trigger_error("Invalid query in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_WARNING);
                    return false;
                }
            }
        } /* end if */

        /**
         * 4) check if address exists
         */
        $result = $dbQuery->sendQuery();
        $this->lastSelectedTable = $dbQuery->getTable();

        if ($this->isError($result)) {
            return false;
        } elseif ($result->numRows() > 0) {
            return true;
        } else {
            return false;
        }

    }

    /**
     * check wether the current database is readonly
     *
     * This returns bool(false) if the database does not
     * exist, or the database property "readonly" is set
     * to bool(true) in the database's structure file.
     * Otherwise the function returns bool(true).
     *
     * @uses $DbStream->isWriteable()
     *
     * @access  public
     * @return  bool
     */
    function isWriteable()
    {
        if ($this->exists() && !$this->structure->isReadonly()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * import SQL from a file
     *
     * The input parameter $sqlFile can wether be filename,
     * or a numeric array of SQL statements.
     *
     * Returns bool(true) on success or bool(false) on error.
     * Note that the statements are executed within a transaction.
     * If the function fails,
     *
     * An error is encountered and an E_USER_NOTICE is issued, if:
     * <ul>
     * <li> the file does not exist or is not readable </li>
     * <li> the $sqlFile parameter is empty </li>
     * <li> the database connection is not available </li>
     * <li> the parameter "readonly" on the database structure file is set to "true" </li>
     * <li> at least one database statement failed (does not issue an E_USER_NOTICE) </li>
     * <li> there are uncommited statements in the queue </li>
     * </ul>
     *
     * @uses $DbStream->importSQL('some_file.sql')
     *
     * @access  public
     * @return  bool
     * @name    DbStream::importsql()
     * @param   string|array  $sqlFile
     */
    function importSQL($sqlFile)
    {
        global $YANA;
        if (!is_string($sqlFile) && !is_array($sqlFile)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String or array', gettype($sqlFile)), E_USER_WARNING);
            return false;
        }
        if (empty($sqlFile)) {
            trigger_error("Argument \$sqlFile is empty in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_NOTICE);
            return false;
        }
        if (!empty($this->queue)) {
            trigger_error("Cannot import SQL statements in " . __CLASS__ . "->" . __FUNCTION__ . "().\n\t\tThere is a pending transaction that needs to be commited before proceeding.", E_USER_NOTICE);
            return false;
        }

        if ($this->_isWriteable() === true) {
            if ($this->structure->isReadonly()) {
                trigger_error("Database is readonly. SQL import aborted.", E_USER_NOTICE);
                return false;
            } elseif (is_string($sqlFile) && !is_readable($sqlFile)) {
                trigger_error("The file '${sqlFile}' is not readable.", E_USER_NOTICE);
                return false;
            } elseif (is_array($sqlFile)) {
                $this->queue = $sqlFile;
                if ($this->write()) {
                    if (isset($YANA)) {
                        $YANA->report(new Log("SQL import was successful.", $sqlFile));
                    }
                    return true;
                } else {
                    if (isset($YANA)) {
                        $YANA->report(new Log("SQL import failed.", $sqlFile));
                    }    
                    return false;
                }
            } elseif (!empty($sqlFile)) {
                $raw_data = file_get_contents($sqlFile);
                /* remove comments and line breaks */
                $raw_data = preg_replace("/\s*\#[^\n]*/i", "", $raw_data);
                $raw_data = preg_replace("/\s*\-\-[^\n]*/i", "", $raw_data);
                $raw_data = preg_replace("/;\s*\n\s*/i", "[NEXT_COMMAND]", $raw_data);
                $raw_data = preg_replace("/\s/", " ", $raw_data);
                /* add items */
                $this->queue = explode("[NEXT_COMMAND]", $raw_data);
                if ($this->write()) {
                    if (isset($YANA)) {
                        $YANA->report(new Log("SQL import was successful.", $raw_data));
                    }
                    return true;
                } else {
                    if (isset($YANA)) {
                        $YANA->report(new Log("SQL import failed.", $raw_data));
                    }    
                    return false;
                }
            } else {
                return true;
            }
        } else {
            trigger_error("Database connection is not available. Check your connection settings.", E_USER_NOTICE);
            return false;
        }
    }

    /**
     * get CSV string from a table
     *
     * This exports the data of $table as a comma-seperated list of values.
     * If $table is not provided, the least recently used table is chosen instead.
     *
     * This function returns bool(false) on error.
     *
     * @uses $DbStream->toString('foo')
     *
     * @access  public
     * @param   string  $table  name of a table
     * @return  string
     */
    function toString($table = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $table = (string) $table;

        /* lower-case input */
        $table = strtolower($table);

        if ($this->exists === YANA_DB_FAILURE || $this->exists === YANA_DB_NOT_AVAILABLE) {
            trigger_error("No Database Connection: Database structure file not available or not loaded", E_USER_ERROR);
            return false;
        }

        if ($table == "" || $table == "*") {
            if (empty($this->lastSelectedTable)) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }

        assert('!isset($text); /* cannot redeclare variable $text */');
        $text = "";

        if (!$this->exists($table)) {
            trigger_error("The selected table '${table}' does not exist.", E_USER_WARNING);
            return "";
        } else {

            assert('!isset($a); /* cannot redeclare variable $a */');
            $a = $this->get($table);
            if(is_array($a)) {
                assert('!isset($k); /* cannot redeclare variable $k */');
                assert('!isset($e); /* cannot redeclare variable $e */');
                foreach ($a as $k => $e)
                {
                    assert('!isset($i); /* cannot redeclare variable $i */');
                    for ($i = 0; $i < count($e); $i++)
                    {
                        $b = current($e);
                        /* settype to STRING */
                        $b = (string) $b;
                        $text .= $b;
                        if (next($e)) {
                              $text .= ", ";
                        }
                    } /* end for */
                    unset($i);
                    $text .= "\n";
                } /* end foreach */
                unset($k,$e);
            } /* end if */
            unset($a);
            return $text;
        }
    }

    /**
     * export database structure to a file
     *
     * Note: if the specified file already exists it will get
     * overwritten.
     *
     * Structure files need to have the extension '.config' and are
     * to be stored at the directory 'config/db/*.config'.
     *
     * @uses $DbStream->exportStructure('some_file.config')
     *
     * @access  public
     * @param   string  $filename  relative path to output-file
     * @return  bool
     */
    function exportStructure($filename)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($filename); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ $filename = (string) $filename;

        if (!preg_match("/\.config$/", $filename)) {
            trigger_error("Unexpected file extension. The filename '${filename}' should end with '.config'.", E_USER_WARNING);
            return false;
        }
        $file = new SML($filename, CASE_UPPER);
        if (!$file->exists()) {
            $file->create();
        }
        $file->read();
        if (isset($this->structure) && $file->set($this->structure->getStructure()) && $file->write()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * get last reported error message
     *
     * Returns the last report error string or bool(false) if there
     * is none.
     *
     * @access  public
     * @return  string
     * @since   2.9.8
     */
    function getErrorMessage()
    {
        if (!empty($this->errorMessage)) {
            return (string) $this->errorMessage;
        } else {
            return false;
        }
    }

    /**
     * handle error
     *
     * @access  public
     * @param   mixed   $error
     * @param   mixed   $additionalInfo (optional)
     * @return  bool
     * @ignore
     */
    function handleError($error, $additionalInfo = "")
    {
        global $YANA;

        if ($this->isError($error)) {
            $this->errorMessage = $error->getMessage();
            if (isset($YANA)) {
                $YANA->report( new Log("$additionalInfo", $error->getMessage()) );
            }
            if (is_string($additionalInfo)) {
                trigger_error($error->getMessage()." $additionalInfo", E_USER_WARNING);
            } else {
                trigger_error($error->getMessage(), E_USER_WARNING);
            }
            return true;

        } elseif (is_string($error)) {
            if (isset($YANA)) {
                $YANA->report( new Log($error, $additionalInfo) );
            } else {
                if (is_string($additionalInfo)) {
                    trigger_error("$error $additionalInfo", E_USER_WARNING);
                } else {
                    trigger_error($error, E_USER_WARNING);
                }
            }
            return true;

        } else {
            return false;
        }
    }

    /**
     * build structure information
     *
     * {@internal
     *
     * WARNING: be aware that this won't work for all DBMS since
     * the type and ammount of information provided might vary.
     * ALSO: be aware that this won't work with PEAR-DB prior
     * to version 1.6.0
     *
     * ... or in other words: avoid this function in any productive
     * environment.
     *
     * It is only there as a "last-resort" in case the structure file
     * is not there as well as for debugging, testing and reverse
     * engineering your own database to a structure file when writing
     * new plugins (and by doing so possibly making your work a bit easier).
     *
     * }}
     *
     * @access  private
     */
    function buildStructure()
    {
        if (!isset($this->structure)) {
            $this->structure = new DbStructure($this->getPath());
            $this->structure->read();
        }
        $tables = $this->database->getListOf('tables');
        if (!$this->isError($tables)) {
            assert('!isset($table); /* cannot redeclare variable $table */');
            foreach ($tables as $table)
            {
                /* remove prefix */
                $table = substr($table, strlen(YANA_DATABASE_PREFIX));
                assert('!isset($tableInfo); // Cannot redeclare var $tableInfo');
                $tableInfo = $this->getTableInfo($table);
                $this->structure->addStructure($table, $tableInfo);
                unset($tableInfo);
            } /* end foreach */
            unset($table);            
        } else {
            $this->handleError("DB WARNING: unable to read structure of database.");
        }
    }

    /**
     * get table information
     *
     * This function returns detailed information on a table and it's columns.
     *
     * It returns an associative array containing the following keys:
     * <ul>
     *  <li> primarykey: Name of primary key column </li>
     *  <li> foreignkeys: Array, listing all foreign keys (MySQL only), with the following values:
     *    <ul>
     *      <li> column: Name of column in this table </li>
     *      <li> foreigntable: Name of referenced foreign table </li>
     *      <li> foreigncolumn: Name of referenced column in foreign table </li>
     *    </ul>
     *  </li>
     *  <li> length: Number of columns in current table </li>
     *  <li> The columns are arrays, numbered 0 through N, with each of them having the following values:
     *    <ul>
     *      <li> table: Name of current table </li>
     *      <li> name: Name of column </li>
     *      <li> type: Data type of column </li>
     *      <li> length: Maximum length</li>
     *      <li> nullable: True, when the column may contain NULL as value, false otherwise </li>
     *      <li> primarykey: True, when the column is the primary key, false otherwise </li>
     *      <li> foreignkey: True, when the column is a foreign key, false otherwise </li>
     *      <li> references: If column is a foreign key, this is a numeric array with the following values:
     *        <ul>
     *          <li> table: name of referenced foreign table </li>
     *          <li> column: name of referenced column in foreign table </li>
     *        </ul>
     *      </li>
     *      <li> unique: True, when the column must not contain duplicate values, false otherwise </li>
     *      <li> index: True, when there is an index on this column, false otherwise (MySQL only) </li>
     *      <li> auto: True, when the column uses auto increment or auto number, false otherwise </li>
     *      <li> default: Default value of this column, or NULL if there is none (MySQL only) </li>
     *      <li> unsigned: True, when the column is an unsigned number type, false otherwise </li>
     *      <li> zerofill: True, when the column has the zerofill flag set, false otherwise </li>
     *      <li> comment: The comment on this column, or NULL if there is none (MySQL only) </li>
     *      <li> select: True, if SELECT statements are allowed on this column and false otherwise (MySQL only) </li>
     *      <li> update: True, if UPDATE/DELETE statements are allowed on this column and false otherwise (MySQL only) </li>
     *      <li> insert: True, if INSERT statements are allowed on this column and false otherwise (MySQL only) </li>
     *    </ul>
     *  </li>
     * </ul>
     *
     * Example:
     * <code>
     * $db = new DbStream();
     * print_r($db->getTableInfo('foo'));
     * </code>
     *
     * The output may look something like:
     * <pre>
     * Array
     * (
     *     [length] => 6
     *     [primarykey] => foo_id
     *     [foreignkeys] => Array
     *         (
     *             [0] => Array
     *                 (
     *                     [column] => bar_id
     *                     [foreigntable] => bar
     *                     [foreigncolumn] => bar_id
     *                 )
     * 
     *         )
     *     [0] => Array
     *         (
     *             [table] => foo
     *             [name] => foo_id
     *             [type] => int
     *             [nullable] =>
     *             [primarykey] => 1
     *             [unique] => 
     *             [index] => 
     *             [auto] => 1
     *             [default] => 
     *             [unsigned] => 
     *             [zerofill] => 
     *             [comment] => id
     *             [length] => 8
     *             [foreignkey] => 
     *             [references] => 
     *             [update] => 1
     *             [insert] => 1
     *             [select] => 1
     *         )
     *
     *     [...]
     * 
     *     [5] => Array
     *         (
     *             [table] => foo
     *             [name] => bar_id
     *             [type] => int
     *             [nullable] =>
     *             [primarykey] => 
     *             [unique] => 
     *             [index] => 1
     *             [auto] => 
     *             [default] => 
     *             [unsigned] => 
     *             [zerofill] => 
     *             [comment] => some label
     *             [length] => 10
     *             [foreignkey] => 1
     *             [references] => Array
     *                 (
     *                     [0] => bar
     *                     [1] => bar_id
     *                 )
     *             [update] => 1
     *             [insert] => 1
     *             [select] => 1
     *         )
     * )
     * </pre>
     *
     * @access  public
     * @param   string  $table  name of table
     * @return  array
     * @since   2.9.6
     */
    function getTableInfo($table)
    {
        switch (YANA_DATABASE_DBMS)
        {
            /**
             * MySQL (custom solution)
             */
            case 'mysql':
            case 'mysqli':

                $tableInfo = new DbInfoTable($table);

                /**
                 * 1) prepare list of indexes
                 */
                assert('!isset($indexes); // Cannot redeclare var $indexes');
                $indexes = array();
                assert('!isset($result); // Cannot redeclare var $result');
                $result = $this->database->query('SHOW INDEX FROM ' . $this->quoteId(YANA_DATABASE_PREFIX . $table));
                if (!$this->isError($result)) {
                    assert('!isset($i); // Cannot redeclare var $i');
                    assert('!isset($col); // Cannot redeclare var $col');
                    for ($i = 0; $i < $result->numRows(); $i++)
                    {
                        $col = $result->fetchRow(DB_FETCHMODE_ASSOC, $i);
                        if (is_array($col)) {
                            $col = Hashtable::changeCase($col, CASE_LOWER);
                            
                            /**
                             * ignore auto-created indexes on primary keys
                             */
                            if ($col['key_name'] === 'PRIMARY') {
                                continue;
                            } else {
                                $indexes[] = strtolower($col['column_name']);
                            }

                        } /* end if */
                    } /* end for */
                    unset($i, $col);
                    $result->free();
                } else {
                    trigger_error("Unable to read list of indexes. Incompatible MySQL-version.", E_USER_ERROR);
                } /* end if */

                /**
                 * 2) prepare list of foreign keys
                 */
                assert('!isset($foreignKeys); // Cannot redeclare var $foreignKeys');
                $foreignKeys = array();
                $result = $this->database->query('SHOW CREATE TABLE ' . $this->quoteId(YANA_DATABASE_PREFIX . $table));
                if (!$this->isError($result)) {
                    assert('!isset($i); // Cannot redeclare var $i');
                    assert('!isset($tableDef); // Cannot redeclare var $tableDef');
                    for ($i = 0; $i < $result->numRows(); $i++)
                    {
                        $tableDef = $result->fetchRow(DB_FETCHMODE_ASSOC, $i);
                        $tableDef = Hashtable::changeCase($tableDef, CASE_LOWER);
                        $tableDef = $tableDef['create table'];
                    } /* end for */
                    unset($i);
                    assert('!isset($m); // Cannot redeclare var $m');
                    if (preg_match_all('/FOREIGN KEY \(`?(.*?)`?\) REFERENCES `?(.*?)`? \(`?(.*?)`?\)/is', $tableDef, $m)) {
                        assert('!isset($i); // Cannot redeclare var $i');
                        for ($i = 0; $i < count($m[0]); $i++)
                        {
                            $foreignKeys[strtolower($m[1][$i])] = array($m[2][$i], $m[3][$i]);
                        } /* end for */
                        unset($i); /* clean up garbage */
                    }
                    unset($m, $tableDef);
                    $result->free();
                } else {
                    trigger_error("Unable to read DDL-statements for table '{$table}'. Incompatible MySQL-version.", E_USER_ERROR);
                } /* end if */

                /**
                 * 3) get column definition
                 */
                $result = $this->database->query('SHOW FULL COLUMNS FROM ' . $this->quoteId(YANA_DATABASE_PREFIX . $table));

                /**
                 * 3.1) auto-fall back to describe on error (e.g. missing privileges)
                 */
                if ($this->isError($result)) {
                    $result = $this->database->query('DESCRIBE ' . $this->quoteId(YANA_DATABASE_PREFIX . $table));
                }
                if ($this->isError($result)) {
                    $this->handleError("DB WARNING: unable to read structure of table '${table}'.");
                    trigger_error("Unable to describe table '{$table}'. Incompatible MySQL-version.", E_USER_ERROR);

                } else {
                    assert('!isset($i); // Cannot redeclare var $i');
                    assert('!isset($col); // Cannot redeclare var $col');
                    for ($i = 0; $i < $result->numRows(); $i++)
                    {
                        $col = $result->fetchRow(DB_FETCHMODE_ASSOC, $i);
                        if (is_array($col)) {
                            $col = Hashtable::changeCase($col, CASE_LOWER);

                            $columnInfo = new DbInfoColumn($col['field']);
                            $columnInfo->setType(strtolower(trim(preg_replace('/\(.*$/', '', $col['type']))));
                            $columnInfo->setNullable($col['null'] === 'YES');
                            $columnInfo->setPrimaryKey($col['key'] === 'PRI');
                            $columnInfo->setUnique($col['key'] === 'UNI');
                            $columnInfo->setIndex(in_array(strtolower($col['field']), $indexes));
                            $columnInfo->setAuto(strpos((string) $col['extra'], 'auto_increment') !== false);
                            $columnInfo->setDefault($col['default']);
                            $columnInfo->setUnsigned(strpos($col['type'], 'unsigned') !== false);
                            $columnInfo->setZerofill(strpos($col['type'], 'zerofill') !== false);
                            @$columnInfo->setComment($col['comment']);

                            /**
                             * extract length (if any)
                             */
                            assert('!isset($m); // Cannot redeclare var $m');
                            if (preg_match('/\(([^\)]+)\)/', $col['type'], $m)) {
                                $columnInfo->setLength( (int) $m[1] );
                            }
                            unset($m);

                            /**
                             * foreign key
                             */
                            if (isset($foreignKeys[$columnInfo->getName()])) {
                                $columnInfo->setForeignKey(true);
                                $columnInfo->setReference($columnInfo->getName(), $foreignKeys[$columnInfo->getName()][0], $foreignKeys[$columnInfo->getName()][1]);
                            } else {
                                $columnInfo->setForeignKey(false);
                            }

                            /**
                             * check privileges and set flags
                             */
                            if (isset($col['privileges'])) {
                                $columnInfo->setUpdate(strpos($col['privileges'], 'update') !== false);
                                $columnInfo->setInsert(strpos($col['privileges'], 'insert') !== false);
                                $columnInfo->setSelect(strpos($col['privileges'], 'select') !== false);
                            }
                            
                            $tableInfo->addColumn($columnInfo);
                        } /* end if */
                    } /* end for */
                    unset($i, $col);
                    $result->free();
                } /* end if */
                unset($indexes, $foreignKeys, $result);
                return $tableInfo->toArray();
            break;

            /**
             * other (PEAR DB function)
             */
            default:

                $tableInfo = new DbInfoTable($table);

                $result = $this->database->query('SELECT * FROM ' . $this->quoteId(YANA_DATABASE_PREFIX . $table). ' WHERE 1=0');
                /**
                 * error - query failed
                 */
                if ($this->isError($result)) {
                    $this->handleError("DB WARNING: unable to read structure of table '${table}'.");

                /**
                 * get table info from result object
                 */
                } else {

                    $result = $this->database->tableInfo($result, DB_TABLEINFO_ORDER);
                    /**
                     * error - query failed
                     */
                    if (!is_array($result)) {
                        $this->handleError("DB WARNING: unable to read structure of table '${table}'.");

                    /**
                     * build answer
                     */
                    } else {
                        $result = Hashtable::changeCase($result, CASE_LOWER);

                        /**
                         * iterate through columns
                         */
                        assert('!isset($i); // Cannot redeclare var $i');
                        assert('!isset($col); // Cannot redeclare var $col');
                        for ($i = 0; $i < $result['num_fields']; $i++)
                        {
                            $col = $result[$i];

                            /**
                             * set column properties
                             */
                            assert('!isset($columnInfo); // Cannot redeclare var $columnInfo');
                            $columnInfo = new DbInfoColumn($col['name']);
                            $columnInfo->setType($col['type']);
                            $columnInfo->setLength( (int) $col['len']);
                            if (strpos($col['flags'], 'not_null') === false) {
                                $columnInfo->setNullable(true);
                            } else {
                                $columnInfo->setNullable(false);
                            }
                            if (strpos($col['flags'], 'primary_key') !== false) {
                                $columnInfo->setPrimaryKey(true);
                            } else {
                                $columnInfo->setPrimaryKey(false);
                            }
                            if (strpos($col['flags'], 'multiple_key') !== false) {
                                $columnInfo->setForeignkey(true);
                            } else {
                                $columnInfo->setForeignkey(false);
                            }
                            if (strpos($col['flags'], 'unique_key') !== false) {
                                $columnInfo->setUnique(true);
                            } else {
                                $columnInfo->setUnique(false);
                            }
                            if (strpos($col['flags'], 'auto_increment') !== false) {
                                $columnInfo->setAuto(true);
                            } else {
                                $columnInfo->setAuto(false);
                            }
                            if (strpos($col['flags'], 'unsigned') !== false) {
                                $columnInfo->setUnsigned(true);
                            } else {
                                $columnInfo->setUnsigned(false);
                            }
                            if (strpos($col['flags'], 'zerofill') !== false) {
                                $columnInfo->setZerofill(true);
                            } else {
                                $columnInfo->setZerofill(false);
                            }

                            /**
                             * add column to table
                             *
                             * Note: foreign keys and default values cannot be resolved,
                             * as the required information is not included in the table info
                             * as it is provided by PEAR DB.
                             */
                            $tableInfo->addColumn($columnInfo);
                            unset($columnInfo);

                        } /* end for */
                        unset($i, $col);
                    } /* end if */
                } /* end if */
                unset($result);
                return $tableInfo->toArray();
            break;
        } /* end switch */
        
        /* error - unable to retrieve result */
        trigger_error("Unable to retrieve result (may be a missing return value?).", E_USER_WARNING);
        return array();
    }

    /**
     * available
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function available()
    {
        if (empty($this->database)) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * get path and name of current structure file
     *
     * @access  protected
     * @return  string
     * @ignore
     */
    function getPath()
    {
        return $this->path;
    }

    /**
     * isError
     *
     * @access  public
     * @param   mixed   $result
     * @return  bool
     * @ignore
     */
    function isError($result)
    {
        return DB::isError($result);
    }

    /**
     * quote a value
     *
     * Returns the quoted values as a string
     * surrounded by delimiters, depending on
     * the DBMS selected.
     *
     * @access  public
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quote($value)
    {
        if (is_null($value)) {
            return 'NULL';
        } else {
            if (!is_scalar($value)) {
                trigger_error("Your SQL statement contains an unexpected non-scalar value: " .
                "(" . gettype($value) .") " . "'" . print_r($value, true) . "'", E_USER_NOTICE);
                $value = (string) $value;
            }
            return $this->database->quoteSmart($value);
        }
    }

    /**
     * smart id quoting
     *
     * Returns the quotes Id as a string
     * surrounded by delimiters, depending on
     * the DBMS selected.
     *
     * Implements a blacklist approach to automated quoting.
     * This will only quote such ids, which are a known
     * SQL keyword.
     *
     * @access  public
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quoteId($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($value); // Wrong argument type for argument 1. String expected, found \''.gettype($value).'\' instead.');

        /* optimistic type casting */
        /* argument 1 */ $value = (string) $value;

        /**
         * check DBMS
         *
         * In general, quoting is required, where the identifier is identical to a
         * reserved keyword of the database software.
         *
         * Under other circumstances it depends on the DMBS. Either you "may",
         * or you "should not", or you "must not" quote the id - depending on the
         * DBMS you use.
         *
         * So this will take the shortcut for DBMS where you don't need to care,
         * while taking the long path only for all the other DBMS, where this is required.
         */
        switch (YANA_DATABASE_DBMS)
        {
            /**
             * 1) always quote
             */
            case 'mysql':
            case 'mysqli':
            case 'postgresql':
            case 'mssql':
                return $this->database->quoteIdentifier($value);
            break;
            /**
             * 2) quote only where necessary
             *
             * Note that "isSqlKeyword()" has O(log(n)) running time.
             */
            default:
                if (strpos($value, ' ') !== false || $this->_isSqlKeyword($value) === true) {
                    return $this->database->quoteIdentifier($value);
        
                } else {
                    return $value;
                }
            break;
        } /* end switch */
    }

    /**
     * check if database is available
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function _isAvailable()
    {
        if ($this->available() === false) {
            trigger_error("Datebase is not available. In structure file '".$this->getPath()."'.", E_USER_WARNING);
            return false;
        } else {
            return true;
        }
    }

    /**
     * check if database is writeable
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function _isWriteable()
    {
        if ($this->_isAvailable() === false) {
            return false;
        } elseif (!$this->isWriteable()) {
            trigger_error("Datebase is not writeable. Check structure file '".$this->getPath()."' if attribute readonly is set to true.", E_USER_WARNING);
            return false;
        } else {
            return true;
        }
    }

    /**
     * returns true if $name is a known SQL keyword and false otherwise
     *
     * implements quick-search
     * + assume that the input is sorted
     * + assume that the input is upper case
     *
     * this algorithm has O(log(n)) running time
     *
     * @access  private
     * @param   string  $name
     * @return  bool
     * @ignore
     */
    function _isSqlKeyword($name)
    {
        if (!is_string($name)) {
            trigger_error("SQL syntax error. Invalid id '".gettype($name)."(".print_r($name,true).")'.", E_USER_WARNING);
            return true;
        } elseif (is_numeric($name)) {
            return true;
        }

        if (is_null($this->reservedSqlKeywords)) {
            global $YANA;
            /* Load list of reserved SQL keywords (required for smart id quoting) */
            if (isset($YANA)) {
                $file = $YANA->registry->get('reserved_sql_keywords');
                $this->reservedSqlKeywords = file($file->getPath());
            } else {
                $this->reservedSqlKeywords = array();
            }
            if (!is_array($this->reservedSqlKeywords)) {
                $this->reservedSqlKeywords = array();
            }
        } elseif (empty($this->reservedSqlKeywords)) {
            return false;
        }

        $name = strtoupper($name);
        if (qSearchArray($this->reservedSqlKeywords, $name) === false) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * return when the row was last modified
     *
     * Check when the data has been last modified by any other than the current user.
     * Use this to handle conditions, where two or more users request the same form.
     * One of them submits his changes and recieves an affirmation. Now the form of
     * the other user contains invalid data, because meanwhile the data he is currently
     * editing has been changed.
     * If then the other also submits the form, the previously made changes of the other
     * user would be overwritten. We call this a "dirty-write".
     *
     * So instead, this function is to be used to check for any such conditions and
     * present the second user an appropriate error message, telling him that his form
     * data is no longer valid and ensure, the "dirty" data is not written to the
     * database.
     *
     * @access  private
     * @param   string  $table
     * @param   string  $row
     * @return  int
     * @ignore
     */
    function _lastModified($table, $row)
    {
        assert('is_string($table); // Wrong argument type argument 1. String expected.');
        assert('is_string($row); // Wrong argument type argument 2. String expected.');

        /* settype to STRING */
        $table = (string) $table;
        $row = (string) $row;

        if ($row === '*') {
            $row = '';
        }

        global $YANA;

        /**
         * 1) load file
         */
        $file_exists = false;

        if (is_null($this->lastModified)) {
            if (isset($YANA)) {
                $this->lastModified = new SML($YANA->getVar('TEMPDIR').'db_last_modified.tmp', CASE_UPPER);
            } else {
                $this->lastModified = new SML('cache/db_last_modified.tmp', CASE_UPPER);
            }
            $this->lastModified->read();
            if (!$this->lastModified->exists()) {
                $this->lastModified->create();
                $file_exists = false;
            } else {
                $file_exists = true;
            }
        }

        /**
         * 2) get IP to identify user
         */
        if (isset($_SERVER['REMOTE_ADDR'])) {
            $userId = $_SERVER['REMOTE_ADDR'];
        } else {
            $userId = '*';
        }

        /**
         * 3) get time
         */
        if ($file_exists === true) {
            if ($row !== '') {
                $lastModified = $this->lastModified->get("$table.$row");
            } elseif ($table !== '') {
                $array = $this->lastModified->get($table);
                if (is_array($array)) {
                    sort($array);
                    $lastModified = array_pop($array);
                } else {
                    $lastModified = 0;
                }
            } else {
                return 0;
            }
        } else {
            $lastModified = 0;
        }

        /**
         * 4) check if changed by different user
         */
        if (is_array($lastModified)) {
            assert('count($lastModified) === 2; /* unexpected result, $lastModified. Should have 2 values: timestamp and user id */');
            if ($userId === array_pop($lastModified)) {
                $lastModified = 0;
            } else {
                $lastModified = array_pop($lastModified);
            }
        }

        /**
         * 5) update time
         */
        if (($file_exists === false || $lastModified != time()) && $row !== '') {
            $this->lastModified->insert(array($table => array($row => array(time(), $userId))));
        }

        /**
         * 6) return data
         */
        if ($lastModified > time()) {
            return 0;
        } elseif (!empty($lastModified)) {
            return (int) $lastModified;
        } else {
            return 0;
        }
    }


    /**
     * Reset the object to default values
     *
     * Resets the history for the last selected table,
     * resets the queue of pending SQL statements and
     * resets the database cache.
     *
     * @access  public
     * @name    DbStream::reset()
     */
    function reset()
    {
        $this->lastSelectedTable = "";
        $this->queue = array();
        $this->cache = array();
    }

    /**
     * Alias of DbStream::reset()
     *
     * @access  public
     * @see     DbStream::reset()
     */
    function rollback()
    {
        return $this->reset();
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same structure file and use equal database connections.
     *
     * @access   public
     * @param    object $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject, 'getClass') && $anotherObject->getClass() === $this->getClass()) {
            if ($this->getPath() === $anotherObject->getPath() && $this->database == $anotherObject->database) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool
     * @ignore
     */
    function test($nr = 1)
    {
        global $YANA;
        switch ($nr)
        {
            case 1:
                $db = new DbStream('check');
            case 2:
                if (!isset($db)) {
                    $db = new FileDb('check');
                }

                // drop all previous entries
                foreach ($db->get('i.*.iid') as $id)
                {
                    $test = @$db->remove("i.${id}");
                    if ($test !== true) {
                        return 'remove i failed';
                    }
                }

                foreach ($db->get('t.*.tid') as $id)
                {
                    $test = @$db->remove("t.${id}");
                    if ($test !== true) {
                        return 'remove t failed';
                    }
                }

                foreach ($db->get('ft.*.ftid') as $id)
                {
                    $test = @$db->remove("ft.${id}");
                    if ($test !== true) {
                        return 'remove ft failed';
                    }
                }

                if ($db->write() !== true) {
                    return 'write to database failed';
                }

                // init database
                $test = $db->insert('ft.1', array('ftvalue' => 1));
                if ($test !== true) {
                    return 'init ft.1 failed';
                }

                $test = @$db->insert('t.foo1', array('tvalue' => 1)); // supposed to fail
                if ($test !== false) {
                    return 'init t.foo1 failed';
                }

                $test = @$db->insertOrUpdate('t.foo2', array('tvalue' => 1, 'ftid' => 2)); // supposed to fail
                if ($test !== false) {
                    return 'init t.foo2 failed';
                }

                $test = $db->insert('t.foo', array('tvalue' => 1, 'ftid' => 1, 'tb' => true));
                if ($test !== true) {
                    return 'init t.foo failed';
                }

                $test = $db->insert('t.foo3', array('tvalue' => 3, 'ftid' => 1, 'tb' => false));
                if ($test !== true) {
                    return 'init t.foo3 failed';
                }

                $test = @$db->insertOrUpdate('i.foo2', array('ta' => array(1 => 1))); // supposed to fail
                if ($test !== false) {
                    return 'init i.foo2 failed';
                }

                $test = $db->insert('i.foo', array('ta' => array('1' => '1')));
                if ($test !== true) {
                    return 'init i.foo failed';
                }

                $test = $db->update('i.foo.ta.1.a', 2); // supposed to succeed
                if ($test !== true) {
                    return '"set array content" failed';
                }

                if ($db->write() !== true) {
                    return 'write to database failed';
                }

                $test = $db->insert('i.foo', array('ta' => array('1' => '1'))) && @$db->write(); // supposed to fail
                if ($test !== false) {
                    return 'duplicate key test (1) failed';
                }

                $test = $db->insert('i', array('iid' => 'foo', 'ta' => array('1' => '1'))) && @$db->write(); // supposed to fail
                if ($test !== false) {
                    return 'duplicate key test (2) failed';
                }

                // exists database
                $test = $db->exists();
                if ($test !== true) {
                    return '"exists database" test failed';
                }

                // exists table
                $test = $db->exists('t');
                if ($test !== true) {
                    return '"exists table" test failed';
                }

                // exists row
                $test = $db->exists('t.fOo');
                if ($test !== true) {
                    return '"exists row" test failed';
                }

                // exists cell
                $test = $db->exists('t.fOo.tid');
                if ($test !== true) {
                    return '"exists cell" test failed';
                }

                // exists column
                $test = $db->exists('i.*.ta');
                if ($test !== true) {
                    return '"exists column" test failed';
                }

                // exists (supposed to fail)
                $test = $db->exists('t.fooBar.tid');
                if ($test !== false) {
                    return '"exists failure" test failed';
                }

                // get table
                $test = $db->get('T');
                if (!is_array($test) || !isset($test['FOO']) || !is_array($test['FOO'])) {
                    return '"get table" test failed';
                }

                // get row
                $test = $db->get('t.fOo');
                if (!is_array($test) || !isset($test['TVALUE'])) {
                    return '"get row" test failed';
                }

                // get cell (1)
                $test = $db->get('T.Foo.TValue');
                if ($test !== 1) {
                    return '"get cell 1" test failed';
                }

                // get cell (2)
                $test = $db->get('t.fOo.tB');
                if ($test != true) {
                    return '"get cell 2" test failed';
                }

                // resolve foreign key
                $test = $db->get('t.fOo.ftid.ftvalue');
                if ($test !== 1) {
                    return '"resolving foreign key" test failed';
                }

                // get column
                $test = $db->get('T.*.fTid');
                $test2 = $db->length('t');
                if (!is_array($test) || count($test) !== $test2) {
                    return '"get column" test failed';
                }

                // get last entry
                $test = $db->get('t.?.tValue');
                if ($test !== 3) {
                    return '"get last entry" test failed';
                }

                // test foreign key constraint
                $test = @$db->insertOrUpdate('t.foo.ftid', 2); // supposed to fail
                if ($test !== false) {
                    return '"foreign key" test failed';
                }

                // test buffer
                $test = $db->update('ft.3', array('ftvalue' => 3));
                if ($test !== true) {
                    return '"buffer test" insert failed';
                }

                $test = $db->update('t.FOO3.ftid', 3); // supposed to succeed
                if ($test !== true) {
                    return '"buffer key" read of buffer content failed';
                }

                $test = $db->get('i.foo.ta.1.a'); // supposed to succeed
                if ($test != 2) {
                    return '"get array content" failed';
                }

                // length table
                $test = $db->length('T');
                if ($test !== 2) {
                    return '"get length" test failed';
                }

                $test = $db->join('t', 'ft');
                if ($test !== true) {
                    return '"join" test failed';
                }

                $test = @$db->get('t.fOo');
                if (!isset($test['FTVALUE']) || $test['FTVALUE'] != 1) {
                    return '"join t+ft" test failed';
                }
                // rollback
                $db->reset();

                $test = @$db->get('i.foo');
                $temp1 = array(1 => 2, 2 => 3);
                $temp2 = 2;
                $test['ta'] = $temp1 ;
                $test['tvalue'] = $temp2;
                if (!$db->update('i.foo', $test) || !$db->write()) {
                    return '"update inheritance 1" test failed';
                }
                if ($db->get('i.foo.ta') != $temp1 || $db->get('t.foo.tvalue') != $temp2) {
                    return '"update inheritance 2" test failed';
                }
                unset($temp1, $temp2);

                // rollback
                $db->reset();
                
                // multiple columns
                $dbQuery = new DbQuery($db);
                $dbQuery->setType(YANA_DB_SELECT);
                $dbQuery->setTable('i');
                $dbQuery->setRow('foo');
                $dbQuery->setJoin('t');
                $dbQuery->setColumns(array('i.iid', 'ta', 't.tvalue'));
                $test = $db->get($dbQuery);
                $test2 = array (
                  "IID" => "FOO",
                  "TA" => array (
                      1 => "2",
                      2 => "3"
                  ),
                  "TVALUE" => 2
                );
                if ($test != $test2) {
                    return '"get multiple columns" test failed';
                }

                $dbQuery->setType(YANA_DB_EXISTS);
                if (!$db->exists($dbQuery)) {
                    return '"exists multiple columns" test failed';
                }

                // test for property "unsigned"
                // column t.tf has unsigned constraint
                if (@$db->update('t.foo.tf', -1)) { // supposed to fail
                    return '"unsigned" test failed';
                }

                // test for property "zerofill"
                // column t.ti is a zerofilled integer with length 4
                if (!$db->update('t.foo.ti', 1) || !$db->write()) {
                    return '"set zerofill" test failed';
                }

                if ($db->get('t.foo.ti') !== '0001') {
                    return '"get zerofill" test failed';
                }

                // @todo  should anything else be tested?

                return true;
            break;
        }
    } /* end of function */
}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
