<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * Manipulate a directory
 *
 * This class represents a directory.
 * You may use this to get a list of contents, or remove,
 * or create a directory.
 *
 * @access      public
 * @package     yana
 * @subpackage  file_system
 */
class Dir extends FileSystemResource
{

    /**#@+
     * @ignore
     * @access  protected
     */
    /** @var array  */ var $content = array();
    /** @var array  */ var $options = array();
    /** @var string */ var $filter  = "";
    /**
     * @static
     * @var array   */ var $size    = null;
    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  string  $path
     */
    function __construct($path)
    {
        parent::__construct($path);
        /* PHP4 workaround for creating static vars */
        static $size;
        $this->size = &$size;
    }

    /**
     * read contents and put results in cache (filter settings will be applied)
     *
     * @uses    $Dir->read()
     * @access  public
     * @return  bool
     */
    function read()
    {
        if (!$this->exists()) {
            return false;
        } else {
            $this->content = dirlist($this->getPath(), $this->filter);
            sort($this->content);
            $this->options = $this->attributes;
            return true;
        }
    }


    /**
     * return list of files within the directory
     *
     * NOTE: will only return filenames with the path stripped
     *
     * @uses    $Dir->get()
     * @access  public
     * @return  array
     */
    function get()
    {
        if (!$this->isEmpty()) {
            return $this->content;
        } elseif ($this->read()) {
            return $this->content;
        } else {
            return array();
        }
    }

    /**
     * create this directory
     *
     * Tries to create the directory.
     * Check the developer's cookbook for an example
     * to this function.
     *
     * @uses    $Dir->create(660)
     *
     * @access  public
     * @param   int  $mode  access mode
     * @return  bool
     * @name    Dir::create()
     */
    function create($mode = 777)
    {
        if (!is_int($mode)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($mode)), E_USER_WARNING);
            return false;
        }

        if (!is_int($mode) || $this->exists() || !isset($this->path)) {
            return false;
        } elseif (!preg_match('/^[1-7]{1}[0-7]{2}$/',"$mode")) {
            trigger_error("Invalid mode (access level) '".$mode."' in ".__CLASS__."->".__FUNCTION__."().\n\t\tThis argument needs to be an integer of 3 or 4 digits between 0-7.\n\t\tNote: when providing 4 digits, the first digit needs to be 0.", E_USER_WARNING);
            return false;
        } else {
            $mode = decoct($mode);
            $path = $this->getPath();
            if (@mkdir($path)) {
                chmod($path, $mode);
                return true;
            } else {
                return false;
            }
        }
    }

    /**
     * remove this directory
     *
     * By option you may choose to also recursivly remove all files and subdirectories inside.
     * Otherwise the directory will only be removed if it is empty.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @uses    $Dir->delete(false)
     *
     * @access  public
     * @param   bool  $isRecursive  triggers wether to remove directories even if they are not empty, default = false
     * @return  bool
     */
    function delete($isRecursive = false)
    {
        if (!is_bool($isRecursive)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Bool', gettype($isRecursive)), E_USER_WARNING);
            return false;
        }

        if ($isRecursive === true) {
            $content = dirlist($this->getPath(), '');
            assert('!isset($element); /* cannot redeclare variable $element */');
            foreach ($content as $element)
            {
                $element = '/'.$element;
                if (is_file($this->getPath().$element)) {
                    $file = new File($this->getPath().$element);
                    if ($file->delete() === false) {
                        return false;
                    }
                } elseif (is_dir($this->getPath().$element)) {
                    $dir = new Dir($this->getPath().$element);
                    if ($dir->delete(true) === false) {
                        return false;
                    }
                } else {
                    /* intentionally left blank */
                }
            } /* end foreach */
            unset($element);
        }
        if (@rmdir($this->getPath()) === false) {
            trigger_error("Unable to delete directory '".$this->getPath()."'.", E_USER_ERROR);
            exit(1);
            return false;
        } else {
            return true;
        }
    }

    /**
     * return a string representation of this directory
     *
     * Returns a string with the contents of this directory.
     * Entries are seperated by line-breaks.
     *
     * @uses    $Dir->toString()
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        if (!$this->exists()) {
            return "Directory ".$this->getPath()." does not exist\n";
        } elseif ($this->isEmpty()) {
            return "Directory ".$this->getPath()." is not loaded\n";
        } else {
            return implode("\n", $this->get());
        }
    }

    /**
     * check wether the directory has no contents
     *
     * Returns bool(true) if there are no files that
     * match the current filter and bool(false) if there
     * is at least 1 file that matches.
     *
     * @uses    $Dir->isEmpty()
     *
     * @access  public
     * @return  bool
     */
    function isEmpty()
    {
        return empty($this->content);
    }

    /**
     * get the number of files inside the directory
     *
     * This returns a positive integer.
     * Note that this functions counts the files in respect
     * to the currently set file filter. So the number
     * of files reported here and the number in total
     * may vary.
     *
     * @uses    $Dir->length()
     *
     * @access  public
     * @return  int
     */
    function length()
    {
        return count($this->content);
    }

    /**
     * list contents of a directory
     *
     * The argument $filter may contain multiple file extension,
     * use a pipe '|' sign to seperate them.
     * Example: "*.xml|*.html" will find all xml- and html-files
     *
     * @uses $Dir->dirlist('*.cfg')
     *
     * @access  public
     * @param   string  $filter
     * @return  array
     */
    function dirlist($filter)
    {
        if (!is_string($filter)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filter)), E_USER_WARNING);
            return false;
        }

        $this->filter = (string) $filter;
        $this->read();
        return $this->content;
    }

    /**
    * getContent
    *
    * @uses    $Dir->getContent()
    *
    * @access  public
    * @return  array
    * @ignore
    */
    function getContent()
    {
         if ($this->isEmpty()) {
             $this->read();
         }
         $REF = array();
         for ($i = 0; $i < count($this->content); $i++)
         {
             $REF[$i]['VALUE'] = $this->content[$i];
             $REF[$i]['NAME']  = substr($this->content[$i], 0, strrpos($this->content[$i], "."));
         }
         return $REF;
    }

    /**
    * get size of directory
    *
    * Returns the size of $directory in bytes.
    *
    * This function gets the sum of the sizes of all files in
    * a directory.
    *
    * If $directory is not provided or NULL, the current directory is used.
    * If $copySubDirs is not provided or true, the result will
    * include all subdirectories.
    *
    * If $useCache is not provided or true, the result is cached for
    * the current directory.
    *
    * Returns bool(false) if the directory does not exist and issues an E_USER_WARNING.
    *
    * @uses    $Dir->getSize()
    * @uses    $Dir->getSize('foo/bar')
    *
    * @access  public
    * @param   string    $directory         directory name
    * @param   bool      $copySubDirs    on / off
    * @param   bool      $useCache          on / off
    * @return  int|bool(false)
    * @since   2.8.8
    */
    function getSize($directory = null, $copySubDirs = true, $useCache = true)
    {
        assert('is_string($directory); // Wrong argument type for argument 1. String expected.');
        assert('is_bool($copySubDirs); // Wrong argument type for argument 2. Boolean expected.');
        assert('is_bool($useCache);    // Wrong argument type for argument 3. Boolean expected.');

        /* use default value */
        if (is_null($directory)) {
            if ($this->exists()) {
                $dir = $this->getPath();
            } else {
                trigger_error("The directory '".$this->getPath()."' does not exist.", E_USER_WARNING);
                return false;
            }

        /* directory does not exist */
        } elseif (!is_string($directory) || !is_dir($directory)) {
            trigger_error("Argument 1 '".print_r($directory, true)."' is not a directory.", E_USER_WARNING);
            return false;

        /* add slash */
        } else {
            $dir = $directory.'/';
        }

        /* if value is already cached, return the cached size */
        if ($useCache === true && isset($this->size[$dir])) {
            return $this->size[$dir];
        }

        /* else determine the size */
        $d = dir($dir);
        $dirsize = 0;
        while (false !== ($filename = $d->read()))
        {
            if ($filename == '.' || $filename == '..') {
                continue;

            /* accumulate file sizes */
            } elseif (is_file($dir.$filename)) {
                $dirsize += filesize($dir.$filename);

            /* only recurse if subdirs are to be included */
            } elseif ($copySubDirs === true) {
                $dirsize += $this->getSize($dir.$filename, true, false);
            }
        }
        $d->close();

        /* cache result for later requests */
        if ($useCache === true) {
           $this->size[$dir] = $dirsize;
        }

        /* return result */
        return $dirsize;
    }

    /**
     * reset statistic data
     *
     * @uses $Dir->resetStats()
     *
     * @access  protected
     * @ignore
     */
    function resetStats()
    {
        if (!preg_match('/.*\/$/', $this->path)) {
            $this->path .= '/';
        }
        if ($this->exists()) {
            clearstatcache();
            $this->attributes['MTIME'] = filemtime($this->path);
            $this->size = null;
            clearstatcache();
        }
    }

    /**
     * check if directory exists and is readable
     *
     * Returns bool(true) on success and bool(false) otherwise.
     *
     * @access  public
     * @return  bool
     */
    function exists()
    {
        $name = $this->getPath();
        return is_dir($name) && file_exists($name) && is_readable($name);
    }

    /**
     * copy the directory to some destination
     *
     * This will create a copy of this directory and its contents on the filesystem.
     * Bool(true) will be returned on success and bool(false) on error.
     *
     * Since version 2.9.6 you can also use this function with some regular expression.
     * To do so, set $useRegExp to bool(true). It defaults to bool(false).
     * See the PHP manual at php.net/docs for an in-depth introduction on the use of regular expressions.
     *
     * If you don't want to use regular expressions, you may still use the following
     * wildcards for file- and directory patterns:
     * <ul>
     * <li> ? = match 1 symbol, example: f?o, matches 'foo', as well as 'fao' </li>
     * <li> * = match any symbols, example: *foo*, matches 'foo', or 'foobar', or 'barfoo', or 'barfoobar' </li>
     * <li> | = seperate 2 choices, example: foo|bar, matches either 'foo', or 'bar' (but NOT 'foobar') </li>
     * </ul>
     *
     * See the following examples:
     * <code>
     * $dir = new Dir('foo/');
     *
     * // copy directory foo/ to destination bar/
     * $dir->copy('bar/');
     *
     * // try to copy to bar2/ but don't overwrite if it already exists
     * $dir->copy('bar2/', false);
     *
     * // copy again and make files write- and executeable
     * $dir->copy('bar3/', true, 777);
     *
     * // copy to bar4/ and recurse sub-dirs
     * $dir->copy('bar4/', true, 766, true);
     *
     * // copy all *.xml and *.xhtml files to bar3/
     * $dir->copy('bar5/', true, 766, true, '*.xml|*.xhtml');
     *
     * // copy all sub-directories, whose names end with 'bar/'
     * $dir->copy('bar6/', true, 766, true, null, '*bar');
     *
     * // copy all but directory 'foobar/' (use regular expression)
     * $dir->copy('bar7/', true, 766, true, null, '/^(?!foobar$)/i', true);
     * </code>
     *
     * @access   public
     * @param    string   $destDir      destination to copy the file to
     * @param    bool     $overwrite    setting this to false will prevent existing files from getting overwritten
     * @param    int      $mode         the access restriction that applies to the copied file, defaults to 0766
     * @param    bool     $copySubDirs  setting this to true will cause sub-directories to be copied as well
     * @param    string   $fileFilter   use this to limit the copied files to a specific extension
     * @param    string   $dirFilter    use this to limit the copied directories to those matching the filter
     * @param    bool     $useRegExp    set this to bool(true) if you want filters to be treated as a regular expression
     * @return   bool
     *
     * @since    2.9.4
     */
    function copy($destDir, $overwrite = true, $mode = 766, $copySubDirs = false, $fileFilter = null, $dirFilter = null, $useRegExp = false)
    {
        if (!is_string($destDir)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($destDir)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($overwrite)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($overwrite)), E_USER_WARNING);
            return false;
        }
        if (!is_int($mode)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'Integer', gettype($mode)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($copySubDirs)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'Boolean', gettype($copySubDirs)), E_USER_WARNING);
            return false;
        }
        if (!is_string($fileFilter) && !is_null($fileFilter)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 5, 'String', gettype($fileFilter)), E_USER_WARNING);
            return false;
        }
        if (!is_string($dirFilter) && !is_null($dirFilter)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 6, 'String', gettype($dirFilter)), E_USER_WARNING);
            return false;
        }
        if (!is_bool($useRegExp)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 7, 'Boolean', gettype($useRegExp)), E_USER_WARNING);
            return false;
        }

        /* validity checking */
        if (empty($destDir) || strlen($destDir) > 512) {
            trigger_error("Invalid directory name '".$destDir."' in ".__CLASS__."->".__FUNCTION__."().", E_USER_WARNING);
            return false;
        }
        if (!preg_match('/^[1-7]{1}[0-7]{2}$/', "$mode")) {
            trigger_error("Invalid mode (access level) '".$mode."' in ".__CLASS__."->".__FUNCTION__."().\n\t\tThis argument needs to be an integer of 3 or 4 digits between 0-7.\n\t\tNote: when providing 4 digits, the first digit needs to be 0.", E_USER_WARNING);
            return false;
        }

        /* check if directory already exists */
        if ($overwrite === false && file_exists($destDir) === true) {
            trigger_error("Unable to copy directory '${destDir}'. Another directory with the same name does already exist.", E_USER_NOTICE);
            return false;
        } elseif ($overwrite === true && file_exists($destDir) === true && is_writeable($destDir) === false) {
            trigger_error("Unable to copydirectory '${destDir}'. The directory does already exist and is not writeable.", E_USER_WARNING);
            return false;
        }

        /* argument $fileFilter */
        if (!is_null($fileFilter) && !$useRegExp) {
            if (strpos($fileFilter, '|') !== false) {
                $fileFilter = preg_replace("/[^\.\-\_\w\d\|]/", "", $fileFilter);
                assert('!isset($tok); // cannot redeclare variable $tok');
                $tok = strtok($fileFilter, "|");
                $fileFilter = "";
                while ($tok !== false)
                {
                    $fileFilter .= toRegExp($tok);
                    $tok = strtok("|");
                    if ($tok !== false) {
                        $fileFilter .= "|";
                    }
                } /* end while */
                unset($tok);
            } else {
                $fileFilter = preg_replace("/[^\.\-\_\w\d]/", "", $fileFilter);
                $fileFilter = toRegExp($fileFilter);
            } /* end if */
            $fileFilter = '/' . $fileFilter . '$/i';
        } /* end if */

        /* argument $dirFilter */
        if (!is_null($dirFilter) && !$useRegExp) {
            $dirFilter = preg_replace('/[\.\\\+\[\^\]\$\(\)\{\}\=\!\<\>\:\/]/', '\\${0}', $dirFilter);
            $dirFilter = str_replace('*', ".*", $dirFilter);
            $dirFilter = str_replace('?', ".?", $dirFilter);
            $dirFilter = '/' . $dirFilter . '$/i';
        } /* end if */

        assert('is_string($destDir); // Unexpected result: $destDir. String expected.');

        /* recursively create directories */
        if (!empty($destDir) && !is_dir($destDir)) {
            assert('!isset($currentDir); // cannot redeclare variable $currentDir');
            $currentDir = '';
            assert('!isset($current); // cannot redeclare variable $dir');
            assert('!isset($dir); // cannot redeclare variable $dir');
            foreach (explode('/', $destDir) as $dir)
            {
                if (!is_dir($currentDir . $dir)) {
                    $current = new Dir($currentDir . $dir);
                    if ($current->create($mode) === false) {
                        trigger_error("Unable to create directory '".$current->getPath()."'.", E_USER_NOTICE);
                        return false;
                    }
                }
                $currentDir .= $dir.'/';
            } /* end foreach */
            unset($dir,$current);
        } /* end if */

        /* copy directory */
        assert('!isset($dir); // cannot redeclare variable $dir');
        assert('!isset($item); // cannot redeclare variable $item');
        assert('!isset($handle); // cannot redeclare variable $handle');
        $handle = opendir($this->path);
        while ($item = readdir($handle))
        {
            /**
             * recurse sub-directories
             */
            if (is_dir($this->path . $item)) {
                /* if sub-dirs are to be handled recursively ... */
                if ($copySubDirs && $item !== '.' && $item !== '..') {
                    /* if sub-dir matches the directory pattern ... */
                    if (is_null($dirFilter) || preg_match($dirFilter, $item)) {
                        $dir = new Dir($this->path . $item);
                        if ($dir->copy($destDir . $item . '/', $overwrite, $mode, $copySubDirs, $fileFilter, $dirFilter, true) === false) {
                            return false;
                        } elseif (chmod($destDir . $item, decoct($mode)) === false) {
                            trigger_error("Unable to set mode (access level) for directory '${destDir}${item}'.", E_USER_NOTICE);
                            return false;
                        }
                    } /* end if */
                } else {
                    continue;
                } /* end if */

            /**
             * handle files
             */
            } elseif (is_file($this->path . $item)) {
                if (is_null($fileFilter) || preg_match($fileFilter, $item)) {
                    if (copy($this->path . $item, $destDir . $item) === false) {
                        trigger_error("Unable to copy file '${destDir}${item}'.", E_USER_NOTICE);
                        return false;
                    } elseif (chmod($destDir . $item, decoct($mode)) === false) {
                        trigger_error("Unable to set mode (access level) for file '${destDir}${item}'.", E_USER_NOTICE);
                        return false;
                    } else {
                        continue;
                    }
                } else {
                    continue;
                } /* end if */

            /**
             * error
             */
            } else {
                continue;
            } /* end if */
        } /* end while */
        closedir($handle);
        unset($dir, $item, $handle);

        return true;
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
