<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * simulate a database
 *
 * this class simulates a sql-database on a flat-file
 *
 * Example:
 * <code>
 * $db = new FileDb('log');
 * print $db->get('log.*');
 * </code>
 *
 * You might also want to see the factory function
 * {@link Yana::connect()}.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 */
class FileDb extends DbStream
{
    /**
     * database connection information
     *
     * Note: only "database" name is used.
     * All other entries are static.
     *
     * @access  private
     * @var     array
     */
    var $dsn = array(
        'USE_ODBC' => false,
        'DBMS'     => 'filedb',
        'HOST'     => false,
        'PORT'     => false,
        'USERNAME' => false,
        'PASSWORD' => false,
        'DATABASE' => 'yana'
    );

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param  string  $filename  name of the structure-file
     */
    function __construct($filename = "")
    {
        /**
         * 1) Input checking
         */
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_ERROR);
        }

        /* settype to STRING */
        $filename = (string) $filename;
        $this->dsn['DATABASE'] = $filename;
        $this->database = new FileDbConnection($filename);
        $this->database->autoCommit(false);
        $this->path = $filename;

        /**
         * 2) Init database connection
         */
        if (empty($this->database)) {
            /* connection failed - abort */
            $this->exists = YANA_DB_FAILURE;
        } elseif (empty($filename)) {
            /* no structure file available, estimate structure */
            $this->buildStructure();
            $this->exists = YANA_DB_NO_FILE;
        } else {
            $this->structure = new DbStructure($this->getPath());
            $test = $this->structure->read();
            if ($test === false) {
                /* given file does not exist */
                unset($this->structure);
                $this->exists = YANA_DB_NOT_AVAILABLE;
                trigger_error("Unable to initialize database. Structure file '${filename}' not found.", E_USER_ERROR);
                exit(1);
            } else {
                /* all fine, proceed */
                $this->exists = YANA_DB_OK;
                $this->database->setStructure($this->structure);
            }
        }

        /* 3) Init IO/Filter
         */
        if (is_object($this->structure)) {
            $this->datatype = new DbDatatype($this->structure);
        }

        /* 4) Init trigger
         */
        if (is_object($this->structure)) {
            $this->exec = new DbTrigger($this->structure);
        }
    }

    /**
     * optional API bypass
     *
     * Send a sinle sql-statement directly to FileDB, bypassing the query-builder.
     *
     * The function will return bool(false) if the database connection is not
     * available and otherwise will return whatever FileDB sends back as the
     * result of your statement.
     *
     * Note: FileDB was designed to understand the statements generated
     * by the query-builder and is not intended to be called directly.
     * This means it will understand only SQL statements that comply with the syntax
     * that is provided by the query-builder.
     *
     * @access  public
     * @param   string|object  $sqlStmt  one SQL statement to execute
     * @param   int            $offset   the row to start from
     * @param   int            $limit    the maximum numbers of rows in the resultset
     * @return  mixed
     * @since   2.8.8
     * @ignore
     */
    function query($sqlStmt, $offset = 0, $limit = 0)
    {
        /**
         * 1) check sql statement
         */
        if (is_object($sqlStmt)) {
            if (!is_a($sqlStmt, 'DbQuery') && !is_subclass_of($sqlStmt, 'DbQuery')) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Instance of DbQuery', gettype($sqlStmt)), E_USER_WARNING);
                return false;
            } else {
                return $this->database->dbQuery($sqlStmt);
            }
        } elseif (is_string($sqlStmt)) {
            if (is_int(strpos($sqlStmt, ';')) && preg_match("/;.*(?:select|insert|delete|update|create|alter|grant|revoke).*$/is", $sqlStmt)) {
                trigger_error("A semicolon has been found in the current input '${sqlStmt}', indicating multiple queries.\n\t\t" .
                "As this might be the result of a hacking attempt it is prohibited for security reasons and the queries won't be executed.", E_USER_WARNING);
                return false;
            }
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($sqlStmt)), E_USER_WARNING);
            return false;
        }

        /**
         * 2) check other arguments
         */
        if (!is_int($offset) || $offset < 0) {
            trigger_error("Invalid argument 2. Must be a positive integer.", E_USER_WARNING);
            return false;
        } elseif (!is_int($limit) || $limit < 0) {
            trigger_error("Invalid argument 3. Must be a positive integer.", E_USER_WARNING);
            return false;
            
        /**
         * 3) check if database connection is available
         */
        } elseif (empty($this->database)) {
            return false;

        /**
         * 4) send query to database
         */
        } else {
            return $this->database->limitQuery($sqlStmt, $offset, $limit);
        }
    }

    /**
     * «not implemented» importSQL
     *
     * file database is created automatically on first use - nothing to do here!
     *
     * @access  public
     * @name    FileDb::importSQL()
     * @param   string  $sqlFile
     * @return  bool
     *
     * @ignore
     */
    function importSQL($sqlFile)
    {
        /* intentionally left blank */
        return true;
    }

    /**
     * build structure
     *
     * @access  private
     */
    function buildStructure()
    {
        foreach ($this->database->getListOf('databases') as $db)
        {
            if (empty($this->structure)) {
                $this->structure = new DbStructure($db);
                $this->structure->read();
                $this->database = new FileDbConnection($db);
                $this->database->autoCommit(false);
                $this->path = $db;
            } else {
                $this->structure->addFile($db);
            }
        }
    }

    /**
     * isError
     *
     * @access  private
     * @param   mixed  $result
     * @return  bool
     * @ignore
     */
    function isError($result)
    {
        if (method_exists($result, 'isError') && $result->isError()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * quote a value
     *
     * Returns the quoted values as a string
     * surrounded by double-quotes.
     *
     * @access  private
     * @param   mixed  $value
     * @return  string
     * @ignore
     */
    function quote($value)
    {
        return DbExtractor::quoteValue($value);
    }

    /**
     * quote an identifier
     *
     * Returns the quotes Id as a string
     * surrounded by double-quotes.
     *
     * @access  private
     * @param   string  $value
     * @return  string
     * @ignore
     */
    function quoteId($value)
    {
        assert('is_string($value); // Invalid argument 1 in ' . __CLASS__ . '->' . __FUNCTION__ . '(). String expected.');
        return "${value}";
    }

    /**
     * get table information
     *
     * This function returns detailed information on a table and it's columns as
     * an multidimensional, associative array.
     *
     * See {@link DbStream::getTableInfo()} for an in depth introduction to this
     * function.
     *
     * @access  public
     * @param   string  $table  name of table
     * @return  array
     * @since   2.9.6
     * @see     DbStream::getTableInfo()
     */
    function getTableInfo($table)
    {
        if (!$this->structure->isTable($table)) {
            return array();
        }

        /**
         * create table
         */
        $tableInfo = new DbInfoTable($table);

        assert('!isset($i); // Cannot redeclare var $i');
        assert('!isset($column); // Cannot redeclare var $column');
        foreach ($this->structure->getColumns($table) as $i => $column)
        {
            /**
             * column properties
             */
            $columnInfo = new DbInfoColumn();
            $columnInfo->name = $column;
            $columnInfo->type = $this->structure->getType($table, $column);
            $columnInfo->length = $this->structure->getLength($table, $column);
            $columnInfo->nullable = $this->structure->isNullable($table, $column);
            $columnInfo->unique = $this->structure->isUnique($table, $column);
            $columnInfo->index = $this->structure->hasIndex($table, $column);
            $columnInfo->auto = $this->structure->isAuto($table, $column);
            $columnInfo->default = $this->structure->getDefault($table, $column);
            $columnInfo->unsigned = null;
            $columnInfo->zerofill = null;
            $columnInfo->comment = $this->structure->getDescription($table, $column);
            $columnInfo->select = $this->structure->isVisible($table, $column);
            $columnInfo->update = !$this->structure->isReadonly($table, $column);
            $columnInfo->insert = !$this->structure->isReadonly($table);
            $tableInfo->addColumn($columnInfo);

        } /* end foreach */
        unset($i, $column); /* clean up garbage */

        /**
         * foreign keys
         */
        assert('!isset($col); // Cannot redeclare var $col');
        assert('!isset($fTable); // Cannot redeclare var $fTable');
        assert('!isset($fKey); // Cannot redeclare var $fKey');
        foreach ($this->structure->getForeignKeys($table) as $col => $fTable)
        {
            if ($this->structure->isTable($fTable)) {
                $fKey = $this->structure->getDefault($table, $col);
                if (is_array($fKey)) {
                    $fKey = array_shift(array_keys($fKey));
                } else {
                    $fKey = $this->structure->getPrimaryKey($fTable);
                }
                $tableInfo->setForeignKey($col, $fTable, $fKey);
            }
        } /* end foreach */
        unset($fKey, $col, $fTable); /* clean up garbage */
        
        $tableInfo->setPrimaryKey( $this->structure->getPrimaryKey($table) );
        @$tableInfo->setComment( $this->structure->getDescription($table) );

        $array = $tableInfo->toArray();
        if (is_array($array)) {
            return $array;

        } else {
            return array();

        }
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool
     * @ignore
     */
    function test($nr = 1)
    {
        global $YANA;
        switch ($nr)
        {
            case 1:
                return parent::test(2);
            break;
        }
    } /* end of function */
}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
