<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 *
 * @ignore
 */

/**#@+
 * @ignore
 */
if (!defined('DB_OK')) {
    define('DB_OK', 1);
}
if (!defined('DB_ERROR_UNSUPPORTED')) {
    define('DB_ERROR_UNSUPPORTED', -6);
}
/**#@-*/

/**
 * FileDbConnection
 *
 * Mapper for sql statements to datfile commands.
 * It implements only a required subset of the interface
 * of PEAR-DB as needed by the DbStream class
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 *
 * @ignore
 */
class FileDbConnection extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var Counter     */ var $autoIncrement = null;
    /** @var DbStructure */ var $structure     = null;
    /** @var string      */ var $database      = "";
    /** @var string      */ var $table         = "";
    /** @var array       */ var $src           = array();
    /** @var array       */ var $idx           = array();
    /** @var string      */ var $sort          = "";
    /** @var bool        */ var $desc          = false;
    /** @var bool        */ var $autoCommit    = true;
    /** @var array       */ var $cache         = array();
    /** @var string      */ var $baseDir       = 'config/db/';
    /** @var DbQuery     */ var $query         = null;

    /**#@-*/

    /**
     * constructor
     *
     * @param  string  $database
     */
    function __construct($database)
    {
        if (!is_string($database)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($database)), E_USER_ERROR);
            exit(1);
        }

        global $YANA;

        $database = strtolower(preg_replace('/^.*?([\w\d_]+)\.config$/', '${1}', $database));
        $database = preg_replace('/[\.\*]/', '', $database);

        if (isset($YANA)) {
            $this->baseDir = $YANA->getVar('DBDIR');
        }

        if (!empty($database)) {
            if (!is_dir($this->baseDir . $database)) {
                mkdir($this->baseDir . $database);
                chmod($this->baseDir . $database, 0700);
            }
            if (is_readable($this->baseDir . $database)) {
                $this->database = $database;
                $this->src[$database] = array();
                $this->idx[$database] = array();
                $this->cache = array();
            }
        }
    }

    /**
     * activate/deactive auto-commit
     *
     * @access  public
     * @param   bool  $mode  on / off
     * @return  int
     */
    function autoCommit($mode = false)
    {
        assert('is_bool($mode); // Wrong argument type argument 1. Boolean expected.');

        if ($mode) {
            $this->autoCommit = true;
        } else {
            $this->autoCommit = false;
        }
        return DB_OK;
    }

    /**
     * rollback current transaction
     *
     * @access  public
     * @return  int
     */
    function rollback()
    {
        $this->cache = array();
        $this->src[$this->database][$this->table]->reset();
        $this->idx[$this->database][$this->table]->rollback();
        return DB_OK;
    }

    /**
     * commit current transaction
     *
     * @access  public
     * @return  bool
     */
    function commit()
    {
        return $this->_write(true);
    }

    /**
     * get list of database objects
     *
     * Returns a list of all 'tables' inside the current database,
     * or a list of all available 'databases' as a numeric array.
     * Returns DB_ERROR_UNSUPPORTED otherwise.
     *
     * @access  public
     * @param   string  $type  valid values are 'tables' and 'databases'
     * @return  array
     */
    function getListOf($type)
    {
        assert('is_string($type); // Wrong argument type argument 1. String expected.');

        switch($type)
        {
            case 'tables':
                return $this->structure->getTables();
            break;
            case 'databases':
                $databases = dirlist($this->baseDir, '*.config');
                for ($i = 0; $i < count($databases); $i++)
                {
                    $databases[$i] = preg_replace('/\.config$/i', '', $databases[$i]);
                }
                return $databases;
            break;
            default:
                return DB_ERROR_UNSUPPORTED;
            break;
        }
    }

    /**
     * execute a single query
     *
     * @access  public
     * @param   DbQuery  &$dbQuery  query object
     * @return  FileDbResult
     * @since   2.9.3
     */
    function dbQuery(&$dbQuery)
    {
        global $YANA;
        /**
         * Add this line for debugging purposes
         *
         * error_log($dbQuery->toString() . "\n", 3, 'test.log');
         */
        $this->query =& $dbQuery;
        $offset = $this->query->getOffset();
        $limit  = $this->query->getLimit();

        switch ($this->query->getType())
        {
            /**
             * 1) SELECT statement
             */
            case YANA_DB_SELECT:
                $id = $this->query->toId();
                /**
                 * 1.1) result is cached
                 */
                if (isset($this->cache[$id])) {
                    /**
                     * 1.1.1) return result
                     */
                    return $this->cache[$id];

                /**
                 * 1.2) result is not cached
                 */
                } elseif (!$this->_select($this->query->getTable())) {
                    return new FileDbResult(null, "SQL ERROR: Table '" . $this->query->getTable() . "' is unknown.");
                } else {
                    /**
                     * 1.2.1) analyse query object
                     */
                    $this->sort = $this->query->getOrderBy();
                    $this->desc = $this->query->isDescending();
                    $columns = $this->query->getColumns();
                    $where = $this->query->getWhere();
                    $joins = $this->query->getJoin();
                    /**
                     * 1.2.2) no joined tables (simple request)
                     */
                    if (empty($joins)) {
                        /**
                         * 1.2.2.1) fetch result
                         */
                        $result = $this->_get($columns, $where, $offset, $limit);

                    /**
                     * 1.2.3) contains joined tables (complex request)
                     */
                    } else {
                        /**
                         * {@internal
                         * Note:
                         * Natural joins cannot occur, because DbQuery always adds
                         * column associations when two tables are joined.
                         * Joining tables without an valid association between them
                         * is prohibited and will issue an error.
                         *
                         * As a result we can skip scanning for natural joins here
                         * to improve performance. :-)
                         * }}
                         */

                        /**
                         * 1.2.3.1) resolve where clause
                         */
                        assert('!isset($tableB); // Cannot redeclare variable $tableB');
                        assert('!isset($clause); // Cannot redeclare variable $tableB');
                        assert('!isset($resultset); // Cannot redeclare variable $resultset');
                        assert('!isset($listOfResultSets); // Cannot redeclare variable $listOfResultSets');
                        $listOfResultSets = array();
                        foreach ($joins as $tableB => $clause)
                        {
                            $resultset = $this->_join($this->table, $tableB, $clause[0], $clause[1], $columns, $where);
                            if (!empty($resultset)) {
                                $listOfResultSets[] = $resultset;
                            }
                        } /* end foreach */
                        /**
                         * 1.2.3.2) merge resultsets
                         */
                        $result = array();
                        if (!empty($listOfResultSets)) {
                            assert('!isset($item); // Cannot redeclare variable $item');
                            foreach ($listOfResultSets as $item)
                            {
                                if (!empty($item)) {
                                    $result = Hashtable::merge($result, $item);
                                }
                            }
                            unset($item);
                        } /* end if */
                        unset($tableB, $clause, $listOfResultSets, $resultset);
                        /**
                         * 1.2.3.3) sorting and limiting
                         */
                        $this->_doSort($result);
                        $this->_doLimit($result, $offset, $limit);
                    } /* end if */
                    $result = new FileDbResult($result);

                    /**
                     * 1.2.4) move to cache
                     */
                    $this->cache[$id] = $result;

                    /**
                     * 1.2.5) return result
                     */
                    return $result;
                } /* end if */
            break;
            /**
             * 2) SELECT 1 statement
             */
            case YANA_DB_EXISTS:
                $id = $this->query->toId();
                /**
                 * 1.1) result is cached
                 */
                if (isset($this->cache[$id])) {
                    /**
                     * 1.1.1) return result
                     */
                    return $this->cache[$id];

                /**
                 * 1.2) result is not cached
                 */
                } elseif (!$this->_select($this->query->getTable())) {
                    return new FileDbResult(null, "SQL ERROR: Table '" . $this->query->getTable() . "' is unknown.");
                } else {
                    $where = $this->query->getWhere();
                    $columns = $this->query->getColumns();

                    /**
                     * 1.2.1) look up
                     */
                    $result = $this->_get($columns, $where);

                    if (count($result) > 0) {
                        $result = new FileDbResult(array(1));
                    } else {
                        $result = new FileDbResult(array());
                    }
                    /**
                     * 1.2.3) move to cache
                     */
                    $this->cache[$id] = $result;
                    /**
                     * 1.2.4) return result
                     */
                    return $result;
                }
            break;
            /**
             * 3) SELECT count(*) statement
             */
            case YANA_DB_LENGTH:
                $id = $this->query->toId();
                /**
                 * 1.1) result is cached
                 */
                if (isset($this->cache[$id])) {
                    /**
                     * 1.1.1) return result
                     */
                    return $this->cache[$id];

                /**
                 * 1.2) result is not cached
                 */
                } elseif (!$this->_select($this->query->getTable())) {
                    return new FileDbResult(null, "SQL ERROR: Table '" . $this->query->getTable() . "' is unknown.");
                } else {
                    $where = $this->query->getWhere();

                    /**
                     * 1.2.1) look up
                     */
                    $length = $this->_length($where);

                    /**
                     * 1.2.3) create result
                     */
                    if (is_int($length)) {
                        $result = new FileDbResult(array(array($length)));
                        /**
                         * 1.2.4) move to cache
                         */
                        $this->cache[$id] = $result;
                        /**
                         * 1.2.5) return result
                         */
                        return $result;
                        $this->cache[$id] = $result;
                    } else {
                        return new FileDbResult(null, 'SQL ERROR: Syntax error.');
                    }
                }
            break;
            /**
             * 4) INSERT INTO statement
             */
            case YANA_DB_INSERT:
                if (!$this->_select($this->query->getTable())) {
                    return new FileDbResult(null, "SQL ERROR: Table '" . $this->query->getTable() . "' is unknown.");
                }
                $set = $this->query->getValues();
                assert('!isset($primaryKey); /* Cannot redeclare var $primaryKey */');
                $primaryKey = $this->structure->getPrimaryKey($this->table);
                if (empty($set)) {
                    return new FileDbResult(null, 'SQL ERROR: The statement contains illegal values.');
                } elseif ($this->structure->isAutonumber($this->table, $primaryKey)) {
                    $this->_increment($set);
                }

                if (isset($set[$primaryKey])) {
                    $primaryValue = $set[$primaryKey];
                    unset($set[$primaryKey]);
                } else {
                    $primaryValue = $this->query->getRow();
                    if ($primaryValue === '*') {
                        return new FileDbResult(null, "SQL ERROR: Cannot insert entry. No primary key provided.");
                    }
                }

                /* get reference to SML file */
                assert('!isset($smlfile); /* Cannot redeclare var $smlfile */');
                $smlfile =& $this->_getSmlFile();
                /* get reference to Index file */
                assert('!isset($idxfile); /* Cannot redeclare var $idxfile */');
                $idxfile =& $this->_getIndexFile();

                /* create column */
                assert('!isset($column); /* Cannot redeclare variable $column */');
                foreach ($this->structure->getColumns($this->table) as $column)
                {
                    $colum = strtoupper($column);
                    if (isset($set[$column])) {
                        /* check unique constraint
                         *
                         * Unique constraint is breached if:
                         * 1) an entry with the same key already exists
                         */
                        if ($this->structure->isUnique($this->table, $column) === true) {
                            assert('!isset($tmp); /* Cannot redeclare variable $tmp */');
                            $tmp = $idxfile->get($column, $set[$column]);
                            /**
                             * error - unique constraint has already been breached by some previous operation
                             * This may occur, when a unique constraint is added, that has not existed before.
                             */
                            if (is_array($tmp) && count($tmp) > 1) {
                                if (isset($YANA)) {
                                    $YANA->report( new Log("SQL WARNING: The column ${column} has an unique constraint, but multiple rows with the same name have been found. This conflict can not be solved automatically. Please edit and update the affected table.", array('affected rows:' => $tmp)) );
                                }
                            }
                            /**
                             * error - constraint is breached
                             */
                            if (!empty($tmp)) {
                                return new FileDbResult(null, "SQL ERROR: Cannot insert entry with column ${column} = ".$set[$column].". The column has an unique constraint and another entry with the same value already exists.");
                            }
                            unset($tmp);
                        } /* end if */
                        /* update indexes
                         *
                         * Index needs to be updated if:
                         * 1) any column which explicitely has an index is updated OR
                         * 2) any column which implicitely requires an index is updated
                         *   (currently that is the case for columns using unique constraints)
                         */
                        if ($this->structure->hasIndex($this->table, $column) === true || $this->structure->isUnique($this->table, $column) === true) {
                            $idxfile->create($column, array($primaryValue, $set[$column]));
                        } /* end if */
                    } /* end if */
                } /* end foreach */
                unset($column);
                $set = array($primaryKey => array($primaryValue => $set));
                if (is_array($smlfile->get("${primaryKey}.${primaryValue}"))) {

                    return new FileDbResult(null, "SQL ERROR: Cannot insert entry with primary key = ${primaryValue}. Another entry with this key already exists.");
                } else {
                    if ($smlfile->insert($set)) {
                        return $this->_write();
                    } else {
                        return new FileDbResult(null, "unable to save changes");
                    } /* end if */
                } /* end if */
            break;
            /**
             * 5) UPDATE statement
             */
            case YANA_DB_UPDATE: 
                if (!$this->_select($this->query->getTable())) {
                    return new FileDbResult(null, "SQL ERROR: Table '" . $this->query->getTable() . "' is unknown.");
                }
                $set = $this->query->getValues();
                $where = $this->query->getWhere();
                $this->sort = $this->query->getOrderBy();
                $this->desc = $this->query->isDescending();

                $row = strtoupper($this->query->getRow());
                if ($row === '*') {
                    return new FileDbResult(null, "SQL ERROR: Cannot update entry. No primary key provided.");
                }

                /* update cell */
                if ($this->query->getExpectedResult() === YANA_DB_CELL) {
                    $column = strtoupper($this->query->getColumn());
                    if ($column !== '*') {
                        $set = array($column => $set);
                    } else {
                        return new FileDbResult(null, "SQL ERROR: Syntax error.");
                    }
                    unset($column);

                /* update row */
                } else {
                    $set = Hashtable::changeCase($set, CASE_UPPER);
                }

                assert('!isset($primaryKey); /* Cannot redeclare var $primaryKey */');
                $primaryKey = strtoupper($this->structure->getPrimaryKey($this->table));

                /* get reference to Index file */
                assert('!isset($idxfile); // Cannot redeclare var $idxfile');
                $idxfile =& $this->_getIndexFile();

                assert('!isset($column); /* cannot redeclare variable $column */');
                foreach ($this->structure->getColumns($this->table) as $column)
                {
                    $column = strtoupper($column);
                    if (isset($set[$column])) {
                        /* check unique constraint
                         *
                         * Unique constraint is breached if:
                         * 1) an entry with the same key already exists AND
                         * 2) the existing entry is not the same as the one which is currently being updated
                         */
                        if ($this->structure->isUnique($this->table, $column) === true) {
                            assert('!isset($tmp); /* Cannot redeclare variable $tmp */');
                            $tmp = $idxfile->get($column, $set[$column]);
                            /**
                             * error - unique constraint has already been breached by some previous operation
                             * This may occur, when a unique constraint is added, that has not existed before.
                             */
                            if (is_array($tmp) && count($tmp) > 1) {
                                if (isset($YANA)) {
                                    $YANA->report( new Log("SQL WARNING: The column ${column} has an unique constraint, but multiple rows with the same name have been found. This conflict can not be solved automatically. Please edit and update the affected table.", array('affected rows:' => $tmp)) );
                                }    
                            }
                            /**
                             * error - constraint is breached
                             */
                            if (!empty($tmp) && strcasecmp($tmp, $row) !== 0) {
                                return new FileDbResult(null, "SQL ERROR: Cannot update entry with column ${column} = ".$set[$column].". The column has an unique constraint and another entry with the same value already exists.");
                            }
                            unset($tmp);
                        }
                    } /* end if */
                } /* end foreach */
                unset($column);

                $smlfile =& $this->_getSmlFile();
                
                /* if primary key is renamed, the old one has to be replaced */
                if (isset($set[$primaryKey]) && strcasecmp($row, $set[$primaryKey]) !== 0) {
                    $smlfile->remove("$primaryKey.$row");
                    $row = strtoupper($set[$primaryKey]);
                    unset($set[$primaryKey]);
                    $set = array($row => $set);
                } elseif (!empty($row)) {
                    $set = array($row => $set);
                } else {
                    return new FileDbResult(null, "SQL ERROR: Syntax error.");
                } /* end if */

                /**
                 * remove old array values -
                 * but only if a new value is set
                 */
                foreach ($this->structure->getColumnsByType($this->table, 'array') as $column)
                {
                    $column = strtoupper($column);
                    if (isset($set[$row][$column])) {
                        $smlfile->remove("$primaryKey.$row.$column");
                    }
                } /* end foreach */

                if ($smlfile->insert(array($primaryKey => $set))) {
                    /* after data has been changed, reorganize all indexes */
                    $idxfile->create();

                    return $this->_write();
                } else {

                    return new FileDbResult(null, "unable to save changes");
                } /* end if */
            break;
            /**
             * 6) DELETE statement
             */
            case YANA_DB_DELETE:
                if (!$this->_select($this->query->getTable())) {
                    return new FileDbResult(null, "SQL ERROR: Table '" . $this->query->getTable() . "' is unknown.");
                }

                $where = $this->query->getWhere();
                $this->sort = $this->query->getOrderBy();
                $this->desc = $this->query->isDescending();

                $smlfile =& $this->_getSmlFile();
                $idxfile =& $this->_getIndexFile();

                $i = $this->_get(array($this->structure->getPrimaryKey($this->table)), $where, 0, 1);

                if (empty($i)) {
                    /* error */

                    return new FileDbResult(null, "SQL-ERROR: unable to delete; entry does not exist");
                } else {
                    $i = $i[0][strtoupper($this->structure->getPrimaryKey($this->table))];
                    if ($smlfile->remove($this->structure->getPrimaryKey($this->table).'.'.$i)) {
                        /* after data has been changed, reorganize all indexes */
                        $idxfile->create();

                        return $this->_write();
                    } else {

                        return new FileDbResult(null, "unable to save changes");
                    } /* end if */
                } /* end if */
            break;
            /**
             * 7) type is undefined
             */
            case YANA_DB_UNKNOWN:
                trigger_error("FileDB error: Unable to compute your query, because no type has been set for your request.\n\t\tSet the type by using the function DbQuery->setType().", E_USER_ERROR);
                return new FileDbResult(null, "FileDB error: Unable to compute the query because the type of request is undefined.");
            break;
            /**
             * 8) invalid or unknown statement
             */
            default:
                global $YANA;
                $sqlStmt = $this->query->toString();
                if (isset($YANA)) {
                    $YANA->setVar("__PHP_DBASE_ERROR__", "Ress#id: @74a34b2b616b34f8eb3f5b2fbce7187c");
                }
                trigger_error("Invalid or unknown SQL statement: ${sqlStmt}.", E_USER_ERROR);
                return new FileDbResult(null, "SQL ERROR: Syntax error.");
            break;
        }

    } /* end of function */

    /**
     * execute a single query
     *
     * @access  public
     * @param   string  $sqlStmt  one SQL statement to execute
     * @param   int     $offset   the row to start from
     * @param   int     $limit    the maximum number of rows in the resultset
     * @return  FileDbResult
     */
    function limitQuery($sqlStmt, $offset = 0, $limit = 0)
    {
        /* settype to STRING
         *            INTEGER
         *            INTEGER
         */
        $sqlStmt = (string) $sqlStmt;
        $offset  = (int)    $offset;
        $limit   = (int)    $limit;

        /**
         * 1) parse SQL
         */
        $dbQuery = new DbQueryParser($this->database);
        /**
         * 2) error - throw a line if invalid or (possibly) hazardous statement is encountered
         */
        if (!$dbQuery->parseSQL($sqlStmt)) {
            global $YANA;
            if (isset($YANA)) {
                $YANA->setVar("__PHP_DBASE_ERROR__", "Ress#id: @74a34b2b616b34f8eb3f5b2fbce7187c");
            }
            trigger_error("Invalid or unknown SQL statement: $sqlStmt.", E_USER_ERROR);

        /**
         * 3) route to query handling
         */
        } else {
            $dbQuery->setOffset($offset);
            $dbQuery->setLimit($limit);
            return $this->dbQuery($dbQuery, $offset, $limit);
        }
    }

    /**
     * execute a single query
     *
     * alias of limitQuery() with $offset and $limit params stripped
     *
     * @access  public
     * @param   string  $sqlStmt
     * @return  FileDbResult
     */
    function query($sqlStmt)
    {
        assert('is_string($sqlStmt);');
        /* settype to STRING */
        $sqlStmt = (string) $sqlStmt;
        return $this->limitQuery($sqlStmt);
    }

    /**
     * setStructure
     *
     * @access  public
     * @param   DbStructure  &$structure
     */
    function setStructure(&$structure)
    {
        /* this is a simple work-around to use PHP5-like type-checking in PHP4 */
        assert('is_a($structure, "DbStructure"); // Wrong argument type argument 1. Instance of DbStructure expected');
        $this->structure =& $structure;
    }

    /**
     * _select
     *
     * loads a table and returns table name
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  private
     * @param   string  $select
     * @return  bool
     * @ignore
     */
    function _select($select)
    {

        assert('is_string($select);');
        /* settype to STRING */
        $select = (string) $select;
        $select = strtolower(trim($select));

        if (empty($select)) {
            return false;
        } elseif (!isset($this->src[$this->database][$select])) {

            /* set current database */
            assert('!isset($database); // Cannot redeclare $database');
            $database = $this->structure->getAssociation($select);

            if (is_null($database)) {
                $database = $this->database;
            }
            assert('is_string($database) && !empty($this->database); // Unexpected result: $database');

            /* set current table */
            $this->table = $select;
            if ($this->structure->isStrict() && !$this->structure->isTable($this->table)) {
                trigger_error("There is no table '".$this->table."' in current structure file.", E_USER_WARNING);
                return null;
            }
            assert('is_string($this->table) && !empty($this->table);// Unexpected result: $this->table');

            /* open source file(s) */
            $this->_setSmlFile($database);
            $this->_setIndexFile($database);

            return true;

        } else {
            $this->table = $select;
            assert('is_string($this->table);');
            return true;
        } /* end if */
    }

    /**
     * _increment
     *
     * simulate MySQL's auto-increment feature
     *
     * @access  private
     * @param   array  &$set
     * @ignore
     */
    function _increment(&$set)
    {
        assert('is_array($set);  /* Wrong argument type for argument 1. Array expected. */');

        /**
         * 1) initialize counter
         */
        if (is_null($this->autoIncrement)) {
            $this->autoIncrement = new Counter('file_db_auto_increment', YANA_COUNTER_IGNORE_IP);
        }

        /**
         * 2) simulate auto-increment
         */
        $primary_key = $this->structure->getPrimaryKey($this->table);
        if (empty($set[$primary_key])) {
            $index = $this->autoIncrement->count($this->database . '__' . $this->table);
            $set[$primary_key] = $index;
        }
    }

    /**
     * _length
     *
     * simulate select count(*) from ... where ...
     *
     * @access  private
     * @param   array  $where
     * @return  int
     * @ignore
     */
    function _length($where)
    {
        assert('is_array($where);');
        /* settype to ARRAY*/
        $where = (array) $where;

        /* if table does not exist, then there is nothing to get */
        if (!isset($this->src[$this->database][$this->table])) {
            return 0;
        }

        if (empty($where)) {
            $smlfile =& $this->src[$this->database][$this->table];
            /* note: the first index is the primary key - NOT the table name
             */
            return $smlfile->length($this->structure->getPrimaryKey($this->table));
        } else {
            $result = $this->_get(array(), $where);
            return count($result);
        }
    }

    /**
     * _get
     *
     * @access  private
     * @param   array  $columns
     * @param   array  $where
     * @param   int    $offset
     * @param   int    $limit
     * @param   int    $type
     * @return  array
     * @ignore
     */
    function _get($columns = array(), $where = array(), $offset = 0, $limit = 0)
    {
        assert('is_array($columns); // Wrong type for argument 1. String expected');
        assert('is_array($where);   // Wrong type for argument 2. String expected');
        assert('is_int($offset);    // Wrong type for argument 3. Integer expected');
        assert('is_int($limit);     // Wrong type for argument 4. Integer expected');
        assert('$offset >= 0;       // Invalid argument 3. Must be a positive integer');
        assert('$limit >= 0;        // Invalid argument 4. Must be a positive integer');
        /* settype to STRING
         *            ARRAY
         *            INTEGER
         *            INTEGER
         */
        $columns = (array) $columns;
        $where   = (array) $where;
        $limit   = (int)   $limit;
        $offset  = (int)   $offset;

        /* if table does not exist, then there is nothing to get */
        if (!isset($this->src[$this->database][$this->table])) {
            return array();
        }

        /* initialize vars */
        $result     =  array();
        $smlfile    =& $this->_getSmlFile();
        $primaryKey =  strtoupper($this->structure->getPrimaryKey($this->table));
        $data       =  $smlfile->get($primaryKey);

        /* if there is data */
        if (is_array($data)) {

            /* 1) add primary key column */
            assert('!isset($i); // Cannot redeclare variable $i');
            /* implements a table-scan */
            foreach (array_keys($data) as $i)
            {
                $data[$i][$primaryKey] = $i;
            }
            unset($i);

            /* 2) pre-sort data */
            $this->_doSort($data);

            /* 3) apply where clause */
            /* encode where clause */
            $this->_where($where);
            $doCollapse = $this->query->getExpectedResult() === YANA_DB_TABLE && count($columns) > 1;
            assert('!isset($i);       // Cannot redeclare var $i');
            assert('!isset($current); // Cannot redeclare variable $current');
            foreach ($data as $i => $current)
            {
                /* implements the resultset */
                if ($this->_doWhere($current, $where) === true) {

                    if (empty($columns)) {
                        $result[] = $current;
                    } else {
                        if ($doCollapse) {
                            $lastResult = count($result);
                        }
                        assert('!isset($column); // Cannot redeclare var $column');
                        foreach ($columns as $column)
                        {
                            if (is_array($column)) {
                                $column = $column[1];
                            }
                            $column = strtoupper($column);
                            if (isset($current[$column])) {
                                if ($doCollapse) {
                                    if (!isset($result[$lastResult])) {
                                        $result[$lastResult] = array($column => $current[$column]);
                                    } else {
                                        $result[$lastResult][$column] = $current[$column];
                                    }
                                } else {
                                    $result[] = array($column => $current[$column]);
                                }
                            }
                        }
                        unset($column);
                    }
                }
            } /* end foreach */
            unset($i, $current);

            /* 4) limit results */
            $this->_doLimit($result, $offset, $limit);

            /* 5) return resultset */
            return $result;

        } else {
            /* result is empty */
            return array();
        }
    }

    /**
     * sort by column
     *
     * @access  private
     * @param   array  &$result
     * @ignore
     */
    function _doSort(&$result)
    {
        /* If no sort criteria is provided, the array will be returned order in which the entries
         * where inserted. If descending order is requested, the order will be reversed, so the
         * latest entries will come first and the oldest last.
         */
        if ($this->desc === true && empty($this->sort)) {
            $result = array_reverse($result);
        /* If a column has been provided to sort entries by, then the resultset will get sorted
         * by it.
         */
        } elseif (!empty($this->sort)) {
            uasort($result, array($this, '_sort'));
        }
    }

    /**
     * _sort
     *
     * @access  private
     * @param   array  $a
     * @param   array  $b
     * @return  int
     * @ignore
     */
    function _sort($a, $b, $columns = null)
    {
        if (is_null($columns)) {
            $columns = $this->sort;
        }
        if (!is_array($columns) || count($columns) === 0) {
            return 0;
        } else {
            $sort = array_shift($columns);
            $sort = strtoupper($sort[1]);

            if (!is_array($a) || !is_array($b)) {
                return 0;
            } elseif (!isset($a[$sort]) && !isset($b[$sort])) {
                $result = $this->_sort($a, $b, $columns);
            } elseif (!isset($a[$sort])) {
                $result = -1;
            } elseif (!isset($b[$sort])) {
                $result = 1;
            } elseif ($a[$sort] < $b[$sort]) {
                $result = -1;
            } elseif ($a[$sort] > $b[$sort]) {
                $result = 1;
            } else {
                if (count($columns) > 0) {
                    $result = $this->_sort($a, $b, $columns);
                } else {
                    $result = 0;
                }
            } /* end if */
            assert('is_int($result) && $result >= -1 && $result <= 1; // unexpected result');

            if ($this->desc) {
                return - $result;
            } else {
                return $result;
            } /* end if */
        } /* end if */
    }

    /**
     * _where
     *
     * @access  private
     * @param   array  $where
     * @ignore
     */
    function _where(&$where)
    {
        foreach (array_keys($where) as $i)
        {
            if (isset($where[$i][2]) && $where[$i][2] === 'like') {
                if (is_scalar($where[$i][1])) {
                    $where[$i][1] = toRegExp($where[$i][1]);
                    $where[$i][1] = str_replace('%', '.*', $where[$i][1]);
                    $where[$i][1] = str_replace('_', '.?', $where[$i][1]);
                }
            }
        }
    }

    /**
     * limits and offsets
     *
     * @access  private
     * @param   array  &$result
     * @param   int    $limit
     * @param   int    $offset
     * @ignore
     */
    function _doLimit(&$result, $offset, $limit)
    {
        if ($limit > 0 || $offset > 0) {
            if ($limit <= 0) {
                $limit = count($result);
            }
            if ($offset < 0) {
                $offset = 0;
            }
            $result = array_slice($result, $offset, $limit);
        }
    }

    /**
     * commit transaction
     *
     * @access  private
     * @param   bool  $commit on / off
     * @return  FileDbResult
     * @ignore
     */
    function _write($commit = false)
    {
        /* pessimistic scanning */
        assert('is_bool($commit); /* Wrong argument type for argument 1. Boolean expected. */');

        /* optimistic type casting */
        /* settype to BOOLEAN */
        $commit = (bool) $commit;

        $this->cache = array();
        if (!$commit && !$this->autoCommit) {
            /* wait for commit */
            return new FileDbResult(array());
        } else {
            /* auto commit */
            reset($this->src[$this->database]);
            while ($table = key($this->src[$this->database]))
            {
                if (!$this->src[$this->database][$table]->write()) {
                    /* error */
                    reset($this->src[$this->database]);
                    return new FileDbResult(null, 'unable to save changes');
                } else {
                    $this->idx[$this->database][$table]->commit();
                }
                next($this->src[$this->database]);
            } /* end for */
            unset($table);
            /* success */
            return new FileDbResult(array());
        }
    }

    /**
     * _join
     *
     * This implements joining two tables.
     *
     * @access  private
     * @param   string  $tableA
     * @param   string  $tableB
     * @param   string  $columnA
     * @param   string  $columnB
     * @param   array   $columns
     * @param   array   $where
     * @return  FileDbResult
     * @ignore
     */
    function _join($tableA, $tableB, $columnA, $columnB, $columns = null, $where = array())
    {
        /* pessimistic scanning */
        assert('is_string($tableA);  /* Wrong argument type for argument 1. String expected. */');
        assert('is_string($tableB);  /* Wrong argument type for argument 2. String expected. */');
        assert('is_string($columnA); /* Wrong argument type for argument 3. String expected. */');
        assert('is_string($columnB); /* Wrong argument type for argument 4. String expected. */');

        /* optimistic type casting */
        /* settype to STRING */
        $tableA  = (string) $tableA;
        $tableB  = (string) $tableB;
        $columnA = (string) $columnA;
        $columnB = (string) $columnB;

        /* prepare input */
        $tableA  = strtoupper($tableA);
        $columnA = strtoupper($columnA);
        $tableB  = strtoupper($tableB);
        $columnB = strtoupper($columnB);

        /* encode where clause */
        $this->_where($where);

        /* this information is needed optimize performance */
        $aHasIndex = $this->structure->hasIndex($tableA, $columnA);
        $bHasIndex = $this->structure->hasIndex($tableB, $columnB);
        $aIsFk     = $this->structure->isForeignKey($tableA, $columnA);
        $aIsPk     = $this->structure->isPrimaryKey($tableA, $columnA);
        $bIsPk     = $this->structure->isPrimaryKey($tableB, $columnB);
        $bIsUnique = $this->structure->IsUnique($tableB, $columnB);
        $typeOfA   = $this->structure->getType($tableA, $columnA);
        $typeOfB   = $this->structure->getType($tableB, $columnB);
        $pkA       = strtoupper($this->structure->getPrimaryKey($tableA));
        $pkB       = strtoupper($this->structure->getPrimaryKey($tableB));

        if (empty($columns)) {
            $columns = null;
        }

        /* notify me if results are not valid */
        assert('is_bool($aHasIndex); /* unexpected result $aHasIndex */');
        assert('is_bool($bHasIndex); /* unexpected result $bHasIndex */');
        assert('is_bool($aIsFk);     /* unexpected result $aIsFk     */');
        assert('is_bool($bIsPk);     /* unexpected result $bIsPk     */');
        assert('is_string($typeOfA); /* unexpected result $typeOfA   */');
        assert('is_string($typeOfB); /* unexpected result $typeOfB   */');
        assert('is_string($pkA);     /* unexpected result $pkA       */');
        assert('is_string($pkB);     /* unexpected result $pkB       */');

        /* strict key checking */
        if ($this->structure->isStrict()) {
            if (!$aIsFk) {
                trigger_error("Joining table '${tableA}' with '${tableB}' might fail.\n\t\tColumn '${columnA}' is not a foreign key in table '${tableA}'.", E_USER_NOTICE);
            }
            if (!$bIsPk && !$bHasIndex) {
                trigger_error("Joining table '${tableA}' with '${tableB}' might fail.\n\t\tColumn '${columnB}' is not a primary key in table '${tableB}'.", E_USER_NOTICE);
            }
            if (!$bIsPk && !$bIsUnique) {
                trigger_error("Joining table '${tableA}' with '${tableB}' might return ambigious results.\n\t\tColumn '${columnB}' is not unique in table '${tableB}'.", E_USER_NOTICE);
            }
            if (strcasecmp($typeOfA, 'select') !== 0 && $typeOfA != $typeOfB) {
                trigger_error("Note: data type of column '${tableA}.${columnA}' (${typeOfA}) does not match column '${tableB}.${columnB}' (${typeOfB}).\n\t\tThis might lead to unexpected results.", E_USER_NOTICE);
            }
        }

        /* the following implements the productive process */

        /* get a cursor on A */
        $this->_select($tableA);
        $SMLA    =& $this->_getSmlFile();
        $cursorA =  $SMLA->getByReference($pkA);
        if ($aHasIndex) {
            $indexA =& $this->idx[$this->database][$this->table];
        } else {
            $indexA =  null;
        }

        /* get a cursor on B */
        $this->_select($tableB);
        $SMLB    =& $this->_getSmlFile();
        $cursorB =  $SMLB->getByReference($pkB);
        if ($bHasIndex) {
            $indexB =& $this->idx[$this->database][$this->table];
        } else {
            $indexB =  null;
        }

        /* clean up */
        $this->_select($tableA);

        $resultSet = array();

        /**
         * iterate through tables
         *
         * {@internal
         *
         * Note: there are several cases for table A and B
         *
         * - case 1: $columnB is a primary key
         *           enables direct look-up of values in B
         *           with the best possible performance - this should be the default
         *           O(n) running time
         * - case 2: $columnB has an index
         *           enables indirect look-up of values in B via the index
         *           still good performance (currently we assume $columnB is unique)
         *           O(n) running time
         * - case 3: other
         *           will result in expensive recursive table-scans - this scenario should be avoided!
         *           worst case O(n*m) running time
         * }}
         */
        if (!is_null($cursorA)) {
            assert('!isset($row); /* cannot redeclare variable $row */');
            assert('!isset($id);  /* cannot redeclare variable $id */');
            foreach ($cursorA as $id => $row)
            {
                if ($aIsPk || (isset($row[$columnA]) && is_scalar($row[$columnA]))) {

                    /* get value of foreign key */
                    if ($aIsPk) {
                        $row[$columnA] = $id;
                        $keyA = strtoupper($id);
                    } else {
                        $keyA = strtoupper($row[$columnA]);
                    }

                    /* case 1: $columnB is a primary key */
                    if ($bIsPk === true) {
                        if (isset($cursorB[$keyA])) {
                            $joinedValueExists = true;
                            assert('!isset($_value); /* Cannot redeclare variable $_value */');
                            $_value = Hashtable::merge($row, $cursorB[$keyA]);
                            $_value[$columnB] = $keyA;
                            $value[] = $_value;
                            unset($_value);
                        } else {
                            $joinedValueExists = false;
                        }
                    /* case 2: $columnB has an index */
                    } elseif ($bHasIndex === true && !is_null($indexB)) {
                        $keyA = $indexB->get($columnB, $keyA);
                        if (is_scalar($keyA) && isset($cursorB[$keyA])) {
                            $joinedValueExists = true;
                            $value             = array($cursorB[$keyA]);
                        } elseif (is_array($keyA) && !empty($keyA)) {
                            $joinedValueExists = true;
                            $value = array();
                            foreach ($keyA as $currentKey)
                            {
                                $value[]       =& $cursorB[$currentKey];
                            }
                        } else {
                            $joinedValueExists = false;
                            $value             = null;
                        }
                    /* other */
                    } else {
                        $value             = array();
                        $joinedValueExists = false;
                        foreach ($cursorB as $idB => $rowB)
                        {
                            if (isset($rowB[$columnB]) && strcmp($rowB[$columnB], $row[$columnA]) === 0) {
                                $joinedValueExists = true;
                                $row[$pkB]         = $idB;
                                $value[]           = $rowB;
                            }
                        }
                        if (empty($value)) {
                            $value         = null;
                        }
                    }

                    if ($joinedValueExists === true) {

                        assert('!isset($newResult);');
                        assert('!isset($currentValue);');
                        foreach ($value as $currentValue)
                        {
                            /* add primary key to result - otherwise it would be missing */
                            $currentValue[$pkA] = $id;
                            if ($this->_doWhere($currentValue, $where) === true) {
                                if (!is_null($columns)) {
                                    assert('!isset($currentRow); // Cannot redeclare var $currentRow');
                                    $currentRow = array();

                                    assert('!isset($column); // Cannot redeclare var $column');
                                    foreach ($columns as $column)
                                    {
                                        if (is_array($column)) {
                                            $column = $column[1];
                                        }
                                        $column = strtoupper($column);
                                        if (isset($currentValue[$column])) {
                                            $currentRow[$column] = $currentValue[$column];
                                        } elseif (isset($row[$column])) {
                                            $currentRow[$column] = $row[$column];
                                        }
                                    }
                                    unset($column);
                                    $resultSet[] = $currentRow;

                                    unset($currentRow);

                                } else {
                                    $resultSet[] = Hashtable::merge($row, $currentValue);
                                }
                            }
                        } /* end foreach value */
                        unset($newResult, $currentValue);

                    }
                } /* end if */
            } /* end foreach row */
            unset($id, $row);
        } /* end if */

        /* create output */
        if (!empty($resultSet)) {
            return $resultSet;
        } else {
            return array();
        }

    }

    /**
     * implements where-clause
     *
     * Returns bool(true) if where clause matches $current,
     * returns bool(false) otherwise.
     *
     * @access  private
     * @param   array  &$current
     * @param   array  &$where
     * @return  bool
     * @ignore
     */
    function _doWhere(&$current, &$where)
    {
        $test = true;
        if (!empty($where)) {
            assert('!isset($value); /* cannot redeclare variable $value */');
            foreach ($where as $value)
            {
                if (isset($value[0]) && is_array($value[0])) {
                    $value[0] = $value[0][1];
                }
                /* if all required information is provided */
                if (count($value) === 3 && is_string($value[0]) && (is_scalar($value[1]) || is_null($value[1]))) {
                    assert('!isset($currentValue); /* cannot redeclare variable $currentValue */');
                    /* get current value */
                    if (isset($current[strtoupper($value[0])])) {
                        $currentValue = $current[strtoupper($value[0])];
                    } else {
                        $currentValue = null;
                    }
                    /* handle non-scalar values */
                    if (!is_scalar($currentValue)) {
                        $currentValue = SML::encode($currentValue);
                    }
                    /* switch by operator */
                    switch ($value[2])
                    {
                        case '=':
                            if (is_null($value[1])) {
                                if (!is_null($currentValue)) {
                                    $test = false;
                                }
                            } elseif ($this->structure->isPrimaryKey($this->table, $value[0]) || $this->structure->isForeignKey($this->table, $value[0])) {
                                if (strcasecmp($currentValue, $value[1]) !== 0) {
                                    $test = false;
                                }
                            } else {
                                if (strcmp($currentValue, $value[1]) !== 0) {
                                    $test = false;
                                }
                            }
                        break;

                        case 'like':
                        case 'regexp':
                            if (!preg_match('/^' . $value[1] . '$/i', $currentValue)) {
                                $test = false;
                            }
                        break;

                        case '<':
                            if (!($currentValue < $value[1])) {
                                $test = false;
                            }
                        break;

                        case '>':
                            if (!($currentValue > $value[1])) {
                                $test = false;
                            }
                        break;

                        case '!=':
                            if (is_null($value[1])) {
                                if (is_null($currentValue)) {
                                    $test = false;
                                }
                            } elseif ($this->structure->isPrimaryKey($this->table, $value[0]) || $this->structure->isForeignKey($this->table, $value[0])) {
                                if (strcasecmp($currentValue, $value[1]) === 0) {
                                    $test = false;
                                }
                            } else {
                                if (strcmp($currentValue, $value[1]) === 0) {
                                    $test = false;
                                }
                            }
                        break;

                        case '<=':
                            if (!($currentValue <= $value[1])) {
                                $test = false;
                            }
                        break;

                        case '>=':
                            if (!($currentValue >= $value[1])) {
                                $test = false;
                            }
                        break;

                        default:
                            /* intentionally left blank */
                        break;
                    }
                    unset($currentValue);
                }
            } /* end foreach */
            unset($value);
        } /* end if */
        if ($test) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * initialize and return current index file by reference
     *
     * @access  private
     * @return  FileDbIndex
     * @ignore
     */
    function &_getIndexFile()
    {
        $idxfile =& $this->idx[$this->database][$this->table];
        if (!is_object($idxfile)) {
            trigger_error("No index-file found.", E_USER_ERROR);
        }
        return $idxfile;
    }

    /**
     * create index file
     *
     * @access  private
     * @param   string  $database
     * @ignore
     */
    function _setIndexFile($database = null)
    {
        $filename = $this->_getFilename($database, 'idx');
        $smlfile =& $this->_getSmlFile();
        $idxfile = new FileDbIndex($this->structure, $this->table, $smlfile, $filename);
        $this->idx[$this->database][$this->table] =& $idxfile;
    }

    /**
     * initialize and return current SML file by reference
     *
     * @access  private
     * @return  SML
     * @ignore
     */
    function &_getSmlFile()
    {
        return $this->src[$this->database][$this->table];
    }

    /**
     * create SML file
     *
     * @access  private
     * @param   string  $database
     * @ignore
     */
    function _setSmlFile($database = null)
    {
        $filename = $this->_getFilename($database, 'sml');
        $smlfile = new SML($filename, CASE_UPPER);
        $smlfile->read();
        if (!$smlfile->exists()) {
            $smlfile->create();
            if ($database != $this->database) {
                $filename = $this->_getFilename($this->database, 'sml');
                $smlfile = new SML($filename, CASE_UPPER);
            }
        }

        $this->src[$this->database][$this->table] =& $smlfile;
    }

    /**
     * return filename
     *
     * @access  private
     * @param   string  $database
     * @param   string  $extension
     * @return  string
     * @ignore
     */
    function _getFilename($database, $extension)
    {
        return (string) $this->baseDir . $database . '/' . $this->table . '.' . $extension;
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same structure file.
     *
     * @access   public
     * @param    object  $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject, 'getClass') && $anotherObject->getClass() === $this->getClass()) {
            if (!isset($this->structure) || !isset($anotherObject->structure)) {
                return isset($this->structure) === isset($anotherObject->structure);
            } elseif ($this->structure->equals($anotherObject->structure)) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
