<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 *
 * @ignore
 */

/**
 * FileDbIndex
 *
 * This implements index files for table indexes,
 * for the FileDB database.
 *
 * @access      public
 * @package     yana
 * @subpackage  database
 * @since       2.8.6
 * @ignore
 */
class FileDbIndex extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var DbStructure */ var $structure = null;
    /** @var string      */ var $table     = "";
    /** @var SML         */ var $data      = null;
    /** @var SML         */ var $src       = null;

    /**#@-*/

    /**
     * constructor
     *
     * @param  DbStructure  &$structure
     * @param  string       &$table
     * @param  SML          &$data
     * @param  string       $filename
     */
    function __construct(&$structure, &$table, &$data, $filename)
    {
        if (!is_object($structure) || !is_string($table) || !is_object($data) || !is_string($filename)) {
            trigger_error('Invalid database input.', E_USER_ERROR);
            exit(1);
        } else {
            $this->structure =& $structure;
            $this->table     =& $table;
            $this->data      =& $data;
            $this->src       =  new SML($filename, CASE_UPPER);
            $this->src->read();
            /* create file if not exists
             *
             * If no index exists yet, the file needs to be created
             * and initialized before using it.
             * Otherwise there you would risk, working with invalid
             * data and possibly cause inconsistent data to be written
             * to the database.
             */
            if ($this->src->exists() === false) {
                $this->src->create();
                $this->src->reset();
                $this->create();
                $this->src->write();
            }
        }
    }

    /**
     * create an index
     *
     * Returns bool(true) on success,
     * returns bool(false) on error.
     *
     * Note: This will commit any uncommitted data.
     *
     * @access  public
     * @param   string  $column
     * @param   array   $update
     * @return  bool
     */
    function create($column = null, $update = null)
    {
        assert('is_null($column) || is_string($column);');
        assert('is_null($update) || is_array($update);');

        /* autoscan */
        if (is_null($column)) {
            if (!is_null($update)) {
                trigger_error('No column name provided. Unable to build index', E_USER_WARNING);
                return false;
            } else {
                $columnList = $this->structure->getIndexes($this->table);
                $columnList = Hashtable::merge($columnList,$this->structure->getUniqueConstraints($this->table));
                if (is_array($columnList) && count($columnList) > 0) {
                    foreach ($columnList as $columnName)
                    {
                        if ($this->create($columnName) !== true) {
                            return false;
                        }
                    } /* end foreach */
                    return true;
                } else {
                    return false;
                }
            }

        /* no scan */
        } else {

            if (!isset($this->structure)) {
                return false;
            } elseif (!is_null($update) && (!is_array($update) || count($update)!==2 || !is_scalar($update[0]) || !is_scalar($update[1]))) {
                return false;
            } elseif (!$this->structure->isColumn($this->table,$column)) {
                trigger_error("SQL syntax error. The column '".$column."' does not exist in table '".$this->table."'.", E_USER_WARNING);
                return false;
            } else {
                $smlfile    =& $this->src;
                $primaryKey = $this->structure->getPrimaryKey($this->table);

                /* no need to index primary key, it is an index by itself */
                if (strcasecmp($primaryKey,$column)===0) {
                    return false;
                }

                $column = strtoupper($column);

                /* create / recreate index */
                if (is_null($update) || !$smlfile->exists($column)) {

                    $dataset = $this->data->get($primaryKey);
                    $data    = array();
                    if (!is_null($update)) {
                        $dataset = Hashtable::merge($dataset, array(strtoupper($update[0]) => array($column => $update[1])));
                    }
                    assert('is_array($dataset);');

                /* update index */
                } else {

                    $dataset = array(strtoupper($update[0]) => array($column => $update[1]));
                    $data    = $smlfile->get($column);
                    if (!is_array($data)) {
                        $data = array();
                    }

                }

                /* process index */
                foreach ($dataset as $value => $row)
                {
                    /* NULL values are to be ignored */
                    if (isset($row[$column])) {
                        $key = strtoupper($row[$column]);
                        if (isset($data[$key])) {
                            if (is_array($data[$key])) {
                                array_push($data,$value);
                            } else {
                                $data[$key] = array($data[$key],$value);
                            }
                        } else {
                            $data[$key] = $value;
                        }
                    }
                } /* end foreach */
                ksort($data);
                $data = array($column => $data);
                $smlfile->remove($column);

                if (!$smlfile->insert($data)) {
                    return false;
                } else {
                    return true;
                }
            } /* end if */
        } /* end if */

        return false;
    }

    /**
     * get index for some column
     *
     * Same as FileDbIndex::get(), but retuns
     * result by reference;
     *
     * @access  public
     * @param   string  $column
     * @param   scalar  $value
     * @return  array()
     * @see     FileDbIndex::get()
     */
    function &getByReference($column, $value = null)
    {
        assert('is_string($column); // Wrong argument type for argument 1 in ' . __CLASS__ . '->' . __FUNCTION__ . '(). String expected.');
        assert('is_null($value) || is_scalar($value); // Wrong argument type for argument 2 in ' . __CLASS__ . '->' . __FUNCTION__ . '(). Scalar expected.');

        $isError    = false;
        $emptyIndex = array();
        if (!isset($this->structure)) {
            return $isError;
        } elseif (!is_null($value) && !is_scalar($value)) {
            return $isError;
        } elseif (!$this->structure->isColumn($this->table, $column)) {
            trigger_error("SQL syntax error. The column '".$column."' does not exist in table '".$this->table."'.", E_USER_WARNING);
            return $isError;
        } elseif (!$this->structure->hasIndex($this->table, $column) && !$this->structure->isUnique($this->table, $column)) {
            return $emptyIndex;
        } else {
            $smlfile    =& $this->src;
            $primaryKey = $this->structure->getPrimaryKey($this->table);

            if (strcasecmp($primaryKey,$column)===0) {
                return $emptyIndex;
            }

            if (is_null($value)) {
                $index = $smlfile->getByReference($column);
                if (!is_array($index)) {
                    return $isError;
                } else {
                    return $index;
                }
            } else {
                $value = strtoupper($value);
                $index = $smlfile->getByReference($column.'.'.$value);
                if (!is_scalar($index) && !is_array($index)) {
                    return $isError;
                } else {
                    return $index;
                }
            }
        }
    }

    /**
     * get index for some column
     *
     * Returns an associative array on success,
     * returns bool(false) on error.
     *
     * The resulting array uses the following syntax:
     * array (
     *     column_value => primary_key_value
     * )
     *
     * {@internal
     *
     * Exampe of usage:
     * E.g. given the following SQL-statement:
     *  <code>  SELECT * from myTable where indexedRow = "test"  </code>
     *
     * Handle this request as follows:
     * <code>
     * $primaryKeys = $this->get('indexedRow','test');
     *
     * </code>
     * }}
     *
     * @access  public
     * @param   string  $column
     * @param   scalar  $value
     * @return  array()
     */
    function get($column, $value = null)
    {
        assert('is_string($column); // Wrong argument type for argument 1 in ' . __CLASS__ . '->' . __FUNCTION__ . '(). String expected.');
        assert('is_null($value) || is_scalar($value); // Wrong argument type for argument 2 in ' . __CLASS__ . '->' . __FUNCTION__ . '(). Scalar expected.');
        return $this->getByReference($column, $value);
    }

    /**
     * write index changes to disk
     *
     * Returns bool(true) on success,
     * returns bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function commit()
    {
        return $this->src->write();
    }

    /**
     * reset index contents
     *
     * Returns bool(true) on success,
     * returns bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function rollback()
    {
        return $this->src->reset();
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
