<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * checked reading of file sources
 *
 * This class adds functionality to compute the checksum
 * of a resource and failsafe reading of resources.
 *
 * Note: This class does not implement methods for writing
 * on files.
 *
 * @access      public
 * @package     yana
 * @subpackage  file_system
 */
class FileReadonly extends FileSystemResource
{

    /**#@+
     * @access  private
     * @ignore
     */

    /** @var string */ var $md5     = null;
    /** @var array  */ var $content = null;
    /** @var array  */ var $options = array();

    /**#@-*/

    /**
     * read file contents
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @return  bool
     */
    function read()
    {
        if (!$this->exists()) {
            return false;
        } else {
            $this->content = file($this->getPath());
            $this->options = $this->attributes;
            /* calculate hash */
            $this->md5 = md5(implode('', $this->content));
            return true;
        }
    }

    /**
     * read file contents
     *
     * Automatically restarts reading if the file-resource
     * is temporarily not available after waiting for 0.5 seconds.
     *
     * The process is aborted if it fails 3 times.
     *
     * @access  public
     * @return  bool
     */
    function failSafeRead()
    {
        for ($i = 0; $i < 3; $i++)
        {
            if ($this->read()) {
              return true;
              break;
            } else {
              sleep(0.5);
            }
        }
        return false;
    }

    /**
     * get file contents
     *
     * @access  public
     * @return  string
     */
    function get()
    {
        if (!$this->isEmpty()) {
            return implode("", $this->content);
        } else {
            return "";
        }
    }

    /**
     * return file contents as string
     *
     * @access  public
     * @return  string
     */
    function getFileContent()
    {
        if (is_string($this->content)) {
            return (string) $this->content;
        } elseif (is_array($this->content)) {
            return implode("", $this->content);
        } else {
            return "";
        }
    }

    /**
     * alias of get()
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        if (!$this->exists()) {
            return "File ".$this->getPath()." does not exist\n";
        } elseif ($this->isEmpty()) {
            return "File ".$this->getPath()." is wether empty or not loaded\n";
        } else {
            return $this->get();
        }
    }

    /**
     * returns bool(true) if the source is empty or not loaded
     *
     * @access  public
     * @return  bool
     */
    function isEmpty()
    {
        if (is_null($this->content)) {
            return true;
        } elseif (is_array($this->content)) {
            return count($this->content) == 0;
        } elseif (is_scalar($this->content)) {
            return empty($this->content);
        } else {
            /* the source is invalid */
            return true;
        }
    }

    /**
     * return crc32 checksum for this file
     *
     * The filename parameter became available
     * in version 2.8.5
     *
     * This function has two synopsis. You may decide
     * either to call it statically or based on a current instance.
     * <ul>
     *  <li> $md5 = FileReadonly::getMd5(string $filename) </li>
     *  <li> $md5 = $file->getMd5(); </li>
     * </ul>
     * Both will return a MD5 hash. The first one for the file with the
     * name you provided, the second for the file currently represented
     * by the object $file (where $file is an instance of FileReadonly
     * or a derived sub-class).
     *
     * Note: This function has been renamed in version 2.9 RC3 from
     * "FileReadonly::checksum", to better comply with the
     * framework's naming convention
     *
     * If you prefer a hash value over a checksum, you may want to have
     * a look at {@link FileReadonly::getMd5()} instead.
     *
     * @access  public
     * @param   string  $filename
     * @return  int
     * @name    FileReadonly::getCrc32()
     * @see     FileReadonly::getMd5()
     */
    function getCrc32($filename = "")
    {
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        } elseif (!empty($filename)) {
            if (!is_file($filename)) {
                trigger_error("Unable to calculate checksum. The file '${filename}' does not exist.", E_USER_NOTICE);
                return false;
            }
        } else {
            $filename = $this->getPath();
        }

        $source = implode("",file($filename));

        if (empty($source)) {
            return 0;
        } elseif (is_string($source)) {
            return crc32($source);
        } else {
            /* the source is invalid */
            return 0;
        }
    }

    /**
     * return md5 hash for this file
     *
     * This function calculates and returns the MD5
     * hash-string for a file.
     *
     * Usefull to check whether or not a
     * file has been changed since last access.
     *
     * This function has two synopsis. You may decide
     * either to call it statically or based on a current instance.
     * <ul>
     *  <li> $md5 = FileReadonly::getMd5(string $filename) </li>
     *  <li> $md5 = $file->getMd5(); </li>
     * </ul>
     * Both will return a MD5 hash. The first one for the file with the
     * name you provided, the second for the file currently represented
     * by the object $file (where $file is an instance of
     * FileReadonly or a derived sub-class).
     *
     * Note: results are cached for the current file.
     *
     * Note: This function has been renamed in version 2.9 RC3 from
     * "FileReadonly::md5Checksum", as the MD5 algorithm does
     * not really return a checksum, but a hash value.
     *
     * If you prefer a checksum over a hash value, you may want to have
     * a look at {@link FileReadonly::getCrc32()} instead.
     *
     * @access  public
     * @param   string  $filename  name of file (using current file if left blank)
     * @return  string
     * @since   2.8.5
     * @name    FileReadonly::getMd5()
     * @see     FileReadonly::getCrc32()
     */
    function getMd5($filename = "")
    {
        /* input checking*/
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        }

        if (!empty($filename)) {
            if (!is_file($filename)) {
                trigger_error("Unable to calculate MD5 hash. The file '${filename}' does not exist.", E_USER_NOTICE);
                return false;
            }
        } else {
            if (isset($this->md5)) {
                assert('is_string($this->md5);');
                return $this->md5;
            } else {
                $filename = $this->getPath();
                assert('is_file($filename);');
            }
        }
        $md5 = md5_file($filename);

        if (!is_string($md5)) {
            return false;
        } else {
            return $md5;
        }
    }

    /**
     * reload file contents
     *
     * You accidently did something wrong with the file?
     * Calling this will reload the file from disk while
     * reseting its current state.
     *
     * @access  public
     */
    function reset()
    {
        $this->read();
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
