<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * Form-generator
 *
 * This class is a form-generator, that produces HTML-code
 * from database structure files.
 *
 * Example of usage:
 * <code>
 * $form = new FormCreator();
 * // use database structure in file 'project'
 * $form->setFile('project');
 * // use table 'effort' from file 'project'
 * $form->setTable('effort');
 * // choose which form to create (here: a form to edit entries)
 * $form->setTemplate(FORM_EDIT);
 * // set which columns to show
 * $form->setColumns(array (
 *     "effort_id",
 *     "effort_date",
 *     "effort_description",
 *     "effort_state"
 * ));
 * // limit entries to show by adding a where clause
 * $form->setWhere( array (
 *     array("effort_duration", 2, '>'),
 *     array("effort_duration", 20, '<')
 * ));
 * // select a column to sort the entries
 * $form->setSort('effort_id');
 * // set the number of entries to show per page
 * $form->setEntriesPerPage(5);
 * // set what action (function) should be triggered,
 * // when the user clicks the "edit" button
 * $form->setEditAction('project_edit_effort');
 * // set what action (function) should be triggered,
 * // when the user clicks the "new" button
 * $form->setNewAction('project_new_effort');
 * // generate the HTML code of the requested form
 * $string = $form->createForm();
 * </code>
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class FormCreator extends Object
{

    /**#@+
     * the following information is MANDATORY
     *
     * @access  private
     */

    /** @var SmartTemplate */ var $template        = null;
    /** @var int           */ var $templateId      = 0;
    /** @var int           */ var $layoutId        = 0;
    /** @var string        */ var $file            = null;
    /** @var string        */ var $table           = null;
    /** @var array         */ var $columns         = array();

    /**#@-*/
    /**#@+
     * the following information is OPTIONAL
     *
     * @access  private
     */

    /** @var string|array  */ var $where          = array();
    /** @var string        */ var $orderBy        = "";
    /** @var bool          */ var $desc           = false;
    /** @var int           */ var $page           = 0;
    /** @var int           */ var $firstPage      = 0;
    /** @var int           */ var $lastPage       = 0;
    /** @var bool          */ var $arrayKeys      = true;

    /**
     * the value of the primary key
     * of the current page
     * @var mixed          */ var $currentIndex   = null;

    /**
     * number of entries per page
     * @var int            */ var $entries        = 5;

    /**
     * only applies to some templates
     * @var bool           */ var $titles         = true;

    /**#@-*/
    /**#@+
     * the following are values for the ACTION attribute
     *
     * @access  private
     */

    /** @var string        */ var $newAction      = null;
    /** @var string        */ var $editAction     = null;
    /** @var string        */ var $deleteAction   = null;
    /** @var string        */ var $searchAction   = null;
    /** @var string        */ var $downloadAction = 'download_file';

    /**#@-*/
    /**#@+
     * the following are STATIC constants
     *
     * @access  private
     * @static
     */

    /** @var array         */ var $fileMap        = null;
    /** @var array         */ var $tmplMap        = null;

    /**#@-*/
    /**#@+
     * the following are INSTANCE-SPECIFIC var
     *
     * @access  private
     */

    /** @var DbStream       */ var $db             = null;
    /** @var array          */ var $refColumns     = array();
    /** @var bool           */ var $cacheResults   = true;
    /** @var bool           */ var $isAdvSearch    = null;
    /** @var bool           */ var $allowNewEntry  = true;

    /**#@-*/

    /**
     * create a new instance
     *
     * This creates and initializes a new
     * instance of this class.
     */
    function __construct()
    {
        /**
         * mapping of ids to filenames
         */
        if (is_null($this->fileMap)) {
            global $YANA;

            $smallSearch = $YANA->skin->getFile("gui_small_search");
            $advSearch = $YANA->skin->getFile("gui_search");

            static $fileMap;
            $fileMap = array(
                FORM_VIEW => array(
                    0 => $YANA->skin->getFile("gui_view"),
                    1 => $YANA->skin->getFile("gui_view_seperated"),
                    2 => $YANA->skin->getFile("gui_view_details"),
                    3 => $YANA->skin->getFile("gui_view_list")
                ),
                FORM_EDIT => array(
                    0 => $YANA->skin->getFile("gui_edit")
                ),
                FORM_NEW => array(
                    0 => $YANA->skin->getFile("gui_new"),
                ),
                FORM_SEARCH => array(
                    0 => $smallSearch,
                    1 => $advSearch,
                    2 => $advSearch,
                    3 => $advSearch
                ),
                FORM_SEPERATED => array(
                    0 => $YANA->skin->getFile("gui_view_seperated") /* this option is deprecated */
                ),
                FORM_DETAILS => array(
                    0 => $YANA->skin->getFile("gui_view_details")   /* this option is deprecated */
                )
            );
            $this->fileMap =& $fileMap;
        } /* end if */

        /**
         * mapping of strings to ids
         */
        if (is_null($this->tmplMap)) {
            static $tmplMap;
            $tmplMap = array(
                'view'           => FORM_VIEW,
                'edit'           => FORM_EDIT,
                'new'            => FORM_NEW,
                'search'         => FORM_SEARCH,
                'view_seperated' => FORM_SEPERATED, /* this option is deprecated */
                'view_details'   => FORM_DETAILS    /* this option is deprecated */
            );
            $this->tmplMap =& $tmplMap;
        } /* end if */

        /**
         * providing references to the template
         */
        $this->template = new SmartTemplate();        

        /**
         * initialize session
         */
        if (!isset($_SESSION['sort']) || !is_array($_SESSION['sort'])) {
            $_SESSION['sort'] = array();
        }
        if (!isset($_SESSION['desc']) || !is_array($_SESSION['desc'])) {
            $_SESSION['desc'] = array();
        }
        if (!isset($_SESSION['page']) || !is_array($_SESSION['page'])) {
            $_SESSION['page'] = array();
        }
        if (!isset($_SESSION['where']) || !is_array($_SESSION['where'])) {
            $_SESSION['where'] = array();
        }
        if (!isset($_SESSION['entries']) || !is_array($_SESSION['entries'])) {
            $_SESSION['entries'] = array();
        }
        if (!isset($_SESSION['layout']) || !is_array($_SESSION['layout'])) {
            $_SESSION['layout'] = array();
        }
        if (!isset($_SESSION['searchform']) || !is_array($_SESSION['searchform'])) {
            $_SESSION['searchform'] = array();
        }
        if (!isset($_SESSION['advanced_search']) || !is_array($_SESSION['advanced_search'])) {
            $_SESSION['advanced_search'] = array();
        }
    }

    /**
     * select a template for output
     *
     * You may use either a filename, or one of the following constants:
     *
     * <ul>
     *     <li>  FORM_VIEW    </li>
     *     <li>  FORM_EDIT    </li>
     *     <li>  FORM_NEW     </li>
     *     <li>  FORM_SEARCH  </li>
     * </ul>
     *
     * Returns bool(true) on success and bool(false) on error.
     * Issues an E_USER_WARNING, if the template is unknown.
     *
     * Note that the argument $layout can be overruled by 'get' and 'session' vars.
     *
     * @access  public
     * @param   int|string  $template
     * @param   int         $layout    does only apply if $template is a string
     * @return  bool
     */
    function setTemplate($template, $layout = 0)
    {
        /**
         * 1) check pre-conditions
         */
        if (!is_string($this->table)) {
            trigger_error('Unable to set property, need to set table property first.', E_USER_WARNING);
            return false;
        } elseif (!isset($this->db->structure)) {
            trigger_error('Unable to set property, no database found.', E_USER_WARNING);
            return false;
        }

        switch (true)
        {
            /**
             * 2) template is constant value
             */
            case is_int($template):
                if (!isset($this->fileMap[$template])) {
                    /* error: unknown template */
                    trigger_error("The template (int) '${template}' is unknown.", E_USER_WARNING);
                    return false;
                } else {
                    /* intentionally left blank */
                }
            break;

            /**
             * 3) template is an identifier
             */
            case is_string($template):
                $template = strtolower($template);
                if (!isset($this->tmplMap[$template])) {
                    /* error: unknown template */
                    trigger_error("The template (string) '${template}' is unknown.", E_USER_WARNING);
                    return false;
                } else {
                    $template = $this->tmplMap[$template];
                }
            break;

            /**
             * 4) invalid input
             */
            default:
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String or integer (constant)', gettype($template)), E_USER_WARNING);
                return false;
            break;

        } /* end switch */

        /**
         * 5) default values
         */
        if (isset($_GET['layout']) && is_numeric($_GET['layout'])) {
            $layout = (int) $_GET['layout'];
        } elseif (isset($_SESSION['layout'][$this->table][$template]) && is_numeric($_SESSION['layout'][$this->table][$template])) {
            $layout = (int) $_SESSION['layout'][$this->table][$template];
        } else {
            /* keep input settings */
        }

        /**
         * 6) check layout parameter
         */
        if (!is_int($layout)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Integer', gettype($layout)), E_USER_WARNING);
        } elseif (!isset($this->fileMap[$template][$layout])) {
            /* error: unknown template */
            $layout = 0;
        }

        /**
         * 6) mapping of constant to filename
         */
        $filename = $this->fileMap[$template][$layout];

        /**
         * 7) set attributes
         */
        if ($this->template->setPath($filename)) {
            /* success */
            $this->templateId = $template;
            $this->layoutId = $layout;
            if ($this->cacheResults) {
                $_SESSION['layout'][$this->table][$template] = $layout;
            }

            /**
             * 8) initialize table's properties
             */
            assert('!isset($test); // Cannot redeclare var $test');
            $test = true;

            /**
             * 1) by default sort by primary key
             */
            $test = $test && $this->setSort();

            /**
             * 2) by default open page 0
             */
            $test = $test && $this->setPage();
            
            /**
             * 3) by default set to 5 entries per page
             */
            $test = $test && $this->setEntriesPerPage();

            /**
             * 4) if new where clause exists, use cache - otherwise set to default
             */
            if (isset($_GET['where']) || !$this->cacheResults) {
                $test = $test && $this->setWhere();
            } else {
                /* don't cache results */
                $this->cacheResults = false;
                $test = $test && $this->setWhere();
                /* restore result cache */
                $this->cacheResults = true;
            }

            if ($test) {
                return true;
            } else {
                return false;
            }


        } else {
            /* error: e.g. file does not exist or is not readable */
            return false;
        }
    }

    /**
     * get the currently selected template file
     *
     * Returns the filename of the template that is currently,
     * selected, or bool(false) if none has been set.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getTemplate()
    {
        if (!is_null($this->template)) {
            $filename = $this->template->getPath();
            if (is_string($filename)) {
                /* success */
                return $filename;
            } else {
                /* error: e.g. file does not exist or is not readable */
                return false;
            }
        } else {
            /* error: no template has been selected, yet */
            return false;
        }
    }

    /**
     * set source file
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $file
     * @return  bool
     */
    function setFile($file)
    {
        if (is_string($file)) {
            global $YANA;
            $file = strtolower($file);
            $this->db = $YANA->connect($file);
            if (is_object($this->db)) {
                if ($this->db->exists()) {
                    $this->file = $file;
                    return true;
                } else {
                    trigger_error("Invalid file argument. Database connection failed.", E_USER_WARNING);
                    return false;
                }
            } else {
                return false;
            }
        } elseif (is_object($file) && (is_a($file, 'DbStream') || is_subclass_of($file, 'DbStream'))) {
            $this->db = $file;
            if ($this->db->exists()) {
                $this->file = $this->db->getPath();
                return true;
            } else {
                trigger_error("Invalid file argument. Database connection failed.", E_USER_WARNING);
                return false;
            }
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($file)), E_USER_WARNING);
            return false;
        }
    }

    /**
     * get the currently selected file
     *
     * Returns the lower-cased name of the currently
     * selected file, or bool(false) if none has been
     * selected yet.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getFile()
    {
        if (is_string($this->file)) {
            return $this->file;
        } else {
            /* error: no file has been selected, yet */
            return false;
        }
    }

    /**
     * turn advanced search on / off
     *
     * Triggers wether advanced search is to be used (true=yes, false=no).
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $advancedSearch
     * @return  bool
     */
    function enableAdvancedSearch($advancedSearch = true)
    {
        /**
         * 1) check pre-conditions
         */
        if (!is_string($this->table)) {
            trigger_error('Unable to set property, need to set table property first.', E_USER_WARNING);
            return false;
        } elseif (!isset($this->db->structure)) {
            trigger_error('Unable to set property, no database found.', E_USER_WARNING);
            return false;
        }

        /**
         * 2) no columns of type text (simple mode not possible)
         */
        if (count($this->db->structure->getColumnsByType($this->table, 'text')) === 0) {
           $this->isAdvSearch = true;
           return true;

        /**
         * 3) select by input
         */
        } elseif (is_bool($advancedSearch)) {
            if (isset($_GET['advanced_search'])) {
                $this->isAdvSearch = (bool) $_GET['advanced_search'];
            } elseif (isset($_SESSION['advanced_search'][$this->table])) {
                $this->isAdvSearch = (bool) $_SESSION['advanced_search'][$this->table];
            } else {
                $this->isAdvSearch = $advancedSearch;
            }
            $_SESSION['advanced_search'][$this->table] = $this->isAdvSearch;
            return true;
        } else {
            return false;
        }
    }

    /**
     * get state of advanced search (on / off)
     *
     * Returns bool(true) if enabled, or bool(false) if disabled.
     *
     * @access  public
     * @return  bool
     */
    function getAdvancedSearch()
    {
        if (is_bool($this->isAdvSearch)) {
            return $this->isAdvSearch;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * trigger wether the user should be allowed to create new entries
     *
     * Use this function to trigger the "new entry" row on edit-forms.
     * If $allowNewEntries is true, the row will be shown, else it will be hidden.
     *
     * Note that this row can be hidden automatically, if a not-nullable column
     * is not visible to the user, making it practically impossible to add entries.
     * However: you may overrule this behavior by calling this function and setting
     * the value to true.
     *
     * Returns the old value of the property.
     * If you just want to read the current setting, call this function without
     * providing any argument.
     *
     * @access  public
     * @param   bool  $allowNewEntry
     * @return  bool
     * @since   2.9.5
     */
    function allowNewEntry($allowNewEntry = null)
    {
        if (is_null($allowNewEntry)) {
            return $this->allowNewEntry;
        } else {
            $previousValue = $this->allowNewEntry;
            if ($allowNewEntry) {
                $this->allowNewEntry = true;
            } else {
                $this->allowNewEntry = false;
            }
            return $previousValue;
        }
    }

    /**
     * set source table
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $table
     * @return  bool
     */
    function setTable($table)
    {
        if (!is_null($this->db->structure) && !$this->db->structure->isTable($table)) {
            return false;
        } elseif (!is_string($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($table)), E_USER_WARNING);
            return false;
        }

        /**
         * 1) set table
         */
        $table = strtolower($table);
        $this->table = $table;
        
        return true;
    }

    /**
     * select columns to view in form
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   array  $columns
     * @return  bool
     */
    function setColumns($columns = null)
    {
        if (is_null($columns)) {
            if (!is_null($this->db->structure) && !is_null($this->table)) {
                $this->columns = $this->db->structure->getColumns($this->table);
                return true;
            } else {
                return false;
            }
        } elseif (is_array($columns)) {
            $columns = array_map("strtolower", $columns);
            $this->columns = $columns;
            if (!is_null($this->db->structure) && !is_null($this->table) && $this->templateId === FORM_EDIT) {
                assert('!isset($column); // Cannot redeclare var $column');
                foreach (array_diff($this->db->structure->getColumns($this->table), $columns) as $column)
                {
                    if (!$this->db->structure->isNullable($this->table, $column)) {
                        $this->allowNewEntry = false;
                    }
                } /* end foreach */
                unset($column); /* clean up garbage */
            }
            return true;
        } else {
            return false;
        }
    }

    /**
     * get selected columns
     *
     * @access  public
     * @return  array
     */
    function getColumns()
    {
        if (empty($this->columns)) {
            $this->setColumns();
        }
        return $this->columns;
    }

    /**
     * get the currently selected table
     *
     * Returns the lower-cased name of the currently
     * selected table, or bool(false) if none has been
     * selected yet.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getTable()
    {
        if (is_string($this->table)) {
            return $this->table;
        } else {
            /* error: no table has been selected, yet */
            return false;
        }
    }

    /**
     * set action for an event
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $name
     * @param   string  $value
     * @return  bool
     */
    function setAction($name, $value)
    {
        /* input checking */
        assert('is_string($name);  // Wrong argument type for argument 1. String expected.');
        assert('is_string($value); // Wrong argument type for argument 2. String expected.');
        $attribute  = strtolower( (string) $name  );
        $action     = strtolower( (string) $value );
        $attribute .= 'Action';

        /* set action to value */
        if (property_exists($this, $attribute)) {
            $this->$attribute = $action;
            return true;
        } else {
            /* error: no such attribute */
            return false;
        }
    }

    /**
     * get action for an event
     *
     * Returns the lower-cased name of the currently
     * selected action, or bool(false) if none has been
     * selected yet.
     *
     * @access  public
     * @param   string  $name  name of event
     * @return  bool(false)|string
     */
    function getAction($name)
    {
        assert('is_string($name); // Wrong argument type for argument 1. String expected.');
        $attribute  = strtolower( (string) $name );
        $attribute .= 'Action';

        if (isset($this->$attribute) && is_string($this->$attribute)) {
            return $this->$attribute;
        } else {
            /* error: attribute has not been set, yet */
            return false;
        }
    }

    /**
     * set download action
     *
     * @see     FormCreator::setAction()
     * @param   string  $action
     * @return  bool
     */
    function setDownloadAction($action)
    {
        return $this->setAction('download', $action);
    }

    /**
     * get download action
     *
     * @see     FormCreator::getAction()
     * @return  bool(false)|string
     */
    function getDownloadAction()
    {
        return $this->getAction('download');
    }

    /**
     * set edit action
     *
     * @see     FormCreator::setAction()
     * @param   string  $action
     * @return  bool
     */
    function setEditAction($action)
    {
        return $this->setAction('edit', $action);
    }

    /**
     * get edit action
     *
     * @see     FormCreator::getAction()
     * @return  bool(false)|string
     */
    function getEditAction()
    {
        return $this->getAction('edit');
    }

    /**
     * set new action
     *
     * @see     FormCreator::setAction()
     * @param   string  $action
     * @return  bool
     */
    function setNewAction($action)
    {
        return $this->setAction('new', $action);
    }

    /**
     * get new action
     *
     * @see     FormCreator::getAction()
     * @return  bool(false)|string
     */
    function getNewAction()
    {
        return $this->getAction('new');
    }

    /**
     * set delete action
     *
     * @see     FormCreator::setAction()
     * @param   string  $action
     * @return  bool
     */
    function setDeleteAction($action)
    {
        return $this->setAction('delete', $action);
    }

    /**
     * get delete action
     *
     * @see     FormCreator::getAction()
     * @return  bool(false)|string
     */
    function getDeleteAction()
    {
        return $this->getAction('delete');
    }

    /**
     * set search action
     *
     * @see     FormCreator::setAction()
     * @param   string  $action
     * @return  bool
     */
    function setSearchAction($action)
    {
        return $this->setAction('search', $action);
    }

    /**
     * get search action
     *
     * @see     FormCreator::getAction()
     * @return  bool(false)|string
     */
    function getSearchAction()
    {
        return $this->getAction('search');
    }

    /**
     * get namespace
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @return  string|bool(false)
     * @since   2.9.2
     */
    function getNamespace()
    {
        switch ($this->templateId)
        {
            case FORM_NEW:
                return "new";
            break;
            case FORM_EDIT:
                return "edit";
            break;
            case FORM_SEARCH:
            case FORM_VIEW:
            case FORM_SEPERATED:
            case FORM_DETAILS:
                return "select";
            break;
            default:
                return false;
            break;
        } /* end switch */
    }

    /**
     * turn titles on / off
     *
     * Triggers wether fieldnames are to be shown (true=yes, false=no).
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $titles
     * @return  bool
     */
    function enableTitles($titles = true)
    {
        if (is_bool($titles)) {
            $this->titles = $titles;
            return true;
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Boolean', gettype($titles)), E_USER_WARNING);
            return false;
        }
    }

    /**
     * get state of titles (on / off)
     *
     * Triggers wether fieldnames are to be shown (true=yes, false=no).
     * Returns bool(true) if enabled, or bool(false) if disabled.
     *
     * @access  public
     * @return  bool
     */
    function getTitles()
    {
        if (is_bool($this->titles)) {
            return $this->titles;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * turn array keys on / off
     *
     * Triggers wether for columns of type array the keys are to be used (true=yes, false=no).
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   bool  $arrayKeys
     * @return  bool
     * @since   2.9.4
     * @name    FormCreator::enableArrayKeys()
     * @see     FormCreator::getArrayKeys()
     */
    function enableArrayKeys($arrayKeys = true)
    {
        if (is_bool($arrayKeys)) {
            $this->arrayKeys = $arrayKeys;
            return true;
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Boolean', gettype($arrayKeys)), E_USER_WARNING);
            return false;
        }
    }

    /**
     * get state of array keys (on / off)
     *
     * This option triggers how to display columns of type "array".
     * If enabled the keys are visible and editable (on edit forms).
     * If disabled the keys are hidden.
     *
     * Returns bool(true) if enabled, or bool(false) if disabled.
     *
     * @access  public
     * @return  bool
     * @since   2.9.4
     * @name    FormCreator::getArrayKeys()
     * @see     FormCreator::enableArrayKeys()
     */
    function getArrayKeys()
    {
        if (is_bool($this->arrayKeys)) {
            return $this->arrayKeys;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * set column to sort the resultset by
     *
     * Note that 'get' and 'session' vars overrule this setting.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string  $orderBy  column name (defaults to primary key)
     * @param   bool    $desc     sort descending (true=yes, false=no, defaults to false)
     * @return  bool
     */
    function setSort($orderBy = '', $desc = null)
    {
        /**
         * decide if results should be cached (if option is activated)
         */
        if ($desc == null) {
            $doCacheDesc = false;
            $desc = false;
        } else {
            $doCacheDesc = true;
        }
        /**
         * 1) check pre-conditions
         */
        if (!is_string($this->table)) {
            trigger_error('Unable to set property, need to set table property first.', E_USER_WARNING);
            return false;
        } elseif (!isset($this->db->structure)) {
            trigger_error('Unable to set property, no database found.', E_USER_WARNING);
            return false;
        }

        /**
         * 2) apply defaults for $orderBy
         */
        if (isset($_GET['sort']) && is_string($_GET['sort'])) {
            $orderBy = $_GET['sort'];

        } elseif (isset($_SESSION['sort'][$this->table]) && is_string($_SESSION['sort'][$this->table])) {
            $orderBy = $_SESSION['sort'][$this->table];

        } elseif ($orderBy === '') {
            $orderBy = $this->db->structure->getPrimaryKey($this->table);

        } else {
           /* keep input setting */
        }

        /**
         * 3) apply defaults for $desc
         */
        if (isset($_GET['desc'])) {
            if ($_GET['desc']) {
                $desc = true;
            } else {
                $desc = false;
            }
        } elseif (isset($_SESSION['desc'][$this->table])) {
            if ($_SESSION['desc'][$this->table]) {
                $desc = true;
            } else {
                $desc = false;
            }
        } else {
           /* keep input setting */
        }

        /**
         * 4) check validity
         */
        /* error - invalid type */
        if (!is_string($orderBy)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($orderBy)), E_USER_WARNING);
            return false;

        /* error - not a column */
        } elseif (!$this->db->structure->isColumn($this->table, $orderBy)) {
            return false;

        /* error invalid type */
        } elseif (!is_bool($desc)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($desc)), E_USER_WARNING);
            return false;

        } else {
           /* input is valid */
        }

        /**
         * 5) apply changes
         */
        $this->orderBy = strtolower($orderBy);
        $this->desc = $desc;
        if ($this->cacheResults) {
            $_SESSION['sort'][$this->table] = $orderBy;
            if ($doCacheDesc === true) {
                $_SESSION['desc'][$this->table] = $desc;
            }
        }
        return true;
    }

    /**
     * get the name of the column the resultset is ordered by
     *
     * Returns the lower-cased column name, or bool(false)
     * if none has been set yet.
     *
     * @access  public
     * @return  bool(false)|string
     */
    function getSort()
    {
        if (is_string($this->orderBy) && !empty($this->orderBy)) {
            return $this->orderBy;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * check if resultset is sorted in descending order
     *
     * Returns bool(true) if sorted descending and bool(false) otherwise.
     *
     * @access  public
     * @return  bool
     */
    function isDescending()
    {
        if (is_bool($this->desc)) {
            return $this->desc;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * set current page
     *
     * The first page is 0, the second is 1, aso., defaults to 0.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note that 'get' and 'session' vars may overrule this setting.
     *
     * @access  public
     * @param   int  $page     number of start page

     * @return  bool
     */
    function setPage($page = 0)
    {
        /**
         * 1) get default parameters
         */
        if (isset($_GET['page']) && is_numeric($_GET['page'])) {
            $page = $_GET['page'];
        } elseif (!isset($_SESSION['page'][$this->table])) {
            $_SESSION['page'][$this->table] = array();
        } elseif (isset($_SESSION['page'][$this->table][$this->templateId]) && is_numeric($_SESSION['page'][$this->table][$this->templateId])) {
            $page = $_SESSION['page'][$this->table][$this->templateId];
        } else {
            /* keep input settings */
        }

        /**
         * 2) error - invalid argument
         */
        if (!is_numeric($page)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($page)), E_USER_WARNING);
            return false;
        } else {

            /**
             * 3) check input
             */

            /* settype to INTEGER */
            $page = (int) $page;

            /* default values */
            if ($page < 0) {
                trigger_error("Page number was negative. Will use the default value 0 instead.", E_USER_NOTICE);
                $page = 0;
            }

            /**
             * 4) set values
             */
            $this->page = $page;
            $this->firstPage = $page + 1;
            if ($this->cacheResults) {
                $_SESSION['page'][$this->table][$this->templateId] = $page;
            }    
            return true;
        }
    }

    /**
     * set number of entries per page
     *
     * The number of entries per page needs to greater 0, defaults to 5.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * Note that 'get' and 'session' vars may overrule this setting.
     *
     * @access  public
     * @param   int  $entries  number of entries per page
     * @return  bool
     */
    function setEntriesPerPage($entries = null)
    {
        /**
         * 1) get default parameters
         */
        if (isset($_GET['entries']) && is_numeric($_GET['entries'])) {
            $entries = $_GET['entries'];
        } elseif (!isset($_SESSION['entries'][$this->table])) {
            $_SESSION['entries'][$this->table] = array();
        } elseif (isset($_SESSION['entries'][$this->table][$this->templateId]) && is_numeric($_SESSION['entries'][$this->table][$this->templateId])) {
            $entries = $_SESSION['entries'][$this->table][$this->templateId];
        } else {
            /* keep input settings */
        }

        /**
         * 2) error - invalid argument
         */
        if (is_null($entries)) {
            $this->entries = 5;
            return true;
        } elseif (!is_numeric($entries)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Integer', gettype($entries)), E_USER_WARNING);
            return false;
        } else {

            /**
             * 3) check input
             */

            /* settype to INTEGER */
            $entries = (int) $entries;

            /* default values */
            if ($entries < 1) {
                trigger_error("Number of entries per page was less than 1. Will use the default value 5 instead.", E_USER_NOTICE);
                $entries = 5;
            }

            /**
             * 4) set values
             */
            $this->entries = $entries;
            if ($this->cacheResults) {
                $_SESSION['entries'][$this->table][$this->templateId] = $entries;
            }    
            return true;
        }
    }
    
    /**
     * get the currently selected page
     *
     * Returns the number of the current page, or bool(false)
     * on error.
     *
     * @access  public
     * @return  bool(false)|int
     */
    function getPage()
    {
        if (is_int($this->page)) {
            return $this->page;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * get number of entries to show per page
     *
     * Returns the number of the current page, or bool(false)
     * on error.
     *
     * @access  public
     * @return  bool(false)|int
     */
    function getEntriesPerPage()
    {
        if (is_int($this->entries)) {
            return $this->entries;
        } else {
            /* error: unexpected value */
            return false;
        }
    }

    /**
     * set where clause (filter)
     *
     * The syntax is as follows:
     * column1=value1,column2=value2,...,columnN=valueN
     *
     * The alternative is as follows:
     * array(array(0=>column1,1=>value1,2=>operator1),...)
     * Where "operator" can be one of the following:
     * '=', 'LIKE', '<', '>', '!=', '<=', '>='
     *
     * To unset the where clause, call this function without
     * providing a parameter.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   string|array  $where
     * @param   bool          $replace
     * @return  bool
     */
    function setWhere($where = null, $replace = false)
    {

        /**
         * 1) get prev. where clause for search forms
         *
         * This is to make sure, the where clause remains active
         * even will browsing through several pages of the resultset.
         */
        if ($this->templateId === FORM_SEARCH) {

            if (count($this->columns) > 0) {
                if ($this->isAdvSearch === true) {
                    $where = FormCreator::getFormdata(FORM_ADV_SEARCH, $this->db->structure, $this->table, $_POST, $this->columns);
                } else {
                    $where = FormCreator::getFormdata(FORM_SEARCH, $this->db->structure, $this->table, $_POST, $this->columns);
                }
            }

            /**
             * 1.1) set where clause
             */
            if (empty($where) && !empty($_GET['where'])) {
                $this->_whereToSession($_GET['where'], true);
                $where = $this->_whereFromSession();

            } elseif (empty($where) && !empty($_SESSION['where'][$this->table][$this->templateId])) {

                $where = $this->_whereFromSession();

            } elseif (!empty($where)) {

                $this->_whereToSession($where);

            } else {
                /* where clause is empty - nothing to do */
            }

        /**
         * 2) get prev. where clause for all other forms
         *
         * Get values for where-clause from user input or session data, if available.
         */
        } else {

            /**
             * 2.1) reset where clause on demand
             */
            if (!empty($_GET['replacewhere'])) {
                $_SESSION['where'][$this->table][$this->templateId] = array();
            }

            if (isset($_GET['where'])) {

                /**
                 * 2.2) empty where clause found
                 */
                if ($_GET['where'] === '') {
                    $where = "";
                    $_SESSION['where'][$this->table][$this->templateId] = array();

                /**
                 * 2.3) move from URI-string to session vars
                 */
                } elseif (isset($_GET['where']) && is_array($_GET['where'])) {
                    $this->_whereToSession($_GET['where'], true);

                }
            }
            
            /**
             * 2.4) get from session vars
             */
            if (isset($_SESSION['where'][$this->table][$this->templateId]) && is_array($_SESSION['where'][$this->table][$this->templateId])) {
                if (is_array($where)) {
                    $where = $this->_whereFromSession($where);
                } else {
                    $where = $this->_whereFromSession();
                }

            } else {
               /* keep input setting */
            }
        } /* end if (FORM_SEARCH) else */

        /**
         * 3.1) initialize only
         */
        if (is_null($where)) {

            $this->where = array();
            return true;

        /**
         * 3.2) reset where clause
         */
        } elseif (empty($where)) {

            if ($this->cacheResults) {
                $_SESSION['where'][$this->table][$this->templateId] = array();
            }

            $this->where = array();
            return true;

        /**
         * 3.3) compute string
         */
        } elseif (is_string($where)) {

            /* get primary key */
            if (is_object($this->db->structure) && !empty($this->table)) {
                $primaryKey = $this->db->structure->getPrimaryKey($this->table);
            } else {
                $primaryKey = '';
            }

            if (empty($primaryKey)) {
                trigger_error("Unable to get primary key for current table while setting 'where'. Result might be corrupt.", E_USER_NOTICE);
            }

            /**
             * Note:
             * when searching for a primary key, page numbering is irrelevant
             * so it needs to be reseted
             */
            if (strpos(strtolower($where), strtolower($primaryKey).'=') > -1) {
                $this->page = 0;
            }

            $this->where = $where;
            return true;

        /**
         * 3.4) compute array
         */
        } elseif (is_array($where)) {

            /* get primary key */
            if (is_object($this->db->structure) && !empty($this->table)) {
                $primaryKey = $this->db->structure->getPrimaryKey($this->table);
            } else {
                $primaryKey = '';
            }

            if (empty($primaryKey)) {
                trigger_error("Unable to get primary key for current table while setting 'where'. Result might be corrupt.", E_USER_NOTICE);
            }

            /**
             * Note:
             * This array scan is not much of a performance issue, as $where
             * has in up to 90% of all cases at most 1 item.
             */
            assert('!isset($temp); // Cannot redeclare variable $temp');
            foreach ($where as $temp)
            {
                if (isset($temp[0]) && strcasecmp($temp[0], $primaryKey) === 0) {
                    $this->page = 0;
                    break;
                }
            } /* end foreach */
            unset($temp);

            $this->_whereToSession($where);

            $this->where = $where;
            return true;

        /**
         * 3.5) error - wrong argument type (neither string, nor array)
         */
        } else {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($where)), E_USER_WARNING);
            return false;

        }
    }

    /**
     * get the currently set where clause
     *
     * Returns the current where clause, or bool(false)
     * if none has been set yet.
     *
     * @access  public
     * @return  bool(false)|string|array
     */
    function getWhere()
    {
        if (empty($this->where)) {
            return false;
        } elseif (is_string($this->where) || is_array($this->where)) {
            return $this->where;
        } else {
            /* error: no where clause has been selected, yet */
            return false;
        }
    }

    /**
     * check if a filter is set on a column
     *
     * Returns bool(true) if a filter has been set on the column
     * and bool(false) otherwise.
     *
     * {@internal
     * Will always return false if $cacheResults is set to false.
     * }}
     *
     * @access  public
     * @param   string  $column  name of column to check
     * @return  bool
     * @since   2.9.5
     */
    function hasFilter($column)
    {
        assert('is_string($column); // Wrong argument type argument 1. String expected');
        $column = strtolower($column);
        if (!$this->cacheResults) {
            return false;

        } elseif (isset($_SESSION['where'][$this->table][$this->templateId][$column])) {
            return true;

        } else {
            return false;

        }
    }

    /**
     * get the currently set filter on a certain column
     *
     * Returns bool(false) on error.
     *
     * {@internal
     * Will always return false if $cacheResults is set to false.
     * }}
     *
     * @access  public
     * @return  mixed
     * @ignore
     * @since   2.9.5
     */
    function getFilter($column)
    {
        assert('is_string($column); // Wrong argument type argument 1. String expected');
        $column = strtolower($column);
        if (!$this->cacheResults) {
            return false;

        } elseif (isset($_SESSION['where'][$this->table][$this->templateId][$column])) {
            return $_SESSION['where'][$this->table][$this->templateId][$column];

        } else {
            return false;

        }
    }

    /**
     * remove filter from a column
     *
     * Returns bool(false) on error.
     *
     * {@internal
     * Will always return false if $cacheResults is set to false.
     * }}
     *
     * @access  public
     * @return  bool
     * @since   2.9.5
     * @ignore
     */
    function unsetFilter($column = '*')
    {
        assert('is_string($column); // Wrong argument type argument 1. String expected');
        $column = strtolower($column);
        if (!$this->cacheResults) {
            return false;

        } elseif ($column === '*') {
            if (isset($_SESSION['where'][$this->table][$this->templateId])) {
                unset($_SESSION['where'][$this->table][$this->templateId]);
            }
            return true;

        } elseif (isset($_SESSION['where'][$this->table][$this->templateId][$column])) {
            unset($_SESSION['where'][$this->table][$this->templateId][$column]);
            return true;

        } else {
            return false;

        }
    }

    /**
     * get where clause from session vars
     *
     * @access  private
     * @param   array  $where
     * @return  array
     */
    function _whereFromSession($where = array())
    {
        assert('is_array($where); // Wrong argument type argument 1. Array expected');
        if (!isset($_SESSION['where'][$this->table]) || !is_array($_SESSION['where'][$this->table])) {
            $_SESSION['where'][$this->table] = array($this->templateId => array());
        } elseif (!isset($_SESSION['where'][$this->table][$this->templateId]) || !is_array($_SESSION['where'][$this->table][$this->templateId])) {
            $_SESSION['where'][$this->table][$this->templateId] = array();
        }

        assert('!isset($column); // Cannot redeclare var $column');
        assert('!isset($clause); // Cannot redeclare var $clause');
        foreach ($_SESSION['where'][$this->table][$this->templateId] as $column => $clause)
        {
            if (is_array($clause) && isset($clause[0]) && $clause[1]) {
                $where[] = array($column, $clause[0], $clause[1]);
            }
        }
        unset($column, $clause); /* clean up garbage */

        return $where;
    }

    /**
     * move where clause to session vars
     *
     * @access  private
     * @param   array  $where
     * @param   bool   $type
     */
    function _whereToSession($where, $type = false)
    {
        assert('is_array($where); // Wrong argument type argument 1. Array expected');
        assert('is_bool($type); // Wrong argument type argument 2. Boolean expected');
        if (!is_array($where)) {
            return;
        }

        if (!isset($_SESSION['where'][$this->table]) || !is_array($_SESSION['where'][$this->table])) {
            $_SESSION['where'][$this->table] = array($this->templateId => array());
        } elseif (!isset($_SESSION['where'][$this->table][$this->templateId]) || !is_array($_SESSION['where'][$this->table][$this->templateId])) {
            $_SESSION['where'][$this->table][$this->templateId] = array();
        }

        if ($this->cacheResults) {
            if ($type) {
                assert('!isset($column); // Cannot redeclare var $column');
                assert('!isset($clause); // Cannot redeclare var $clause');
                foreach ($where as $column => $clause)
                {
                    if ($clause === '') {
                        unset($_SESSION['where'][$this->table][$this->templateId][$column]);

                    } elseif (is_string($clause)) {
                        $clause = preg_replace('/[^\w\*\?]/', '?', $clause);
                        if (strpos($clause, '*') !== false || strpos($clause, '?') !== false) {
                            $clause = str_replace('%', '_', $clause);
                            $clause = str_replace('*', '%', $clause);
                            $clause = str_replace('?', '_', $clause);
                            $_SESSION['where'][$this->table][$this->templateId][$column] = array($clause, 'like');
                        } else {
                            $_SESSION['where'][$this->table][$this->templateId][$column] = array($clause, '=');
                        }

                    } else {
                        /* error - invalid argument type */
                    }
                }
                unset($column, $clause); /* clean up garbage */
            } else {
                assert('!isset($clause); // Cannot redeclare var $clause');
                foreach ($where as $clause)
                {
                    if (is_array($clause) && isset($clause[0]) && isset($clause[1]) && isset($clause[2])) {
                        $_SESSION['where'][$this->table][$this->templateId][$clause[0]] = array($clause[1], $clause[2]);
                    }
                }
                unset($clause);
            } /* end if */
        } /* end if */
    }

    /**
     * resolve foreign keys
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  private
     * @return  bool
     */
    function _getForeignKeys()
    {
        /* get a list of all foreign keys */
        if (is_object($this->db->structure)) {
            $foreignKeys = $this->db->structure->getForeignKeys($this->table);
            if (empty($foreignKeys)) {
                /* no foreign keys found */
                return true;
            }
        } else {
            /* error: structure not initialized */
            return false;
        }

        /* loop through foreign keys */
        foreach ($foreignKeys as $key => $foreignTable)
        {
            /* if is foreign key of type select ... */
            if (strcasecmp($this->db->structure->getType($this->table, $key), 'select') == 0) {

                /* get default values */
                assert('!isset($default); /* Cannot redeclare var $default */');
                $default = $this->db->structure->getDefault($this->table, $key);

                /* 1) if default is not set */
                if (is_null($default)) {
                    assert('!isset($defaultList); /* Cannot redeclare var $defaultList */');
                    $defaultList = $this->db->get("${foreign_table}.*.${key}");
                    if (empty($defaultList)) {
                        unset($default, $defaultList);
                        continue;
                    } else {
                        $default = array();
                        assert('!isset($defaultValue);');
                        foreach ($defaultList as $defaultValue)
                        {
                            assert('is_scalar($defaultValue); // unexpected value: $defaultValue');
                            $default[$defaultValue] = $defaultValue;
                        }
                        unset($defaultList, $defaultValue);
                        $this->db->structure->setDefault($this->table, $key, $default);
                        unset($default);
                    }

                /* 2) or default is an array of table references */
                } elseif (is_array($default) && count($default) === 1) {

                    assert('!isset($defaultKeys); /* Cannot redeclare var $defaultKeys */');
                    assert('!isset($defaultValues); /* Cannot redeclare var $defaultValues */');

                    if ($this->db->structure->isColumn($foreignTable, key($default))) {
                        $defaultKeys = $this->db->get($foreignTable.".*.".key($default), array(), array($this->db->structure->getPrimaryKey($foreignTable)));
                    } else {
                        $defaultKeys = array();
                        trigger_error("The foreign key '${key}' in table '${table}' references a non-existing column '".key($default)."' in table '${foreignTable}'.",E_USER_ERROR);
                    }
                    if ($this->db->structure->isColumn($foreignTable, current($default))) {
                        $defaultValues = $this->db->get($foreignTable.".*.".current($default), array(), array($this->db->structure->getPrimaryKey($foreignTable)));
                    } else {
                        $defaultValues = array();
                        trigger_error("The foreign key '${key}' in table '${table}' references a non-existing column '".current($default)."' in table '${foreignTable}'.",E_USER_ERROR);
                    }
                    /* check unique constraint */
                    if (count($defaultKeys) !== count(array_unique($defaultKeys))) {
                        trigger_error("The column '".key($default)."' in table '${foreignTable}' referenced as foreign key from table '${table}' contains ambigious values. Possible reasons are that your data is inconsistent or your database scheme contains errors.", E_USER_WARNING);
                    /* check if result is empty */
                    } elseif (count($defaultKeys) === 0 && count($defaultValues) === 0) {
                        $this->db->structure->setDefault($this->table, $key, array());
                        unset($default, $defaultKeys, $defaultValues);
                        continue;
                    }
                    assert('is_array($defaultKeys) && is_array($defaultValues) && count($defaultKeys) === count($defaultValues);');

                    $default = array_combine($defaultKeys, $defaultValues);
                    assert('is_array($default); // Unexpected result: $default is not an array');
                    $this->db->structure->setDefault($this->table, $key, $default);
                    unset($default, $defaultKeys, $defaultValues);

                /* 3) ignore empty values */
                } elseif (empty($default)) {

                    continue;

                /* 4) error: encountered an unexpected value */
                } else {

                    return false;

                }
                unset($foreignTable, $foreignKeys);
            } /* end if */
        } /* end foreach */
        unset($key, $foreignTable);
        /* success */
        return true;
    }

    /**
     * get result rows
     *
     * This function returns a reference to the rows that will be displayed.
     * You may use this to modify the result before it is send to the browser.
     *
     * Example:
     * <code>
     * $rows =& $form->getRows();
     * foreach ($rows as $i => $row)
     * {
     *     $rows[$i]['FOO'] = $row['BAR'] . "\n" . $row['FOO'];
     *     unset($rows[$i]['BAR']);
     * }
     * </code>
     *
     * Returns an empty array on error.
     *
     * @access  public
     * @return  array
     * @since   2.9.6
     */
    function &getRows()
    {
        if (!is_string($this->table)) {
            trigger_error("Unable to retrieve resultset. No table has been selected.", E_USER_WARNING);
            $rows = array();
            return $rows;
        } elseif (!is_object($this->db)) {
            trigger_error("Unable to retrieve resultset. No database connection available.", E_USER_WARNING);
            $rows = array();
            return $rows;
        } elseif (($rows = $this->template->get('ROWS')) && !empty($rows)) {
            $this->template->insertByReference('ROWS', $rows);
            return $rows;
        }

        /**
         * get last page
         */
        if (empty($this->where)) {
            $this->lastPage = $this->db->length($this->table);
        } else {
            $this->lastPage = $this->db->length($this->table, $this->where);
        }
        if (!is_int($this->lastPage)) {
            /* error: e.g. table does not exist */
            $this->lastPage = 0;
            $rows = array();
            return $rows;
        }

        /* make sure the user can't go further than the last page */
        if ($this->page >= $this->lastPage) {
            if ($this->lastPage > 0) {
                $this->page = (int) $this->lastPage - 1;
            } else {
                $this->page = 0;
            }
        }

        /**
         * get resultset
         */
        $rows = $this->db->get($this->table, $this->where, $this->orderBy, $this->page, $this->entries, $this->desc);

        if (empty($rows)) {
            $this->currentIndex = 0;
            $rows = array();
            return $rows;
        } else {
            $this->currentIndex = array_shift(array_keys($rows));
            /* resolve page number */
            if (!empty($this->where)) {
                assert('!isset($temp); // Cannot redeclare variable $test');
                $temp = $this->db->get($this->table.'.*.'.$this->db->structure->getPrimaryKey($this->table), $this->where, $this->orderBy, 0, 0, $this->desc);
                assert('is_array($temp); // Unexpected result: query did not match anything');
                if (is_array($temp)) {
                    $temp = array_search($this->currentIndex, $temp);
                    assert('is_int($temp); // Unexpected result: current index not found in resultset');
                    if ($temp === false) {
                        $temp = 0;
                    }
                    if ($this->cacheResults) {
                        $_SESSION['page'][$this->table][$this->templateId] = $temp;
                    }    
                    $this->page = $temp;
                }
                unset($temp);
            }
        }

        /**
         * get offset
         */
        $length = count($rows);
        if ($length > 0) {
            $offsetPage = $this->page + $length;
            $this->template->insert('OFFSET_PAGE', $offsetPage);
            if ($this->firstPage > $offsetPage) {
                $this->firstPage = $offsetPage;
            }
        } else {
            $this->template->insert('OFFSET_PAGE', 0);
        }

        /**
         * get foreign keys
         *
         * (not for edit forms)
         */
        if ($this->templateId !== FORM_EDIT) {
            assert('!isset($foreignKeys); // Cannot redeclare variable $foreignKeys');
            $foreignKeys = $this->db->structure->getForeignKeys($this->table);
            if (is_array($foreignKeys)) {
                assert('!isset($foreignKey); /* Cannot redeclare var $foreignKey */');
                assert('!isset($foreignTable); /* Cannot redeclare var $foreignTable */');
                foreach ($foreignKeys as $foreignKey => $foreignTable)
                {
                    $foreignKey = strtoupper($foreignKey);
                    assert('!isset($default); /* Cannot redeclare var $default */');
                    $default = $this->db->structure->getDefault($this->table, $foreignKey);
                    assert('!isset($id); /* Cannot redeclare var $id */');
                    assert('!isset($row); /* Cannot redeclare var $row */');
                    foreach ($rows as $id => $row)
                    {
                        if (isset($rows[$id][$foreignKey]) && isset($default[strtoupper($row[$foreignKey])])) {
                            $this->refColumns[strtolower($id)][strtolower($foreignKey)] = $rows[$id][$foreignKey];
                            $rows[$id][$foreignKey] = $default[strtoupper($row[$foreignKey])];
                        }
                    }
                    unset($id, $row, $default);
                }
                unset($foreignKey, $foreignTable);
            }
        }

        /**
         * prepare files and images
         *
         * @todo  review this section and find a better way to do this
         */
        $columnList = $this->db->structure->getFiles($this->table);
        if (!empty($columnList)) {
            global $YANA;
            assert('!isset($dbBlob); /* Cannot redeclare var $dbBlob */');
            $dbBlob = $YANA->getVar('DBBLOB');
            assert('!isset($a); /* Cannot redeclare var $a */');
            assert('!isset($b); /* Cannot redeclare var $b */');
            foreach ($rows as $a => $b)
            {
                assert('!isset($column); /* Cannot redeclare var $column */');
                foreach ($columnList as $column)
                {
                    $column = strtoupper($column);
                    if (!empty($b[$column])) {
                        $string = new StringObject($b[$column]);
                        $string->replace($dbBlob, '');
                        /* applies to images only: use thumbnails instead of original image */
                        $string->replace('image.', 'thumb.');
                        /* encrypt path */
                        $string->encrypt('xor', YANA_INSTANCE_ID . session_id());
                        /* unbox string */
                        $string = $string->get();

                        /* add file to SESSION-vars */
                        if (!isset($_SESSION['filelist']) || !is_array($_SESSION['filelist'])) {
                            $_SESSION['filelist'] = array();
                            $i = false;
                        } else {
                            $i = array_search($string, $_SESSION['filelist']);
                        }
                        if ($i === false) {
                            $i = count($_SESSION['filelist']) + 1;
                            $_SESSION['filelist'][$i] = $string;
                        }
                        $rows[$a][$column] = $i;
                    }
                } /* end foreach */
                unset($column);
            } /* end foreach */
            unset($a, $b);
        } /* end if */

        $this->template->insertByReference('ROWS', $rows);
        return $rows;
    }

    /**
     * get data from posted form
     *
     * This function can be used to collect the data submitted
     * by a form, which was created using the Yana form generator.
     *
     * It returns an array containing the converted form data or
     * bool(false) on error.
     *
     * If the argument $data is not provided, the function automatically
     * takes it from the $_POST array instead.
     *
     * Example of usage:
     * <code>
     * $args = FormCreator::getFormdata(FORM_NEW, "guestbook", "guestbook");
     * </code>
     *
     * @access  public
     * @static
     * @param   int                 $form     choose the form, the data came from (FORM_NEW, FORM_EDIT, FORM_SEARCH, FORM_ADV_SEARCH)
     * @param   string|DbStructure  $source   name of the database structure file to use
     * @param   string              $table    name of the database table, the data is related to
     * @param   array               $data     provide the form data
     * @param   array               $columns  you may limit the results to these columns (if not provided, all columns are returned)
     * @return  array|bool(false)
     */
    function getFormdata($form, $source, $table, $data = null, $columns = null)
    {
        global $YANA;
        /**
         * Input checking
         */
        if (!is_int($form)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($form)), E_USER_WARNING);
            return false;
        }
        if (!is_string($source) && !is_object($source)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($source)), E_USER_WARNING);
            return false;
        }
        if (!is_string($table)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($table)), E_USER_WARNING);
            return false;
        }
        if (!is_array($data) && !is_null($data)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 4, 'Array', gettype($data)), E_USER_WARNING);
            return false;
        }
        if (!is_array($columns) && !is_null($columns)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 5, 'Array', gettype($columns)), E_USER_WARNING);
            return false;
        }

        /**
         * 1) get data
         */
        if (is_null($data)) {
            $data = $YANA->getRequestVar('*', YANA_POST);
        } else {
            $data = Hashtable::changeCase($data, CASE_LOWER);
        }

        /**
         * 1.1) non-standard case: use edit-form to create new entry or copy existing entry
         */
        if ($form === FORM_NEW && isset($data['selected_entries']) && $YANA->getRequestVar('namespace') === 'edit') {
            $form = FORM_EDIT;
            $isSaveAs = true;
        } else {
            $isSaveAs = false;
        }

        /**
         * 2) load structure file
         */
        if (is_string($source)) {

            $structure = new DbStructure($source);

            /* check if file exists */
            if (!$structure->exists()) {
                trigger_error("Invalid structure file '${source}'.", E_USER_WARNING);
                return false;
            } else {
                $structure->read();
            }

        } elseif (!is_a($source, 'DbStructure')) {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Instance of DbStructure', get_class($source)), E_USER_WARNING);
            return false;

        } else {

            $structure =& $source;
            $structure->read();

        }

        /**
         * 3) check if table exists
         */
        if (!$structure->isTable($table)) {
            trigger_error("Invalid table name '${table}'.", E_USER_WARNING);
            return false;
        }

        /* declare $result */
        assert('!isset($result); // Cannot redeclare $result');
        $result = array(); /* used to store returned vallue */

        /* declare $prefixes */
        assert('!isset($prefixes); // Cannot redeclare $prefixes');
        $prefixes = array("");

        /**
         * 4) get list of selected entries
         */
        switch ($form)
        {
            case FORM_EDIT:

                /* check if any entries have been selected */
                if (isset($data['selected_entries'])) {

                    /* declare $selectedEntries */
                    assert('!isset($selectedEntries); // Cannot redeclare $selectedEntries');
                    $selectedEntries = $data['selected_entries'];
                    settype($selectedEntries, 'array');

                    /* fill $result and $prefixes */
                    assert('!isset($entry); // Cannot redeclare $entry');
                    $prefixes = array();
                    foreach ($selectedEntries as $entry)
                    {
                        if (is_scalar($entry)) {
                            $entry = strtolower($entry);
                            $result[$entry] = array();
                            $prefixes[$entry] = "id_${entry}_";
                        }
                    }
                    unset($entry, $selectedEntries);

                /* error - no entry selected */
                } else {
                    return array();
                }

            /* fall through*/

            case FORM_NEW:
                /* avoid junk entries */
                if ($YANA->getVar("PROFILE.SPAM.CAPTCHA") === true && !$YANA->handle("security_check_image", $data)) {
                    $YANA->report(new Alert('INPUT_INVALID', 'Security check failed!'));
                    $YANA->report(new Log('SPAM: CAPTCHA not solved, entry has not been created.'));
                    return false;
                }
                /* lowercase file array */
                if (isset($_FILES) && is_array($_FILES)) {
                    $FILES = Hashtable::changeCase($_FILES, CASE_LOWER);
                } else {
                    $FILES = array();
                }
            break;
            
            default:
                $prefixes = array("");
                $FILES = array();
            break;
        } /* end switch */

        /**
         * 5) get columns
         */
        if (!is_array($columns) || count($columns) === 0) {
            $columns = $structure->getColumns($table);
        }

        /**
         * 6) iterate through columns
         */
        assert('!isset($column); // Cannot redeclare $column');
        foreach ($columns as $column)
        {
            $column = strtolower($column);
            $type = strtolower($structure->getType($table, $column));

            /**
             * 5.1) skip certain columns
             */
            switch ($form)
            {
                case FORM_SEARCH:
                    /* if column is visible (in search form) */
                    if ($type === 'text') {
                        if (!empty($data[$column])) {
                            $result[] = array($column, '%' . $data[$column] . '%', 'like');
                            $_SESSION['searchform'][$table][$column] = $data[$column];
                        }
                    } /* end if */
                    continue 2;
                break;

                case FORM_ADV_SEARCH:
                    /* if column is not selected */
                    if (isset($data[$column])) {
                        switch ($type)
                        {
                            case 'bool': case 'boolean':
                                if ($data[$column] !== '*') {
                                    $result[] = array($column, ($data[$column] == "true" || $data[$column] == "1"), '=');
                                }
                                $_SESSION['searchform'][$table][$column] = $data[$column];
                            break;

                            /**
                             * The next passage applies to columns of type TIME
                             * only. To search for a date / time a range needs to be
                             * created.
                             *
                             * The following code will create such a range when necessary.
                             */
                            case 'time':
                                assert('!isset($time); // Cannot redeclare variable $time');
                                assert('!isset($timestamp); // Cannot redeclare variable $timestamp');
                                $_SESSION['searchform'][$table][$column] = array();

                                /**
                                 * 1) start time
                                 */
                                $time = array(
                                    'Y' => (int) @$data["${column}_start_year"],
                                    'M' => (int) @$data["${column}_start_month"],
                                    'D' => (int) @$data["${column}_start_day"],
                                    'h' => (int) @$data["${column}_start_hour"],
                                    'm' => (int) @$data["${column}_start_minute"],
                                    's' => (int) @$data["${column}_start_second"]
                                );
                                $_SESSION['searchform'][$table][$column][0] = $time;

                                /**
                                 * 1.a) create filter rule for start time
                                 */
                                if (isset($data[$column]) && $data[$column] === "true") {
                                    assert('!isset($timestamp); // Cannot redeclare var $timestamp');
                                    $timestamp = mktime($time['h'], $time['m'], $time['s'], $time['M'], $time['D'], $time['Y']);
                                    if (!empty($timestamp) && is_int($timestamp)) {
                                        assert('is_array($result); /* unexpected input $result */');
                                        $result[] = array($column, $timestamp, '>=');
                                    }
                                    unset($timestamp);
                                }

                                /**
                                 * 2) end time
                                 */
                                $time = array(
                                    'Y' => (int) @$data["${column}_end_year"],
                                    'M' => (int) @$data["${column}_end_month"],
                                    'D' => (int) @$data["${column}_end_day"],
                                    'h' => (int) @$data["${column}_end_hour"],
                                    'm' => (int) @$data["${column}_end_minute"],
                                    's' => (int) @$data["${column}_end_second"]
                                );
                                $_SESSION['searchform'][$table][$column][1] = $time;

                                /**
                                 * 2.a) create filter rule for end time
                                 *
                                 * Note:
                                 * to create a valid range, the hours, minutes and
                                 * seconds need to be set to the end of the day.
                                 *
                                 * This does not apply to the start time, because
                                 * the default values are int(0), which happens to
                                 * be interpreted as midnight (start of the day).
                                 * So we will leave these to the defaults.
                                 */
                                if (empty($time['h'])) {
                                    $time['h'] = 23;
                                }
                                if (empty($time['m'])) {
                                    $time['m'] = 59;
                                }
                                if (empty($time['s'])) {
                                    $time['s'] = 59;
                                }

                                if (isset($data[$column]) && $data[$column] === "true") {
                                    assert('!isset($timestamp); // Cannot redeclare var $timestamp');
                                    $timestamp = mktime($time['h'], $time['m'], $time['s'], $time['M'], $time['D'], $time['Y']);
                                    if (!empty($timestamp) && is_int($timestamp)) {
                                        assert('is_array($result); /* unexpected input $result */');
                                        $result[] = array($column, $timestamp, '<=');
                                    }
                                    unset($timestamp);
                                }

                                /* success */
                                unset($time, $timestamp);
                            break;

                            case 'select':
                                if ($data[$column] !== "") {
                                    $result[] = array($column, $data[$column], '=');
                                }
                                $_SESSION['searchform'][$table][$column] = $data[$column];
                            break;

                            /* for other create match */
                            default:
                                if (!empty($data[$column])) {
                                    assert('!isset($value); // Cannot redeclare variable $value');
                                    $value = $data[$column];
                                    $value = str_replace('%', '_', $value);
                                    $value = str_replace(' ', '%', $value);
                                    $value = str_replace('?', '_', $value);
                                    $value = str_replace('*', '%', $value);
                                    $value = '%' . $value . '%';
                                    $result[] = array($column, $value, 'like');
                                    unset($value);
                                }
                                $_SESSION['searchform'][$table][$column] = $data[$column];
                            break;
                        } /* end switch */
                    } /* end if */
                    continue 2;
                break;
                
                case FORM_NEW:

                    /* check if column can be edited */
                    if (!$structure->isVisible($table, $column, 'new')) {
                        continue 2;
                    } else {
                        $current = array(&$result);
                        $i = $column;
                    } /* end if */

                break;

                case FORM_EDIT:

                    /* check if column can be edited */
                    if (!$structure->isEditable($table, $column, 'edit')) {
                        continue 2;
                    } else {
                        $current =& $result;
                        $i = "";
                    } /* end if */

                break;

                default:
                    trigger_error("The selected form is not supported.\n\t\tPlease choose one of the following " .
                    "constants: FORM_NEW, FORM_EDIT, FORM_SEARCH, FORM_ADV_SEARCH.", E_USER_WARNING);
                    return false;
                break;
            } /* end switch */

            /* create result rows
             *
             * Note: FORM_NEW has only one row as
             * only one row can be added at a time
             */
            assert('!isset($id); // Cannot redeclare $id');
            foreach (array_keys($current) as $id)
            {
                /* initialize where necessary */
                if (!isset($current[$id]) || !is_array($current[$id])) {
                    $current[$id] = array();
                }

                /* get form name */
                $i = $prefixes[$id] . $column;

                /* initialize where necessary */
                switch ($type)
                {
                    case 'boolean': case 'bool':
                        if (isset($data[$i]) && $data[$i]) {
                            $current[$id][$column] = true;
                        } else {
                            $current[$id][$column] = false;
                        }
                    break;

                    case 'time':
                        if (isset($data[$i.'_day'])) {
                            $current[$id][$column] = SmartUtility::toTimestamp($i.'_', $data);
                        }
                    break;

                    case 'image': case 'file':
                        if (!empty($FILES[$i]['name'])) {
                            $current[$id][$column] = $i;

                        } elseif (!empty($data[$i.'_delete'])) {
                            $current[$id][$column] = -1;
                        }
                    break;

                    case 'array':
                        /* get names and values */
                        assert('!isset($names); // Cannot redeclare $names');
                        if (isset($data[$i.'_names']) && is_array($data[$i.'_names'])) {
                            $names = $data[$i.'_names'];
                        } else {
                            $names = array();
                        }
                        assert('!isset($values); // Cannot redeclare $values');
                        if (isset($data[$i.'_values']) && is_array($data[$i.'_values'])) {
                            $values = $data[$i.'_values'];
                        } else {
                            $values = array();
                        }
                        if (empty($names)) {
                            unset($names, $values);
                            continue;
                        }

                        /* build array */
                        assert('!isset($param); // Cannot redeclare $param');
                        $param = array();
                        if (count($names) === count($values)) {
                            $param = array();
                            assert('!isset($j); // Cannot redeclare $j');
                            for ($j = 0; $j < count($names); $j++)
                            {
                                if ($names[$j] !== '') {
                                    $values[$j] = untaintInput($values[$j], "string", 255, YANA_ESCAPE_LINEBREAK);
                                    Hashtable::set($param, $names[$j], $values[$j]);
                                }
                            }
                            unset($j);
                        }
                        unset($names, $values);

                        if (empty($param) && $form !== FORM_EDIT) {
                            unset($param);
                            continue;
                        }

                        $current[$id][$column] = $param;
                        unset($param);
                    break;

                    case 'set':
                        if (isset($data[$i]) && is_array($data[$i])) {
                            assert('!isset($default); // Cannot redeclare var $default');
                            $default = $structure->getDefault($table, $column);
                            if (is_array($default)) {
                                $current[$id][$column] = array();
                                assert('!isset($element); // Cannot redeclare var $element');
                                foreach ($data[$i] as $element)
                                {
                                    if (isset($default[$element])) {
                                        $current[$id][$column][] = $element;
                                    }
                                } /* end foreach */
                                unset($element); /* clean up garbage */
                            }
                            unset($default);
                        }
                    break;

                    case 'text':
                        if (isset($data[$i])) {
                            $current[$id][$column] = untaintInput($data[$i], "string", 0, YANA_ESCAPE_USERTEXT, true);
                        }
                    break;

                    case 'int':
                    case 'integer':
                        if (isset($data[$i]) && $data[$i] !== "") {
                            if (is_numeric($data[$i])) {
                                $current[$id][$column] = untaintInput($data[$i], 'int');
                            } else {
                                $current[$id][$column] = untaintInput($data[$i], 'string', 0, YANA_ESCAPE_LINEBREAK);
                                $YANA->report( new Warning('INVALID_CHARACTERS', array('VALUE' => (string) $data[$i], 'FIELD' => $structure->getDescription($table, $column), 'VALID' => '0-9')) );
                            }

                        }
                    break;

                    case 'float':
                    case 'double':
                        if (isset($data[$i]) && $data[$i] !== "") {
                            if (preg_match('/^[\+\-]?[\d\.,]+$/s', $data[$i])) {
                                $current[$id][$column] = untaintInput($data[$i], 'double');
                            } else {
                                $current[$id][$column] = untaintInput($data[$i], 'string', 0, YANA_ESCAPE_LINEBREAK);
                                $YANA->report( new Warning('INVALID_CHARACTERS', array('VALUE' => (string) $data[$i], 'FIELD' => $structure->getDescription($table, $column), 'VALID' => '0-9 .')) );
                            }
                        }
                    break;

                    default:
                        if (isset($data[$i])) {
                            $current[$id][$column] = untaintInput($data[$i], "string", 0, YANA_ESCAPE_LINEBREAK);
                        }
                    break;
                } /* end switch */
            } /* end foreach */
            unset($id);

        } /* end foreach */
        unset($column);

        if ($isSaveAs === true) {
            return array_pop($result);
        } else {
            return $result;
        }
    }    

    /**
     * «smarty function» create
     *
     * This is a generator function to create dynamic HTML forms from database
     * schema files and database values.
     * Forms with several layouts can be generated to view, edit, insert, delete
     * or search database values.
     *
     * If necessary it communicates with the database to retrieve values.
     * This communication is limited to Select-Statements. No data is changed.
     *
     * For security reasons this function does NOT provide functionality to
     * write changes made in the forms to the database on it's own.
     * This might have introduced the possibility to compromise the security settings
     * of third-party plugIns and thus has never been implemented and never will be.
     *
     * Instead you need to provide the name of a function in your plugin,
     * that will handle the requests. This gives you the chance to do additional
     * security checks and filter the provided data as you see fit.
     *
     * <pre>
     * This function takes the following arguments:
     *
     * string  $template        (mandatory) template that should be used (view_details, view_seperated, view, new, edit, search)
     * string  $file            (mandatory) path and name of structure file
     * string  $table           (mandatory) table to choose from structure file
     * string  $show            (optional)  comma seperated list of columns, that should be shown in the form
     * string  $hide            (optional)  comma seperated list of columns, that should NOT be shown in the form
     * string  $where           (optional)  sequence for SQL-where clause <FIELDNAME1>=<VALUE1>[,<FIELDNAME2>=<VALUE2>[,...]]
     * string  $sort            (optional)  name of column to sort entries by
     * boolean $desc            (optional)  sort entries in descending (true) or ascending (false) order
     * integer $page            (optional)  number of 1st entry to show
     * integer $entries         (optional)  number of entries to show on each page
     * boolean $titles          (optional)  should fieldnames be shown (true=yes, false=no)
     * string  $on_new          (optional)  name of action (plugin-function) to activate in order to compute the event
     * string  $on_edit         (optional)  name of action (plugin-function) to activate in order to compute the event
     * string  $on_delete       (optional)  name of action (plugin-function) to activate in order to compute the event
     * string  $on_search       (optional)  name of action (plugin-function) to activate in order to compute the event
     * string  $on_download     (optional)  name of action (plugin-function) to activate in order to compute the event
     * int     $layout          (optional)  where multiple layouts are available to present the result, this allows to choose the prefered one
     * boolean $advanced_search (optional)
     * </pre>
     *
     * Example of usage:
     * <code>
     * [%create template="view" layout="1" file="guestbook" table="guestbook" sort="guestbook_date" desc="true"%]
     * </code>
     *
     * @access  public
     * @static
     * @param   array   $params
     * @return  string|bool(false)
     *
     * @ignore
     */
    function create($params)
    {
        global $YANA;

        /* type checking */
        if (!is_array($params)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($params)), E_USER_WARNING);
            return false;
        }

        /**
         * create new FormCreator-instance
         */
        $GENERATOR = new FormCreator();

        /**
         * 1) Build database query
         */

        /**
         * 1.1) parameter 'file'
         *
         * This parameter is mandatory.
         */
        if (isset($params['file']) && is_string($params['file'])) {
            if (!$GENERATOR->setFile($params['file'])) {
                /* error - structure file not found */
                return false;
            }
        } else {
            /* error - this parameter is mandatory */
            return false;
        }

        /**
         * 1.2) parameter 'table'
         *
         * This parameter is mandatory.
         */
        if (isset($params['table']) && is_string($params['table'])) {
            if (!$GENERATOR->setTable($params['table'])) {
                /* error - table unknown */
                return false;
            }
        } else {
            /* error - this parameter is mandatory */
            return false;
        }

        /**
         * 1.3) parameter 'advanced_search'
         */
        if (isset($params['advanced_search']) && $params['advanced_search'] == 1) {
            $GENERATOR->enableAdvancedSearch(true);
        } else {
            $GENERATOR->enableAdvancedSearch(false);
        }

        /**
         * 1.4) parameters 'template', 'layout'
         */
        if (isset($params['layout'])) {
            if (!$GENERATOR->setTemplate(@$params['template'], (int) $params['layout'])) {
                /* error: template + layout was invalid */
                return false;
            }
        } else {
            if (!$GENERATOR->setTemplate(@$params['template'], 0)) {
                /* error: template was invalid */
                return false;
            }
        }
        unset($template);

        /**
         * 1.5) parameters 'sort', 'desc'
         *
         * set 'orderBy' - clause
         */
        if (isset($params['sort'])) {
            if (!$GENERATOR->setSort($params['sort'], !empty($params['desc']))) {
                /* error: parameter was invalid */
                return false;
            }
        } else {
            if (!$GENERATOR->setSort('', !empty($params['desc']))) {
                /* error: parameter was invalid */
                return false;
            }
        }

        /**
         * 1.6) set event handler functions
         */
        if (isset($params['on_edit'])) {
            if (!$GENERATOR->setEditAction($params['on_edit'])) {
                return false;
            }
        }
        if (isset($params['on_delete'])) {
            if (!$GENERATOR->setDeleteAction($params['on_delete'])) {
                return false;
            }
        }
        if (isset($params['on_new'])) {
            if (!$GENERATOR->setNewAction($params['on_new'])) {
                return false;
            }
        }
        if (isset($params['on_search'])) {
            if (!$GENERATOR->setSearchAction($params['on_search'])) {
                return false;
            }
        }
        if (isset($params['on_download'])) {
            if (!$GENERATOR->setDownloadAction($params['on_download'])) {
                return false;
            }
        }

        /**
         * 1.7) parameter 'titles'
         */
        if (isset($params['titles'])) {
            if ($params['titles'] === 'false' || $params['titles'] === false) {
                $GENERATOR->enableTitles(false);
            } else {
                $GENERATOR->enableTitles(true);
            }
        }

        /**
         * 1.8) parameter 'keys'
         */
        if (isset($params['keys'])) {
            if ($params['keys'] === 'false' || $params['keys'] === false) {
                $GENERATOR->enableArrayKeys(false);
            } else {
                $GENERATOR->enableArrayKeys(true);
            }
        }

        /**
         * 1.9) parameters 'show', 'hide'
         */
        if (isset($params['show'])) {
            if (!is_array($params['show'])) {
                $showColumns = explode(',', $params['show']);
            } else {
                $showColumns = $params['show'];
            }
        } else {
            $showColumns = false;
        }
        if (isset($params['hide'])) {
            if (!is_array($params['hide'])) {
                $hideColumns = explode(',', $params['hide']);
            } else {
                $hideColumns = $params['hide'];
            }
        } else {
            $hideColumns = false;
        }
        /**
         * 1.10) set columns
         */
        if (empty($hideColumns)) {
            if (empty($showColumns)) {
                $GENERATOR->setColumns();
            } else {
                $GENERATOR->setColumns($showColumns);
            }
        } else {
            if (empty($showColumns)) {
                $showColumns = $GENERATOR->db->structure->getColumns($GENERATOR->table);
                $GENERATOR->setColumns(array_diff($showColumns, $hideColumns));
            } else {
                $GENERATOR->setColumns(array_diff($showColumns, $hideColumns));
            }
        }
        unset($showColumns, $hideColumns);

        /**
         * 1.11) parameter 'page'
         */
        if (isset($params['page'])) {
            if (!$GENERATOR->setPage($params['page'])) {
                return false;
            }
        }

        /**
         * 1.12) parameter 'entries'
         */
        if (isset($params['entries'])) {
            if (!$GENERATOR->setEntriesPerPage($params['entries'])) {
                return false;
            }
        }

        /**
         * 1.13) parameter 'where'
         */
        if (!empty($params['where'])) {
            if (!$GENERATOR->setWhere(@$params['where'])) {
                return false;
            }
        }

        return $GENERATOR->createForm();
    }

    /**
     * create a form from the current instance and return it
     *
     * Returns the HTML-code for this form.
     *
     * @access  public
     * @return  string
     */
    function createForm()
    {
        if (!is_object($this->template) || empty($this->table) || empty($this->templateId)) {
            trigger_error("Unable to create form. Don't have all required data.", E_USER_WARNING);
            return "";
        }

        /**
         * initialize table's properties
         */
        if ($this->getSort() === false) {
            $this->setSort();
        }
        if ($this->getPage() === false) {
            $this->setPage();
        }
        if ($this->getEntriesPerPage() === false) {
            $this->setEntriesPerPage();
        }
        if ($this->getWhere() === false) {
            $this->setWhere();
        }


        /**
         * create references
         */
        $this->template->insertByReference('ON_NEW',            $this->newAction);
        $this->template->insertByReference('ON_EDIT',           $this->editAction);
        $this->template->insertByReference('ON_DELETE',         $this->deleteAction);
        $this->template->insertByReference('ON_SEARCH',         $this->searchAction);
        $this->template->insertByReference('TABLE',             $this->table);
        $this->template->insertByReference('SHOW_FIELD_TITLES', $this->titles);
        $this->template->insertByReference('SORT',              $this->orderBy);
        $this->template->insertByReference('DESC',              $this->desc);
        $this->template->insertByReference('PAGE',              $this->page);
        $this->template->insertByReference('FIRST_PAGE',        $this->firstPage);
        $this->template->insertByReference('LAST_PAGE',         $this->lastPage);
        $this->template->insertByReference('ENTRIES',           $this->entries);
        $this->template->insertByReference('CURRENT_INDEX',     $this->currentIndex);
        $this->template->insertByReference('FORM',              $this);
        $this->template->insertByReference('DB',                $this->db->structure);
        $this->template->insert('PRIMARY_KEY', strtoupper($this->db->structure->getPrimaryKey($this->getTable())));
        if (isset($GLOBALS['YANA'])) {
            $this->template->insert('SELECTED_LANGUAGE',        $GLOBALS['YANA']->language->getSelectedLanguage());
        } else {
            $this->template->insert('SELECTED_LANGUAGE',        'en');
        }

        $this->_getForeignKeys();

        switch ($this->templateId)
        {
            case FORM_SEARCH:
                /**
                 * preselect hitlist template, by the number of entries per page
                 */
                if ($this->isAdvSearch === true) {
                    $this->template->setPath($this->fileMap[FORM_SEARCH][1]);
                } else {
                    $this->template->setPath($this->fileMap[FORM_SEARCH][0]);
                }
                
                if (isset($this->fileMap[FORM_VIEW][$this->layoutId])) {
                    $this->template->insert('HITLIST_TEMPLATE', $this->fileMap[FORM_VIEW][$this->layoutId]);
                } elseif ($this->entries > 1) {
                    $this->template->insert('HITLIST_TEMPLATE', $this->fileMap[FORM_VIEW][0]);
                } else {
                    $this->template->insert('HITLIST_TEMPLATE', $this->fileMap[FORM_VIEW][1]);
                }
                /* refresh */
                $this->setWhere(array());
            /* fall through */
            case FORM_VIEW:
            case FORM_SEPERATED:
            case FORM_DETAILS:
            case FORM_EDIT:
                $this->getRows();
                $this->_getListOfEntries();
            /* fall through */

            case FORM_NEW:
                /* intentionally left blank */
            break;

            default:
                return "";
            break;
        } /* end switch */

        return $this->template->get(null, false);
    }

    /**
     * create links to other pages
     *
     * @access  private
     * @return  array
     */
    function _getListOfEntries()
    {
        if ($this->lastPage > ($this->entries * 2)) {
            $listOfEntries = array();
            $dots = false;

            $isTooLong = $this->lastPage > (10 * $this->entries);
            for ($k = 0; $k < ceil( $this->lastPage / $this->entries ); $k++)
            {
                /**
                 * if more than 10 pages exist and current page is not first page or last page
                 * and is not current page or previous or next 3 pages
                 */
                $isNearCurrent = (floor($this->page / $this->entries) - 3) < $k && $k < (floor($this->page / $this->entries) + 3);

                $isFirstOrLast = $k <= 1 || $k >= (ceil($this->lastPage / $this->entries) - 2 );

                if ($isTooLong && !$isFirstOrLast && !$isNearCurrent) {
                    /* this marks an elipsis */
                    if ($dots === false) {
                        $dots = true;
                        array_push($listOfEntries, false);
                    } else {
                        /* ignore this page */
                        continue;
                    }
                } else {
                    $first = ($k * $this->entries);
                    if (($k + 1) * $this->entries < $this->lastPage) {
                        $last = ($k + 1) * $this->entries;
                    } else {
                        $last = $this->lastPage;
                    }
                    assert('is_array($listOfEntries); /* unexpected input $listOfEntries */');
                    $listOfEntries[] = array('FIRST' => ($first + 1), 'LAST' => $last);
                    if ($isNearCurrent) {
                        $dots = false;
                    }
                }
            } /* end for */

            assert('is_array($listOfEntries); /* Unexpected output */');
            $this->template->insert('LIST_OF_ENTRIES', $listOfEntries);
        } else {
            /* only one page exists, so the result is empty */
            return array();
        }
    }

    /**
     * create a reference link (where available)
     *
     * Returns the HTML-code for this field.
     *
     * @access  public
     * @param   string  $column
     * @param   string  $value
     * @param   string  $id
     * @param   string  $primaryKey
     * @return  string
     *
     * @ignore
     */
    function createReference($column, $value, $id, $primaryKey)
    {
        if (isset($value) ) {
            $action = $this->db->structure->getAction($this->table, $column, $this->getNamespace());
            $linkText = $this->db->structure->getLinkText($this->table, $column, $this->getNamespace());
            $linkTitle = $this->db->structure->getTitle($this->table, $column, $this->getNamespace());

            if (!is_null($action)) {
                assert('!isset($class); // Cannot redeclare variable $class');
                assert('!isset($title); // Cannot redeclare variable $title');
                assert('!isset($href); // Cannot redeclare variable $href');
                assert('!isset($text); // Cannot redeclare variable $text');
                $class = 'class="gui_generator_int_link"';
                if (is_string($linkTitle)) {
                    $title = 'title="' . $linkTitle . '"';
                } else {
                    $linkTitle = '%LANGUAGE.BUTTON_OPEN%';
                    $title = 'title="' . $GLOBALS['YANA']->getVar('LANGUAGE.DB_ENTITY_LINK') . '"';
                }

                if (isset($id) && strcasecmp($column, $primaryKey) !== 0) {
                    $id = strtolower($id);
                    if (is_null($primaryKey)) {
                        $primaryKey = $this->db->structure->getPrimaryKey($this->table);
                    }
                    $primaryKey = strtolower($primaryKey);
                    if (isset($this->refColumns[$id][$primaryKey])) {
                        $id = $this->refColumns[$id][$primaryKey];
                    }
                    $target = "&target[$primaryKey]=$id";
                } else {
                    $target = "";
                }
                $column = strtolower($column);
                if (isset($this->refColumns[$id][$column])) {
                    $value = $this->refColumns[$id][$column];
                }
                $href = 'href="' . SmartUtility::url("action=${action}${target}&target[${column}]=${value}&replacewhere=1") . '"';
                if (is_string($linkText)) {
                    if (is_file($linkText)) {
                        $text  = '<img src="' . $linkText . '" alt="' . $linkTitle . '"/>';
                    } else {
                        $text = $linkText;
                    }
                } else {
                    $text  = '<img src="%DATADIR%symbol_key.gif" alt="' . $linkTitle . '"/>';
                }
                return "<a $class $title $href>$text</a>";
                unset($class, $title, $href, $text);
            } else {
                return "";
            }
        } else {
            return "";
        }
    }

    /**
     * create a field for a certain column
     *
     * Returns the HTML-code for this field.
     *
     * {@internal
     * This function creates HTML for special cases
     * where templates
     * }}
     *
     * @access  public
     * @param   string  $column
     * @param   string  $value
     * @param   string  $prefix
     * @param   bool    $ignoreText
     * @param   bool    $isEditable
     * @param   bool    $isSearch
     * @return  string
     *
     * @ignore
     */
    function createHTML($column, $value, $prefix = "", $ignoreText = false, $isEditable = null, $isSearch = false)
    {
        $name = "${prefix}${column}";
        $id = $this->createId($column, $prefix);
        $name = htmlentities($name);

        /**
         * {@internal
         *
         * Due to the differences in reference handling between PHP 4 and PHP 5,
         * PHP 4 might have lost the variable reference to the structure file
         * at this point. In this case the following code will restore the
         * reference.
         *
         * }}
         */
        if (!is_object($this->db->structure)) {
            if (is_object($this->db->structure)) {
                $this->db->structure = $this->db->structure;
            } elseif (is_object($this->template)) {
                $this->db->structure = $this->template->get('DB');
            } elseif (isset($this->table)) {
                $this->setFile($this->table);
            }
        }

        /**
         * Restore search arguments
         */
        if (empty($value) && $isSearch) {
            if (isset($_SESSION['searchform'][$this->table][$name])) {
                $value = $_SESSION['searchform'][$this->table][$name];
            } elseif (isset($_POST[$name])) {
                $value = $_POST[$name];
            } else {
                $value = $this->db->structure->getDefault($this->table, $column);
            }
            if (is_scalar($value)) {
                $value = htmlentities($value);
            }
        }

        /**
         * Check if column is editable
         */
        if (!is_bool($isEditable)) {
            $isEditable = $this->db->structure->isEditable($this->table, $column, $this->getNamespace());
        }

        /**
         * Switch by column's type
         */
        switch ($this->db->structure->getType($this->table, $column))
        {
            case 'text':
                if ($isEditable) {
                    $length = $this->db->structure->getLength($this->table, $column);
                    if (!$ignoreText) {
                        return '<textarea id="' . $id . '" name="' . $name . '" ' .
                        ( ($length > 0 ) ? 'onkeypress="if (yanaMaxLength) yanaMaxLength(this, ' . $length . ', event)"' : '' ) .
                        ( ($length > 2000 ) ? ' cols="30" rows="3">' : ' cols="20" rows="3">' ) .
                        $value . '</textarea>';
                    } else {
                        return '<span class="roundbtn">' .
                        '<input id="' . $id . '" name="' . $name . '" type="text" value="' . $value . '" ' .
                        ( ($length > 0 ) ? 'maxlength="' . $length . '"' : '' ) .
                        ( ($length < 22 ) ? 'size="' . $length . '"' : '' ) .
                        '/></span>';
                    }
                } else {
                    $result = SmartUtility::smilies(SmartUtility::embeddedTags($value));
                    if ($ignoreText) {
                        $result = strip_tags($result);
                        if (strlen($result) > 60) {
                            return substr($result, 0, 56) . ' ...';
                        } elseif (!empty($result)) {
                            return $result;
                        } else {
                            return '&nbsp;';
                        }
                    } elseif (!empty($result)) {
                        if (strlen($result) > 25) {
                            return '<div class="gui_generator_readonly_textarea">' .
                            $result . '</div>';
                        } else {
                            return $result;
                        }
                    } else {
                        return '&nbsp;';
                    }
                }
            break;
            case 'boolean': case 'bool':
                if ($isSearch) {
                    return
                    
                    '<label class="gui_generator_radio_none">' .
                    '<input class="gui_generator_radio" id="' . $id . '_none" name="' . $name . '" type="radio" value="*" ' .
                    ( ($value !== "true" && $value !== "false") ? 'checked="checked"' : '' ) . '/>'.
                    '%LANGUAGE.ANY%'.
                    '</label>' .

                    '<label class="gui_generator_radio_true">%LANGUAGE.YES%'.
                    '<input class="gui_generator_radio" id="' . $id . '_true" name="' . $name . '" type="radio" value="true" ' .
                    ( ($value === "true") ? 'checked="checked"' : '' ) . '/></label>' .

                    '<label class="gui_generator_radio_false">%LANGUAGE.NO%'.
                    '<input class="gui_generator_radio" id="' . $id . '_false" name="' . $name . '" type="radio" value="false" ' .
                    ( ($value === "false") ? 'checked="checked"' : '' ) . '/></label>';

                } elseif ($isEditable) {
                    return '<input id="' . $id . '" name="' . $name . '" type="checkbox" value="true" ' .
                    ( ($value) ? 'checked="checked"' : '' ) . '/>';

                } elseif ($value) {
                    return '<img class="gui_generator_bool" alt="true" src=\'%DATADIR%boolean_true.gif\'/>';

                } else {
                    return '<img class="gui_generator_bool" alt="false" src=\'%DATADIR%boolean_false.gif\'/>';

                }
            break;
            case 'set':
                if ($isEditable) {
                    $result = "";
                    $i = 0;

                    $default = $this->db->structure->getDefault($this->table, $column);

                    if (is_array($default)) {
                        if (is_array($value)) {
                            assert('!isset($optText); // Cannot redeclare variable $optText');
                            assert('!isset($optValue); // Cannot redeclare variable $optValue');
                            foreach ($default as $optValue => $optText)
                            {
                                $result .= '<label class="gui_generator_set"><input type="checkbox" id="' . $id . '_' . $i . '" name="' . $name . '[]" value="' . $optValue . '"';
                                if (in_array($optValue, $value)) {
                                    $result .= ' checked="checked"';
                                }
                                $result .= '/>' . $optText . '</label>';
                            }
                            unset($optText, $optValue);
                        } else {
                            assert('!isset($optText); // Cannot redeclare variable $optText');
                            assert('!isset($optValue); // Cannot redeclare variable $optValue');
                            foreach ($default as $optValue => $optText)
                            {
                                $result .= '<label class="gui_generator_set"><input type="checkbox" id="' . $id . '_' . $i++ . '" name="' .
                                $name . '[]" value="' . $optValue . '"/>' . $optText . '</label>';
                            }
                            unset($optText, $optValue);
                        }
    
                        return $result;

                    } else {
                        return "";

                    }

                } elseif (empty($value) || !is_array($value)) {
                    return '&ndash;';

                } else {

                    $default = $this->db->structure->getDefault($this->table, $column);
                    $result = "";
                    if (is_array($default)) {
                        assert('!isset($element); // Cannot redeclare var $element');
                        foreach ($value as $element)
                        {
                            if (isset($default[$element])) {
                                if ($result !== "") {
                                    $result .= ", ";
                                }
                                $result .= $default[$element];
                            }
                        } /* end foreach */
                        unset($element); /* clean up garbage */
                    }
                    return $result;
                }
            break;
            case 'select':
                if ($isEditable) {
                    $result = '<select id="' . $id . '" name="' . $name . '">';
                    if ($isSearch) {
                        $result .= '<option value="">%LANGUAGE.ANY%</option>';

                    } elseif ($this->db->structure->isNullable($this->table, $column)) {
                        $result .= '<option value="">%LANGUAGE.CHOOSE_OPTION%</option>';

                    }
                    $isForeignKey = $this->db->structure->isForeignKey($this->table, $column);

                    assert('!isset($optText); // Cannot redeclare variable $optText');
                    assert('!isset($optValue); // Cannot redeclare variable $optValue');
                    foreach ($this->db->structure->getDefault($this->table, $column) as $optValue => $optText)
                    {
                        if ($isForeignKey && is_scalar($value) && strcasecmp($optValue, $value) === 0) {
                            $result .= '<option value="' . $optValue . '" selected="selected">' . $optText . '</option>';
                        } elseif ($optValue == $value) {
                            $result .= '<option value="' . $optValue . '" selected="selected">' . $optText . '</option>';
                        } else {
                            $result .= '<option value="' . $optValue . '">' . $optText . '</option>';
                        }
                    }
                    unset($optText, $optValue);

                    $result .= '</select>';
                    return $result;
                } elseif ($value == "") {
                    return '&ndash;';
                } else {
                    $default = $this->db->structure->getDefault($this->table, $column);
                    if (isset($default[$value])) {
                        return $default[$value];
                    } else {
                        return $value;
                    }
                }
            break;
            case 'time':
                $result = "";

                assert('!isset($display); // $display already declared');
                $display = $this->db->structure->getDisplay($this->table, $column);
                assert('!isset($isDateWithTime); // $isDateWithTime already declared');
                $isDateWithTime = !empty($display['WITH_TIME']);
                unset($display);

                assert('!isset($time); // Cannot redeclare var $time');

                /**
                 * 1) get current day
                 */

                /**
                 * 1.1) use current timestamp if no value provided
                 */
                if (empty($value)) {
                    $time = time();
                } else {
                    $time = $value;
                }
                /**
                 * 1.2) create start + end dates
                 */
                if ($isSearch && isset($_SESSION['searchform'][$this->table][$name])) {
                    $selectedStartDay = (int) $_SESSION['searchform'][$this->table][$name][0]['D'];
                    $selectedStartMonth = (int) $_SESSION['searchform'][$this->table][$name][0]['M'];
                    $selectedStartYear = (int) $_SESSION['searchform'][$this->table][$name][0]['Y'];
                    $selectedEndDay = (int) $_SESSION['searchform'][$this->table][$name][1]['D'];
                    $selectedEndMonth = (int) $_SESSION['searchform'][$this->table][$name][1]['M'];
                    $selectedEndYear = (int) $_SESSION['searchform'][$this->table][$name][1]['Y'];

                } elseif ($isSearch && isset($_POST[$name . '_START_Day'])) {
                    $selectedStartDay = (int) $_POST[$name . '_START_Day'];
                    $selectedStartMonth = (int) $_POST[$name . '_START_Month'];
                    $selectedStartYear = (int) $_POST[$name . '_START_Year'];
                    $selectedEndDay = (int) $_POST[$name . '_END_Day'];
                    $selectedEndMonth = (int) $_POST[$name . '_END_Month'];
                    $selectedEndYear = (int) $_POST[$name . '_END_Year'];

                } else {
                    $selectedEndDay = $selectedStartDay = (int) date('j', $time);
                    $selectedEndMonth = $selectedStartMonth = (int) date('n', $time);
                    $selectedEndYear = $selectedStartYear = (int) date('Y', $time);

                }
                unset($time);

                /**
                 * 2) prepare values
                 */
                if ($isSearch || $isEditable) {
                    $startDays = "";
                    $endDays = "";
                    assert('!isset($i); // Cannot redeclare variable $i');
                    for ($i = 1; $i <= 31; $i++)
                    {
                        $startDays .= '<option value="' . $i .
                        ( ($i === $selectedStartDay) ? '" selected="selected">' : '">' ) .
                        ( ($i < 10) ? "0${i}" : $i ) .
                        '</option>';
                        if ($isSearch) {
                            $endDays .= '<option value="' . $i .
                            ( ($i === $selectedEndDay) ? '" selected="selected">' : '">' ) .
                            ( ($i < 10) ? "0${i}" : $i ) .
                            '</option>';
                        }
                    } /* end for */
                    $startMonths = "";
                    $endMonths = "";
                    for ($i = 1; $i <= 12; $i++)
                    {
                        $startMonths .= '<option value="' . $i .
                        ( ($i === $selectedStartMonth) ? '" selected="selected">' : '">' ) .
                        ( ($i < 10) ? "0${i}" : $i ) .
                        '</option>';
                        if ($isSearch) {
                            $endMonths .= '<option value="' . $i .
                            ( ($i === $selectedEndMonth) ? '" selected="selected">' : '">' ) .
                            ( ($i < 10) ? "0${i}" : $i ) .
                            '</option>';
                        }
                    } /* end for */
                    $startYears = "";
                    $endYears = "";
                    for ($i = $selectedStartYear - 5; $i < $selectedStartYear + 5; $i++)
                    {
                        $startYears .= '<option value="' . $i .
                        ( ($i === $selectedStartYear) ? '" selected="selected">' : '">' ) .
                        $i . '</option>';
                        if ($isSearch) {
                            $endYears .= '<option value="' . $i .
                            ( ($i === $selectedEndYear) ? '" selected="selected">' : '">' ) .
                            $i . '</option>';
                        }
                    } /* end for */
                    unset($i);
                }

                /**
                 * 3) create output for ...
                 */
                assert('!isset($calendar); // Cannot redeclare var $calendar');
                $calendar = $GLOBALS['YANA']->getVar('DATADIR') . 'calendar.gif';

                /**
                 * 3.1) Search form
                 */
                if ($isSearch) {
                    $result .= '<span class="gui_generator_time">' .
                    '<label class="gui_generator_radio_true">' .
                    '<input' . (($_POST[$name] !== "true") ? ' checked="checked"' : '' ) .
                    ' class="gui_generator_radio" type="radio" value="false" id="' . $id . '_radio_false"' .
                    ' name="' . $name . '"/>' .
                    '%LANGUAGE.ANY%' .
                    '</label>'.
                    '<input' . (($_POST[$name] === "true") ? ' checked="checked"' : '' ) .
                    ' class="gui_generator_radio" type="radio" value="true" id="' . $id . '_radio_true"' .
                    ' name="' . $name . '"/>' .
                    '<select id="' . $id . '_START_day" name="' . $name . '_START_Day">' . $startDays . '</select>' .
                    '<select id="' . $id . '_START_month" name="' . $name . '_START_Month">' . $startMonths . '</select>' .
                    '<select id="' . $id . '_START_year" name="' . $name . '_START_Year">' . $startYears . '</select>' .
                    '<script type="text/javascript">yanaAddCalendar("' . $calendar . '", "' . $id . '_START", ' . $selectedStartDay . ', ' . ($selectedStartMonth - 1) . ', ' . $selectedStartYear . ');</script>' .
                    '&nbsp;&ndash;&nbsp;' .
                    '<select id="' . $id . '_END_day" name="' . $name . '_END_Day">' . $endDays . '</select>' .
                    '<select id="' . $id . '_END_month" name="' . $name . '_END_Month">' . $endMonths . '</select>' .
                    '<select id="' . $id . '_END_year" name="' . $name . '_END_Year">' . $endYears . '</select>' .
                    '<script type="text/javascript">yanaAddCalendar("' . $calendar . '", "' . $id . '_END", ' . $selectedEndDay . ', ' . ($selectedEndMonth - 1) . ', ' . $selectedEndYear . ');</script>' .
                    '</span>';

                /**
                 * 3.2) Edit form
                 */
                } elseif ($isEditable) {
                    assert('!isset($hours); // $hours already declared');
                    $hours = "";
                    assert('!isset($minutes); // $minutes already declared');
                    $minutes = "";
                    assert('!isset($seconds); // $seconds already declared');
                    $seconds = "";

                    if ($isDateWithTime) {
                        assert('!isset($currentHour); // $currentHour already declared');
                        $currentHour = (int) date('G');
                        for ($i = 0; $i < 24; $i++)
                        {
                            $hours .= '<option value="' . $i . '"' . ($i === $currentHour ? ' selected="selected"' : '')  . '>' .
                                (($i < 10) ? '0' . $i : $i) . '</option>';
                        }
                        unset($i, $currentHour);
                        assert('!isset($currentMinute); // $currentMinute already declared');
                        $currentMinute = (int) date('i');
                        for ($i = 0; $i < 60; $i++)
                        {
                            $minutes .= '<option value="' . $i . '"' . ($i === $currentMinute ? ' selected="selected"' : '')  . '>' .
                                (($i < 10) ? '0' . $i : $i) . '</option>';
                            $seconds .= '<option value="' . $i . '">' . (($i < 10) ? '0' . $i : $i) . '</option>';
                        }
                        unset($i, $currentMinute);
                    }
                    $result .= '<span class="gui_generator_time">' .
                        '<select id="' . $id . '_day" name="' . $name . '_Day">' . $startDays . '</select>' .
                        '<select id="' . $id . '_month" name="' . $name . '_Month">' . $startMonths . '</select>' .
                        '<select id="' . $id . '_year" name="' . $name . '_Year">' . $startYears . '</select>' .
                        '<script type="text/javascript">yanaAddCalendar("' . $calendar . '", "' . $id . '", ' . $selectedStartDay . ', ' . ($selectedStartMonth - 1) . ', ' . $selectedStartYear . ');</script>' .
                        (
                            ($isDateWithTime) ?
                                '<select id="' . $id . '_hour" name="' . $name . '_Hour">' . $hours . '</select>' .
                                '<select id="' . $id . '_minute" name="' . $name . '_Minute">' . $minutes . '</select>' .
                                '<select id="' . $id . '_second" name="' . $name . '_Second">' . $seconds . '</select>'
                                : ''
                        ) .
                        '</span>';
                    unset($hours, $minutes, $seconds);

                /**
                 * 3.3) Catalog view
                 */
                } elseif ($value) {
                    $result = SmartUtility::date($value);

                /**
                 * 3.4) Empty cell
                 */
                } else {
                    $result = '&ndash;';

                }
                unset($calendar, $calendarScript);

                return $result;
            break;
            case 'image':
                if (!$isSearch) {
                    global $YANA;
                    $length = $this->db->structure->getLength($this->table, $column);
                    $result = '<div class="gui_generator_image">';
                    if (!empty($value) && $YANA->session->checkPermission(null, null, $this->downloadAction)) {
                        $result .= '<a onclick="return yanaGuiShowImage(this.href, \'%LANGUAGE.CLOSE%\')" href=' .
                        SmartUtility::href("action=" . $this->downloadAction . "&target=${value}&fullsize=true") . '>' .
                        '<img border="0" alt="Thumbnail" src=' .
                        SmartUtility::href("action=" . $this->downloadAction . "&target=${value}") . '/></a>';

                    } else {
                        $result .= '&nbsp;';
                    }
                    if ($isEditable) {
                        if ($length > 0) {
                            $result .= '<input type="hidden" name="MAX_FILE_SIZE" value="' . $length . '"/>';
                        }
                        $result .= '<span class="roundbtn"><input size="1" type="file" id="' . $id . '" name="' . $name . '"/></span>';
                        if (!empty($value) && $this->db->structure->isNullable($this->table, $column)) {
                            $result .= '<label class="gui_generator_file_delete"><input title="%LANGUAGE.BUTTON_DELETE_ONE%" type="checkbox" id="' . $id . '_delete" name="' . $name . '_delete"/>%LANGUAGE.BUTTON_DELETE_ONE%</label>';
                        }
                    }
                    $result .= '</div>';
                    return $result;
                } else {
                    return '&ndash;';
                }
            break;
            case 'file':
                if (!$isSearch) {
                    global $YANA;
                    $length = $this->db->structure->getLength($this->table, $column);
                    $result = '<div class="gui_generator_file_download">';
                    if (!empty($value) && $YANA->session->checkPermission(null, null, $this->downloadAction)) {
                        $result .= '<a title="%LANGUAGE.TITLE_DOWNLOAD%" href=' .
                        SmartUtility::href("action=" . $this->downloadAction . "&target=${value}") . '>' .
                        '<img border="0" alt="Download" src="%DATADIR%icon_download.gif"/></a>';

                    } else {
                        $result .= '<a name="img_' . $id . '"><img border="0" alt="" src="%DATADIR%blank.gif" ' .
                        'style="width: 21px; height: 1px;"/></a>';
                    }
                    if ($isEditable) {
                        if ($length > 0) {
                            $result .= '<input type="hidden" name="MAX_FILE_SIZE" value="' . $length . '"/>';
                        }
                        $result .= '<span class="roundbtn"><input size="1" type="file" id="' . $id . '" name="' . $name . '"/></span>';
                        if (!empty($value) && $this->db->structure->isNullable($this->table, $column)) {
                            $result .= '<label class="gui_generator_file_delete"><input title="%LANGUAGE.BUTTON_DELETE_ONE%" type="checkbox" id="' . $id . '_delete" name="' . $name . '_delete"/>%LANGUAGE.BUTTON_DELETE_ONE%</label>';
                        }
                    }
                    $result .= '&nbsp;</div>';
                    return $result;
                } else {
                    return '&ndash;';
                }
            break;
            case 'array':
                if ($isSearch) {
                    return '<span class="roundbtn"><input id="' . $id . '" name="' . $name . '" type="text" value="' . $value . '"/></span>';
                } elseif ($isEditable) {
                    $isNumericArray = !$this->arrayKeys || $this->db->structure->isNumericArray($this->table, $column);
                    /* template for new entries */
                    if (!$isNumericArray) {
                        $template = '<span class="roundbtn"><input size="5" type="text" name="' .$name . '_names[]" value="$key"/></span>' .
                        '&nbsp;=&nbsp;<span class="roundbtn"><input size="10" type="text" name="' . $name .
                        '_values[]" value="$value"/></span>' .
                        '<a href="javascript://yanaRemoveItem(this)" onclick="yanaRemoveItem(this)">' .
                        '<img src="%DATADIR%icon_delete.gif" alt="[X]"/>%LANGUAGE.REMOVE%</a>';
                    } else {
                        $template = '<input type="hidden" name="' .$name . '_names[]" value="$key"/>' .
                        '<span class="roundbtn"><input size="21" type="text" name="' . $name .
                        '_values[]" value="$value"/></span>' .
                        '<a href="javascript://yanaRemoveItem(this)" onclick="yanaRemoveItem(this)">' .
                        '<img src="%DATADIR%icon_delete.gif" alt="[X]"/>%LANGUAGE.REMOVE%</a>';
                    }

                    $result = '<div class="gui_generator_array">';
                    
                    /* list of entries*/
                    $result .= '<ol>' .
                    '<li id="' . $id . '" class="gui_generator_array_reference">' .
                    str_replace('$key', '', str_replace('$value', '', $template)) .
                    '</li>';

                    if (!empty($value) && is_array($value)) {
                        $template = '<li>' . $template . '</li>';
                        ksort($value);
                        $result .= SmartUtility::loop('key', 'value', $value, $template);
                    }
    
                    $result .= '</ol>';
    
                    /* link to add new entry */
                    if (!$isNumericArray) {
                        $result .= '<a href="javascript:yanaAddItem(\'' . $id . '\')" title="%LANGUAGE.BUTTON_NEW%">';
                    } else {
                        $result .= '<a href="javascript:yanaAddItemAutonumber(\'' . $id . '\')" title="%LANGUAGE.BUTTON_NEW%">';
                    }
                    $result .= '<img src="%DATADIR%icon_new.gif" alt="[*]"/>%LANGUAGE.BUTTON_NEW%</a>';

                    $result .= '</div>';
                    return $result;
                } elseif (!empty($value)) {
                    $isNumericArray = !$this->arrayKeys || $this->db->structure->isNumericArray($this->table, $column);
                    $result = '<div class="gui_generator_array">';
                    if ($isNumericArray) {
                        $result .= SmartUtility::printUL1($value, 2);
                    } else {
                        $result .= SmartUtility::printUL1($value);
                    }
                    $result .= '</div>';
                    return $result;
                } else {
                    return '&ndash;';
                }
            break;
            case 'url':
                if ($isEditable) {
                    $length = $this->db->structure->getLength($this->table, $column);
                    return '<span class="roundbtn">' .
                    '<input id="' . $id . '" name="' . $name . '" type="text" value="' . $value . '" ' .
                    ( ($length > 0 ) ? 'maxlength="' . $length . '"' : '' ) .
                    ( ($length < 22 ) ? 'size="' . $length . '"' : '' ) .
                    '/></span>';
                } elseif (empty($value)) {
                    return '&ndash;';
                } else  {
                    $class = 'class="gui_generator_ext_link"';
                    $onclick = 'onclick="return confirm(\'' . $GLOBALS['YANA']->getVar('LANGUAGE.CONFIRM_EXT_LINK') . '\')"';
                    $title = 'title="' . $GLOBALS['YANA']->getVar('LANGUAGE.EXT_LINK') . '"';
                    $target = 'target="_blank"';
                    $href = 'href="' . htmlentities($value) . '"';
                    if (strlen($value) > 80) {
                        $text = substr($value, 0, 76) . ' ...';
                    } else {
                        $text = $value;
                    }
                    return "<a $target $class $onclick $title $href>$text</a>";
                }
            break;
            case 'reference':
                return '&nbsp;';
            break;
            case 'float':
            case 'double':
                if ($isEditable) {
                    $length = (int) $this->db->structure->getLength($this->table, $column);
                    $precision = (int) $this->db->structure->getPrecision($this->table, $column);
                    $title = $this->db->structure->getDescription($this->table, $column) . ': ' .
                    (($length < 8) ? str_pad('', $length, '#') : '########') . '.' .
                    (($precision < 8) ? str_pad('', $precision, '#') : '########');
                    $length = $length + $precision + 1;
                    return '<span class="roundbtn">' .
                    '<input id="' . $id . '" name="' . $name . '" type="text" value="' . $value . '" ' .
                    ( ($length > 0 ) ? 'maxlength="' . $length . '"' : '' ) .
                    ( ($length < 22 ) ? ' size="' . $length . '"' : '' ) . ' title="' . $title . '"' .
                    '/></span>';
                } elseif (empty($value)) {
                    return '&ndash;';
                } else  {
                    if (strlen($value) > 80) {
                        return substr($value, 0, 76) . ' ...';
                    } else {
                        return $value;
                    }
                }
            break;
            default:
                if ($isEditable) {
                    $length = $this->db->structure->getLength($this->table, $column);
                    return '<span class="roundbtn">' .
                    '<input id="' . $id . '" name="' . $name . '" type="text" value="' . $value . '" ' .
                    ( ($length > 0 ) ? 'maxlength="' . $length . '"' : '' ) .
                    ( ($length < 22 ) ? ' size="' . $length . '"' : '' ) .
                    '/></span>';
                } elseif (empty($value)) {
                    return '&ndash;';
                } else  {
                    if (strlen($value) > 80) {
                        return substr($value, 0, 76) . ' ...';
                    } else {
                        return $value;
                    }
                }
            break;
        }
    }

    /**
     * create id
     *
     * Creates a unique identifier for given data
     *
     * @access  public
     * @param   string  $column
     * @param   string  $value
     * @param   string  $prefix
     * @return  string
     * @ignore
     */
    function createId($column, $prefix = "")
    {
        return htmlentities( "input_" . $this->getNamespace() . "_" . $prefix . $column );
    }

    /**
     * self test
     *
     * This function exists only for debugging purposes.
     * Do not uncomment in productive environments.
     *
     * @access  public
     * @static
     * @param   int  $nr  number of test to take
     * @return  bool
     * @ignore
     */
    function test($nr = 1)
    {
        switch ($nr)
        {
            case 1:
                $_SESSION['sort'] = array();
                $_SESSION['desc'] = array();
                $_SESSION['page'] = array();
                $_SESSION['where'] = array();
                $_SESSION['entries'] = array();
                $_SESSION['layout'] = array();

                $gui = new FormCreator();
                $gui->cacheResults = false;

                // setting / getting file
                $gui->setFile('GUESTBOOK');
                $test = $gui->getFile();
                if ($test !== "guestbook") {
                    return '"set/get file" test failed';
                }
                $gui->setTable('GUESTBOOK');

                // setting / getting table
                $test = $gui->getTable();
                if ($test !== "guestbook") {
                    return '"set/get table" test failed';
                }

                // setting / getting template
                $gui->setTemplate('edit');
                $test1 = $gui->getTemplate();
                $gui->setTemplate(FORM_EDIT);
                $test2 = $gui->getTemplate();
                if (!is_string($test1) || $test1 !== $test2) {
                    return '"set/get template" test failed';
                }

                // setting / getting action
                $gui->setAction('EDIT', 'SITEMAP');
                $test = $gui->getAction('EDIT');
                if ($test !== 'sitemap') {
                    return '"set/get action" test failed';
                }

                // setting / getting bool values
                $gui->enableTitles();
                $test1 = $gui->getTitles();
                if ($test1 !== true) {
                    return '"set/get titles" test failed';
                }

                // setting / getting sorting
                $test1 = $gui->setSort('foo');
                $gui->setSort('GUESTBOOK_name', true);
                $test2 = $gui->getSort();
                $test3 = $gui->isDescending();
                if ($test1 !== false || $test2 !== 'guestbook_name' || $test3 !== true) {
                    return '"set/get sorting" test failed';
                }

                // setting / getting page + entries
                $err = error_reporting(E_ALL & ~ E_USER_NOTICE);
                $gui->setPage(-1); // this should output an E_USER_NOTICE
                $gui->setEntriesPerPage(0); // this should output an E_USER_NOTICE
                error_reporting($err);
                $test1 = $gui->getPage();
                $gui->setPage(1);
                $gui->setEntriesPerPage(2);
                $test2 = $gui->getPage();
                $test3 = $gui->getEntriesPerPage();
                if ($test1 !== 0 || $test2 !== 1 || $test3 !== 2) {
                    return '"set/get page+entries" test failed';
                }

                // setting / getting where clause
                $gui->setWhere("a=b");
                $test1 = $gui->getWhere();
                $gui->setWhere(array());
                $test2 = $gui->getWhere();
                if ($test1 !== "a=b" || $test2 !== false) {
                    return '"set/get where" test failed';
                }

                // get foreign keys
                $gui->setFile('check');
                $gui->setTable('t');
                $test = $gui->_getForeignKeys();
                if ($test !== true) {
                    return '"get foreign keys" test failed';
                }

                return true;
            break;
        }
    } /* end of function */
}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
