<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * create and send mails from form data
 *
 * @access      public
 * @package     yana
 * @subpackage  mail
 */
class FormMailer extends Object
{

    /**#@+
     * @access public
     */

    /** @var string */  var $subject  = "";
    /** @var string */  var $headline = "";
    /** @var string */  var $footline = "";
    /** @var array  */  var $content  = array();

    /**#@-*/

    /**
     * constructor
     *
     * Creates a new instance of this class.
     */
    function __construct()
    {
        /* intentionally left blank */
    }

    /**
     * send an e-mail
     *
     * Sends the current contents to $recipient,
     * where $recipient should be a valid mail address.
     *
     * Returns bool(true) on success or bool(false) on error.
     *
     * @access  public
     * @param   string  $recipient  mail address
     * @return  bool
     */
    function send($recipient)
    {
        /**
         * Input checking
         */
        if (!is_string($recipient)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($recipient)), E_USER_WARNING);
            return false;
        }         

        assert('$recipient !== ""; /* Recipient cannot be empty */');
        /* settype to STRING */
        $recipient = (string) $recipient;

        if (count($this->content) > 0 && $recipient !== '') {

            /* untaint subject */
            assert('is_string($this->subject);');
            $this->subject = untaintInput($this->subject, "string", 128, YANA_ESCAPE_LINEBREAK);
            $this->subject = strip_tags($this->subject);
            $this->subject = preg_replace("/[^\w \(\)äÄüÜöÖß]/", "", $this->subject);
            /* untaint send mail */
            $test = Mailer::mail($recipient, "[MAILFORM] " . $this->subject, $this->_makeMail());
            assert('is_bool($test); // Unexpected result: $test. Boolean expected.');
            return $test;

        } else {
            return false;
        }

    }

    /**
     * makeMail
     *
     * @access  private
     * @return  string
     * @ignore
     */
    function _makeMail()
    {
        assert('is_string($this->headline);');
        $mail  = $this->headline;

        $mail .= "==========================================\n";
        assert('is_string($this->subject);');
        if (!empty($this->subject)) {
            $mail .= "  " . $this->subject . "\n";
            $mail .= "==========================================\n\n";
        }

        assert('is_array($this->content) || is_string($this->content);');
        /* settype to ARRAY */
        $this->content = (array) $this->content;
        assert('!isset($a); /* cannot redeclare variable $a */');
        assert('!isset($b); /* cannot redeclare variable $b */');
        foreach ($this->content as $a => $b)
        {
            /* 1) check if key is valid */
            if (!preg_match('/^[\w\döäüß\/_\- ]+$/si', $a)) {
                trigger_error("Invalid form data. Key '".$a."' is invalid. Only alpha-numeric characters may be used as form keys.", E_USER_NOTICE);
            /* 2) check if value has valid type */
            } elseif (is_scalar($b)) {
                $mail .= "    ".$a.":\t".$b."\n";
            } elseif (is_array($b)) {
                $mail .= "    ".$a.":\t".print_r($b, true)."\n";
            } else {
                trigger_error("Invalid form data in key '".$a."'. Expected a value of a scalar type or an array, found '".gettype($b)."' instead.", E_USER_NOTICE);
            }
        } /* end foreach */
        unset($a, $b);

        $mail .= "\n==========================================\n";
        $mail .= "Date:\t".date("d.m.Y H:i:s", time())."\n";
        if (isset($_SERVER['REMOTE_ADDR'])) {
            $mail .= "IP:\t".$_SERVER['REMOTE_ADDR']."\n";
        } else {
            $mail .= "IP:\tno IP available\n";
        }

        assert('is_string($this->footline);');
        $mail .= $this->footline;

        return $mail;
    }

    /**
     * create an automatically formated e.mail from an array of form data
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @static
     * @param   string  $recipient  mail address
     * @param   string  $subject    subject of the mail
     * @param   array   &$formdata  form-data submitted by the user
     * @return  bool
     * @since   2.8
     */
    function mail($recipient, $subject, &$formdata)
    {
        /**
         * Input checking
         */
        if (!is_string($recipient)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($recipient)), E_USER_WARNING);
            return false;
        }
        if (!is_string($subject)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($subject)), E_USER_WARNING);
            return false;
        }
        if (!is_array($formdata)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'Array', gettype($formdata)), E_USER_WARNING);
            return false;
        }        
        assert('!empty($recipient); /* Recipient cannot be empty */');

        unset($formdata['action']);

        $formMail = new formMailer();
        $formMail->subject = $subject;
        $formMail->content =& $formdata;

        $recipient = untaintInput($recipient, 'mail');

        if (!empty($recipient)) {
            $test = $formMail->send($recipient);
            assert('is_bool($test);');
            return $test;
        } else {
            trigger_error("Invalid recipient, the provided value is not a valid mail adress.", E_USER_NOTICE);
            return false;
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and their
     * attributes are equal.
     *
     * @access   public
     * @param    object $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject,'getClass') && $anotherObject->getClass() === $this->getClass()) {
            $attrs = array_keys(get_class_vars($this->getClass()));
            foreach ($attrs as $attribute)
            {
                if (!isset($this->$attribute) && !isset($anotherObject->$attribute)) {
                    continue;
                } elseif (!isset($this->$attribute) && isset($anotherObject->$attribute)) {
                    return false;
                } elseif (isset($this->$attribute) && !isset($anotherObject->$attribute)) {
                    return false;
                } elseif ($this->$attribute !== $anotherObject->$attribute) {
                    return false;
                }
            } /* end foreach */
            return true;
        } else {
            return false;
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
