<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «utility» Hashtable
 *
 * This is a static utility class to implement certain
 * operations on hashtables (associative arrays).
 *
 * Example for usage of the $key parameter:
 * <pre>
 * Array {
 *     ID1 => Array {
 *         ID2 => 'value'
 *     }
 * }
 * </pre>
 * To get the string 'value' from the hashtable above
 * use $key = 'ID1.ID2'
 *
 * The wildcard '*' may be used to refer to the hashtable
 * as a whole.
 *
 * @access      public
 * @static
 * @package     yana
 * @subpackage  utilities
 */
class Hashtable extends Utility
{
    /**
     * retrieve a value
     *
     * Finds the value identified by $key and returns it.
     * If the value is not found NULL is returned.
     *
     * @access  public
     * @static
     * @param   array   &$hash   associative array
     * @param   string  $key     address
     * @return  mixed
     */
    static function &get(&$hash, $key)
    {
        assert('is_array($hash); // Wrong argument type for argument 1. Array expected.');
        $result = null;

        if (!Hashtable::_checkKey($key)) {
            $result = null;
            return $result;
        }

        if ($key == '*') {
            return $hash;
        } else {
            $list_of_keys = explode(".", $key);
            assert('is_array($list_of_keys) && count($list_of_keys) > 0;');
            $a =& Hashtable::_get($hash, $list_of_keys);
            return $a;
        }
    }

    /**
     * _get
     *
     * recursively resolve key address
     *
     * @access  private
     * @static
     * @param   array  &$hash          associative array
     * @param   array  &$list_of_keys  list of array keys
     * @return  mixed
     * @ignore
     */
    static function &_get(&$hash, &$list_of_keys)
    {
        $key_name = array_shift($list_of_keys);
        if (is_array($hash) && isset($hash[$key_name]) && !is_null($hash[$key_name])) {
            $result = &$hash[$key_name];
            if (count($list_of_keys) == 0) {
                return $result;
            } else {
                $a =& Hashtable::_get($result, $list_of_keys);
                return $a;
            }
        } else {
            $result = null;
            return $result;
        }
    }

    /**
     * set an element by Reference
     *
     * Sets the element identified by $key
     * to $value by passing it's reference.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete hashtable with a non-array value.
     * It returns bool(true) otherwise.
     *
     * @access  public
     * @static
     * @name    Hashtable::setByReference()
     * @param   array   &$hash   associative array
     * @param   string  $key     address
     * @param   mixed   &$value  some new value
     * @return  bool
     * @see     Hashtable::set()
     * @since   2.8.5
     */
    static function setByReference(&$hash, $key, &$value)
    {
        assert('is_array($hash); // Wrong argument type for argument 1. Array expected.');

        if (!Hashtable::_checkKey($key)) {
            return false;
        }

        if ($key == '*') {
            if (is_array($value)) {
                $hash = &$value;
                return true;
            } else {
                trigger_error("Wrong argument type: only values of type array may be assigned to a Hashtable. Found '".gettype($value)."' instead.", E_USER_WARNING);
                return false;
            }
        } else {
            $list_of_keys = explode(".", $key);
            assert('is_array($list_of_keys) && count($list_of_keys) > 0;');
            $result = &$hash;
            for ($i = 0; $i < (count($list_of_keys) -1); $i++)
            {
                if (!isset($result[$list_of_keys[$i]]) || is_scalar($result[$list_of_keys[$i]])) {
                    $result[$list_of_keys[$i]] = array();
                }
                $result = &$result[$list_of_keys[$i]];
            }
            $result[$list_of_keys[count($list_of_keys)-1]] = &$value;
            return true;
        }
    }

    /**
     * set an element to a value
     *
     * Sets the element identified by $key
     * to $value.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete hashtable with a non-array value.
     * It returns bool(true) otherwise.
     *
     * @access  public
     * @static
     * @name    Hashtable::set()
     * @param   array   &$hash  associative array
     * @param   string  $key    address
     * @param   mixed   $value  some new value
     * @return  bool
     * @see     Hashtable::setByReference()
     */
    static function set(&$hash, $key, $value)
    {
        assert('is_array($hash); // Wrong argument type for argument 1. Array expected.');
        return Hashtable::setByReference($hash, $key, $value);
    }

    /**
     * set the data type of an element
     *
     * Set the data type of the element identified by $key
     * to $type.
     *
     * Returns bool(false) if the element is NULL or does not exist,
     * or the $type parameter is invalid. Returns bool(true) otherwise.
     *
     * @access  public
     * @static
     * @param   array   &$hash  associative array
     * @param   string  $key    address
     * @param   string  $type   data type
     * @return  bool
     */
    static function setType(&$hash, $key, $type)
    {
        assert('is_array($hash); // Wrong argument type for argument 1. Array expected.');
        assert('is_string($type) && !empty($type);');
        if (!is_string($type)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($type)), E_USER_WARNING);
            return false;
        } else {
            $field =& Hashtable::get($hash, $key);
            if (!is_null($field)) {
                return settype($field, $type);
            } else {
                return false;
            }
        }
    }

    /**
     * check if an element exists
     *
     * Returns bool(false) if the element identified by $key
     * can not be found in the given hashtable or it is NULL.
     * Returns bool(true) otherwise.
     *
     * @access  public
     * @static
     * @param   array   &$hash  associative array
     * @param   string  $key    address
     * @return  bool
     */
    static function exists(&$hash, $key)
    {
        assert('is_array($hash); // Wrong argument type for argument 1. Array expected.');
        return !is_null(Hashtable::get($hash, $key));
    }

    /**
     * remove an element
     *
     * Unsets the element identified by $key in the
     * hashtable. Returns bool(false) if the element
     * does not exist or the key is invalid.
     * Returns bool(true) otherwise.
     *
     * @access  public
     * @static
     * @param   array   &$hash  associative array
     * @param   string  $key    address
     * @return  bool
     */
    static function remove(&$hash, $key)
    {
        assert('is_array($hash); // Wrong argument type for argument 1. Array expected.');

        if (!Hashtable::_checkKey($key)) {
            return false;
        }

        if ($key == '*') {
            unset($hash);
            return true;
        } else {
            $list_of_keys = explode(".", $key);
            assert('is_array($list_of_keys) && count($list_of_keys) > 0;');
            $stack = array();
            $stack[0] = &$hash;
            $result = &$hash;
            for ($i = 0; $i < (count($list_of_keys) -1); $i++)
            {
                if (!isset($result[$list_of_keys[$i]])) {
                    return false;
                } else {
                    $result = &$result[$list_of_keys[$i]];
                    $stack[count($stack)] = &$result;
                }
            }
            $j = array_pop($list_of_keys);
            if (!isset($stack[count($stack)-1][$j])) {
                return false;
            } else {
                unset($stack[count($stack)-1][$j]);
                array_pop($stack);
                for ($i = (count($stack) -1); $i > -1; $i--)
                {
                    $j = array_pop($list_of_keys);
                    if (is_array($stack[$i][$j]) && count($stack[$i][$j])==0) {
                        unset($stack[$i][$j]);
                    } else {
                        break;
                    }
                }
                return true;
            }
        }
    }

    /**
     * Lowercase or uppercase all keys of an associative array
     *
     * This is a recursive implementation of the PHP function array_change_key_case().
     * It takes the same arguments: an array $input to work on and an optional
     * argument $case. The argument $case can be one of two constants: CASE_LOWER and
     * CASE_UPPER, where CASE_LOWER is the default.
     *
     * @access  public
     * @static
     * @param   array     $input  input array
     * @param   int|bool  $case   CASE_UPPER or CASE_LOWER
     * @return  array
     */
    static function changeCase($input, $case = CASE_LOWER)
    {
        assert('is_array($input); // Wrong argument type for argument 1. Array expected.');
        assert('is_bool($case) || $case === CASE_UPPER || $case === CASE_LOWER; // Wrong argument type for argument 2. Expected CASE_UPPER or CASE_LOWER.');

        /* argument 1 */

        if (!is_array($input)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in '".__CLASS__.'::'.__FUNCTION__."()", 'Array', gettype($input)), E_USER_WARNING);
            return $input;
        }

        /* argument 2 */

        /* Map boolean input to constant */
        if (is_bool($case)) {
            if ($case === true) {
                $case = CASE_UPPER;
            } else {
                $case = CASE_LOWER;
            }

        /* input is invalid */
        } elseif ($case !== CASE_UPPER && $case !== CASE_LOWER) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "2 in '".__CLASS__.'::'.__FUNCTION__."()", 'Boolean or CASE_LOWER, or CASE_UPPER', gettype($case)), E_USER_WARNING);
            return $input;
        }

        /* input handling */
        $input = array_change_key_case($input, $case);
        foreach ($input as $k => $e)
        {
            if (is_array($e)) {
                $input[$k] = Hashtable::changeCase($e, $case);
            }
        } /* end foreach */
        assert('is_array($input); // Unexpected result: $input. Array expected.');
        return $input;
    }

    /**
     * recursively merge two arrays to one
     *
     * This function is pretty much the same as the php function "array_merge_recursive"
     * except for the way how duplicate keys are treated. Dupplicate keys get replaced
     * in this implementation rather than being appended.
     *
     * You may provide two or more arrays to merge.
     *
     * Example with 2 arrays:
     * <code>
     * $foo = Hashtable::merge(array(0 => 'a', 'a' => 'a'), array(0 => 'b'), 1 => 'c');
     * </code>
     * Result:
     * <pre>
     * array (
     * 0 => 'b',
     * 'a' => 'a',
     * 1 => 'c'
     * )
     * </pre>
     *
     * Example with 4 arrays:
     * <code>
     * $bar = Hashtable::merge($a1, $a2, $a3, $a4);
     * </code>
     *
     * Recursive example:
     * <pre>
     * $a1 = array (
     * 0 => 1,
     * 1 => array(
     *   0 => 1,
     *   'a' => 'b'
     *   )
     * );
     * $b = array (
     * 1 => array(
     *   0 => 'c',
     *   1 => 2,
     *   )
     * );
     * 2 => 3,
     * </pre>
     * Compute this via:
     * <code>
     * $foobar = Hashtable::merge($a1, $a2);
     * </code>
     * Result:
     * <pre>
     * array (
     * 0 => 1,
     * 1 => array(
     *   0 => 'c',
     *   1 => 2,
     *   'a' => 'b'
     *   ),
     * 2 => 3
     * )
     * </pre>
     *
     * @access  public
     * @static
     * @param   array  $A  base array
     * @param   array  $B  merge with this array
     * @return  array
     */
    static function merge($A, $B)
    {
        assert('is_array($A); // Wrong argument type argument 1. Array expected');
        assert('is_array($B); // Wrong argument type argument 2. Array expected');

        /* for better performance check trivial cases first */
        if (count($B) == 0) {
            return $A;

        } elseif (count($A) == 0) {
            return $B;

        /* only if none of this applies, do the real thing */
        } else {
            assert('!isset($i); /* cannot redeclare variable $i */');
            for ($i = 1; $i < func_num_args(); $i++)
            {
                /* check argument $i */

                $array = func_get_arg($i);
                assert('is_array($array); // Wrong argument type argument ' . $i . '. Array expected');
                /* handle argument $i */
                foreach ($array as $k => $e)
                {
                    /* inserting an array */
                    if (is_array($e)) {
                        if (isset($A[$k]) && is_array($A[$k])) {
                            /* recursion */
                            $A[$k] = Hashtable::merge($A[$k], $e);
                        } else {
                            $A[$k] = $e;
                        }

                    /* inserting a scalar value */
                    } else {
                        $A[$k] = $e;
                    }
                } /* end foreach */
            } /* end for */
            assert('is_array($A); /* Unexpected result: $A. */');
            return $A;
        }
    }

    /**
     * _checkKey
     *
     *
     * Note: The following symbols are not allowed. $, <, >, ", ', `
     * Any whitespace character, except space (character \x20).
     * Any character exceeding the set of 128 basic symbols (above character \x7f).
     *
     * @access  private
     * @static
     * @param   string  $key
     * @return  bool
     * @ignore
     */
    static function _checkKey($key)
    {
        assert('is_string($key);');
        if (!is_string($key) || ($key != '*' && !preg_match("/^[\x20 \x21 \x23 \x25 \x26 \x28-\x3b \x3d \x3f-\x5f \x61-\x7e]+$/" ,$key))) {
            trigger_error("Syntax error: '".gettype($key)."(".$key.")' is not a valid hash-key.", E_USER_ERROR);
            return false;
        } else {
            return true;
        }
    }
}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
