<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «utility» controller
 *
 * this implements the main program
 *
 * Note that as of version 2.8.6 the framework also accepts calls
 * from the PHP command line interface.
 * This is intended to be used for cronjobs and maintenance.
 *
 * Example of usage (Windows):
 * <pre>php index.php action=test target=detailed >stdout.log</pre>
 *
 * When running PHP from the command line under UNIX, you need to
 * include the following line on top of the "index.php" file:
 * <pre>#!/usr/bin/php</pre>
 * This is to tell where the PHP binaries can be found.
 *
 * The path might be another on your server.
 * If you don't know the path, enter the following on your command line:
 * <pre>which php</pre>
 * This will output the path where the php binaries are installed.
 *
 * When running the program under Windows this is not necessary.
 * Instead you should change the system path to point to the
 * PHP installation directory, if you have'nt already done that.
 *
 * See the PHP manual for more details.
 *
 * @static
 * @access      public
 * @name        Index
 * @package     yana
 * @subpackage  core
 */
class Index extends Utility
{
    /**
     * YANA controller
     *
     * The Static function "main" is an YANA controller,
     * used to untaint user data and operate an YANA-instance.
     *
     * @static
     * @access  public
     * @uses    YANA::main()
     */
    static function main()
    {
        /* Set error reporting to 'off' by default */
        if (!defined('YANA_ERROR_REPORTING')) {
            ErrorUtility::setErrorReporting(YANA_ERROR_OFF);
        }
        global $YANA;

        /* differentiate between web interface and command line calls */
        if (defined('STDIN') && !isset($_SERVER['REQUEST_METHOD'])) {

            /* command line interface
             *
             * The CLI is for maintenance tasks - e.g. database backups aso.
             * triggered by cronjobs.
             */

            $args = array();
            foreach ($_SERVER['argv'] as $argument)
            {
                if (is_string($argument) && preg_match('/^([\w\d-_]*)=(.*)$/',$argument, $m)) {
                    $args[strtolower($m[1])] = $m[2];
                }
            }

            /* Get a new yana-instance */
            $YANA = new Yana("config/system.config", $args);
            /* Handle the request */
            $YANA->handle();

            /* Since this is expected to be used for cronjobs,
             * no (human readable) output is explicitely created here.
             * Instead the called code should create comprehensive log files,
             * which the administrator can review later.
             */

        } else {
            
            /* session preparation
             *
             * {@internal
             *
             * NOTE on session handling in PHP:
             *
             * As a default PHP marks a session as "expired" after 24 minutes,
             * BUT this does not mean the session data really gets deleted
             * immediatelly after this time!
             *
             * Session data only gets deleted if and only if it is marked
             * as expired AND the garbage collector is started.
             * Which MIGHT be (in best case) as soon as the data expires,
             * or (in worst case) never.
             *
             * The garbage collector is randomly started on an incoming request.
             * The propability in percent that the gc runs on a request
             * is calculated by (session.gc_probability / session.gc_divisor * 100)
             * Where gc_probability defaults to 1 and gc_divisor defaults
             * to 100, which actually results in 1% and means the gc
             * should run once each 100 requests.
             *
             * Of course this value does'nt fit all.
             * While it should work well on servers with frequent requests,
             * on a server with low traffic you may come back after days
             * or even weeks and find your session is still there.
             *
             * If you got low traffic you should consider increasing the value
             * of "session.gc_probability" from default 1 to a higher value
             * e.g. 5. Which actually means the garbage collector will autostart
             * each 5% of all incomming requests.
             *
             * Example: ini_set("session.gc_probability", "5");
             *
             * Don't use too high values either, as this might slow
             * down your server. E.g. running the gc on every request is
             * certainly a bad idea.
             *
             * Abbr: "gc" = garbage collector
             * }}
             */
            /* set session lifetime to 30 minutes */
            ini_set("session.cookie_lifetime", "1800");
            ini_set("session.gc_maxlifetime", "1800");
            session_name(YANA_SESSION_NAME);
            @session_start();

            /* Check language settings */
            if (isset($_GET['language'])) {
                $_SESSION['language'] = $_GET['language'];
                unset($_GET['language']);
            } elseif (!isset($_SESSION['language']) && isset($_SERVER['HTTP_ACCEPT_LANGUAGE'])) {
                $_SESSION['language'] = substr($_SERVER['HTTP_ACCEPT_LANGUAGE'], 0, 2);
            }
            /* {@internal
             *
             * NOTE:
             * PHP's policy on which vars to import in what order to the super-global
             * $_REQUEST-var could have been changed in php.ini. Thus it might contain
             * other values than expected. This work-around ensures the expected behaviour.
             *
             * }}
             */
            $ARGS = Hashtable::merge($_GET, $_POST);
            /* {@internal
             *
             * Things to know BEFORE enabling output compression:
             * 1) headers already sent (e.g. by an PHP message, or another script?)
             * 2) did the browser provide a list of accepted encoding?
             * 3) does the browser accept the desired encoding?
             * 4) you can't use 'ob_gzhandler' if 'zlib.output_compression' is already enabled in php.ini
             *
             * Things to know AFTER enabling output compression:
             * 1) flushes automatically on program termination
             * 2) buffer can be erased using ob_clean() and ob_end_clean()
             * 3) flush buffer to output device using ob_flush() and ob_end_flush()
             * 4) use ob_get_length() === false to check if buffer is disabled
             * 5) use ob_get_length() === 0 to check if buffer is empty
             *
             * }}
             */
            if (false && headers_sent() === false && !ini_get('zlib.output_compression')) {
                if (isset($_SERVER['HTTP_ACCEPT_ENCODING']) && strpos($_SERVER['HTTP_ACCEPT_ENCODING'],"gzip") !== false) {
                    /* output compression level is an integer between -1 (off/default) - 9 (max) */
                    ini_set('zlib.output_compression_level', 6);
                    ob_start("ob_gzhandler");
                }
            }
            /* Get a new yana-instance */
            $YANA = new Yana("config/system.config", $ARGS);
            if (isset($_SESSION['language']) && preg_match('/^[\w\d-_]+$/s', $_SESSION['language'])) {
                if (!$YANA->language->setDirectory($_SESSION['language'])) {
                    unset($_SESSION['language']);
                }
            }

            /* Handle the request */
            $YANA->handle();
            /* Create the output */
            $YANA->writeView();
            /* {@internal
             *
             * Though this should'nt be necessary, one would be
             * best adviced to flush the buffer
             *
             * }}
             */
            if (ob_get_length() !== false) {
                ob_end_flush();
            }
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
