<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «utility» Microsummary
 *
 * This class provides static methods to read and
 * write persistent microsummaries.
 *
 * "Microsummaries" are a Firefox 2.0 feature that allows users
 * to create dynamic bookmark titles that automatically update
 * when the content of the bookmarked page changes.
 *
 * Have a look at what microsummaries can be:
 * <ul>
 *   <li> the numbers of downloads of a file on a download site </li>
 *   <li> the latest news on a news page </li>
 *   <li> current number of unread e-mail in the inbox of a webmail service </li>
 *   <li> current total of donations to a project </li>
 *   <li> the date of latest updates on a database </li>
 *   <li> the latest submission to a guestbook or forum </li>
 *   <li> the number of visitors currently online in a chat room </li>
 *   <li> the latest stock values aso. </li>
 * </ul>
 *
 * Examples of usage:
 * <ol>
 *  <li> Setting a microsummary from a plugin:
 *       <code>Microsummary::set($this->name, 'Summary text');</code>
 *  </li>
 *  <li> Retrieving a microsummary in a plugin:
 *       <code>$microsummary = Microsummary::get($this->name);</code>
 *  </li>
 *  <li> To indicate that a microsummary exists for your plugin
 *       add this to your output function
 *       <code>Microsummary::publish($this->name);</code>
 *  </li>
 *  <li> Calling a microsummary from a browser:
 *       <code>index.php?action=get_microsummary&target=guestbook</code>
 *       (where 'guestbook' is the name of the plugin)
 *  </li>
 * </ol>
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 */
class Microsummary extends Utility
{

    /**
     * get a microsummary
     *
     * Reads the microsummary string identified by $id and returns it.
     * This function returns bool(false) if no corresponding summary exists.
     *
     * Example of usage:
     *
     * Retrieving a microsummary in a plugin:
     * <code>$microsummary = Microsummary::get($this->name);</code>
     *
     * @access  public
     * @param   string  $id  identifies the summary to get
     * @return  bool
     */
    function get($id)
    {
        global $YANA;
        if (!isset($YANA)) {
            return false;
        } elseif (!is_string($id) || strlen($id) > 32) {
            trigger_error("Wrong argument type for argument 1. String of at most 32 charaters expected.", E_USER_NOTICE);
            return false;
        } else {
            $id = strtoupper($id);            
            $db = $YANA->connect('microsummary');
            if (!empty($db)) {
                $result = $db->get("microsummary.$id.microsummary_text");
                if (!empty($result) && is_string($result)) {
                    return $result;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
    }

    /**
     * set a microsummary
     *
     * Saves the microsummary string identified by $id for later use.
     * This function returns bool(false) if the input is invalid or
     * the entry could not be saved.
     *
     * Example of usage:
     *
     * Retrieving a microsummary in a plugin:
     * <code>$microsummary = Microsummary::get($this->name);</code>
     *
     * @access  public
     * @param   string  $id    identifies the summary to get
     * @param   string  $text  the text of the microsummary
     * @return  bool
     */
    function set($id, $text)
    {
        global $YANA;
        if (!isset($YANA)) {
            return false;
        } elseif (!is_string($id) || strlen($id) > 32) {
            trigger_error("Wrong argument type for argument 1. String of at most 32 charaters expected.", E_USER_NOTICE);
            return false;
        } elseif (!is_string($text) || strlen($text) > 200) {
            trigger_error("Wrong argument type for argument 2. String of at most 200 charaters expected.", E_USER_NOTICE);
            return false;
        } else {
            $db = $YANA->connect('microsummary');
            $id = strtoupper($id);
            if (!empty($db)) {
                $result = $db->insertOrUpdate("microsummary.$id", array('microsummary_id' => $id, 'microsummary_text' => $text));
                if (!empty($result)) {
                    if ($db->write()) {
                        return true;
                    } else {
                        /* commit failed */
                        return false;
                    }
                } else {
                    /* insert failed */
                    return false;
                }
            } else {
                /* connection not available */
                return false;
            }
        }
    }

    /**
     * publish a microsummary
     *
     * Adds the microsummary identified by $id to the list of
     * microsummaries to be printed to the browser.
     *
     * Example of usage:
     *
     * Add this to your output function
     * <code>Microsummary::publish($this->name);</code>
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @param   string  $id  identifies the summary to get
     * @return  bool
     */
    static function publish($id)
    {
        global $YANA;
        if (!isset($YANA)) {
            return false;
        } elseif (!is_string($id) || strlen($id) > 32) {
            trigger_error("Wrong argument type for argument 1. String of at most 32 charaters expected.", E_USER_NOTICE);
            return false;
        } else {
            $id = strtoupper($id);
            $db = $YANA->connect('microsummary');
            if (!empty($db)) {
                if ($db->exists("microsummary.$id")) {
                    $microsummary = $YANA->getVar('MICROSUMMARY');
                    if (empty($microsummary)) {
                        $microsummary = array();
                    }
                    array_push($microsummary, $id);
                    array_unique($microsummary);
                    $YANA->setVar('MICROSUMMARY', $microsummary);
                    return true;
                } else {
                    return false;
                }
            } else {
                return false;
            }
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
