<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * Registry
 *
 * This class implements a registry.
 * You can operate on it via keys, identifying
 * scalar values or arrays inside the registry.
 *
 * Example for usage of the $key parameter:
 * <pre>
 * Array {
 *     ID1 => Array {
 *         ID2 => 'value'
 *     }
 * }
 * </pre>
 * To get the string 'value' from the array above
 * use $key = 'ID1.ID2'
 *
 * The wildcard '*' may be used to refer to the array
 * as a whole.
 *
 * @access     public
 * @package    yana
 * @subpackage vdrive
 */
class Registry extends VDrive
{
    /**#@+
     * @ignore
     * @access  private
     */

    /**
     * An index of references to improve performance when accessing values
     * @var array */ var $cache = array();

    /**#@-*/

    /**
     * retrieves var from registry
     *
     * This returns the var identified by $key.
     *
     * Note that you may use the following magic keys FOR THIS FUNCTION ONLY:
     *
     * <ul>
     *    <li> <BOOLEAN> returns an one-dimensional associative array of all values converted to boolean  </li>
     *    <li> <STRING>  returns an one-dimensional associative array of all values converted to string, arrays are mapped to bool(true) </li>
     *    <li> <ARRAY>   returns an array of all arrays, scalar values are mapped to an empty array </li>
     * </ul>
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @param   string  $key  (optional)
     * @return  mixed
     */
    function getVar($key = "*")
    {
        /* argument 1 */ assert('is_string($key); // Wrong argument type argument 1. String expected');
        /* settype to STRING */
        $key = (string) $key;

        assert('is_array($this->options); // Unexpected type for instance property "options". Array expected');
        /**
         * 1) return whole registry
         */
        if (empty($key) || $key === "*") {
            return $this->options;

        /**
         * 2) return value from index
         */
        } elseif (isset($this->cache[$key])) {
            if (is_null($this->cache[$key])) {
                return false;
            } else {
                return $this->cache[$key];
            }

        /**
         * 3) use magic keys
         */
        } elseif (strspn($key, '<') == 1) {
            switch (strtoupper($key))
            {

                case '<BOOLEAN>':
                    if (is_array($this->options)) {
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        foreach (array_keys($this->options) as $a)
                        {
                            $RESULT[$a] = true;
                        }
                        unset($a);
                    } else {
                        $RESULT = true;
                    }
                    return $RESULT;
                break;

                case '<STRING>':
                    if (is_array($this->options)) {
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        assert('!isset($b); /* cannot redeclare variable $b */');
                        foreach ($this->options as $a => $b)
                        {
                            if (!is_array($b)) {
                                $RESULT[$a] = $b;
                            } else {
                                $RESULT[$a] = true;
                            }
                        }
                        unset($a, $b);
                    } else {
                      $RESULT = $this->options;
                    }
                    return $RESULT;
                break;

                case '<ARRAY>':
                    if (is_array($this->options)) {
                        assert('!isset($a); /* cannot redeclare variable $a */');
                        assert('!isset($b); /* cannot redeclare variable $b */');
                        foreach ($this->options as $a => $b)
                        {
                            if (!is_array($b)) {
                                $RESULT[$a] = array();
                            } else {
                                $RESULT[$a] = $b;
                            }
                        }
                        unset($a, $b);
                    } else {
                      $RESULT = array();
                    }
                    return $RESULT;
                break;

                default:
                    return false;
                break;

            } /* end switch */

        /**
         * 4) return value specified by key
         */
        } else {
            
            $this->cache[$key] =& Hashtable::get($this->options, $key);
            if (is_null($this->cache[$key])) {
                return false;
            } else {
                return $this->cache[$key];
            }
        } /* end if */

    } /* end getVar */

    /**
     * retrieves var from registry and returns it by reference
     *
     * This returns the var identified by $key.
     * Returns null (not bool(false)) on error.
     *
     * Note: this function may return false but also
     * other values that evaluates to false.
     * To check for an error use: is_null($result).
     * To check for bool(false) use: $result === false.
     *
     * @access  public
     * @param   string  $key  (optional)
     * @return  mixed
     * @since   2.9.5
     */
    function &getVarByReference($key = "*")
    {
        /**
         * 1) return all
         */
        if (empty($key) || $key === "*") {
            return $this->options;

        /**
         * 2) return value from index
         */
        } elseif (isset($this->cache[$key])) {
            return $this->cache[$key];

        /**
         * 3) return value by key
         */
        } else {
            /* returns NULL on error */
            $this->cache[$key] =& Hashtable::get($this->options, $key);
            return $this->cache[$key];

        }
    }

    /**
     * sets var on registry by Reference
     *
     * Sets the element identified by $key
     * to $value by passing it's reference.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete registry with a non-array value.
     * It returns bool(true) otherwise.
     *
     * If $overwrite is set to false and the $key does already exist,
     * the current value is not replaced and the function returns
     * bool(false).
     *
     * @access  public
     * @param   string  $key
     * @param   mixed   &$value
     * @param   bool    $overwrite
     * @return  bool
     * @since   2.8.5
     */
    function setVarByReference($key, &$value, $overwrite = true)
    {
        /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');
        /* settype to STRING */
        $key = (string) $key;

        if ($key == "" || $key == "*") {

            $this->options =& $value;
            $this->cache[$key] = array();
            return true;

        } elseif ($overwrite || !Hashtable::exists($this->options, $key)) {

            if (Hashtable::setByReference($this->options, $key, $value)) {
                $this->cache[$key] =& $value;
                return true;

            } else {
                return false;
            }

        } else {
            return false;

        }

    }

    /**
     * sets var on registry
     *
     * Sets the element identified by $key
     * to $value.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete registry with a non-array value.
     * It returns bool(true) otherwise.
     *
     * If $overwrite is set to false and the $key does already exist,
     * the current value is not replaced and the function returns
     * bool(false).
     *
     * @access  public
     * @param   string  $key        key of updated element
     * @param   mixed   $value      new value
     * @param   bool    $overwrite  true = replace, false = throw error
     * @return  bool
     */
    function setVar($key, $value, $overwrite = true)
    {
        /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

        return $this->setVarByReference($key, $value, $overwrite);
    }

    /**
     * merges the value at adresse $key with the provided array data
     *
     * Merges the element identified by $key with the array $value.
     *
     * Returns bool(false) on error.
     *
     * If $overwrite is set to false, then the values of keys that already exist are ignored.
     * Otherwise these values get updated to the new ones.
     *
     * @access  public
     * @param   string  $key        key of updated element
     * @param   array   $array      new value
     * @param   bool    $overwrite  true = update, false = ignore
     * @return  bool
     */
    function merge($key, $array, $overwrite = true)
    {
        /* argument 1 */ assert('is_string($key);  // Wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_array($array); // Wrong argument type for argument 2, array expected');

        /* settype to STRING
         *            ARRAY
         */
        $key   = (string) $key;
        $array = (array)  $array;

        if ($key == "" || $key == "*") {
            $this->options = Hashtable::merge($this->options, $array);
            $this->cache[$key] = array();
            return true;

        } elseif ($overwrite) {
            $options =& Hashtable::get($this->options, $key);
            if (is_null($options)) {
                return Hashtable::set($this->options, $key, $array);
            } else {
                $options = Hashtable::merge($options, $array);
                return true;
            }

        } else {
            $options =& Hashtable::get($this->options, $key);
            if (is_null($options)) {
                return Hashtable::set($this->options, $key, $array);
            } else {
                $array = array_diff_key($array, $options);
                $options = Hashtable::merge($options, $array);
                return true;
            }

        } /* end if */
    }

    /**
     * removes var from registry
     *
     * Unsets the element identified by $key in the
     * registry. Returns bool(false) if the element
     * does not exist or the key is invalid.
     * Returns bool(true) otherwise.
     *
     * @access  public
     * @param   string  $key
     * @return  bool
     */
    function unsetVar($key)
    {
        /* argument 1 */ assert('is_string($key);   // wrong argument type for argument 1, string expected');

        /* settype to STRING */
        $key = (string) $key;

        if ($key == "" || $key == "*") {
            $this->options = $this->attributes;
            return true;
        } else {
            return Hashtable::remove($this->options, $key);
        }

    }

    /**
     * sets the type of a var on registry
     *
     * Set the data type of the element identified by $key
     * to $type.
     *
     * Returns bool(false) if the element is NULL or does not exist,
     * or the $type parameter is invalid. Returns bool(true) otherwise.
     *
     * @access  public
     * @param   string  $key
     * @param   string  $type
     * @return  bool
     */
    function setType($key, $type)
    {
        /* argument 1 */ assert('is_string($key);   // wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_string($type);  // wrong argument type for argument 2, string expected');
        /* settype to STRING */
        $key  = (string) $key;
        $type = (string) $type;
        return Hashtable::setType($this->options, $key, $type);
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
