<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «abstract» Report
 *
 * This is an abstract super class for all kinds of messages.
 *
 * @abstract
 * @access      public
 * @package     yana
 * @subpackage  error_reporting
 */
class Report
{
    /**#@+
     * @access  public
     */

    /**
     * the message that should be reported
     * @var string */ var $message = "";

    /**
     * any kind of data that might help to understand the context in which the message was created
     * @var mixed */  var $data    = null;

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance, representing a system message.
     * To report this message to the framework, just call $YANA->report($report_object).
     *
     * @param  string  $message  the message that should be reported
     * @param  mixed   $data     any kind of data that might help to understand context in which the message was created
     */
    function __construct($message, $data = null)
    {
        /* argument 1 */ assert('is_string($message); // wrong argument type for argument 1, string expected');

        /* settype to STRING */
        $message = (string) $message;

        $this->message = $message;
        $this->data    = $data;
    }

    /**
     * create log
     *
     * This creates a log entry that is supposed to be written
     * to a log-file.
     *
     * The result is an associative array or bool(false) on error.
     *
     * @access public
     * @param  string $prefix string to put before each key string
     * @return string
     */
    function getLog($prefix = "")
    {
        /* 0. check input */
        if (!is_string($prefix)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($prefix)), E_USER_WARNING);
            return false;
        }
        /* nothing to report */
        if (empty($this->message) && empty($this->data)) {
            return false;
        }

        /* 1. initialize vars */
        global $YANA;
        $report = array();
        $prefix = strtoupper($prefix);

        /* 2. create entry */
        if (!empty($this->message)) {
            $report[$prefix.'MESSAGE']= $this->message;
        }
        if (!is_null($this->data)) {
            if (is_array($this->data)) {
                $report[$prefix.'DATA'] = $this->data;
            } else {
                $report[$prefix.'DATA'] = array($this->data);
            }
        }
        if (isset($YANA)) {
            $report[$prefix.'USER']   = $YANA->getVar("SESSION_USER_ID");
        }
        if (empty($report[$prefix.'USER'])) {
            $report[$prefix.'USER']   = '*GUEST';
        }
        if (isset($YANA)) {
            $report[$prefix.'ACTION'] = $YANA->getVar("ACTION");
        } elseif (isset($_GET['action']) && is_string($_GET['action']) && strlen($_GET['action']) < 100) {
            $report[$prefix.'ACTION'] = $_GET['action'];
        } else {
            $report[$prefix.'ACTION'] = "*default";
        }
        assert('is_array($report); // Unexpected result: $report should be an array.');
        return $report;
    }

    /**
     * returns the message string of this event
     *
     * This creates a string of at most 1000 characters
     * containing the text in the message attribute and a serialized
     * representation of the data attribute.
     *
     * @access public
     * @return string
     */
    function getMessage()
    {
        $message = $this->message;
        if (!empty($this->data)) {
            if (is_scalar($this->data)) {
                $message .= ";".$this->data;
            } elseif (is_array($this->data)) {
                $message .= ";".str_replace("\n", '', SML::encode($this->data));
            } else {
                $message .= ";".print_r($this->data, true);
            }
        }
        $message = str_replace('"', '&quot;',substr(preg_replace('/\s/', ' ', $message),0,1000));
        return $message;
    }

    /**
     * test for equality
     *
     * The parameter $otherLog needs to be an array
     * created by Report::getLog().
     *
     * Checks if the input array and the log created
     * from the current object are the same.
     *
     * The entries are considered identical if and
     * only if: the textual representation of
     * the "message" and "data" values are the
     * same.
     *
     * @access  public
     * @param   array   $otherLog
     * @param   string  $prefix  string to put before each key string
     * @return  bool
     * @ignore
     */
    function logEquals($otherLog, $prefix = "")
    {
        /* 0. check input */
        if (!is_array($otherLog)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($otherLog)), E_USER_NOTICE);
            return false;
        }
        if (!is_string($prefix)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($prefix)), E_USER_NOTICE);
            return false;
        }

        /* 1. initialize vars */
        $prefix = strtoupper($prefix);

        /* 2. get reference entry */
        $thisLog = $this->getLog();

        /* 3. convert data log */
        if (isset($thisLog['DATA'])) {
            settype($thisLog['DATA'], 'array');
        } else {
            $thisLog['DATA'] = null;
        }

        /* 4. compare strings and return result */
        if (print_r(@$thisLog['MESSAGE'], true) === print_r(@$otherLog[$prefix.'MESSAGE'], true) && strcasecmp(print_r(@$thisLog['DATA'], true), print_r(@$otherLog[$prefix.'DATA'], true)) === 0) {
            return true;
        } else {
            return false;
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
