<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «entity» RSS item
 *
 * This class represents an item of an RSS feed.
 *
 * Note: all attributes are optional, but at least one of title or description must be set.
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 */
class RSSitem extends Object
{
    /**#@+
     * this field is optional
     *
     * @access  public
     */

    /**
     * title of the item
     * @var string */ var $title       = "";
    /**
     * URL of the item
     * @var string */ var $link        = "";
    /**
     * item synopsis, note that this may contain HTML
     * @var string */ var $description = "";
    /**
     * Email address of the author
     * @var string */ var $author      = "";
    /**
     * include item in one or more categories
     * @var array  */ var $category    = array();
    /**
     * URL of a page for comments
     * @var string */ var $comments    = "";
    /**
     * indicates when the item was published (auto-generated if not provided)
     * @var string */ var $pubDate     = "";

    /**#@-*/
    /**#@+
     * this field is optional
     *
     * @access  private
     */

    /**
     * unique identity (always auto-generated)
     * @var string */ var $guid        = "";

    /**#@-*/

    /**
     * constructor
     *
     * creates and initializes a new instance of this class
     *
     */
    function __construct()
    {
        $this->pubDate = date('r',time());
    }
    
    /**
     * return item as associative array
     *
     * This checks the values of all attributes, creates
     * a unique id for this item, and returns
     * an associative array of all object vars.
     *
     * Returns an empty array and issues an E_USER_NOTICE on error.
     *
     * @access  public
     * @return  array
     */
    function get()
    {
        /* check optional fields */
        if (!is_string($this->title)) {
            trigger_error("Unable to create RSS item. Field 'title' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->link)) {
            trigger_error("Unable to create RSS item. Field 'link' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->description)) {
            trigger_error("Unable to create RSS item. Field 'description' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->author)) {
            trigger_error("Unable to create RSS item. Field 'author' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->author;
            $this->author = untaintInput($this->author, 'mail');
            if ($tmp != $this->author) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->author."'.", E_USER_NOTICE);
            }
        }
        if (!is_array($this->category)) {
            trigger_error("Unable to create RSS item. Field 'category' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->comments)) {
            trigger_error("Unable to create RSS item. Field 'comments' is invalid.", E_USER_NOTICE);
            return array();
        } else {
            $tmp = $this->comments;
            $this->comments = untaintInput($this->comments, 'url');
            if ($tmp != $this->comments) {
                trigger_error("RSS item has been auto-converted.\n\t\tThe input value '".$tmp."' did not seem to be valid.\n\t\tThe value has been changed to '".$this->comments."'.", E_USER_NOTICE);
            }
        }
        if (!is_string($this->guid)) {
            trigger_error("Unable to create RSS item. Field 'guid' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (!is_string($this->pubDate)) {
            trigger_error("Unable to create RSS item. Field 'pubDate' is invalid.", E_USER_NOTICE);
            return array();
        }
        if (empty($this->title) && empty($this->description)) {
            trigger_error("Unable to create RSS item. Need at least 'title' or 'description' to be present.", E_USER_NOTICE);
            return array();
        }

        /* auto-fill fields, that can be generated */
        $this->guid = md5($this->title.$this->description);

        /* return result */
        return get_object_vars($this);
    }

}

// © 2006, Thomas Meyer www.All-Community.de
// Copyright Hinweis nicht entfernen!
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
