<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 *
 * @ignore
 */

/**
 * «Singleton» SessionManager
 *
 * This is a manager class to handle user data and
 * permission levels.
 *
 * @access      public
 * @name        SessionManager
 * @package     yana
 * @subpackage  core
 *
 * @ignore
 */
class SessionManager extends Singleton
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var mixed   */ var $database    = null;
    /** @var mixed   */ var $defaultUser = array('GROUP_ID' => 0, 'ACTIVE' => true);
    /** @var string  */ var $profileId   = "";
    /** @var string  */ var $userId      = "";
    /** @var array   */ var $cache       = array();

    /**#@-*/

    /**
     * «Singleton» constructor
     *
     * creates a new instance of this class
     *
     * @param  DbStream  $database
     * @param  array     $user
     * @ignore
     */
    function __construct($database, $user)
    {
        $this->checkSingletonConstraint();
        $this->database = $database;
        if (is_array($user)) {
            $this->defaultUser = $user;
        }
    }

    /**
     * get current profile id
     *
     * Returns the id of the profile the
     * data of the current profile is to be associated with.
     *
     * @access  public
     * @return  string
     * @ignore
     */
    function getId()
    {
        if (empty($this->profileId)) {
            global $YANA;
            /**
             * {@internal
             * When switching from files to database, the behaviour of switching
             * from profile settings to id and then to a default value is lost,
             * since it was part of the file configuration.
             *
             * So this is why it has to be included here.
             * }}
             */
            if (isset($YANA)) {
                $this->profileId = $YANA->getVar('PROFILE.PROFILE_ID');
                if (empty($this->profileId)) {
                    $this->profileId = $YANA->getVar('ID');
                }
                if (empty($this->profileId)) {
                    $this->profileId = $YANA->getDefault('PROFILE');
                }
            }    
            if (empty($this->profileId)) {
                $this->profileId = 'default';
            }
        }
        return $this->profileId;
    }

    /**
     * check login data
     *
     * Returns bool(true) if the pair of user name and password
     * are correct an bool(false) otherwise.
     *
     * @access  public
     * @param   string  $userName
     * @param   string  $userPwd
     * @return  bool
     * @ignore
     */
    function checkLogin($userName, $userPwd)
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }
        if (!is_string($userPwd)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($userPwd)), E_USER_WARNING);
            return false;
        }

        $userName = strtoupper($userName);

        if (!$this->database->exists("USER.$userName")) {
            return false;
        }

        $currentPwd = sessionManager::_encode($userName, $userPwd);
        $savedPwd = $this->_getPassword($userName);

        if ($savedPwd === "UNINITIALIZED") {
            return true;
        } elseif ($currentPwd === $savedPwd) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * check if user is logged in
     *
     * Returns bool(true) if the user is currently
     * logged in and bool(false) otherwise.
     *
     * @access  public
     * @return  bool
     */
    function isLoggedIn()
    {
        /**
         * Note on security:
         * SHA-1 produces a 20 bytes long string, resulting in 40 characters hexadecimal representation.
         * However this framework introduces SHA-1 encoded session-ids only to logged in users and issues
         * standard md5 encoded ids, which are faster to calculate, than others.
         * Also md5 encoded ids are only 16 bytes long and in hexadecimal representation with 32 characters.
         * So: having an id that is not of the expected length is an obvious hint for either a user that
         * has not logged in, or an invalid session id.
         */
        if (function_exists('sha1') && strlen(session_id()) !== strlen(sha1(''))) {
            return false;
        } elseif (empty($_SESSION['prog_id']) || $_SESSION['prog_id'] !== YANA_INSTANCE_ID || !isset($_SESSION['user_pwd']) || !isset($_SESSION['user_name'])) {
            return false;
        } else {
            $sesPwd = sessionManager::_createKey2($_SESSION['user_pwd']);
            $dbPwd = $this->_getPassword($_SESSION['user_name']);
            if ($dbPwd === "UNINITIALIZED") {
                return true;
            } else {
                return $dbPwd == $sesPwd;
            }
        }
    }

    /**
     * get password of some user
     *
     * @access  private
     * @param   string  $userName
     * @return  string
     */
    function _getPassword($userName)
    {
        return $this->database->get("USER.$userName.USER_PWD");
    }

    /**
     * set password of some user
     *
     * @access  private
     * @param   string  $userName
     * @param   string  $userPwd
     * @return  bool
     */
    function _setPassword($userName, $userPwd)
    {
        return $this->database->update("USER.$userName.USER_PWD", sessionManager::_encode($userName, $userPwd));
    }

    /**
     * get name of group
     *
     * @access  private
     * @param   string  $groupId
     * @return  string
     */
    function _getGroupName($groupId)
    {
        $result = $this->database->get("usergroup.$groupId.group_name");
        if (is_string($result)) {
            return $result;
        } else {
            trigger_error("Unable to get group name. The provided user group '${groupId}' does not exist.", E_USER_WARNING);
            return 0;
        }
    }

    /**
     * get permission value of group
     *
     * @access  private
     * @param   string  $groupId
     * @return  int
     */
    function _getGroupValue($groupId)
    {
        $result = $this->database->get("usergroup.$groupId.group_value");
        if (is_numeric($result)) {
            return (int) $result;
        } else {
            trigger_error("Unable to get permission level. The provided user group '${groupId}' does not exist.", E_USER_WARNING);
            return 0;
        }
    }

    /**
     * login
     *
     * This function is used to handle user logins.
     *
     * It takes the following steps.
     *
     * 1) destroy any previous session (to prevent session fixation)
     * 2) check user name and password
     *
     * on success:
     *
     * 3.a) create new session id
     * 4.a) return bool(true)
     *
     * on error:
     *
     * 3.b) bool(false)
     *
     * @access  public
     * @param   string  $userName  user name
     * @param   string  $userPwd   password
     * @return  bool
     */
    function login($userName, $userPwd)
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }
        if (!is_string($userPwd)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($userPwd)), E_USER_WARNING);
            return false;
        }

        $userName = strtoupper($userName);

        /* backup language setting before destroying old session */
        if (isset($_SESSION['language']) && preg_match('/^[\w\d-_]+$/s', $_SESSION['language'])) {
            $language = $_SESSION['language'];
        } else {
            $language = false;
        }

        /* never reuse old sessions, to prevent injection of data or session id */
        $this->logout();

        /* create new session with brand new session id */
        $sessionId = new StringObject(uniqid(YANA_INSTANCE_ID.rand(), true));
        $encryptedId = $sessionId->encrypt('sha1');
        /* if sha1 is not supported, fall back to default encryption method (whatever it is) */
        if ($encryptedId === STRING_UNSUPPORTED_ENCRYPTION) {
            $encryptedId = $sessionId->encrypt();
        }
        session_id($encryptedId);
        @session_start();

        /* check username and password */
        if ($this->checkLogin($userName, $userPwd)) {
            $userPwd = sessionManager::_createKey1($userName, $userPwd);
            $_SESSION = array();
            $_SESSION["user_name"] = $userName;
            $_SESSION["user_pwd"] = $userPwd;
            $_SESSION["prog_id"] = YANA_INSTANCE_ID;

            /**
             * initialize language settings
             *
             * Note: if language settings have been set in
             * a previous session, then they are taken from
             * the these settings and not from the database.
             */
            if (empty($language)) {
                $language = $this->getLanguage($userName);
            }
            if (!empty($language)) {
                global $YANA;
                if (isset($YANA) && $YANA->language->setDirectory($language)) {
                    $_SESSION['language'] = $language;
                }
            } /* end if */
            return true;
        } else {
            return false;
        } /* end if */
    }

    /**
     * logout
     *
     * Destroy the current session and clear all session data.
     *
     * @access  public
     * @return  bool
     */
    function logout()
    {

        if (isset($_SESSION['user_name'])) {
            $this->updateLanguage($_SESSION['user_name']);
        }
        /* make session cookie expire (get's deleted) */
        if (isset($_COOKIE[@session_name()])) {
            @setcookie(@session_name(), '', time() - 42000, '/');
        }
        /* unset session data */
        $_SESSION = array();
        /* kill session */
        @session_destroy();
        /* get rid of the old sesion id - just in case */
        @session_regenerate_id();
        return true;
    }

    /**
     * update language
     *
     * Set $userName's prefered language by using
     * the value set in $_SESSION['language'] if
     * there is any.
     *
     * Returns bool(true) on success and
     * bool(false) on error.
     *
     * @access  public
     * @param   string  $userName
     * @return  bool
     * @ignore
     */
    function updateLanguage($userName)
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        $userName = strtoupper($userName);
        /* 1) error: language not set */
        if (empty($_SESSION['language'])) {
            return false;
        /* 2) error: user does not exist */
        } elseif (!$this->userExists($userName)) {
            trigger_error("The user '$userName' does not exist", E_USER_NOTICE);
            return false;
        /* 3) write changes to database */
        } else {
            if (!$this->database->update("user.$userName.user_language", $_SESSION['language'])) {
                /* 3.a) error: internal database error */
                trigger_error("Database does not accept the language value to be changed for user '$userName'.", E_USER_NOTICE);
                return false;
            }

            if (!$this->database->write()) {
                /* 3.b) error: internal database error */
                trigger_error("Database did return an error when trying to commit changes to the language value.", E_USER_NOTICE);
                return false;
            } else {
                /* 3.c) success */
                return true;
            }
        } /* end if */

    }

    /**
     * get prefered language
     *
     * Get $userName's prefered language from database.
     *
     * Returns bool(false) on error.
     *
     * @access  public
     * @param   string  $userName
     * @return  bool
     * @ignore
     */
    function getLanguage($userName)
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        $userName = strtoupper($userName);
        /* 1) error: user does not exist */
        if (!$this->userExists($userName)) {
            trigger_error("The user '$userName' does not exist", E_USER_NOTICE);
            return false;
        /* 2) get value from database */
        } else {
            return $this->database->get("USER.$userName.USER_LANGUAGE");
        }

    }

    /**
     * change password
     *
     * Set $userName's login password to $userPwd
     *
     * @access  public
     * @param   string  $userName
     * @param   string  $userPwd
     * @return  bool
     * @ignore
     */
    function changePwd($userName, $userPwd)
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }
        if (!is_string($userPwd)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($userPwd)), E_USER_WARNING);
            return false;
        }

        if ($this->isLoggedIn()) {
            if ($this->_setPassword($userName, $userPwd)) {
                return $this->database->write();
            } else {
                return false;
            }
        } else {
            return false;
        } /* end if */
    }

    /**
     * set new (random) password
     *
     * Creates a new random password for $userName and returns it.
     * The encoded string of the password is stored to the database.
     *
     * In case of an error, bool(false) is returned.
     *
     * @access  public
     * @param   string  $userName
     * @return  string|bool(false)
     * @ignore
     */
    function setPwd($userName)
    {
        global $YANA;

        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        /* upper-case user's name */
        $userName = strtoupper($userName);

        if (!$this->userExists($userName)) {
            trigger_error("The user '$userName' does not exist", E_USER_NOTICE);
            return false;
        }

        /* create new (random) password */
        $userPwd = sessionManager::generatePassword();

        /**
         * To set a password the user does either:
         * 1) need to be logged in OR
         * 2) no password has been set yet.
         */
        if ($this->isLoggedIn() || $this->_getPassword($userName) === "UNINITIALIZED") {

            /* error - unable to update password settings */
            if (!$this->_setPassword($userName, $userPwd)) {
                if (isset($YANA)) {
                    $YANA->report( new Log("A new password was requested for user '${userName}'. But the database entry could not be updated.") );
                }    
                return false;
            }

            /* error - internal database error, unable to commit changes */
            if (!$this->database->write()) {
                if (isset($YANA)) {
                    $YANA->report( new Log("A new password was requested for user '${userName}'. But has not been saved, as data could not be commited.") );
                }    
                return false;
            } else {
                return $userPwd;
            }

        } else {
            return false;
        } /* end if */
    }

    /**
     * create or update user data
     *
     * @access  public
     * @param   string  $userName
     * @param   array   $userData
     * @return  bool
     * @ignore
     */
    function insertUser($userName, $userData)
    {
        global $YANA;

        /* error - wrong data type: argument 1 */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        } else {
            $userName = strtoupper($userName);
        }
        /* error - wrong data type: argument 2 */
        if (!is_array($userData)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Array', gettype($userData)), E_USER_WARNING);
            return false;
        }

        /* insert user data */
        if (!$this->database->insertOrUpdate("user.$userName", $userData)) {
            return false;
        } elseif (!$this->database->write()) {
            if (isset($YANA)) {
                $YANA->report( new Log("Unable to commit changes to the database server while trying to update settings for user '${userName}'.") );
            }    
            return false;
        } else {
            return true;
        }
    }

    /**
     * remove user
     *
     * @access  public
     * @param   string  $userName
     * @return  bool
     * @ignore
     */
    function removeUser($userName)
    {
        global $YANA;

        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        if ($userName !== '') {

            $userName = strtoupper($userName);
            if (!$this->database->remove("USER.$userName")) {
                return false;
            } elseif (!$this->database->write()) {
                if (isset($YANA)) {
                    $YANA->report( new Log("Unable to commit changes to the database server while trying to remove user '${userName}'.") );
                }    
                return false;
            } else {
                return true;
            }

        /* error - no user name provided */
        } else {
            return false;
        }
    }

    /**
     * get user permission
     *
     * Returns the $userName's security level
     * for the profile identified by $profileId
     * as an integer (-1 through 100).
     *
     * There are some default user groups:
     * <ul>
     * <li> -1 = blocked </li>
     * <li> 0 = guest </li>
     * <li> 1 = registered </li>
     * <li> 30 = moderator </li>
     * <li> 75 = owner </li>
     * <li> 100 = administrator </li>
     * </ul>
     * Note that this table may be subject to changes.
     *
     * @access  public
     * @param   string  $userName   name of user
     * @param   string  $profileId  name of profile
     * @return  int
     * @ignore
     */
    function getUserPermission($userName, $profileId = "")
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return 0;

        /* return 0 if no user name provided */
        } elseif (empty($userName)) {
            return 0;
        }

        if (!is_string($profileId)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($profileId)), E_USER_WARNING);
            return 0;
        }

        $userName = strtoupper($userName);
        $profileId = strtoupper($profileId);

        /* get user activity state and return 0 if user is inactive */
        if (!$this->database->get("USER.$userName.ACTIVE")) {
            return 0;
        } elseif (empty($this->defaultUser['ACTIVE'])) {
            return 0;
        } else {
            $active = 0;
        }

        /* get current mode */
        global $YANA;
        if (isset($YANA)) {
            $mode = (int) $YANA->getMode();
        } else {
            $mode = 0;
        }

        /* get user permission level */
        if ($profileId != '') {
            $where = array(
                array('user_id', $userName, '='),
                array('profile_id', strtolower($profileId), '=')
            );
            if (! $permission = $this->database->get("useraccess.?.group_id", $where)) {
                $permission = null;
            } elseif (! $permission = $this->_getGroupValue($permission)) {
                $permission = null;
            }
        } else {
            $permission = null;
        }

        if ($this->database->exists("USER.$userName.GROUP_ID")) {
            $type = $this->database->get("USER.$userName.GROUP_ID");
        } elseif (isset($this->defaultUser["GROUP_ID"])) {
            $type = $this->defaultUser["GROUP_ID"];
        } else {
            $type = null;
        }

        if ($mode === 0 && $permission != false) {
            $perm = $permission;
        } elseif (!is_null($type)) {
            $perm = $type;
        } elseif (isset($this->defaultUser["GROUP_ID"])) {
            $perm = $this->defaultUser["GROUP_ID"];
        }

        if (is_int($perm)) {
            return $perm;
        } elseif (is_string($perm) && $perm !== '') {
            return $this->_getGroupValue($perm);
        } else {
            return 0;
        }
    }

    /**
     * check permission
     *
     * Check if user has permission to apply changes
     * to the profile identified by the argument
     * $profileId.
     *
     * Returns bool(true) if the user's permission
     * level is high enough to execute the changes and
     * bool(false) otherwise.
     *
     * @access  public
     * @param   string  $profileId
     * @param   string  $userName
     * @param   string  $action
     * @return  bool
     * @since   2.9
     * @ignore
     */
    function checkPermission($profileId = null, $userName = null, $action = null)
    {
        global $YANA;

        /* Argument 1 */
        if (is_null($profileId)) {
            if (isset($YANA)) {
                $profileId = $YANA->getId();
            } elseif (isset($_GET['id'])) {
                $profileId = (string) $_GET['id'];
            } else {
                $profileId = '';
            }
        } elseif (!empty($profileId) && !is_string($profileId)) {
            /* error: invalid input */
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($profileId)), E_USER_WARNING);
            return false;
        }
        /* if no id is provided, switch to default instead */
        if (empty($profileId)) {
            if (isset($YANA)) {
                $profileId = $YANA->getDefault('PROFILE');
            } else {
                $profileId = 'DEFAULT';
            }
        }
        $profileId = strtoupper($profileId);

        /* Argument 2 */
        /**
         * {@internal
         * The user id is resolved by the "user" plugin and stored
         * in the registry under a key named "SESSION_USER_ID",
         * so other plugins can look it up.
         * }}
         */
        if (is_null($userName)) {
            if (isset($YANA)) {
                $userName = $YANA->getVar('SESSION_USER_ID');
                if (!is_string($userName)) {
                    $userName = '';
                }
            } else {
                $userName = "";
            }
        } elseif (empty($userName) || !is_string($userName)) {
            /* error: invalid input */
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        /* Argument 3 */
        if (is_null($action)) {
            if (isset($YANA)) {
                $action = $YANA->getVar('ACTION');
            } elseif (isset($_GET['action'])) {
                $action = (string) $_GET['action'];
            } else {
                $action = "";
            }
        } elseif (!is_string($action)) {
            /* error: invalid input */
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($action)), E_USER_WARNING);
            return false;
        } else {
            /* intentionally left blank */
        }
        if (empty($action)) {
            if (isset($YANA)) {
                $action = $YANA->getDefault('HOMEPAGE');
            } else {
                $action = 'SITEMAP';
            }
        }
        $action = strtoupper($action);

        /**
         * {@internal
         * check if value has already been processed and cached
         * and if so, return the cached value instead, for a
         * better performance.
         * }}
         */
        if (isset($this->cache[$profileId][$action])) {
            assert('is_bool($this->cache[$profileId][$action]); /* unexpected result in cached value */');
            return $this->cache[$profileId][$action];
        }

        /* find out what the required permission level is to perform the current action */
        if (isset($YANA)) {
            $requiredAccess = $YANA->plugins->getVar("*", "INTERFACE.${action}.PERMISSION");
        } else {
            /* fall back to highest level, if no input is available */
            $requiredAccess = 100;
        }
        if (is_null($requiredAccess)) {
            if (isset($YANA)) {
                $requiredAccess = $YANA->getDefault("EVENT.PERMISSION");
            } else {
                $requiredAccess = 100;
            }
        }
        if (!is_numeric($requiredAccess)) {
            trigger_error("Action '${action}' is unspecified in " . __CLASS__ . "->" . __FUNCTION__ . "().", E_USER_NOTICE);
            return false;
        } else {
            /* settype to INTEGER */
            $requiredAccess = (int) $requiredAccess;
        }

        /* find out what the permission level the user has for the current profile id */
        $currentAccess = $this->getUserPermission($userName, $profileId);

        if (!is_int($currentAccess)) {
            /* error: profile or user is unknown */
            return false;
        }

        /* return true if permission level is high enough and false otherwise */
        if ($currentAccess >= $requiredAccess) {
            $result = true;
        } else {
            $result = false;
        }

        /* cache the result and return it */
        if (!isset($this->cache[$profileId])) {
            $this->cache[$profileId] = array();
        }
        $this->cache[$profileId][$action] = $result;
        return $result;
    }

    /**
     * check if user table is empty
     *
     * Returns bool(true) if the user table is empty.
     * (Actually this means, it is certainly in some state of error.)
     *
     * Otherwise returns bool(false).
     *
     * @access  public
     * @return  bool
     * @ignore
     */
    function isEmpty()
    {
        return $this->database->isEmpty('user');
    }

    /**
     * get user
     *
     * Returns the user's data record as an associative array.
     *
     * @access  public
     * @param   string  $userName
     * @return  array
     * @ignore
     */
    function getUser($userName = '*')
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        if (empty($userName) || $userName === '*' ) {
            return $this->database->get("USER");
        } else {
            $userName = strtoupper($userName);
            $userData = $this->database->get("USER.$userName");
            return $this->_formatUserData($userName, $userData, true);
        }
    }

    /**
     * check if user exists
     *
     * Returns bool(true) if a user named $userName
     * can be found in the current database.
     * Returns bool(false) otherwise.
     *
     * @access  public
     * @param   string  $userName
     * @return  bool
     * @ignore
     */
    function userExists($userName)
    {
        /* input check */
        if (!is_string($userName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($userName)), E_USER_WARNING);
            return false;
        }

        $userName = strtoupper($userName);
        return $this->database->exists("USER.$userName");
    }

    /**
     * generate a random password
     *
     * creates a random password string of length 10
     *
     * @access public
     * @static
     * @return string
     */
    function generatePassword()
    {
        return substr(md5(uniqid(rand(), true)), 0, 10);
    }

    /**
     * encode
     *
     * @access  private
     * @static
     * @param   string  $a
     * @param   string  $b
     * @return  string
     * @ignore
     */
    function _encode($a, $b)
    {
        assert('is_scalar($a); // Wrong argument type for argument 1. String expected.');
        assert('is_scalar($b); // Wrong argument type for argument 2. String expected.');
        /* settype to STRING */
        $a = (string) $a;
        $b = (string) $b;
        return sessionManager::_createKey2(sessionManager::_createKey1($a, $b));
    }

    /**
     * _createKey1
     *
     * @access  private
     * @static
     * @param   string  $salt
     * @param   string  $text
     * @return  string
     * @ignore
     */
    function _createKey1($salt, $text)
    {
        assert('is_scalar($salt); // Wrong argument type for argument 1. String expected.');
        assert('is_scalar($text); // Wrong argument type for argument 2. String expected.');
        /* settype to STRING */
        $salt = (string) $salt;
        $text = (string) $text;
        $salt = substr(strtoupper($salt), 0, 2);

        $string = new StringObject($salt . $text);
        /* encrypt using sha1 */
        $key1 = $string->encrypt('sha1');
        /* if sha1 is not supported, fall back to default encryption method (whatever it is) */
        if ($key1 === STRING_UNSUPPORTED_ENCRYPTION) {
            $key1 = $string->encrypt();
        }

        return $key1;
    }

    /**
     * _createKey2
     *
     * @access  private
     * @static
     * @param   string  $text
     * @return  string
     * @ignore
     */
    function _createKey2($text)
    {
        assert('is_scalar($text); // Wrong argument type for argument 1. String expected.');
        /* settype to STRING */
        $string = new StringObject($text);
        /* encrypt using sha1 */
        $key2 = $string->encrypt('sha1');
        /* if sha1 is not supported, fall back to default encryption method (whatever it is) */
        if ($key2 === STRING_UNSUPPORTED_ENCRYPTION) {
            $key2 = $string->encrypt();
        }

        return $key2;
    }

    /**
     * _formatUserData
     *
     * @access  private
     * @param   string   $userName
     * @param   array    $userData
     * @param   bool     $i
     * @return  array
     * @ignore
     */
    function _formatUserData($name, $data, $i)
    {
        /* settype to STRING
         *            ARRAY
         */
        $name = (string) $name;
        $data = (array)  $data;
        $name = strtoupper($name);

        /* 1) remove password for security reasons */
        unset($data['USER_PWD']);
        $REF = $data;

        /* 2) convert access restriction information */
        unset($REF['GROUP_ID']);

        $REF['USER_ID'] = $name;

        /* 2.a) type parameter */
        if (!isset($data['GROUP_ID'])) {
            $REF['GROUP_ID']['NAME'] = "";
            $REF['GROUP_ID']['VALUE'] = 0;
        } elseif (is_numeric($data['GROUP_ID'])) {
            $REF['GROUP_ID']['NAME'] = $data['GROUP_ID'];
            $REF['GROUP_ID']['VALUE'] = $data['GROUP_ID'];
        } else {
            $REF['GROUP_ID']['NAME'] = $this->_getGroupName($data['GROUP_ID']);
            $REF['GROUP_ID']['VALUE'] = $data['GROUP_ID'];
        }


        return $REF;
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
