<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «decorator» SmartTemplate
 *
 * This implements a decorator class for the smarty
 * template engine. The use of the "decorator" pattern
 * actually means it "implements an API" on top
 * of the original.
 *
 * Note that this does not touch or even change the
 * engine itself.
 *
 * This class provides two things: most obviously it
 * provides some features, that smarty does not have and
 * in addition it does some more type checking and
 * automates the initialization process.
 *
 * {@internal
 *
 * The global token %INSERT_CONTENT_HERE% has been
 * dropped as of version 2.9.2.
 *
 * This feature has been replaced by the following
 * two system vars:
 *
 * <ol>
 *   <li> %SYSTEM_TEMPLATE% = id of current base template </li>
 *   <li> %SYSTEM_INSERT% = id of current extensional template </li>
 * </ol>
 *
 * Note that this will break support for skins before version 2.9.2,
 * including the stable version 2.9.1.
 *
 * Note: as PHP4 did not support static vars, a workaround is required.
 * Thus, instead of refering to the class name as namespace for class
 * vars, one should always refer to $GLOBALS['YANA']->view, which is
 * an unique instance that is always present, if an instance of the
 * framework has been created.
 * This is not the solution I had prefered, but I'm afraid, there is
 * no alternative if maintaining backwards compatibility is an issue.
 *
 * }}
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class SmartTemplate extends Object
{

    /**#@+
     * @ignore
     *
     * @access  private
     */

    /** @var string */  var $path   = "";
    /** @var bool */    var $exists = false;
    /** @var smarty */  var $smarty = null;

    /**#@-*/
    /**#@+
     * @ignore
     *
     * @access  private
     * @static
     */

    /** @var array  */ var $functions = array();
    /** @var array  */ var $modifiers = array();
    /** @var array  */ var $blocks = array();

    /**#@-*/

    /**
     * create an instance
     *
     * You may enter a filename of a template you want to use.
     *
     * @param  string  $filename  (optional)
     */
    function __construct($filename = "")
    {
        /* settype to STRING */
        $filename = (string) $filename;
        global $YANA;

        /**
         * 1) Config Smarty
         */
        $this->smarty = new Smarty();

        /**
         * 1.1) delimiters
         */
        $this->smarty->left_delimiter = YANA_LEFT_DELIMITER;
        $this->smarty->right_delimiter = YANA_RIGHT_DELIMITER;

        /**
         * 1.2) directories
         */
        if (isset($YANA)) {
            SmartUtility::loadSmilies($YANA->getVar("PROFILE.SMILEYDIR"));
            $this->smarty->template_dir = '.';
            $this->smarty->config_dir = $YANA->getVar('SKINCONFIGDIR');
            $this->smarty->compile_dir = $YANA->getVar('TEMPDIR');
        } else {
            $this->smarty->template_dir = '.';
            $this->smarty->config_dir = 'skins/.config/';
            if (YANA_CDROM === true) {
                $this->smarty->compile_dir = YANA_CDROM_DIR;
            } else {
                $this->smarty->compile_dir = 'cache/';
            }
        }

        /**
         * 1.3) security settings
         */
        $this->smarty->php_handling = SMARTY_PHP_REMOVE;
        $this->smarty->security = true;

        /**
         * 1.4) caching behaviour
         */
        $this->smarty->caching = YANA_TPL_CACHE;
        $this->smarty->use_sub_dirs = YANA_TPL_CACHE_DIR;

        /**
         * 1.4.1) default setting for compile check
         */
        if (!defined('YANA_ERROR_REPORTING') || !defined('YANA_ERROR_OFF')) {
            $this->smarty->compile_check = true;

        /**
         * 1.4.2) distinguish between debug-mode and production-mode
         */
        } else {
            if (YANA_ERROR_REPORTING !== YANA_ERROR_OFF) {
                $this->smarty->compile_check = true;
            } else {
                $this->smarty->compile_check = false;
            }
            if (YANA_ERROR_REPORTING === YANA_ERROR_ON) {
                $this->smarty->register_function('varDump',     array('SmartUtility', 'varDump'));
            }
        }

        /**
         * 1.5) extensions
         */
        $this->smarty->register_modifier('embeddedTags',        array('SmartUtility', 'embeddedTags'));
        $this->smarty->register_modifier('replaceToken',        array('SmartUtility', 'replaceToken'), false);
        $this->smarty->register_modifier('css',                 array('SmartUtility', 'css'));
        $this->smarty->register_modifier('date',                array('SmartUtility', 'date'));
        $this->smarty->register_modifier('entities',            array('SmartUtility', 'entities'));
        $this->smarty->register_modifier('href',                array('SmartUtility', 'href'), false);
        $this->smarty->register_modifier('scanForAt',           array('SmartUtility', 'scanForAt'));
        $this->smarty->register_modifier('smilies',             array('SmartUtility', 'smilies'));
        $this->smarty->register_modifier('url',                 array('SmartUtility', 'url'), false);
        $this->smarty->register_modifier('urlEncode',           array('SmartUtility', 'urlEncode'));
        $this->smarty->register_function('printArray',          array('SmartUtility', 'printArray'));
        $this->smarty->register_function('printUnorderedList',  array('SmartUtility', 'printUnorderedList'));
        $this->smarty->register_function('rss',                 array('SmartUtility', 'rss'));
        $this->smarty->register_function('import',              array('SmartUtility', 'import'), false);
        $this->smarty->register_function('smilies' ,            array('SmartUtility', 'guiSmilies'));
        $this->smarty->register_function('embeddedTags',        array('SmartUtility', 'guiEmbeddedTags'));
        $this->smarty->register_function('create',              array('FormCreator',  'create'), false);
        $this->smarty->register_function('captcha',             array('SmartUtility', 'captcha'));
        $this->smarty->register_function('sizeOf',              array('SmartUtility', 'sizeOf'));
        $this->smarty->register_function('preview',             array('SmartUtility', 'preview'));
        $this->smarty->register_function('sml_load',            array('SmartUtility', 'smlLoad'));
        $this->smarty->register_function('smlLoad',             array('SmartUtility', 'smlLoad'));
        $this->smarty->register_function('updateCheck',         array('SmartUtility', 'updateCheck'));
        $this->smarty->register_block('loop',                   array('SmartUtility', 'loopArray'), false);
        $this->smarty->register_block('isDynamic',              array('SmartUtility', 'isDynamic'), false);
        $this->smarty->default_modifiers =                      array('replaceToken', 'scanForAt');
        $this->smarty->register_prefilter(                      array('SmartUtility', 'processLanguageInstructions'));
        $this->smarty->register_prefilter(                      array('SmartUtility', 'htmlPreProcessor'));
        $this->smarty->register_postfilter(                     array('SmartUtility', 'htmlPostProcessor'));
        $this->smarty->register_outputfilter(                   array('SmartUtility', 'outputFilter'));

        /**
         * 2) set selected template (if any)
         */
        $this->setPath($filename);
    }

    /**
     * toString
     *
     * @access public
     * @return string
     *
     * @ignore
     */
    function toString()
    {
        if (!$this->exists()) {
            return "The template file '".$this->getPath()."' does not exist\n";
        } else {
            return "Template class '".get_class($this)."' using template file '".$this->getPath()."'.\n";
        }
    }

    /**
     * fetch a template or template var
     *
     * There are three ways to call this function:
     *
     * If you call $smartTemplate->get($varName) it will get the
     * template var $varName and return it.
     *
     * If you call $smartTemplate->get("*") with the wildcard '*'
     * or an empty string '' it will return an associative array
     * containing all template vars.
     *
     * If you call $smartTemplate->get() without a parameter it
     * will fetch the current template and return it as a string.
     *
     * If the $key parameter is NULL, the second parameter can be
     * used to change the way how variables are imported from the
     * global registry to the template.
     * If set to false, existing template vars will not be replaced
     * by vars of the same name in the registry. Otherwise template
     * vars will get overwritten if the global registry has another
     * value.
     *
     * @access  public
     * @param   string  $key        variable-name
     * @param   bool    $overwrite  true = on / false = off
     * @return  mixed
     */
    function get($key = null, $overwrite = true)
    {
        global $YANA;

        /* 1) fetch template */
        if (is_null($key)) {
            /* 1.1) import vars from global registry */
            if (isset($YANA)) {
                if ($overwrite === true) {
                    /* import and overwrite */
                    $this->smarty->assign($YANA->getVar('*'));
                } elseif ($overwrite === false) {
                    /* import only non-existing vars */
                    $vars = $YANA->getVar('*');
                    $vars = Hashtable::merge($vars, $this->smarty->get_template_vars());
                    $this->smarty->assign($vars);
                } else {

                    trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Boolean', gettype($overwrite)), E_USER_WARNING);
                    return false;
                }
            }
            /* 1.2) set functions, modifiers and block functions */
            if (is_array($GLOBALS['YANA']->view->functions)) {
                assert('!isset($name); // Cannot redeclare var $name');
                assert('!isset($code); // Cannot redeclare var $code');
                foreach ($GLOBALS['YANA']->view->functions as $name => $code)
                {
                    $this->smarty->register_function($name, $code);
                }
                unset($name, $code);
            }
            if (is_array($GLOBALS['YANA']->view->modifiers)) {
                assert('!isset($name); // Cannot redeclare var $name');
                assert('!isset($code); // Cannot redeclare var $code');
                foreach ($GLOBALS['YANA']->view->modifiers as $name => $code)
                {
                    $this->smarty->register_modifier($name, $code);
                }
                unset($name, $code);
            }
            if (is_array($GLOBALS['YANA']->view->blocks)) {
                assert('!isset($name); // Cannot redeclare var $name');
                assert('!isset($code); // Cannot redeclare var $code');
                foreach ($GLOBALS['YANA']->view->blocks as $name => $code)
                {
                    $this->smarty->register_block($name, $code);
                }
                unset($name, $code);
            }

            /* 1.3) fetch and return template */
                if ($this->smarty->caching > 0) {
                assert('!isset($cacheId); // Cannot redeclare var $cacheId');
                if (isset($YANA)) {
                    $cacheId = md5($YANA->language->getDirectory() . '/' . $YANA->getVar('QUERY_STRING'));
                } else {
                    $cacheId = md5($this->getPath());
                }
                return $this->smarty->fetch($this->getPath(), $cacheId);
                unset($cacheId);
            } else {
                return $this->smarty->fetch($this->getPath());
            }

        /* 2) get all template vars */
        } elseif (empty($key) || $key === '*') {

            return $this->smarty->get_template_vars();

        /* 3) get one template var, identified by $key */
        } elseif (is_scalar($key)) {

            $resource = $this->smarty->get_template_vars();
            assert('is_array($resource); /* unexpected result: $resource should be an array */');
            return Hashtable::get($resource, $key);

        /* 4) error: invalid input */
        } else {

            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($key)), E_USER_WARNING);
            return false;

        }

    }

    /**
     * output the template to screen
     *
     * This function will fetch and print
     * the current template to the screen.
     *
     * @access  public
     */
    function write()
    {
        if (ob_get_length() !== false) {
            ob_flush();
        }
        if ($this->exists()) {
            echo $this->get();
        }    
    }

    /**
     * bypass template class
     *
     * This function is used to unbox the smarty instance inside the
     * object. It may be used to bypass the template class in cases
     * where direct access to the smarty template engine is necessary.
     *
     * @access  public
     * @return  Smarty
     * @since   2.8.9
     */
    function &getSmarty()
    {
        return $this->smarty;
    }

    /**
     * assign a variable by value
     *
     * This assigns the $var to the name $varName.
     *
     * Unlike Smarty's "assign()" this function takes an
     * additional value for $varName:
     *
     * You may use the wildcard '*' to merge an associative array with the template vars.
     * Example of usage: <code>$smartTemplate->insert('*', array  $var) </code>
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * {@internal
     *
     * The following synopsis: <code>$smartTemplate->insert('*', string $var)</code>
     * has been dropped as of version 2.9.2.
     *
     * }}
     *
     * @access  public
     * @param   string  $varName  address
     * @param   mixed   $var      some new value
     * @return  bool
     */
    function insert($varName, $var)
    {
        /* 1) error: invalid input */
        if (!is_scalar($varName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($varName)), E_USER_WARNING);
            return false;

        /* 2) assign to global namespace */
        } elseif ($varName == '*') {

            if (is_array($var)) {
                $this->smarty->assign($var);
                return true;
            } else {
                trigger_error("When using the wildcard '*' with the function 'insert',\n\t\t" .
                "argument 2 is expected to be an array. Found '".gettype($var)."' instead.", E_USER_WARNING);
                return false;
            }

        /* 3) assigned var is an object */
        } elseif (is_object($var)) {

            $this->smarty->assign_by_ref($varName, $var);
            return true;

        /* 4) assign to var identified by $varName */
        } else {
            $this->smarty->assign($varName, $var);
            return true;
        }

    }

    /**
     * assign a variable by reference
     *
     * This assigns the $var to the name $varName.
     *
     * Unlike Smarty's "assign()" this function takes an
     * additional value for $varName:
     *
     * You may use the wildcard '*' to merge an associative array
     * with the template vars.
     *
     * Example of usage:
     * <code>$smartTemplate->insertByReference('*', array  $var) </code>
     *
     * {@internal
     *
     * The following synopsis:
     * <code>$smartTemplate->insertByReference('*', string $var)</code>
     * has been dropped as of version 2.9.2.
     *
     * }}
     *
     * @access  public
     * @param   string  $varName  address
     * @param   mixed   &$var     some new value
     * @return  bool
     */
    function insertByReference($varName, &$var)
    {
        /* 1) error: invalid input */
        if (!is_scalar($varName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($varName)), E_USER_WARNING);
            return false;

        /* 2) assign to global namespace */
        } elseif ($varName === '*') {

            /* 2.a) assign contents to global namespace */
            if (is_array($var)) {

                assert('!isset($key); /* cannot redeclare variable $key */');
                foreach (array_keys($var) as $key)
                {
                    $this->smarty->assign_by_ref($key, $var[$key]);
                }
                unset($key);
                return true;

            /* 2.c) error: invalid input */
            } else {
                trigger_error("When using the wildcard '*' with the function 'insertByReference',\n\t\t" .
                "argument 2 is expected to be an array. Found '".gettype($var)."' instead.", E_USER_WARNING);
                return false;
            }

        /* 3) assign to var identified by $varName */
        } else {
            $this->smarty->assign_by_ref($varName, $var);
            return true;
        }

    }

    /**
     * insert a file
     *
     * Load a file and assign it's content as a string
     * to the provided template var.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * {@internal
     *
     * The following synopsis:
     * <code>$smartTemplate->insertFile('*', string $filename)</code>
     * has been dropped as of version 2.9.2.
     *
     * }}
     *
     * @access  public
     * @param   string  $varName
     * @param   string  $filename
     * @return  bool
     */
    function insertFile($varName, $filename)
    {
        /* input check */
        if (!is_scalar($varName)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Scalar value', gettype($varName)), E_USER_WARNING);
            return false;
        } elseif ($varName == "" || $varName == "*") {
            trigger_error("Argument 1 cannot be empty.", E_USER_WARNING);
            return false;
        }

        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        } elseif (!file_exists($filename) || !is_readable($filename)) {
            trigger_error("The file '${filename}' is not readable.", E_USER_WARNING);
            return false;
        }
        /* settype to STRING */
        $varName  = (string) $varName;

        /* import file */
        $template = new smartTemplate($filename);
        $template->insert('FILE_IS_INCLUDE', 'true');
        $include = $template->get();
        $this->smarty->assign($varName, $include);
        return true;
    }

    /**
     * set filename of current template
     *
     * You may set another filename of a template to fetch.
     *
     * Please note:
     * <ol>
     *   <li>  Template files may not have a reserved extension like
     *         "htaccess", "php", "config" or the like.
     *   </li>
     *   <li>  Files should be adressed from the root.
     *         This is where "index.php" is stored.
     *   </li>
     *   <li>  If you can't access a file, the file does not exist
     *         or is not readable, a template error is thrown.
     *   </li>
     *   <li>  Filenames are case-sensitive!  </li>
     * </ol>
     *
     * @access  public
     * @param   string  $filename  name of the template file
     * @return  bool
     */
    function setPath($filename)
    {
        /* input check */
        if (!is_string($filename)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_WARNING);
            return false;
        }

        /* Reset file binding */
        $this->smarty->clear_assign('MTIME');
        $this->smarty->clear_assign('ATIME');
        $this->smarty->clear_assign('CTIME');
        $this->exists = false;
        $this->path = "";

        if (empty($filename)) {

            return false;

        } elseif (preg_match("/.*\.(register|config|cfg|lock|dat|htaccess|php|inc|conf)/Ui", $filename)) {

            $this->smarty->trigger_error("ERROR 403: Access denied for file '$filename'.");
            return false;

        } elseif (!file_exists($filename) || !is_readable($filename)) {

            $this->smarty->trigger_error("ERROR 404: File '$filename' not found.");
            return false;

        } else {

            $this->path = $filename;

            /* Append Template Data */
            $this->smarty->assign('MTIME', filemtime($filename));
            $this->smarty->assign('ATIME', fileatime($filename));
            $this->smarty->assign('CTIME', filectime($filename));
            $this->smarty->assign_by_ref('FILENAME', $this->path);
            preg_match("/^.*\//", $filename, $basedir);
            $this->smarty->assign('BASEDIR', $basedir[0]);
            clearstatcache();

            $this->exists = true;
            return true;

        }
    }

    /**
     * get path and name of current template
     *
     * Returns a string containing the path and filename
     * of the current template relative to the framework's
     * home directory.
     *
     * @access  public
     * @return  string
     */
    function getPath()
    {
        assert('is_string($this->path);');
        return $this->path;
    }

    /**
     * check if current template exists
     *
     * Returns bool(true) if template exists and bool(false)
     * otherwise.
     *
     * @access  public
     * @return  bool
     *
     * @ignore
     */
    function exists()
    {
        assert('is_bool($this->exists);');
        return $this->exists;
    }

    /**
     * clear cache
     *
     * Deletes all temporary files in the 'cache/' directory.
     *
     * @access  public
     * @static
     * @ignore
     */
    function clearCache()
    {
        global $YANA;

        /* 1) make sure .htaccess does'nt get deleted */
        if (isset($YANA)) {
            $dir = $YANA->getVar('TEMPDIR');
        } else {
            $dir = 'cache/';
        }
        if (is_writeable($dir . '/.htaccess')) {
            chmod($dir . '/.htaccess', 0550);
        }

        /* 2) clear Smarty cache */
        $tpl = new SmartTemplate();
        $tpl->smarty->clear_all_cache();

        /* 2) clear Yana cache */
        $files = dirlist($dir, '.php|.cache|.tmp');
        for ($i = 0; $i < count($files); $i++)
        {
            $current_file = $dir.$files[$i];
            /* If file can't be deleted due to active write-protection
              (e.g. when running under Windows), check wether this can be fixed. */
            if (!is_writeable($current_file)) {
                chmod($current_file, 0666);
            }
            if (is_writeable($current_file)) {
                unlink($current_file);
            } else {
                trigger_error("Unable to delete file '${current_file}', because the file is not writeable.", E_USER_WARNING);
            }
        }
    }

    /**
     * Register function
     *
     * This function registers the function at $code as a template
     * function under the name $name in the template engine.
     *
     * Argument $as can be one of the followin constants:
     * <ul>
     *  <li>  YANA_TPL_FUNCTION
     *        - registers the resource as a template function,
     *          called as [%foo %]
     *  </li>
     *  <li>  YANA_TPL_MODIFIER
     *        - registers the resource as a template modifier,
     *          called as [%$bar|foo%]
     *  </li>
     *  <li>  YANA_TPL_BLOCK
     *        - registers the resource as a template block function,
     *          called as [%foo %]...[%/foo%]
     *  </li>
     * </ul>
     *
     * Argument $name:
     * <ul>
     * <li> Say, you register some function as YANA_TPL_FUNCTION under the name "foo",
     * then you can call this function from within a template using: [%foo %] </li>
     * </ul>
     *
     * Argument $code:
     * <ul>
     * <li>Say, you got a PHP function like foo().
     * To refer to this function use the code-argument with the string
     * "foo".</li>
     * <li>Say, you got a PHP class "Foo" and a static function "bar",
     * which you would call from PHP as Foo::bar().
     * To refer to this function use the code-argument with the array
     * array("Foo", "bar").</li>
     * <li>Say, you got a PHP object $foo and a non-static function "bar",
     * which you would call from PHP as $foo->bar().
     * To refer to this function use the code-argument with the array
     * array($foo, "bar").</li>
     * </ul>
     *
     * Note: For details on how the called function should look like,
     * see the smarty documentation at http://smarty.php.net/docs.php,
     * chapter 16) "Extending smarty", sections: "template functions",
     * "modifiers" and "block functions".
     *
     * @access  public
     * @param   int     $as    one of: YANA_TPL_FUNCTION, YANA_TPL_MODIFIER, YANA_TPL_BLOCK
     * @param   string  $name  name of the function
     * @param   mixed   $code  a callable resource
     * @return  bool
     */
    function setFunction($as, $name, $code)
    {
        assert('is_null($name) || is_string($name); // Wrong argument type for argument 2 in function SmartTemplate->setFunction(). String expected.');
        assert('is_null($name) || !empty($name); // Argument 2 in function SmartTemplate->setFunction() cannot be empty.');
        assert('is_null($code) || is_callable($code); // Invalid value for argument 3 in function SmartTemplate->setFunction(). Not a callable resource.');
        switch ($as)
        {
            case YANA_TPL_FUNCTION:
                if (is_null($GLOBALS['YANA']->view->functions)) {
                    $GLOBALS['YANA']->view->functions = array();
                } elseif (!is_null($name)) {
                    $GLOBALS['YANA']->view->functions[$name] = $code;
                } else {
                    /* intentionally left blank */
                }
                return true;
            break;
            case YANA_TPL_MODIFIER:
                if (is_null($GLOBALS['YANA']->view->modifiers)) {
                    $GLOBALS['YANA']->view->modifiers = array();
                } elseif (!is_null($name)) {
                    $GLOBALS['YANA']->view->modifiers[$name] = $code;
                } else {
                    /* intentionally left blank */
                }
                return true;
            break;
            case YANA_TPL_BLOCK:
                if (is_null($GLOBALS['YANA']->view->blocks)) {
                    $GLOBALS['YANA']->view->blocks = array();
                } elseif (!is_null($name)) {
                    $GLOBALS['YANA']->view->blocks[$name] = $code;
                } else {
                    /* intentionally left blank */
                }
                return true;
            break;
            default:
                trigger_error("Invalid argument 1 in function ".__CLASS__."->".__FUNCTION__.".", E_USER_WARNING);
                return false;
            break;
        }
    }

    /**
     * Unregister function
     *
     * By using this, the function named $name will no longer be
     * available in template. Be cautious: If the unregistered funciton is
     * still used inside the template, this will issue a template error
     * and possibly cause your application to exit.
     *
     * @access  public
     * @param   int     $as    one of: YANA_TPL_FUNCTION, YANA_TPL_MODIFIER, YANA_TPL_BLOCK
     * @param   string  $name  name of the function
     * @return  bool
     */
    function unsetFunction($as, $name)
    {
        assert('is_string($name); // Wrong argument type for argument 2 in function SmartTemplate->setFunction(). String expected.');
        assert('!empty($name); // Argument 2 in function SmartTemplate->setFunction() cannot be empty.');
        switch ($as)
        {
            case YANA_TPL_FUNCTION:
                $this->smarty->unregister_function($name);
                if (isset($GLOBALS['YANA']->view->functions[$name])) {
                    unset($GLOBALS['YANA']->view->functions[$name]);
                }
                return true;
            break;
            case YANA_TPL_MODIFIER:
                $this->smarty->unregister_modifier($name);
                if (isset($GLOBALS['YANA']->view->modifiers[$name])) {
                    unset($GLOBALS['YANA']->view->modifiers[$name]);
                }
                return true;
            break;
            case YANA_TPL_BLOCK:
                $this->smarty->unregister_block($name);
                if (isset($GLOBALS['YANA']->view->blocks[$name])) {
                    unset($GLOBALS['YANA']->view->blocks[$name]);
                }
                return true;
            break;
            default:
                trigger_error("Invalid argument 1 in function ".__CLASS__."->".__FUNCTION__.".", E_USER_WARNING);
                return false;
            break;
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $anotherObject
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same filesystem resource.
     *
     * @access   public
     * @param    object  $anotherObject
     * @return   string
     */
    function equals($anotherObject)
    {
        if (is_object($anotherObject) && method_exists($anotherObject, 'getClass') && $anotherObject->getClass() === $this->getClass()) {
            return $this->getPath() === $anotherObject->getPath();
        } else {
            return false;
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
