<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «datatype» String
 *
 * This is an OO-wrapper for the scalar string type of
 * PHP. This implementation is compatible with PHP 4 and 5.
 *
 * {@internal
 *
 * Note: Unlike Java, in PHP you got no automatic
 * boxing/unboxing as there are no native OO-wrappers
 * for scalar types and arrays. I would appreciate to see
 * this high on the agenda for upcomming releases of PHP.
 * Also there is no way to overload a non-static function
 * with a static function of the same name.
 * All this is making life a lot more complicated.
 * Yes, there is "kind of" unboxing in PHP 5, through the
 * __toString() magic function, BUT first of all, this only
 * works for strings and as of PHP version 5.1 even the
 * support for strings is still very incomplete.
 * So for now it seems we will have to live on it - and you
 * are best advised to do any boxing/unboxing by hand.
 *
 * Note: sadly had to deactivate preg_replace parameter $count
 * again, for there is no way to get it working in PHP 4.
 * Instead the functions will return the $count value.
 * This is kind of unexpected behaviour but it's the best way
 * there is right now without breaking backwards compatibility.
 *
 * }}
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 * @name        String
 */
class StringObject extends Object
{

    /**
     * @access private
     * @var string
     * @ignore
     */
    var $value = "";

    /**
     * create new instance
     *
     * Creates a new string wrapper.
     * The input must be convertable to a string
     * value.
     *
     * @param  string  $value
     *
     * @name   String::__construct()
     */
    function __construct($value)
    {
        $value = StringObject::getString($value);
        $this->value = $value;
    }

    /**
     * get string value
     *
     * Unboxing the object. This function
     * returns the scalar string value of the object.
     *
     * @access  public
     * @return  string
     *
     * @name    String::get()
     */
    function get()
    {
        /* settype to STRING */
        $this->value = (string) $this->value;
        return $this->value;
    }

    /**
     * set string value
     *
     * Assigns a new value to the object.
     * Returns the old value.
     *
     * @access  public
     * @param   string  $value  a new string value
     * @return  string
     *
     * @name    String::set()
     */
    function set($value)
    {
        $value = StringObject::getString($value);
        $prev_value = $this->value;
        $this->value = $value;
        return $prev_value;
    }

    /**
     * Alias of: String::get()
     *
     * @see StringObject::get()
     */
    function toString()
    {
        return $this->get();
    }

    /**
     * return value as int
     *
     * Converts the string value to an integer and
     * returns it. Returns bool(false) if the string
     * is not numeric.
     *
     * @access  public
     * @return  int|bool(false)
     *
     * @name    String::toInt()
     */
    function toInt() {
        if (is_numeric($this->value)) {
            return intval($this->get());
        } else {
            return false;
        }
    }

    /**
     * return value as float
     *
     * Converts the string value to a float and
     * returns it. Returns bool(false) if the string
     * is not numeric.
     *
     * @access  public
     * @return  float|bool(false)
     *
     * @name    String::toFloat()
     */
    function toFloat()
    {
        if (is_numeric($this->value)) {
            return floattval($this->get());
        } else {
            return false;
        }
    }

    /**
     * return value as boolean
     *
     * Returns a boolean value depending on the value of the string.
     * <ul>
     *     <li>    string("false") returns bool(false)    </li>
     *     <li>    string("true")  returns bool(true)     </li>
     *     <li>    any other value returns a boolean value
     *             depending on the result of PHP's
     *             internal conversion mechanism,
     *             BUT also issues an E_USER_NOTICE for on
     *             an invalid string to bool conversion   </li>
     * </ul>
     *
     * Note:
     * If you just want to check wether a string is empty
     * or not, use $string->equals("") instead.
     *
     * @access  public
     * @return  bool
     *
     * @name    String::toBool()
     */
    function toBool()
    {
        if ($this->compareIgnoreCase('false') === 0) {
            return false;
        } elseif ($this->compareIgnoreCase('true') === 0) {
            return true;
        } else {
            trigger_error('Invalid string to bool conversion. If you want to check wether a string is empty or not, use $string->equals("") instead.', E_USER_NOTICE);
            $b = $this->value;
            return (bool) $b;
        }
    }

    /**
     * OO-Alias of: addslashes(), addcslashes()
     *
     * @access  public
     * @param   string  $charlist  a string of characters that should be escaped
     * @return  string
     *
     * @name    String::addSlashes()
     */
    function &addSlashes($charlist = "")
    {
        $charlist = StringObject::getString($charlist);
        if (!empty($charlist)) {
            $this->value = addcslashes($value,$charlist);
        } else {
            $this->value = addcslashes($value);
        }
        return $this->value;
    }

    /**
     * OO-Alias of: stripslashes(), stripcslashes()
     *
     * @access  public
     * @param   string  $charlist  a string of characters that should be unescaped
     * @return  string
     *
     * @name    String::removeSlashes()
     */
    function &removeSlashes()
    {
        $this->value = stripslashes($value);
        return $this->value;
    }

    /**
     * OO-Alias of: $string[$index]
     *
     * Returns bool(false) on error.
     * Issues an E_USER_ERROR if $index is of wrong type.
     * Issues an E_USER_NOTICE if $index is out of bounds.
     *
     * Note that indices are numbered starting with '0'.
     *
     * @access  public
     * @param   int  $index  position of the character (starting with 0)
     *
     * @name    String::charAt()
     */
    function charAt($index)
    {
        /* If it is an object and has a "toInt()" method, try to unbox the object. */
        if (is_object($index) && method_exists($index, 'toint')) {
            $index = $index->toInt();
        }
        /* If unboxing fails, or the input is no integer at all, issue an E_USER_ERROR and abort. */
        if (!is_int($index)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Integer', gettype($index)), E_USER_WARNING);
            return false;
        }
        /* check if $index is in bounds */
        if ($this->_check_bounds($index) === false) {
            return false;
        } else {
            /* all fine, proceed */
            return $this->value[$index];
        }
    }

    /**
     * OO-Alias of: trim(), chop()
     *
     * @access  public
     * @return  string
     *
     * @name    String::trim()
     */
    function &trim()
    {
        $this->value = trim($value);
        return $this->value;
    }

    /**
     * hashing function, encryption, transformation (not revertable)
     *
     * Returns an encrypted version depending on
     * the type of encryption you choose.
     *
     * The input value is not case-sensitive.
     *
     * Note: The result of this function is alwas
     * irreversible. If you are looking for
     * reversible encryption methods see the "encode"
     * function.
     *
     * The following values are available.
     * <ul>
     *     <li>    crc32:    computes the crc32 checksum value of the string    </li>
     *     <li>    md5:      computes the md5 hash-string (128Bit)              </li>
     *     <li>    sha1:     computes the sha1 hash-string (160Bit)             </li>
     *     <li>    crypt:    uses crypt() function, result depending on $salt.
     *                       See PHP-Manual for details.                        </li>
     *     <li>    des:      uses DES encryption algorithm                      </li>
     *     <li>    blowfish: uses BLOWFISH encryption algorithm                 </li>
     *     <li>    soundex:  calculates the soundex hash: While this is not an
     *                       encryption algorithm, it is listed here, because it
     *                       is uses some sort of irreversible hashing and thus
     *                       won't fit to encoding()                            </li>
     *     <li>    metaphone:calculates the metaphone hash: While this is not an
     *                       encryption algorithm, it is listed here, for the
     *                       same reason as "soundex". Note: uses the argument
     *                       $salt as second argument for metaphone() if $salt
     *                       is a numeric value, that can be converted to int.  </li>
     *     <li>              any of the encryption types supported by the PHP
     *                       "hash()" function introduced as of PHP 5.1.2       </li>
     *     <li>    xor:      XOR is a simple revertable block chiffre.
     *                       Use $salt string as password to encrypt the clear
     *                       text. Call encrypt() with the same arguments again
     *                       to revert the ciphered text back to clear text.
     *                       Note that the security of XOR and all other simple
     *                       block chiffres depend on the length and security
     *                       off your password. To be really secure your password
     *                       needs to be minimum as long as the clear text, which
     *                       is not always praticable.                          </li>
     * </ul>
     *
     * This method issues an E_USER_NOTICE and returns the integer constant
     * STRING_UNSUPPORTED_ENCRYPTION if the $encryption parameter has an invalid value.
     *
     * {@internal
     *
     * Note:
     * Add other types of encryption as you see fit.
     * Remember to update not only the source, but
     * also the documentation on this function.
     *
     * }}
     *
     * @access  public
     * @param   string  $encryption  see the list of valid inputs for details
     * @param   string  $salt        only used for certain encryption types
     * @return  string
     *
     * @name    String::encrypt()
     * @see     StringObject::encode()
     */
    function &encrypt($encryption = "md5", $salt = "")
    {
        /* settype to STRING */
        $this->value = (string) $this->value;

        $encryption = new StringObject($encryption);
        $encryption->toLowerCase();

        $salt = new StringObject($salt);
        switch ($encryption->toString())
        {
            case 'crc32':
                if (function_exists('crc32')) {
                    $this->value = crc32($this->value);
                } else {
                    trigger_error("Unsupported encryption method: '".$encryption->toString()."'.", E_USER_NOTICE);
                    $result = STRING_UNSUPPORTED_ENCRYPTION;
                    return $result;
                }
            break;
            case 'md5':
                if ($salt->length() > 8 && CRYPT_MD5 == 1) {
                    $this->value = crypt($this->value, '$1$'.$salt->substring(0,9));
                } else {
                    $this->value = md5($this->value);
                }
            break;
            case 'sha': case 'sha1':
                if (function_exists('sha1')) {
                    $this->value = sha1($this->value);
                } else {
                    trigger_error("Unsupported encryption method: '".$encryption->toString()."'.", E_USER_NOTICE);
                    $result = STRING_UNSUPPORTED_ENCRYPTION;
                    return $result;
                }
            break;
            case 'crypt':
                if ($salt->length() > 0) {
                    $this->value = crypt($this->value, $salt->toString());
                } else {
                    $this->value = crypt($this->value);
                }
            break;
            case 'des':
                if ($salt->length() == 0) {
                    $this->value = NULL;
                } else {
                    if (CRYPT_EXT_DES == 1 && $salt->length() > 8) {
                        $this->value = crypt($this->value, $salt->substring(0,9));
                    } elseif (CRYPT_STD_DES == 1 && $salt->length() > 1) {
                        $this->value = crypt($this->value, $salt->substring(0,2));
                    } else {
                        $this->value = NULL;
                    }
                }
            break;
            case 'blowfish':
                if (CRYPT_BLOWFISH != 1 || $salt->length() < 12) {
                    $this->value = NULL;
                } else {
                    $this->value = crypt($this->value, '$2a$'.$salt->substring(0,12));
                }
            break;
            case 'soundex':
                if (function_exists('soundex')) {
                    $this->value = soundex($this->value);
                } else {
                    trigger_error("Unsupported encryption method: '".$encryption->toString()."'.", E_USER_NOTICE);
                    $result = STRING_UNSUPPORTED_ENCRYPTION;
                    return $result;
                }
            break;
            case 'metaphone':
                if (function_exists('metaphone')) {
                    if ($salt->length() > 0 && is_numeric($salt->get())) {
                        $phones = $salt->get();
                        /* settype to INTEGER */
                        $phones = (int) $phones;
                        $this->value = metaphone($this->value, $phones);
                    } else {
                        $this->value = metaphone($this->value);
                    }
                } else {
                    trigger_error("Unsupported encryption method: '".$encryption->toString()."'.", E_USER_NOTICE);
                    $result = STRING_UNSUPPORTED_ENCRYPTION;
                    return $result;
                }
            break;
            case 'xor':
                $pass = $salt->toString();
                for ($i = 0; $i < strlen($this->value); $i++)
                {
                    $this->value[$i] = $this->value[$i] ^ $pass[$i%strlen($pass)];
                }
            break;
            default:
                if (function_exists('hash_algos') && function_exists('hash')) {
                    $supported_encryption = hash_algos();
                    $supported_encryption = Hashtable::changeCase($supported_encryption, CASE_LOWER);
                } else {
                    $supported_encryption = array();
                }
                if (in_array($encryption->toString(),$supported_encryption)) {
                    $this->value = hash($encryption->toString(),$this->value);
                } else {
                    trigger_error("Unsupported encryption method: '".$encryption->toString()."'.", E_USER_NOTICE);
                    $result = STRING_UNSUPPORTED_ENCRYPTION;
                    return $result;
                }
            break;
        }
        assert('is_scalar($this->value);');
        /* settype to STRING */
        $this->value = (string) $this->value;
        return $this->value;
    }

    /**
     * encoding, or converting a string (revertable)
     *
     * Note: charset applies only to encoding = "entities"
     *
     * Returns an encoded version depending on
     * the type of encoding you choose.
     *
     * The input value is not case-sensitive.
     *
     * Note: The results of this function can be reversed using
     * the "decode()" function with the same values.
     * If you are looking for checksums and hashing-methods
     * see the "encrypt" function.
     *
     * The following values are available.
     * <ul>
     *     <li>    unicode:  uses utf8_encode(), aliases: "utf", "utf8"         </li>
     *     <li>    base64:   uses base64_encode()                               </li>
     *     <li>    url:      uses urlencode()                                   </li>
     *     <li>    rawurl:   uses rawurlencode()                                </li>
     *     <li>    entities: uses htmlentities(), uses $style argument          </li>
     *     <li>    rot13:    does a ROT13 transformation                        </li>
     *     <li>    quote:    quotes meta signs using quotemeta()                </li>
     *     <li>    regexp:   uses preg_quote(), alias: "regular expression"     </li>
     * </ul>
     *
     * {@internal
     *
     * Note:
     * Add other types of encodings as you see fit.
     * Remember to add a corresponding decode method
     * in function "decode()" as well.
     * Remember to update not only the source, but
     * also the documentation on this function.
     *
     * }}
     *
     * @access  public
     * @param   string  $encoding  see the list of valid inputs for details
     * @param   string  $style     used for entity conversion
     * @param   string  $charset  used for entity conversion
     * @return  string
     *
     * @name    String::encode()
     * @see     StringObject::encrypt()
     * @see     StringObject::decode()
     */
    function encode($encoding, $style = ENT_COMPAT, $charset = "")
    {
        /* settype to STRING */
        $this->value = (string) $this->value;

        $encoding = new StringObject($encoding);
        $encoding->toLowerCase();

        /* settype to INTEGER */
        $style    = (int) $style;
        $charset  = new StringObject($charset);

        switch ($encoding->toString()) {
            case 'unicode': case 'utf': case 'utf8':
                $this->value = utf8_encode($this->value);
            break;
            case 'base64':
                $this->value = base64_encode($this->value);
            break;
            case 'url':
                $this->value = urlencode($this->value);
            break;
            case 'rawurl':
                $this->value = rawurlencode($this->value);
            break;
            case 'rot13':
                $this->value = str_rot13($this->value);
            break;
            case 'entities':
             if (($style == ENT_COMPAT || $style == ENT_QUOTES || $style == ENT_NOQUOTES)) {
                 if ($charset->get() != '') {
                     $this->value = htmlentities($this->value, $style, $charset->get());
                 } else {
                     $this->value = htmlentities($this->value, $style);
                 }
             } elseif ($style == ENT_FULL) {
                 $temp = "";
                 for ($i = 0; $i < strlen($this->value); $i++)
                 {
                     $temp .= "&#".ord($this->value[$i]).";";
                 }
                 $this->value = $temp;
                 unset($temp);
             } else {
                 $this->value = htmlentities($this->value);
             }
            break;
            case 'quote':
                $this->value = quotemeta($this->value);
            break;
            case 'regexp': case 'regular expression':
                $this->value = preg_quote($this->value);
            break;
            default:
                trigger_error("The value of the \$encoding parameter (argument 1) is invalid: '".$encoding->toString()."'.",E_USER_WARNING);
            break;
        }
        assert('is_scalar($this->value);');
        /* settype to STRING */
        $this->value = (string) $this->value;
        return $this->value;
    }

    /**
     * decode a string (revertable)
     *
     * Note: charset applies only to encoding = "entities"
     *
     * This function is the opposite of "encode()".
     * See "encode()" for details on the available types of
     * encoding.
     *
     * @access  public
     * @param   string  $encoding
     * @param   int     $style   (optional)
     * @param   string  $charset (optional)
     * @return  string
     *
     * @name    String::decode()
     * @see     StringObject::encrypt()
     * @see     StringObject::encode()
     */
    function &decode($encoding, $style = ENT_COMPAT, $charset = "")
    {
        /* settype to STRING */
        $this->value = (string) $this->value;

        $encoding = new StringObject($encoding);
        $encoding->toLowerCase();

        /* settype to INTEGER */
        $style    = (int) $style;
        $charset  = new StringObject($charset);

        switch ($encoding->toString()) {
            case 'unicode': case 'utf': case 'utf8':
                $this->value = utf8_decode($this->value);
            break;
            case 'base64':
                $this->value = base64_decode($this->value);
            break;
            case 'url':
                $this->value = urldecode($this->value);
            break;
            case 'rawurl':
                $this->value = rawurldecode($this->value);
            break;
            case 'rot13':
                $this->value = str_rot13($this->value);
            break;
            case 'entities':
             if (($style == ENT_COMPAT || $style == ENT_QUOTES || $style == ENT_NOQUOTES)) {
                 if ($charset->toString() != "") {
                     $this->value = html_entity_decode($this->value, $style, $charset->toString());
                 } else {
                     $this->value = html_entity_decode($this->value, $style);
                 }
             } else {
                 $this->value = html_entity_decode($this->value);
             }
            break;
            case 'rot13':
                $this->value = str_rot13($this->value);
            break;
            default:
                trigger_error("The value of the \$encoding parameter (argument 1) is invalid: '".$encoding->toString()."'.",E_USER_WARNING);
            break;
        }
        assert('is_scalar($this->value);');
        /* settype to STRING */
        $this->value = (string) $this->value;
        return $this->value;
    }

    /**
     * return a lower-cased version of the string
     *
     * @access  public
     * @return  string
     *
     * @name    String::toLowerCase()
     * @see     StringObject::toUpperCase()
     */
    function &toLowerCase()
    {
        $this->value = strtolower($this->value);
        return $this->value;
    }

    /**
     * return a upper-cased version of the string
     *
     * @access  public
     * @return  string
     *
     * @name    String::toUpperCase()
     * @see     StringObject::toLowerCase()
     */
    function &toUpperCase()
    {
        $this->value = strtoupper($this->value);
        return $this->value;
    }

    /**
     * extract a substring
     *
     * Returns a substring beginning at character-offset $start with
     * $length characters.
     * See PHP-Manual "string functions" "substr()" for details.
     *
     * @access  public
     * @param   int   $start
     * @param   int   $length (optional)
     * @return  string
     *
     * @name    String::substring()
     */
    function &substring($start, $length = 0)
    {
        /* settype to INTEGER */
        $start  = (int) $start;
        $length = (int) $length;
        assert('abs($length) <= $this->length();');

        if ($length != 0) {
            return substr($this->value, $start, $length);
        } else {
            return substr($this->value, $start);
        }
        return $this->value;
    }

    /**
     * test two strings for equality
     *
     * Returns
     * <ul>
     *     <li>    bool(true)  if and only if the input is of type string and
     *             string representation of the value of this object and the input string are the same,
     *             or the input is an object of the same type and the string values of both are equal
     *     </li>
     *     <li>    bool(false) otherwise
     *     </li>
     * </ul>
     *
     * Note: this is unlike the PHP code ($string == $something) where you
     * might accidently run into comparision in e.g. a boolean context.
     *
     * @access  public
     * @param   mixed  $something  some object that should be compared
     * @return  bool
     *
     * @name    String::equals()
     */
    function equals($something)
    {
        if (is_object($something) && method_exists($something, 'getClass') && $something->getClass() === $this->getClass()) {
            return $this->value === $something->value;
        } elseif (is_string($something)) {
            return $this->value === $something;
        } else {
            return false;
        }
    }

    /**
     * compare two strings
     *
     * Returns
     * <ul>
     *     <li>    int(-1)  if this string < $anotherString      </li>
     *     <li>    int(+0)  if this string === $anotherString    </li>
     *     <li>    int(+1)  if this string > $anotherString      </li>
     * </ul>
     *
     * Note: This function is case-sensitive.
     *
     * @access  public
     * @param   string  $anotherString  some other string
     * @return  int(+1)|int(0)|int(-1)
     *
     * @name    String::compareTo()
     * @see     StringObject::compareToIgnoreCase()
     */
    function compareTo($anotherString)
    {
        $anotherString = StringObject::getString($anotherString);
        return strcmp($this->value, $something);
    }

    /**
     * compare two strings (ignore case)
     *
     * Returns
     * <ul>
     *     <li>    int(-1)  if this string < $anotherString      </li>
     *     <li>    int(+0)  if this string === $anotherString    </li>
     *     <li>    int(+1)  if this string > $anotherString      </li>
     * </ul>
     *
     * Note: This function is NOT case-sensitive.
     *
     * @access  public
     * @param   string $anotherString  some other string
     * @return  int(+1)|int(0)|int(-1)
     *
     * @name    String::compareToIgnoreCase()
     * @see     StringObject::compareTo()
     */
    function compareToIgnoreCase($anotherString)
    {
        $anotherString = StringObject::getString($anotherString);
        return strcasecmp($this->value, $something);
    }

    /**
     * match string against regular expression
     *
     * Matches this string against a given Perl-compatible regular expression.
     * Returns an array containing the FIRST set of matches or bool(false) if
     * the regular expression did not match at all.
     *
     * @access  public
     * @param   string  $regularExpression
     * @return  array|bool(false)
     *
     * @name    String::match()
     * @see     StringObject::matchAll()
     */
    function match($regularExpression)
    {
        $regularExpression = StringObject::getString($regularExpression);
        $numberOfHits = preg_match($regularExpression, $this->value, $matches);
        if (is_int($numberOfHits) && $numberOfHits > 0) {
            assert('is_array($matches);');
            return $matches;
        } else {
            return false;
        }
    }

    /**
     * match string against regular expression (return all results)
     *
     * Matches this string against a given Perl-compatible regular expression.
     * Returns an array containing ALL the matches or bool(false) if
     * the regular expression did not match at all.
     *
     * @access  public
     * @param   string  $regularExpression
     * @return  array|bool(false)
     *
     * @name    String::matchAll()
     * @see     StringObject::match()
     */
    function matchAll($regularExpression)
    {
        $regularExpression = StringObject::getString($regularExpression);
        $numberOfHits = preg_match_all($regularExpression, $this->value, $matches);
        if (is_int($numberOfHits) && $numberOfHits > 0) {
            assert('is_array($matches);');
            return $matches;
        } else {
            return false;
        }
    }

    /**
     * replace a needle with a substitute
     *
     * This will replace all entities of $needle with $substitute.
     * Returns the number of times $needle is replaced.
     *
     * @access  public
     * @param   string  $needle
     * @param   string  $substitute (optional)
     * @return  int
     *
     * @name    String::replace()
     * @see     StringObject::replaceRegExp()
     */
    function replace($needle, $substitute = ""/* no way to make this work for PHP 4 :-( , &$count = null */)
    {
        $count = 0;
        $needle = StringObject::getString($needle);
        $substitute = StringObject::getString($substitute);

        if (/* func_num_args() > 2 */ true) {
            if (version_compare("5.0.0", PHP_VERSION) === 1) {
                /* This simulates the behaviour of the $count parameter
                 * that was introduced in PHP 5.0.0 for PHP 4.
                 */
                $count = substr_count($this->value, $needle);
                assert('is_int($count);');
                $this->value = str_replace($needle, $substitute, $this->value);
            } else {
                /* As of PHP version 5.0.0 and beyond the $count
                 * parameter may be used.
                 */
                $this->value = str_replace($needle, $substitute, $this->value, $count);
                assert('is_int($count);');
            }
        } else {
            $this->value = str_replace($needle, $substitute, $this->value);
        }

        return $count;
    }

    /**
     * replace a substring by using a regular expression
     *
     * This will replace all hits of the Perl-compatible $regularExpression with $substitute.
     *
     * Returns the number of times $needle is replaced.
     *
     * @access  public
     * @param   string  $regularExpression
     * @param   string  $substitute  (optional)
     * @param   int     $limit       (optional) must be a positive integer > 0, defaults to -1 (no limit)
     * @return  int
     *
     * @name    String::replaceRegExp()
     * @see     StringObject::replace()
     */
    function replaceRegExp($regularExpression, $substitute = "", $limit = -1/* no way to make this work for PHP 4 :-( , &$count = null */)
    {
        /* settype to STRING
         *            STRING
         *            INTEGER
         *            INTEGER
         */
        $regularExpression = StringObject::getString($regularExpression);
        $substitute        = StringObject::getString($substitute);
        $limit             = (int) $limit;
        $count             = 0;

        /**
         * Limit must be a positive integer > 0.
         * All other values default to -1 (= no limit).
         */
        if ($limit < 1) {
            $limit = -1;
        }

        if (/* func_num_args() > 3 */ true) {
            /* initialize count variable */
            $count = 0;
            if (version_compare("5.1.0", PHP_VERSION) === 1) {
                /* This simulates the behaviour of the $count parameter
                 * that was introduced in PHP 5.1.0 for PHP prior to this version.
                 */
                assert('!isset($dummy);');
                $count = preg_match_all($regularExpression, $this->value, $dummy);
                /* Since preg_match_all returns bool(false) on error it is wise to ensure
                 * the result has the expected data-type.
                 */
                if ($count === false) {
                    $count = 0;
                }
                if ($limit > 0 && $count > $limit) {
                    $count = $limit;
                }
                unset($dummy);
                assert('is_int($count); /* preg_match_all() returned an unexpected result */');
                $this->value = preg_replace($regularExpression, $substitute, $this->value, $limit);
                assert('is_string($this->value); /* preg_replace() returned an unexpected result */');
            } else {
                /* As of PHP version 5.1.0 and beyond the $count
                 * parameter may be used.
                 */
                $this->value = preg_replace($regularExpression, $substitute, $this->value, $limit, $count);
                assert('is_string($this->value); /* preg_replace() returned an unexpected result */');
                assert('is_int($count); /* preg_replace() did not return an integer for parameter $count */');
            }
        } else {
            $this->value = preg_replace($regularExpression, $substitute, $this->value, $limit);
        }

        return $count;
    }

    /**
     * get the length of the string
     *
     * Returns the number of characters in the string.
     *
     * @access  public
     * @return  int
     *
     * @name    String::length()
     */
    function length()
    {
        return strlen($this->value);
    }

    /**
     * convert string to an array
     *
     * @access  public
     * @param   string  $separator
     * @param   int     $limit
     * @return  array
     *
     * @name    String::split()
     * @see     StringObject::splitRegExp()
     */
    function split($separator, $limit = 0)
    {
        /* settype to STRING
         *            INTEGER
         */
        $separator = StringObject::getString($separator);
        $limit     = (int) $limit;

        if ($limit > 0) {
            return explode($separator, $this->value, $limit);
        } else {
            return explode($separator, $this->value);
        }
    }

    /**
     * convert string to an array by using regular expression to find a speratator
     *
     * @access  public
     * @param   string  $separator
     * @param   int     $limit
     * @return  array
     *
     * @name    String::splitRegExp()
     * @see     StringObject::split()
     */
    function splitRegExp($separator, $limit = 0)
    {
        /* settype to STRING
         *            INTEGER
         */
        $separator = StringObject::getString($separator);
        $limit     = (int) $limit;

        if ($limit > 0) {
            return preg_split($separator, $this->value, $limit);
        } else {
            return preg_split($separator, $this->value);
        }
    }

    /**
     * get position of first occurence of a needle inside the string
     *
     * Returns character-offset of first occurence of $needle within this string.
     * Indices starting with int(0).
     *
     * Returns Java-style int(-1) if $needle is not found, NOT Php-style bool(false).
     * This is because int(0) and bool(false) might get mixed by accident.
     *
     * So while if <code>(strpos($string, $needle) == 0))</code> will return true,
     * even if $needle is not found, the test <code>if ($string->indexOf($needle) == 0)</code>
     * will return false if $needle is not found and true if and only if $string
     * starts with the string $needle.
     *
     * @access  public
     * @param   string  $needle
     * @param   int     $offset
     * @return  int
     *
     * @name    String::indexOf()
     */
    function indexOf($needle, $offset = 0)
    {
        /* settype to STRING
         *            INTEGER
         */
        $needle = StringObject::getString($needle);
        $offset = (int) $offset;

        if ($offset > 0) {
            $result = strpos($this->value, $needle, $offset);
        } else {
            $result = strpos($this->value, $needle);
        }
        if ($result === false) {
            return -1;
        } else {
            return $result;
        }
    }

    /**
     * wrap a long text
     *
     * @access  public
     * @param   int     $width
     * @param   string  $break
     * @param   bool    $cut
     * @return  string
     *
     * @name    String::wrap()
     */
    function &wrap($width = 0, $break = "", $cut = false)
    {
        /* settype to INTEGER
         *            STRING
         *            BOOLEAN
         */
        $width = (int) $width;
        $break = StringObject::getString($break);
        $cut   = (bool) $cut;

        if ($cut === true) {
            $this->value = wordwrap($this->value,$width,$break,$cut);
        } elseif (!empty($break)) {
            $this->value = wordwrap($this->value,$width,$break);
        } elseif ($width > 0) {
            $this->value = wordwrap($this->value,$width);
        } else {
            $this->value = wordwrap($this->value);
        }
        return $this->value;
    }

    /**
     * shuffle the string's characters
     *
     * @access  public
     * @return  string
     *
     * @name    String::shuffle()
     */
    function &shuffle()
    {
        $this->value = str_shuffle($this->value);
        return $this->value;
    }

    /**
     * reverse the string value
     *
     * @access  public
     * @return  string
     *
     * @name    String::reverse()
     */
    function &reverse()
    {
        $this->value = strrev($this->value);
        return $this->value;
    }

    /**
     * clone the string
     *
     * creates a copy of this object
     *
     * @access  public
     * @return  String
     *
     * @name    String::cloneObject()
     */
    function cloneObject()
    {
        return new StringObject($this->value);
    }

    /**
     * alias of cloneObject()
     *
     * @access  public
     * @return  String
     *
     * @name    String::copy()
     */
    function copy()
    {
        return $this->cloneObject();
    }

    /**
     * convert to string
     *
     * This will convert any var to a string and return
     * the result.
     *
     * Unlike PHP's standard conversion, if the var is
     * an object, this will use the magic function
     * '__toString()' or an existing function 'toString()'
     * to get the string representation.
     * If the var is an array, it will implode it's contents
     * to a string instead of just returning the text 'Array',
     * what is the default behaviour of PHPs.
     * Note that this is not done recursively for multidimensional
     * arrays.
     *
     * @access  public
     * @param   mixed  $something
     * @return  string
     *
     * @ignore
     */
    static function getString($something)
    {

        if (is_object($something)) {
            if (method_exists($something, 'toString')) {
                $something = $something->toString();
            } elseif (method_exists($something, '__toString')) {
                $something = $something->__toString();
            } else {
                $something = "";
            }
        } elseif (is_array($something)) {
            $something = (string) implode('', $something);
        }

        return (string) $something;
    }

    /**
     * convert to html entities
     *
     * @access  public
     * @param   string  $input  some string
     * @return  string
     *
     * @name    String::htmlEntities()
     * @see     StringObject::encode()
     */
    static function htmlEntities($input)
    {
        assert('is_string($input);');
        $out = new StringObject($input);
        $out->encode('entities', ENT_FULL);
        return $out->toString();
    }

    /**
     * convert html special characters
     *
     * This function is much like the original htmlspecialchars() function,
     * but it enables you to use the argument $doubleEncode, which was introduced
     * in PHP 5.2.3 even when running on PHP 4.
     *
     * If $doubleEncode is set to false, the function will not encode existing
     * HTML entities. In PHP 5.2.3 and beyound the original function is used.
     * In any previous version, that does not have this parameter, the behavior is
     * emulated.
     *
     * @access  public
     * @param   string  $string        text to encode
     * @param   int     $quoteStyle    ENT_COMPAT, ENT_QUOTES, ENT_NOQUOTES
     * @param   string  $charset       e.g. UTF-8 or ISO-8859-1
     * @param   bool    $doubleEncode  set to true to avoid double encoded string
     * @return  string
     * @since   2.9.6
     */
    static function htmlSpecialChars($string, $quoteStyle = ENT_COMPAT, $charset = 'UTF-8', $doubleEncode = true)
    {
        if (!is_null($string)) {
            assert('is_scalar($string);');
            /* settype to STRING */
            $string = (string) $string;

            if ($doubleEncode) {
                $string = htmlspecialchars($string , $quoteStyle, $charset);
            } else{
                /**
                 * cache result for perfomance reasons
                 */
                if (!defined('STRING_COMPAT_DOUBLEENCODE')) {
                    /**
                     * @ignore
                     */
                    define('STRING_COMPAT_DOUBLEENCODE', (version_compare(phpversion(), "5.2.3") === -1));
                }
                /**
                 * PHP4 emulation of double-encode parameter
                 */
                if (STRING_COMPAT_DOUBLEENCODE) {
                    /**
                     * shortcut for better performance in PHP4
                     */                    
                    if ($charset === 'ISO-8859-1') {
                        $string = preg_replace('/\&(?![\w\#][\w\d]+;)/', "&amp;", $string);
                        if ($quoteStyle === ENT_QUOTES) {
                            $string = str_replace("'", "&#039;", $string);
                            $string = str_replace('"', "&quot;", $string);
                        } elseif ($quoteStyle !== ENT_NOQUOTES) {
                            $string = str_replace('"', "&quot;", $string);
                        }
                        $string = str_replace('<', "&lt;", $string);
                        $string = str_replace('>', "&gt;", $string);
                    } else {
                        $string = preg_split('/\&(?=[\w\#][\w\d]+;)/', $string);
                        for ($i = 0; $i < count($string); $i++)
                        {
                            $string[$i] = htmlspecialchars($string , $quoteStyle, $charset);
                        }
                        $string = join("&", $string);
                    }
                /**
                 * where supported, use original function
                 */
                } else {
                    $string = htmlspecialchars($string , $quoteStyle, $charset, false);
                }
            }

            return $string;
        } else {
            return "";
        }

    }

    /**
     * _check_bounds
     *
     * string array bounds check
     *
     * @access  private
     * @param   int  $index
     * @return  bool
     *
     * @ignore
     */
    function _check_bounds($index)
    {
        /* If the input is no integer at all, issue an E_USER_ERROR and abort. */
        if (!is_int($index)) {
            return false;
        } elseif ($index < 0 || $index >= $this->length()) {
            trigger_error("String index '".$index."' out of bounds.", E_USER_NOTICE);
            return false;
        } else {
            return true;
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
