<?php
/**
 * YANA library
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * virtual Drive
 *
 * Class to abstract from real filesystems by mapping
 * filenames to aliases (mountpoints).
 * Mountpoints may be mounted and unmounted at runtime.
 * When reading a mountpoint, a corresponding file wrapper
 * will be returned to work on the file.
 *
 * Using virtual drives will clean your source from
 * hard-coded file- and directory names, as well as
 * doing all the job of creating and initializing
 * the file wrappers for you.
 *
 * @access     public
 * @package    yana
 * @subpackage vdrive
 * @name       VDrive
 */
class VDrive extends FileReadonly
{

    /**#@+
     * @ignore
     * @access  private
     */
    /** @var array  */ var $drive   = array();
    /** @var string */ var $baseDir = "";
    /** @var int    */ var $mode    = 0;
    /** @var array  */ var $files   = array();
    /**#@-*/

    /**
     * constructor
     *
     * creates a new virtual drive instance
     *
     * @name   VDrive::__construct()
     * @param  string  $path
     * @param  array   $options
     * @param  string  $baseDir
     * @param  int     $mode
     */
    function __construct($path, $options = array(), $baseDir = "", $mode = 0)
    {
        /* settype to STRING
         *            INTEGER
         *            INTEGER
         */
        $path    = (string) $path;
        $baseDir = (string) $baseDir;
        $mode    = (int)    $mode;

        if (is_array($options)) {
            $this->attributes = $options;
        }

        $this->path = $path;
        $this->resetStats();
        $this->baseDir = $baseDir;
        $this->mode = $mode;
    }

    /**
     * add another drive to repository
     *
     * Loads the drive identified by $filename within the directory $baseDir and adds
     * a new mountpoint containing the drive's data to the repository.
     *
     * @access  public
     * @name    VDrive::addDrive()
     * @param   string  $filename
     * @param   string  $baseDir
     * @return  bool
     */
    function addDrive($filename, $baseDir)
    {
        assert('is_string($filename); // Wrong argument type for argument 1. String expected.');
        assert('is_string($baseDir);  // Wrong argument type for argument 2. String expected.');
        /* settype to STRING */
        $filename = (string) $filename;
        $baseDir  = (string) $baseDir;

        if (!is_readable($filename)) {
            return false;
        } else {
            $vDriveData = SML::getFile($filename, CASE_MIXED);
            /* import settings */
            if (isset($vDriveData['SETTINGS']) && is_array($vDriveData['SETTINGS'])) {
                $this->options = $vDriveData['SETTINGS'];
            }
            $this->addMountpoint($vDriveData, $baseDir);
            return true;
        }
    }

    /**
     * mount an unmounted virtual drive
     *
     * Mount the mountpoint identified by $key and copies the contents
     * (if any) to the repository.
     *
     * This function returns bool(true) on success, or bool(false) on error.
     *
     * @access  public
     * @name    VDrive::mount()
     * @param   string  $key  name of the drive to mount
     * @return  bool
     */
    function mount($key)
    {
     assert('is_string($key); // Wrong argument type for argument 1. String expected.');
     /* settype to STRING */
     $key = (string) $key;

        if (isset($this->drive[$key])) {
            $return = $this->drive[$key]->mount();
            if (is_array($return)) {
                $this->options = Hashtable::merge($this->options, $return);
                return true;
            } else {
                return (bool) $return;
            }
        } else {
            return false;
        }
    }

    /**
     * read the virtual drive
     *
     * This (re-)loads the virtual drive and (re-)initializes it's contents.
     *
     * If the file does not exist or is not readable, the functions returns bool(false),
     * else bool(true) is returned.
     *
     * @access  public
     * @name    VDrive::read()
     * @return  bool
     */
    function read()
    {
        /* inherit functionality from parent */
        $test = parent::read();
        if ($test === false) {
            return false;
        } else {
            unset($test);
        }

        /* additional actions */
        $this->content = SML::getFile($this->content, CASE_MIXED);
        if (isset($this->content['SETTINGS']) && is_array($this->content['SETTINGS'])) {
            $this->options = Hashtable::merge($this->options, $this->content['SETTINGS']);
        }
        $this->addMountpoint($this->content, $this->baseDir);
        return true;
    }

    /**
     * get string represenation of a virtual drive
     *
     * This returns a human readable overview of the currently loaded virtual drive and
     * it's contents.
     *
     * You might want to use this for debugging purposes.
     *
     * @uses    print "<pre>".$vDrive->toString()."</pre>";
     *
     * @access  public
     * @name    VDrive::toString()
     * @return  string
     */
    function toString()
    {
        if ($this->drive) {
            assert('is_array($this->drive);');
            $txt = "";
            $drive = null;
            foreach ($this->drive as $key => $element)
            {
                $drive = $element->getMountpoint();
                $txt .= "Mountpoint '${key}':"."\n\t- source = ".$drive->getPath()."\n\t- type = ".$element->getType()."\n\t- is mounted = ".(($element->mountpoint->isEmpty()) ? 'no' : 'yes')."\n\n";
            }
            return $txt;
        } else {
            return "unmounted virtual Drive\n";
        }
    }

    /**
     * get a mountpoint
     *
     * Returns the mountpoint specified by $key or bool(false)
     * if no such mountpoint exists, or the mountpoint was unable to
     * return any contents.
     *
     * @access  public
     * @name    VDrive::get()
     * @param   string  $key  name of the drive to mount
     * @return  bool(false)|Mountpoint
     */
    function get($key = "")
    {
        assert('is_string($key); // Wrong argument type for argument 1. String expected.');
        /* settype to STRING */
        $key = (string) $key;

        if (isset($this->drive[$key])) {
            $mountpoint = @$this->drive[$key]->getMountpoint();
            if (!empty($mountpoint)) {
                return $mountpoint;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * add a mountpoint
     *
     * This function is for internal use inside the class only.
     *
     * @access  private
     * @name    VDrive::addMountpoint()
     * @param   array   $drive    drive information
     * @param   string  $baseDir  base directory
     *
     * @ignore
     */
    function addMountpoint($drive, $baseDir)
    {
        if (!empty($drive)) {
            assert('is_array($drive);    // Wrong argument type for argument 1. Array expected.');
            assert('is_string($baseDir); // Wrong argument type for argument 2. String expected.');
            /* settype to ARRAY
             *            STRING
             */
            $drive   = (array)  $drive;
            $baseDir = (string) $baseDir;

            uasort($drive, array("vDrive", "sortByPriority"));

            if (strpos($baseDir, YANA_UNI_DELIMITER) !== false) {
                $baseDir = SmartUtility::fillTemplate($baseDir, $this->options);
            }

            /**
             * import included scripts
             */
            if (isset($drive['INCLUDE']) && is_array($drive['INCLUDE'])) {
                assert('!isset($element); // Cannot redeclare var $element');
                foreach ($drive['INCLUDE'] as $element)
                {
                    if (!preg_match('/^[\w\.]+\.php$/s', $element)) {
                        trigger_error("Invalid filename to include: '$baseDir$element'.", E_USER_WARNING);
                    } elseif (!is_file("$baseDir$element")) {
                        trigger_error("No such file to include: '$baseDir$element'.", E_USER_WARNING);
                    } else {
                        include_once("$baseDir$element");
                    }
                } /* end foreach */
                unset($element); /* clean up garbage */
            }

            assert('!isset($key); /* cannot redeclare variable $key */');
            assert('!isset($element); /* cannot redeclare variable $element */');
            foreach ($drive as $key => $element)
            {
                /* settype to ARRAY */
                if (isset($element['SRC'])) {
                    $element['SRC'] = (array) $element['SRC'];
                } else {
                    $element['SRC'] = array();
                }
                if ($this->mode == 1 && isset($element['DEFAULT'])) {
                    $element['SRC'] = array($element['DEFAULT']);
                }
                if (isset($element['TYPE']) && $element['TYPE'] == "dir") {
                    if (isset($element['CONTENT'])) {
                        $this->addMountpoint($element['CONTENT'], $baseDir . $element['SRC'][0]);
                    } else {
                        $this->addMountpoint(NULL, $baseDir . $element['SRC'][0]);
                    }
                }

                assert('!isset($i); /* cannot redeclare variable $i */');
                for ($i = 0; $i < count($element['SRC']); $i++)
                {
                    if (strpos($element['SRC'][$i], YANA_UNI_DELIMITER) !== false) {
                        $element['SRC'][$i] = SmartUtility::fillTemplate($element['SRC'][$i], $this->options);
                    } else {
                        /* intentionally left blank */
                    }
                    if (YANA_CDROM) {
                        if (!preg_match('/^\w:/', $element['SRC'][$i])) {
                            $element['SRC'][$i] = $baseDir . $element['SRC'][$i];
                        }
                    } else {
                        $element['SRC'][$i] = $baseDir . $element['SRC'][$i];
                    }
                } /* end for */
                unset($i);
                $element['AUTOMOUNT'] = !empty($element['AUTOMOUNT']);
                if (isset($element['TYPE'])) {
                    if ($element['TYPE'] == "dir") {
                        if (isset($element['FILTER'])) {
                            $this->drive[$key] = new vDrive_dir($element['SRC'], $element['TYPE'], $element['AUTOMOUNT'], $element['FILTER']);
                        } else {
                            $this->drive[$key] = new vDrive_dir($element['SRC'], $element['TYPE'], $element['AUTOMOUNT']);
                        }
                    } else {
                        $this->drive[$key] = new vDrive_file($element['SRC'], $element['TYPE'], $element['AUTOMOUNT']);
                    }
                    $this->files[$key] =& $this->drive[$key]->getMountpoint();
                    if ($this->files[$key]) {
                        if ($this->drive[$key]->getType() === 'SML' || is_subclass_of($this->files[$key], 'SML')) {
                            $ref =& $this->files[$key]->getByReference();
                            assert('is_null($ref) || is_array($ref); /* unexpected result: $ref */');
                            if (is_array($ref)) {
                                $this->options = Hashtable::merge($this->options, $ref);
                            }
                        }
                    } /* end if */
                } else {
                    // trigger_error("Missing property 'type' for mountpount '$key' in file '" . $this->getPath() . "'.", E_USER_WARNING);
                } /* end if */
            } /* end foreach */
        } /* end if */
    }

    /**
     * check a virtual drive for errors and return a report
     *
     * This is to check a virtual drive for syntax errors and missing files.
     *
     * Returns the report as a string. The argument $asHtml allows
     * you to decide, wether the result should be HTML or plain text.
     *
     * HTML will use the following CSS classes:
     * <ul>
     * <li> div.error - encloses an error message </li>
     * <li> div.error span.label - some heading (e.g. "Error") </li>
     * <li> div.error span.description - some text that describes the error </li>
     * </ul>
     *
     * The text will be presented in English, using charset Latin-1.
     *
     * If the resulting string is empty, no errors occured.
     *
     * Example of usage:
     * <code>
     * $errors = $instance->getReport();
     *
     * if (empty($errors)) {
     * print 'all fine';
     * } else {
     * print 'The following errors were reported:'.$errors;
     * }
     * </code>
     *
     * @access  public
     * @param   bool  $asHtml
     * @return  string
     *
     * @name    VDrive::getReport()
     * @ignore
     */
    function getReport($asHtml = true)
    {
        if ($this->isEmpty()) {
             return "Cannot perform check! Drive is not mounted.\n";
        } else {
            $drive[0] = $this->content;
            $i = 0;
            $j = 1;
            $txt = "";

            while ($i < $j)
            {
                assert('!isset($key); /* cannot redeclare variable $key */');
                assert('!isset($element); /* cannot redeclare variable $element */');
                foreach ($drive[$i] as $key => $element)
                {
                    if (!empty($element['TYPE']) && $element['TYPE']=="dir" && isset($element['CONTENT'])) {
                        $drive[count($drive)] = $element['CONTENT'];
                    }

                    if (!empty($this->drive[$key])) {
                        $mountpoint = $this->drive[$key]->getMountpoint();
                    }
                    $filename = (isset($mountpoint) && method_exists($mountpoint, 'getFilename')) ? $mountpoint->getPath() : "";

                    if (!empty($element['SRC'])) {
                        switch ($element['REQUIRED']) {
                            case "exists":
                                if (!file_exists($filename)) {
                                    $txt .= '<div class="error"><span class="label">Error:</span>'.
                                    '<span class="description">File \'' . $filename . '\' is a required ressource, but it does not exist.</span>' .
                                    "</div>\n";
                                }
                                break;
                            case "readable":
                                if (!is_readable($filename)) {
                                    $txt .= '<div class="error"><span class="label">Error:</span>'.
                                    '<span class="description">File \'' . $filename . '\' is not readable. ' .
                                    'Please make sure the file exists AND this program has permission to open it.</span>' .
                                    "</div>\n";
                                } else {
                                    $txt .= vDrive::checkFilePermission($filename, 444);
                                }
                            break;
                            case "writeable":
                                if (!is_writeable($filename)) {
                                    $txt .= '<div class="error"><span class="label">Error:</span>'.
                                    '<span class="description">File \'' . $filename . '\' is not writeable. Please make sure the file exists ' .
                                    'AND this program has permission to write at the file and the directory containing it.</span>' .
                                    "</div>\n";
                                } else {
                                    $txt .= vDrive::checkFilePermission($filename, 666);
                                }
                                break;
                            case "create":
                                if (file_exists($filename) && !is_writeable($filename)) {
                                    $txt .= '<div class="error"><span class="label">Error:</span>'.
                                    '<span class="description">File \'' . $filename . '\' exists but is not writeable. Please make sure the file ' .
                                    'exists AND this program has permission to write at the file and the directory containing it.</span>' .
                                    "</div>\n";
                                } elseif (file_exists($filename)) {
                                    $txt .= vDrive::checkFilePermission($filename, 666);
                                }
                                break;
                            default:
                            break;
                        }
                    }
                } /* end foreach */
                unset($key, $element);
                clearstatcache();
                $i++;
                $j = count($drive);
            } /* end while */

            if ($asHtml) {
                return $txt;
            } else {
                return strip_tags($txt);
            }
        } /* end if */
    }

    /**
     * check a virtual drive for errors
     *
     * This is only to be used by vdrive::getReport() for validation.
     *
     * @access  private
     * @static
     * @name    VDrive::checkFilePermission()
     * @return  string
     *
     * @ignore
     */
    function checkFilePermission($filename, $permission2)
    {
        $txt = "";
        $permission1 = decoct(fileperms($filename));

        /* settype to STRING */
        $permission1 = (string) $permission1;
        $permission2 = (string) $permission2;
        $permission1 = preg_replace("/.*(\d\d\d)$/s", "$1", $permission1);
        $permission2 = preg_replace("/.*(\d\d\d)$/s", "$1", $permission2);

        if (strlen($permission1) > 2) {
            $permission[0] = (int) preg_replace("/.*(\d)\d\d$/s", "$1", $permission1);
            $permission[1] = (int) preg_replace("/.*\d(\d)\d$/s", "$1", $permission1);
            $permission[2] = (int) preg_replace("/.*\d\d(\d)$/s", "$1", $permission1);
        }
        if (strlen($permission2) > 2) {
            $permission[3] = (int) preg_replace("/.*(\d)\d\d$/s", "$1", $permission2);
            $permission[4] = (int) preg_replace("/.*\d(\d)\d$/s", "$1", $permission2);
            $permission[5] = (int) preg_replace("/.*\d\d(\d)$/s", "$1", $permission2);
        }
        if (function_exists("posix_getuid") && fileowner($filename) && fileowner($filename) == posix_getuid()
            && $permission[0] >= $permission[3])
        {
            /* intenionally left blank */
        } elseif ($permission[0] < $permission[3] || $permission[1] < $permission[4] || $permission[2] < $permission[5]) {

            $txt .= '<div class="error"><span class="label">Error:</span>'.
            '<span class="description">This script requires a higher permission level for File \'' . $filename . '\'.' .
            'Permissions are currently set to \'' . $permission1 . '\', required level is \'' . $permission2 . '\'. ' .
            'Use CHMOD command to change permission level. See script documentation for details.</span>' .
            "</div>\n";
        }

        clearstatcache();

        return $txt;
    }

    /**
     * sorting function
     *
     * Internal function for sorting two arrays with sort().
     *
     * @access  private
     * @static
     * @name    VDrive::sortByPriority()
     * @param   array $a
     * @param   array $b
     * @return  int
     *
     * @ignore
     */
    function sortByPriority($a, $b)
    {
        assert('is_array($a); // Wrong argument type for argument 1. Array expected.');
        assert('is_array($b); // Wrong argument type for argument 2. Array expected.');
        if (@$a['PRIORITY'] == @$b['PRIORITY']) {
            return 0;
        } elseif ((int) @$a['PRIORITY'] < (int) @$b['PRIORITY']) {
            return  1;
        } else {
            return -1;
        }
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
