<?php
/**
 * YANA library
 *
 * Primary controller class
 *
 *
 * Software:        Yana Framework 3.0.2
 * Author:          Dipl. Inf. Thomas Meyer
 * Homepage:        http://www.yanaframework.net
 * Copyright:       (c) 2015  Thomas Meyer
 * License:         GNU GPL  http://www.gnu.org/licenses/
 *
 * This program: can be redistributed and/or modified under the
 * terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see http://www.gnu.org/licenses/.
 *
 * You may use this software for free. In case you like it, or want to
 * use it for a commercial purpose, please consider making a donation.
 * http://yanaframework.net/donate.php
 *
 * This notice MAY NOT be removed.
 *
 *
 * @package  yana
 * @license  http://www.gnu.org/licenses/gpl.txt
 */

/**
 * «Singleton» Yana
 *
 * This is a primary controller and application loader for the Yana PHP-Framework.
 *
 * Example:
 * <code>
 * // create instance
 * $YANA = new Yana("config/system.config", $_REQUEST);
 * // handle request
 * $YANA->handle($_REQUEST['action']);
 * // output results
 * $YANA->writeView();
 * </code>
 *
 * @access      public
 * @package     yana
 * @subpackage  core
 */
class Yana extends Singleton
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var string */  var $filename = "";
    /** @var array  */  var $config   = array();
    /** @var int    */  var $mode     = 0;
    /** @var array  */  var $log      = array();
    /** @var array  */  var $post     = null;
    /** @var array  */  var $get      = null;
    /** @var array  */  var $request  = null;
    /** @var string */  var $id       = "";

    /**#@-*/
    /**#@+
     * This variable is «readonly».
     * @access  public
     */

    /**
     * to communicate with plugins
     * @var PluginManager
     */ var $plugins  = null;
    /**
     * to load language strings
     * @var Language
     */ var $language = null;
    /**
     * to load skins and templates
     * @var Skin
     */ var $skin     = null;
    /**
     * to read and write data to the global registry
     * @var Registry
     */ var $registry = null;
    /**
     * to read and write user data and permissions
     * @var SessionManager
     * @since 2.9
     */ var $session  = null;
    /**
     * the currently selected template
     * @var SmartView
     * @since 2.9.3
     */ var $view     = null;

    /**#@-*/

    /**
     * «Singleton» Constructor
     *
     * This function creates a new instance of the framework.
     * Note that you may only operate one instance at a time.
     *
     * Examples:
     *
     * If you want to use a web interface:
     * <code>
     * global $YANA;
     * $YANA = new Yana("config/system.config", $_REQUEST);
     * </code>
     *
     * If you want to use a command line interface:
     * <code>
     * global $YANA;
     * $YANA = new Yana("config/system.config", $_SERVER['argv]);
     * </code>
     *
     * @uses    new Yana("config/system.config", $_REQUEST)
     * @param   string  $filename  path to system.config
     * @param   array   $ARGS      associative array of request vars
     * @access  public
     */
    function __construct($filename, $ARGS)
    {
        /**
         * 0) Check
         *
         * before doing anything else, let's check constraints and user-input first
         */

        {
            /**
             * 0.1) check type of input vars
             */
            if (!is_string($filename)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($filename)), E_USER_ERROR);
                exit(1);
            } elseif (!is_array($ARGS)) {
                trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Array', gettype($ARGS)), E_USER_ERROR);
                exit(1);
            } else {
                $this->checkSingletonConstraint();
            }
            assert('is_file($filename);     /* Invalid argument 1. Input is not a file. */');
            assert('is_readable($filename); /* Invalid argument 1. Configuration file is not readable. */');

            /**
             * 0.2) lower-case all entries
             */
            $this->_setRequestVar($ARGS);
            unset($ARGS);
        }/* END 0) */

        /**
         * 1) Startup-Settings
         *
         * system config file contains default- and startup-settings
         */

        /**
         * 1.1) Get System Config file :
         */
        {
            $this->config = SML::getFile($filename, CASE_UPPER);
        }/* END 1.1) */

        /**
         * 1.2 CD-ROM settings
         */
        if (YANA_CDROM === true) {
            assert('!isset($tmpDir);    /* Cannot redeclare variable $tmpDir */');
            assert('!isset($configDir); /* Cannot redeclare variable $configDir */');
            assert('!isset($configSrc); /* Cannot redeclare variable $configSrc */');
            if (!file_exists(YANA_CDROM_DIR)) {
                mkdir(YANA_CDROM_DIR);
                chmod(YANA_CDROM_DIR, 0777);
            }
            $tmpDir = YANA_CDROM_DIR . "cache\\";
            if (!file_exists($tmpDir)) {
                mkdir($tmpDir);
                chmod($tmpDir, 0777);
            }
            $configDir = YANA_CDROM_DIR . $this->config['CONFIGDIR'];
            if (!file_exists($configDir)) {
                $configSrc = new Dir( $this->config['CONFIGDIR'] );
                $configSrc->copy($configDir, true, 777, true, null, '/^(?!\.blob$)/i', true);
            }

            $this->config['CONFIGDIR'] = $configDir;
            $this->config['PROFILESDIR'] = YANA_CDROM_DIR . $this->config['PROFILESDIR'];
            $this->config['COUNTERDIR'] = YANA_CDROM_DIR . $this->config['COUNTERDIR'];
            $this->config['DBDIR'] = YANA_CDROM_DIR . $this->config['DBDIR'];
            $this->config['PLUGINFILE'] = YANA_CDROM_DIR . $this->config['PLUGINFILE'];
            $this->config['PLUGINCACHE'] = YANA_CDROM_DIR . $this->config['PLUGINCACHE'];
            $this->config['CONFIGDRIVE'] = YANA_CDROM_DIR . $this->config['CONFIGDRIVE'];
            $this->config['TEMPDIR'] = $tmpDir;
            /**
             * Database blobs should not be moved since they might be too large -
             * however if you need to do it anyway, uncomment this option.
             *
             * $this->config['DBBLOB'] = YANA_CDROM_DIR . $this->config['DBBLOB'];
             */

            unset($tmpDir, $configDir, $configSrc);
        }/* END 1.2) */

        /**
         * 1.3) get action var
         */
        {
            if (!empty($this->request['action'])) {
                if (!preg_match('/^[\w][\w\d\_]*$/si', $this->request['action'])) {
                    if (YANA_ERROR_REPORTING === YANA_ERROR_LOG) {
                        trigger_error("Illegal action attribute '" . print_r($this->request['action'], true) . "'. Operation aborted.", E_USER_WARNING);
                        exit(1);
                    } else {
                        print "Illegal action attribute '" . print_r($this->request['action'], true) . "', operation aborted.\n" .
                        "Note to IE users: if you get this message it might be due to a known HTML compatibility issue in some IE versions.\n" .
                        "This software has a JavaScript workaround installed. You should activate JavaScript in your browser and try again.";
                        exit(1);
                    }
                } else {
                    $this->request['action'] = strtolower($this->request['action']);
                }
            } else {
                assert('!empty($this->config["DEFAULT"]["HOMEPAGE"]); /* Configuration error. No default homepage is set. */');
                $this->request['action'] = strtolower($this->config['DEFAULT']['HOMEPAGE']);
            }
        }/* END 1.3) */

        /* END 1) */

        /**
         * 2) Create Repositories
         *
         * pluginManager holds repositories for interfaces and implementations of plugins
         */
        {
            if (YANA_CACHE_ACTIVE === true) {
                assert('!isset($cache_file); // Cannot redeclare var $cache_file');
                $cache_file = $this->config['PLUGINCACHE'];
                if (file_exists($cache_file)) {
                    $this->plugins = unserialize(file_get_contents($cache_file));
                } else {
                    $this->plugins = new PluginManager($this->config['PLUGINDIR'], $this->config['PLUGINFILE']);
                    file_put_contents($this->config['PLUGINCACHE'], serialize($this->plugins));
                }
                unset($cache_file);
            } else {
                $this->plugins = new PluginManager($this->config['PLUGINDIR'], $this->config['PLUGINFILE']);
            }
        } /* END 2) */

        /**
         * 3) Build Registry
         *
         * registry offers coupling of components through common memory
         */

        /**
         * 3.1) get configuration mode
         *
         * 0 = default-mode (use profile settings)
         * 1 = safe-mode    (use default profile)
         *
         * This setting may be read using Yana::getMode().
         */
        {
            $this->mode = $this->plugins->getVar("", "INTERFACE." . $this->request['action'] . ".MODE");
            if (is_null($this->mode)) {
                $this->mode = $this->config['DEFAULT']['EVENT']['MODE'];
            }
            /* settype to INTEGER */
            $this->mode = (int) $this->mode;
        } /* END 3.1) */

        /**
         * 3.2) build configuration
         *
         * These settings may be read using Yana::getVar().
         */
        {
            if (YANA_CACHE_ACTIVE === true) {
                assert('!isset($cache_file); // Cannot redeclare var $cache_file');
                assert('!isset($profileId); // Cannot redeclare var $profileId');
                if (!empty($this->request['id'])) {
                    $profileId = $this->request['id'];
                } else {
                    $profileId = $this->config['DEFAULT']['PROFILE'];
                }
                $cache_file = $this->config['TEMPDIR'] . 'registry_' . $profileId . '.tmp';
                if (file_exists($cache_file)) {
                    $this->registry = unserialize(file_get_contents($cache_file));
                } else {
                    $REF = $this->config;
                    $REF['id'] = $profileId;
                    $this->registry = new Registry($this->config['CONFIGDRIVE'], $REF, "", $this->mode);
                    $this->registry->read();
                    file_put_contents($cache_file, serialize($this->registry));
                }
                unset($cache_file, $profileId);
                $this->registry->merge('*', $this->request, false);
            } else {
                $REF = array();
                $REF = Hashtable::merge($this->config, $this->request);
                assert('is_array($REF);');
                if (count($REF) == 0) {
                    $REF = array ("__PHP_PWD_DBASE__" => "74e12c1b837c67e7fa4f6a2eacd6076f");
                }
                $this->registry = new Registry($this->config['CONFIGDRIVE'], $REF, "", $this->mode);
                $this->registry->read();
            }

        } /* END 3.2) */

        /**
         * 3.3) Save Referer (where available)
         */
        {
            if (!empty($_SERVER['HTTP_REFERER'])) {
                $this->registry->setVar("REFERER", preg_replace("/(.*\/).*(\?.*)?/", "\\1", $_SERVER['HTTP_REFERER']));
            }
        } /* END 3.3) */

        /**
         * 3.4) Save Website Location
         */
        {
            /* fall back implementation */
            if (isset($_SERVER['HTTP_HOST']) && isset($_SERVER['REQUEST_URI'])) {
                $WEBSITE_URL = 'http://' . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'];
            } elseif (isset($_SERVER['SERVER_ADDR']) && isset($_SERVER['PHP_SELF'])) {
                $WEBSITE_URL = 'http://' . $_SERVER['SERVER_ADDR'] . $_SERVER['PHP_SELF'];
            } elseif (isset($_SERVER['argv']) && is_array($_SERVER['argv'])) {
                assert('!isset($argv) && !isset($website_url);');
                $argv = $_SERVER['argv'];
                $WEBSITE_URL = 'php ' . print_r(array_shift($argv), true);
                while (count($argv) > 0)
                {
                    $WEBSITE_URL .= ' "' . print_r(array_shift($argv), true) . '"';
                }
                unset($argv);
            } elseif (isset($_SERVER['PHP_SELF'])) {
                $WEBSITE_URL = $_SERVER['PHP_SELF'];
            } else {
                trigger_error('Unable to resolve website URL. This feature will not be available on this server.', E_USER_NOTICE);
                $REF["__PHP_URL_DBASE__"] = "9c738d55290fdca9eae92126f39e622b";
                $WEBSITE_URL = null;
            }
            /* init registry setting */
            $this->registry->setVar("SETTINGS.WEBSITE_URL", $WEBSITE_URL);
            /* Note: the var $WEBSITE_URL will be needed in 7. again,
             * to resolve the request-uri. Don't remove it yet.
             */
        } /* END 3.4) */

        /* END 3) */

        /**
         * 4) prepare Template
         *
         * view is an auxiliary class that provides access to output-specific functions
         */
        {
            $this->view = new SmartView();
            $this->setVar("ACTION", $this->request['action']);
        } /* END 4) */

        /**
         * 5) create Language Library
         *
         * Creates the translation-repository.
         * This setting can be accessed via Yana::$language
         */
        {
            $this->language = new Language($this->config['LANGUAGEDIR'] . $this->getVar("LANGUAGE_INFO.DIRECTORY"), $this->registry->get("LANG"), $this->config['LANGUAGEDIR'] . $this->config['DEFAULT']['LANGUAGE']['DIRECTORY']);
            assert('!isset($dirlist) && !isset($dir) && !isset($array);');
            $dirlist = dirlist($this->config['LANGUAGEDIR'],'',YANA_GET_DIRS);
            foreach ($dirlist as $dir)
            {
                $array[$dir] = 1;
            }
            $this->registry->setVar('INSTALLED_LANGUAGES', $array);
            if (isset($_SESSION['language'])) {
                $this->registry->setVar('SELECTED_LANGUAGE', $_SESSION['language']);
            }
            unset($dirlist,$dir,$array);
        } /* END 5) */

        /**
         * 6) create Skin Library
         *
         * Creates the skin-repository.
         * This setting can be accessed via Yana::$skin
         */
        {
            if (YANA_CACHE_ACTIVE === true) {
                assert('!isset($cache_file); // Cannot redeclare var $cache_file');
                $cache_file = $this->registry->mount('SKINCACHE');
                $cache_file = $this->registry->get('SKINCACHE');
                if ($cache_file->exists()) {
                    $this->skin = unserialize(file_get_contents($cache_file->getPath()));
                } else {
                    $defaultSkinDir = new Dir($this->config['SKINDIR'].$this->config['DEFAULT']['SKIN']['DIRECTORY']);
                    $defaultSkinDir->filter = ".config";
                    $defaultSkinDir->read();
                    $this->skin = new Skin($this->config['SKINDIR'].$this->getVar("SKIN_INFO.DIRECTORY"), $this->registry->get("SKIN"), $this->registry->get('SKINDIR'), $defaultSkinDir);
                    $cache_file->create();
                    $cache_file->insert(serialize($this->skin));
                    $cache_file->write();
                }
                unset($cache_file);
            } else {
                $defaultSkinDir = new Dir($this->config['SKINDIR'].$this->config['DEFAULT']['SKIN']['DIRECTORY']);
                $defaultSkinDir->filter = ".config";
                $defaultSkinDir->read();
                $this->skin = new Skin($this->config['SKINDIR'].$this->getVar("SKIN_INFO.DIRECTORY"), $this->registry->get("SKIN"), $this->registry->get('SKINDIR'), $defaultSkinDir);
            }
        } /* END 6) */

        /**
         * 7) create uper-cased aliases of required request vars
         *
         * these are used in templates
         */
        {
            if (!empty($this->request['id'])) {
                $this->id = strtolower($this->request['id']);
                $this->setVarByReference('ID', $this->id);
            }
            if (!empty($this->request['page'])) {
                $this->setVar('PAGE', $this->request['page']);
            }
            if (!empty($this->request['target'])) {
                $this->setVar('TARGET', $this->request['target']);
            }
            if (isset($_SERVER['REMOTE_ADDR'])) {
                $this->setVar('REMOTE_ADDR', $_SERVER['REMOTE_ADDR']);
            } else {
                $this->setVar('REMOTE_ADDR', '0.0.0.0');
            }
            /**
             * 7.a) build request uri
             *
             * remember some properties of $_SERVER['REQUEST_URI']:
             * 1) it is not available from CLI.
             * 2) it is only 'complete' if $_SERVER['REQUEST_METHOD'] is 'get', because ...
             * 3) it does not include POST-vars!
             * 4) even when the request method is 'post', there might also have been GET-vars sent.
             */

            /**
             * 7.1) get-request and build "query_string" and "request_uri" reliable environment vars
             *
             * Arguments are already stored in request uri.
             * Note: this assumes the $_POST array is empty!
             */
            if (isset($_SERVER['QUERY_STRING'])) {
                assert('!isset($query);       /* Cannot redeclare variable $query */');
                $query = $this->request;
                ksort($query);
                assert('!isset($queryString); /* Cannot redeclare variable $queryString */');
                unset($query[YANA_SESSION_NAME]);
                assert('is_array($query); /* Array expected: $query */');
                $this->setVar('QUERY_STRING', http_build_query($query));
            }
            if (isset($_SERVER['REQUEST_URI']) && isset($_SERVER['REQUEST_METHOD']) && strcasecmp($_SERVER['REQUEST_METHOD'], 'get') === 0) {
                $this->setVar('REQUEST_URI', str_replace('&', '&amp;', $_SERVER['REQUEST_URI']));

            /**
             * 7.2) other request: arguments are NOT stored in request uri, but in $_POST array.
             */
            } else {
                /**
                 * 7.2.1) Got an uri for the current request? Then use it.
                 */
                if (isset($_SERVER['REQUEST_URI'])) {
                    $WEBSITE_URL = $_SERVER['REQUEST_URI'];
                }

                /**
                 * 7.2.2) append delimiter
                 */
                if (strpos('?', $WEBSITE_URL) !== false) {
                    $WEBSITE_URL .= '&amp;';
                } else {
                    $WEBSITE_URL .= '?';
                }

                /**
                 * 7.2.3) append posted vars
                 */
                foreach ($_POST as $a => $b)
                {
                    if (is_scalar($b)) {
                        $WEBSITE_URL .= urlencode($a).'='.urlencode($b).'&amp;';
                    }
                }

                /**
                 * 7.2.4) store uri
                 */
                $this->setVar('REQUEST_URI', $WEBSITE_URL);
            }
        } /* END 7) */
        /* Ready to go ... */
    }

    /**
     * handle an event
     *
     * Resolves event and calls plugin(s), with the given arguments.
     *
     * Example:
     * <code>
     * // handle current event
     * $YANA->handle();
     * // same as above
     * $YANA->handle($_REQUEST['action'], $_REQUEST);
     * // handle user defined event 'test'
     * $myArgs = array('foo' => 'bar');
     * $success = $YANA->handle('test', $myArgs);
     * if ($success) {
     *     print "Success!\n";
     * } else {
     *     print "Encountered an error.\n";
     * }
     * </code>
     *
     * @access  public
     * @param   string  $event  script action parameter
     * @param   array   $ARGS   array of passed arguments
     * @return  bool
     */
    function handle($event = "", $ARGS = null)
    {
        /**
         * 1) Input checking
         */
        if (!is_string($event)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($event)), E_USER_WARNING);
            return false;
        }
        if (!is_array($ARGS) && !is_null($ARGS)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Array', gettype($ARGS)), E_USER_WARNING);
            return false;
        }

        /**
         * 2) check for default arguments
         */
        if (empty($event)) {
            $event = $this->request['action'];
        }
        if (is_null($ARGS)) {
            $ARGS  = $this->request;
        }

        /**
         * 3) create session manager instance
         *
         * The SessionManager class is used to manage user information
         * and resolve permissions.
         */
        if (!isset($this->session)) {
            $this->session = new SessionManager($this->connect('user'), $this->getDefault('user'));
        }

        /**
         * 4) handle event
         *
         * Returns bool(true) on success and bool(false) otherwise.
         */
        $test = $this->plugins->handle($event, $ARGS);
        assert('is_bool($test);');
        if ($test === true) {
            /* Create timestamp to provide information for read-stability isolation level */
            $_SESSION['transaction_isolation_created'] = time();
            return true;
        } else {
            return false;
        }
    }

    /**
     * get a value from the request vars
     *
     * Returns the value identified by $key from an untainted, merged copy of the $_POST
     * and $_GET arrays. (Or the input arguments, if the framework is run via a command
     * line interface)
     *
     * In this case "untainted" means, the result is either a string with at most
     * 50000 characters, or an array of such strings, or the constant NULL if the var
     * does not exist.
     *
     * If an input string contains a template token, or a '$' character, they are
     * automatically escaped to HTML entities. This is done to reduce the risk of code
     * injection.
     *
     * Note that this version is case-insensitive. This means "var", "Var" and "VAR"
     * all refer to the same input.
     *
     * If you call this function without the $key parameter or you use the wildcard '*'
     * the whole "request" array is returned.
     *
     * The argument $method was added in version 2.9.5.
     * Use this argument to choose wether you want to get vars sent via:
     * <ul>
     *  <li> YANA_POST = post method </li>
     *  <li> YANA_GET = get method </li>
     *  <li> YANA_REQUEST = both </li>
     * </ul>
     *
     * Important note: as a feature this function always returns unquoted
     * input, regardless if "magic quotes" is turned on or off. However this does leave
     * it to you to ensure for yourself that input is quoted where needed.
     * If you do not want this feature to be used, this behaviour can be turned off by
     * setting the constant YANA_AUTODEQUOTE to false.
     *
     * @uses    $YANA->getRequestVar('CONTAINER.DATA')
     * @access  public
     * @param   string  $key     adress of data in memory (case insensitive)
     * @param   int     $method  YANA_REQUEST, YANA_POST, YANA_GET
     * @return  mixed
     * @name    Yana::getRequestVar()
     */
    function getRequestVar($key = null, $method = YANA_REQUEST)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');

        /**
         * 1) switch by method
         */
        switch ($method)
        {
            case YANA_REQUEST:
                $property =& $this->request;
                if (is_null($property)) {
                    if (!$this->_setRequestVar(Hashtable::merge($_GET, $_POST), $method)) {
                        return null;
                    }
                }
            break;
            case YANA_POST:
                $property =& $this->post;
                if (is_null($property)) {
                    if (!$this->_setRequestVar($_POST, $method)) {
                        return null;
                    }
                }
            break;
            case YANA_GET:
                $property =& $this->get;
                if (is_null($property)) {
                    if (!$this->_setRequestVar($_GET, $method)) {
                        return null;
                    }
                }
            break;
            default:
                trigger_error("Argument 2 is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().\n\t\t" .
                "Exected one of the following constants: YANA_REQUEST, YANA_POST, YANA_GET.", E_USER_WARNING);
                return null;
            break;
        }

        /**
         * 2) return result
         */

        /**
         * 2.1) if key is not set, return whole array (default)
         */
        if (is_null($key)) {
            return $property;

        /**
         * 2.2) if key is set, check which index to return
         */
        } else {

            /* settype to STRING */
            $key = (string) $key;
            $key = strtolower($key);
            if ($key === '*') {
                return $property;
            } elseif (isset($property[$key])) {
                return $property[$key];
            } else {
                return Hashtable::get($property, $key);
            }

        }
    }

    /**
     * set request vars
     *
     * @access  private
     * @param   array   $value   request vars
     * @param   int     $method  YANA_REQUEST, YANA_POST, YANA_GET
     * @return  bool
     * @name    Yana::_setRequestVar()
     * @see     Yana::getRequestVar()
     */
    function _setRequestVar($value, $method = YANA_REQUEST)
    {
        assert('is_array($value); /* Wrong argument type for argument 1. Array expected. */');
        /* settype to ARRAY */
        $value = (array) $value;

        /**
         * 1) untaint value
         */
        $value = Hashtable::changeCase($value, CASE_LOWER);
        Yana::_untaintRequestVar($value, YANA_AUTODEQUOTE && get_magic_quotes_gpc());

        /**
         * 2) switch by method
         */
        switch ($method)
        {
            case YANA_REQUEST:
                $this->request = $value;
                return true;
            break;
            case YANA_POST:
                $this->post = $value;
                return true;
            break;
            case YANA_GET:
                $this->get = $value;
                return true;
            break;
            default:
                trigger_error("Argument 2 is invalid in " . __CLASS__ . "->" . __FUNCTION__ . "().\n\t\t" .
                "Exected one of the following constants: YANA_REQUEST, YANA_POST, YANA_GET.", E_USER_WARNING);
                return false;
            break;
        }
    }

    /**
     * untaint request vars
     *
     * @access  private
     * @static
     * @param   array   &$value
     * @param   bool    $unquote
     * @name    Yana::_untaintRequestVar()
     * @see     Yana::_setRequestVar()
     */
    function _untaintRequestVar(&$value, $unquote = false)
    {
        if (is_array($value)) {
            foreach ($value as $i => $item)
            {
                if (is_array($item)) {
                    Yana::_untaintRequestVar($value[$i], $unquote);
                } else {
                    if ($unquote === true) {
                        $item = stripcslashes($item);
                    }
                    $value[$i] = untaintInput($item, "string", 50000, YANA_ESCAPE_TOKEN, true);
                }
            }
        } else {
            $value = array();
        }
    }

    /**
     * returns the current profile id
     *
     * This is a shortcut for $YANA->getVar('ID').
     *
     * However it is important to note a slight difference.
     * <ul>
     *   <li> $YANA->getVar('ID'):
     *     The value you get via  is
     *     available to all plugins and all plugins may read
     *     AND write this setting as the developer sees fit.
     *     This may mean, that this setting has been subject
     *     to changes by some plugin, e.g. to switch between
     *     profiles.
     *   </li>
     *   <li> $YANA->getId():
     *     Always returns the original value, regardless of
     *     changes by plugins.
     *   </li>
     * </ul>
     *
     * You may want to decide for the behaviour you prefer
     * and choose either one or the other.
     *
     * @access  public
     * @return  string
     */
    function getId()
    {
        return $this->id;
    }

    /**
     * get value from registry
     *
     * Returns var from registry (memory shared by all plugins)
     *
     * Example:
     * <code>
     * $YANA->setVar('foo.bar', 'Hello World');
     * // outputs 'Hello World'
     * print $YANA->getVar('foo.bar');
     * </code>
     *
     * @uses    $YANA->getVar('CONTAINER1.CONTAINER2.DATA')
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  mixed
     * @name    Yana::getVar()
     * @see     Yana::setVarByReference()
     * @see     Yana::setVar()
     */
    function getVar($key)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        if (isset($this->registry)) {
            return $this->registry->getVar($key);
        } else {
            return null;
        }
    }

    /**
     * sets var on registry by Reference
     *
     * The "registry" is memory shared by all plugins.
     *
     * Example:
     * <code>
     * $bar = 'Hello';
     * $YANA->setVarByReference('foo.bar', $bar);
     * $bar .= ' World';
     * // outputs 'Hello World'
     * print $YANA->getVar('foo.bar');
     * </code>
     *
     * @uses    $YANA->setVar('CONTAINER1.CONTAINER2.DATA', $value)
     * @access  public
     * @param   string  $key     adress of data in memory (case insensitive)
     * @param   mixed   &$value  new value (may be scalar value or array)
     * @return  bool
     * @name    Yana::setVarByReference()
     * @see     Yana::setVar()
     * @see     Yana::getVar()
     */
    function setVarByReference($key, &$value)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        if (isset($this->registry)) {
            return $this->registry->setVarByReference($key, $value);
        } else {
            return false;
        }
    }

    /**
     * sets var on registry
     *
     * The "registry" is memory shared by all plugins.
     *
     * Example:
     * <code>
     * $YANA->setVar('foo.bar', 'Hello World');
     * // outputs 'Hello World'
     * print $YANA->getVar('foo.bar');
     * </code>
     *
     * @uses    $YANA->setVar('CONTAINER1.CONTAINER2.DATA', $value)
     * @access  public
     * @param   string  $key    adress of data in memory (case insensitive)
     * @param   mixed   $value  new value (may be scalar value or array)
     * @return  bool
     * @name    Yana::setVar()
     * @see     Yana::setVarByReference()
     * @see     Yana::getVar()
     */
    function setVar($key, $value)
    {
        return $this->setVarByReference($key, $value);
    }

    /**
     * sets the type of a var on registry (memory shared by all plugins)
     *
     * @uses    $YANA->setType('CONTAINER1.CONTAINER2.DATA', 'string')
     * @access  public
     * @param   string  $key   adress of data in memory (case insensitive)
     * @param   string  $type  new type of variable
     * @return  bool
     */
    function setType($key, $type)
    {
        assert('is_scalar($key);  /* Wrong argument type for argument 1. String expected. */');
        assert('is_string($type); /* Wrong argument type for argument 2. String expected. */');
        /* settype to STRING */
        $key  = (string) $key;
        $type = (string) $type;
        return $this->registry->setType($key, $type);
    }

    /**
     * remove var from registry
     *
     * Removes var from registry (memory shared by all plugins).
     * Returns bool(true) on success and bool(false) on error.
     *
     * Example:
     * <code>
     * // set foo.bar
     * $YANA->setVar('foo.bar', 'Hello World');
     * // remove foo.bar
     * $YANA->unsetVar('foo.bar');
     * // foo.bar now is false
     * var_dump($YANA->getVar('foo.bar'));
     * </code>
     *
     * @uses    $YANA->unsetVar('CONTAINER1.CONTAINER2.DATA')
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  bool
     */
    function unsetVar($key)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        return $this->registry->unsetVar($key);
    }

    /**
     * merges value in registry
     *
     * Merges the value at adresse $key with the provided array data.
     * If a key already exists, it is replaced by the new data.
     *
     * Example:
     * <code>
     * $bar = array('foo' => 'bar');
     * $YANA->merge('FOO', $bar);
     * // outputs 'bar'
     * print $YANA->get('FOO.FOO');
     * </code>
     *
     * @uses    $YANA->merge('CONTAINER1.CONTAINER2.DATA', $array)
     * @access  public
     * @param   string  $key    adress of data in memory (case insensitive)
     * @param   array   $array  associative array to merge
     * @return  bool
     */
    function merge($key, $array)
    {
        assert('is_scalar($key);  /* Wrong argument type for argument 1. String expected. */');
        assert('is_array($array); /* Wrong argument type for argument 2. Array expected.  */');
        /* settype to STRING */
        $key = (string) $key;
        return $this->registry->merge($key, $array);
    }

    /**
     * adds an entry to the log-queue
     *
     * The entry will be printed to screen or written to log-file, depending on it's base-class.
     *
     * Examples:
     * <code>
     * global $YANA;
     * // report an error level message
     * $msg = new Error("Oups! Something went wrong and I can't fix it.");
     * $db = $YANA->report($msg);
     * // report a warning level message
     * $msg = new Warning("The string 'my dog ate my homework' is not a valid excuse.");
     * $db = $YANA->report($msg);
     * // report an alert level message
     * $msg = new Alert("Swimming in lava might cause your shorts to catch fire.");
     * $db = $YANA->report($msg);
     * // report a status level message
     * $msg = new Message("Thanks for your request. All is fine!");
     * $db = $YANA->report($msg);
     * // report a log entry
     * $msg = new Log("This is a note that goes to the logs.");
     * $db = $YANA->report($msg);
     * </code>
     *
     * @uses    $YANA->report($report)
     * @access  public
     * @param   Report  $log  object containing the log entry
     * @return  void
     */
    function report($log)
    {
        if (is_object($log) && is_subclass_of($log, 'Report')) {
            array_unshift($this->log, $log);
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, "1 in ".__CLASS__."->".__FUNCTION__."()", "\n\t\tInstance of Report or a derived subclass expected", get_class($log)), E_USER_WARNING);
        }
    }

    /**
     * outputs an text message (while terminating the current script)
     *
     * Examples:
     * <code>
     * global $YANA;
     * // error - page not found
     * $YANA->message('ERROR', '404');
     * // internal error
     * $YANA->message('ERROR', '500');
     * // alert - entry allready exists
     * // (the token 'ID' is used in the error message)
     * $YANA->setVar('ID', $existing_id);
     * $YANA->message('ALERT', 'ALLREADY_EXISTS');
     * // OK - changes have been saved
     * $YANA->message('OK', '200');
     * // output message and route to 'login' page
     * $YANA->message('ERROR', 'access denied', 'login');
     * </code>
     *
     * This function is deprecated as of version 2.9.0 RC2.
     * It is strongly recommended, to use the following code instead:
     * <code>
     * global $YANA;
     * // error - page not found
     * $YANA->report( new Error('404') );
     * $YANA->exitTo('');
     * // internal error
     * $YANA->report( new Error('500') );
     * $YANA->exitTo('');
     * // alert - entry allready exists
     * // (the token 'ID' is used in the error message)
     * $YANA->report( new Alert('ALLREADY_EXISTS'), array('ID' => $existing_id) );
     * $YANA->exitTo('');
     * // OK - changes have been saved
     * $YANA->report( new Message('200') );
     * $YANA->exitTo('');
     * // output message and route to 'login' page
     * $YANA->report( new Error('access denied') );
     * $YANA->exitTo('login');
     * </code>
     *
     * @access      public
     * @param       string  $level    type of message ('OK'|'ALERT'|'ERROR')
     * @param       scalar  $errcode  code of corresponding message in message.config
     * @param       string  $event    upcoming event to route to
     * @deprecated  since 2.9.0 RC2
     * @ignore
     */
    function message($level, $errcode, $event = "")
    {
        /**
         * Input checking
         */
        if (!is_string($level)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($level)), E_USER_WARNING);
            return false;
        }
        if (!is_scalar($errcode)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 2, 'Scalar', gettype($errcode)), E_USER_WARNING);
            return false;
        }
        if (!is_string($event)) {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 3, 'String', gettype($event)), E_USER_WARNING);
            return false;
        }
        
        /* argument 1 */
        $errcode = strtoupper("${errcode}");
        /* argument 2 */
        $level = strtolower($level);
        /* argument 3 */
        if (is_null($event)) {
            $event = 'null';
        } else {
            $event = strtolower($event);
        }

        /**
         * save log-files (if any)
         *
         * By default this will output any messages
         * to a table of the database named 'log'.
         */
        $this->_writeLog();
        $this->_prepareMessages($level, $level !== 'message');

        /**
         * is an AJAX request
         */
        if (!empty($this->request['is_ajax_request'])) {
            /**
             * This option will restrict the output
             * to the template's body-tag (if any).
             *
             * This is done by the output post-processor,
             * thus causing minimum side-effects.
             */
            $this->setVar('FILE_IS_INCLUDE', true);
            if (headers_sent() === false) {
                header('Content-Type: text/html; charset=UTF-8');
            }
            $this->view->showMessage($level, $errcode, 'null', 'STDOUT');

        /**
         * output a message and DO NOT RELOCATE, when
         *   1) headers are already sent, OR
         *   2) the template explicitely requests a message, OR
         *   3) the special 'NULL-event' (no event) is requested.
         */
        } elseif ($event === 'null' || $this->getDefault('MESSAGE') === true || headers_sent() === true) {
            $this->view->showMessage($level, $errcode, $event);

        /**
         * save message and relocate.
         */
        } else {
            $stdout = $this->getVar('STDOUT');
            if (!is_array($stdout)) {
                $stdout = array();
            }
            if (empty($stdout) && isset($this->language)) {
                $this->language->read('message');
                array_push($stdout, $this->getVar("TEXT.$errcode"));
            }
            $_SESSION['STDOUT']             = array();
            $_SESSION['STDOUT']['MESSAGES'] = $stdout;
            $_SESSION['STDOUT']['LEVEL']    = $level;

            header("Location: ".SmartUtility::url("action=$event", true));
            exit(0);
        }
    }

    /**
     * exit the current script
     *
     * This will flush error messages and warnings to the screen,
     * write all reported errors (if any) to the framework's logs
     * and then exit the current script.
     * After that it will call itself again to handle the event
     * provided by the argument $event.
     *
     * You may use the special event 'null' to prevent the
     * framework from handling an event. In this case it will just
     * exit.
     *
     * If the argument $event is not provided, the default event
     * will be used instead.
     *
     * Examples:
     * <code>
     * global $YANA;
     *
     * // print an error and go to start page
     * $YANA->report( new Error('404') );
     * $YANA->exitTo();
     *
     * // same as:
     * $YANA->exitTo('');
     *
     * // Use special event 'null' if you just want to
     * // view the error message and exit the script
     * // without handling another event.
     * // ( You may translate this to: "exit to 'nowhere'" )
     * $YANA->report( new Error('500') );
     * $YANA->exitTo('null');
     *
     * // output message and route to 'login' page
     * $YANA->report( new Error('access denied') );
     * $YANA->exitTo('login');
     * </code>
     *
     * Please note: any code followed after a call to this function
     * will never be executed.
     *
     * @access  public
     * @param   string  $event  upcoming event to route to
     * @param   array   $args   list of arguments to pass to the event
     * @since   2.9.0 RC2
     */
    function exitTo($event = '', $args = array())
    {
        if (is_string($event)) {
            $event = strtolower($event);

        } elseif (is_null($event)) {
            $event = 'null';

        } elseif (is_numeric($event)) {
            $event = (string) $event;

        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'String', gettype($event)), E_USER_ERROR);
            exit(1);

        }

        if (is_array($args)) {
            $args = array_change_key_case($args, CASE_LOWER);
        } else {
            trigger_error(sprintf(YANA_ERROR_WRONG_ARGUMENT, 1, 'Array', gettype($args)), E_USER_ERROR);
            exit(1);
        }

        $args["action"] = $event;

        /**
         * save log-files (if any)
         *
         * By default this will output any messages
         * to a table of the database named 'log'.
         */
        $this->_writeLog();
        $level = '';
        $this->_prepareMessages($level);


        /**
         * is an AJAX request
         */
        if (!empty($this->request['is_ajax_request'])) {
            /**
             * This option will restrict the output
             * to the template's body-tag (if any).
             *
             * This is done by the output post-processor,
             * thus causing minimum side-effects.
             */
            $this->setVar('FILE_IS_INCLUDE', true);
            if (headers_sent() === false) {
                header('Content-Type: text/html; charset=UTF-8');
            }
            $this->view->showMessage($level, '', 'null', 'STDOUT');

        /**
         * output a message and DO NOT RELOCATE, when
         *   1) headers are already sent, OR
         *   2) the template explicitely requests a message, OR
         *   3) the special 'NULL-event' (no event) is requested.
         */
        } elseif ($event === 'null' || $this->getDefault('MESSAGE') === true || headers_sent() === true) {

            $this->view->showMessage($level, '', $event);

        /**
         * save message and relocate.
         */
        } else {

            $stdout = $this->getVar('STDOUT');
            if (!is_array($stdout)) {
                $stdout = array();
            }
            if (empty($stdout) && isset($this->language)) {
                $this->language->read('message');
                $errorCode = $this->getVar('SETTINGS.LAST_RESULT') ? '200' : '500';
                array_push($stdout, $this->getVar("TEXT." . $errorCode));
            }
            $_SESSION['STDOUT']             = array();
            $_SESSION['STDOUT']['MESSAGES'] = $stdout;
            $_SESSION['STDOUT']['LEVEL']    = $level;

            header("Location: " . SmartUtility::url(http_build_query($args), true));

        }
        exit(0);

    }

    /**
     * provides GUI from current data
     *
     * @uses    $YANA->writeView()
     * @access  public
     */
    function writeView()
    {
        /* 0 initialize vars */
        $event = $this->getVar("ACTION");
        $result = $this->getVar('SETTINGS.LAST_RESULT');
        $baseDoc = $this->plugins->getVar("", "INTERFACE.".$event.".TEMPLATE");
        $includeDoc = $this->plugins->getVar("", "INTERFACE.".$event.".INSERT");

        /**
         * 1. AJAX-check
         *
         * If this is an AJAX request we should produce shortened output.
         * "Shortened" here means, we leave off the static document "frame".
         * Instead we will only return the changed content.
         */

        /**
         * 1.1) is an AJAX request
         */
        if (!empty($this->request['is_ajax_request'])) {
            if (!is_null($includeDoc)) {
                $baseDoc = $includeDoc;
                $includeDoc = null;
            }
            /**
             * This option will restrict the output
             * to the template's body-tag (if any).
             *
             * This is done by the output post-processor,
             * thus causing minimum side-effects.
             */
            $this->setVar('FILE_IS_INCLUDE', true);
            if (headers_sent() === false) {
                header('Content-Type: text/html; charset=UTF-8');
            }


        /**
         * 1.2) is NOT an AJAX request
         */
        } else {
            if (is_null($baseDoc)) {
               $baseDoc = (!empty($this->config['DEFAULT']['EVENT']['TEMPLATE'])) ? $this->config['DEFAULT']['EVENT']['TEMPLATE'] : null;
            }
            if (is_null($includeDoc)) {
               $includeDoc = (!empty($this->config['DEFAULT']['EVENT']['INSERT'])) ? $this->config['DEFAULT']['EVENT']['INSERT'] : null;
            }
        }

        /**
         * 2) resolve selected templates
         */
         
        /**
         * 2.1) the reserved template 'NULL' is an alias
         * for 'no template' and will surpress any output.
         *
         * This may mean the plugin will create some output
         * itself using print(), or it is a triggered
         * cron-job that is not meant to produce any output
         * at all.
         */
        if ($baseDoc === 'NULL') {

            $baseDoc = null;

        /**
         * 2.1) the reserved template 'MESSAGE' is a special
         * template that produces a text message.
         *
         * The text usually is an ID of some text.
         * The actual message is stored in the language files
         * and the translated message will be read from there
         * depending on the user's prefered language setting.
         */
        } elseif ($baseDoc === 'MESSAGE' || ($result === false && !empty($this->log))) {

            /* a. get vars */
            $route  = $this->getVar('SETTINGS.EVENT_ROUTING');

            if (!empty($route['TEMPLATE'])) {
                $message_type = $route['TEMPLATE'];
            } else {
                $message_type = '';
            }

            /* b. check vars */
            if (empty($this->log) && empty($route['TEXT'])) {
                if ($result) {
                    $this->report(new Message('200'));
                } else {
                    $this->report(new ErrorMessage('500'));
                }
            }
            if (!is_array($route)) {
                $route = array();
            }

            $this->_prepareMessages($message_type, !$result);

            if (empty($message_type))  {
                $message_type  = ($result) ? 'ok' : 'error';
            }
            if (empty($route['TEXT'])) {
                $route['TEXT'] = ($result) ? '200' : '500';
            }

            /* 1. routing is not set = use default */
            if (empty($route['GOTO'])) {

                $this->message($message_type, $route['TEXT']);

            /* 2. routing is so to null = no route */
            } elseif (!empty($route['GOTO']) && strcasecmp($route['GOTO'], 'null') === 0) {

                $this->message($message_type, $route['TEXT'], null);


            /* 3. routing is set to another event */
            } else {

                $this->message($message_type, $route['TEXT'], $route['GOTO']);

            }


        /**
         * 2.2) all other template settings go here
         */
        } else {

            /* register templates with view sub-system */
            $this->view->setTemplates($baseDoc, $includeDoc);
            /* there is a special var called 'STDOUT' that is used to output messages */
            if (!empty($_SESSION['STDOUT']) && !empty($_SESSION['STDOUT']['MESSAGES']) && is_array($_SESSION['STDOUT']['MESSAGES'])) {
                $this->setVar('STDOUT', $_SESSION['STDOUT']);
                unset($_SESSION['STDOUT']);
            }
            /* build the output page using the current settings */
            $this->view->buildPage();
            /* print the page to the client */
            $this->view->write();
            /**
             * save log-files (if any)
             *
             * By default this will output any messages
             * to a table of the database named 'log'.
             */
            $this->_writeLog();

        }

    }

    /**
     * get mode of current action
     *
     * Actions can run in two different modes.
     * They can either be:
     * <ul>
     *  <li> 0 = using profile settings, or </li>
     *  <li> 1 = using default settings </li>
     * </ul>
     *
     * Actions may use mode = 1 if they manipulate
     * global system settings, or for security reasons,
     * e.g. when you need global admin rights to
     * execute a certain command.
     *
     * Example:
     * <code>
     * $mode = $YANA->getMode();
     * if ($mode === 1) {
     *     print "Global settings are active.\n";
     * } else {
     *     print "Local settings are active.\n";
     * }
     * </code>
     *
     * @access  public
     * @return  int
     */
    function getMode()
    {
        assert('is_int($this->mode); // Invalid value: mode');
        return (int) $this->mode;
    }

    /**
     * get default configuration value
     *
     * Returns the default value for a given var if any,
     * returns NULL (not false!) if there is none.
     *
     * Example 1:
     * <code>
     * $YANA->getDefault('CONTAINER1.CONTAINER2.DATA');
     * </code>
     *
     * Example 2:
     * <code>
     * if (!isset($foo)) {
     *     $foo = $YANA->getDefault('FOO');
     * }
     * </code>
     *
     * Note: system default values are typically defined in the
     * 'default' section of the 'config/system.config' configurations file.
     *
     * @access  public
     * @param   string  $key  adress of data in memory (case insensitive)
     * @return  mixed
     */
    function getDefault($key)
    {
        assert('is_scalar($key); /* Wrong argument type for argument 1. String expected. */');
        /* settype to STRING */
        $key = (string) $key;
        $key = strtoupper("DEFAULT.${key}");
        return Hashtable::get($this->config, $key);
    }

    /**
     * clear system cache
     *
     * Deletes all temporary files in the 'cache/' directory.
     *
     * This includes templates and preinitialized instances of system objects.
     * Use this function where system settings or profile systems are changed,
     * to make sure changes are applied without delay.
     *
     * @access  public
     * @static
     */
    function clearCache()
    {
        SmartTemplate::clearCache();
    }

    /**
     * «factory» connect()
     *
     * Returns a ready-to-use database connection.
     *
     * If you enter more than 1 argument, each additional database structure file
     * will be loaded as well and merged with the others.
     *
     * Example:
     * <code>
     * global $YANA;
     * // Connect to database using 'config/db/user.config'
     * $db = $YANA->connect('user');
     * // Connect to database using multiple files
     * $db = $YANA->connect('user','foo','bar');
     * </code>
     *
     * @access  public
     * @param   string  $source  name of the database *.config file that describes the names and structure of tables within the database (see config/db/*.config)
     * @return  mixed
     */
    function &connect($source/* ... */)
    {
        $numargs = func_num_args();
        for ($i = 0; $i < $numargs; $i++)
        {
            $source = func_get_arg($i);
            assert('is_string($source);');
            /* settype to STRING */
            $source = (string) $source;
            if (!(is_file($source) && preg_match('/^[\w\d\_\/\.]+\.config$/', $source))) {
                if (!preg_match('/^[\w\d\_\-]*$/', $source)) {
                    trigger_error("Illegal database name '${source}'.", E_USER_ERROR);
                    $connection = null;
                    return $connection;
                }
            }
            if ($i === 0) {
                if (YANA_DATABASE_ACTIVE) {
                  $connection = new DbStream($source);
                } else {
                  $connection = new FileDb($source);
                }
            } else {
                $connection->structure->addFile($source);
            }
        } /* end for */
        return $connection;
    }

    /**
     * iterate through message queue
     *
     * @access  private
     * @param   bool   $isError
     * @param   mixed  &$message_type
     * @ignore
     */
    function _prepareMessages(&$message_type, $isError = null)
    {
        $stdout = array();

        /* need those strings - load 'em */
        $this->language->read('message');

        /* 1. event logging */
        assert('!isset($log); /* cannot redeclare variable $log */');
        foreach ($this->log as $log)
        {
            if (is_subclass_of($log, 'Report')) {

                /* checks whether type of report is known and pushes message into output-queue */
                switch (strtolower(get_class($log)))
                {
                    case 'message':
                        if (is_null($isError)) {
                            $isError = false;
                        }
                        $stdout[] = $log->getMessage();
                        if (!$isError) {
                            $message_type = 'message';
                        }
                    break;
                    case 'warning':
                        $stdout[] = $log->getMessage();
                        $message_type = 'warning';
                    break;
                    case 'alert':
                        $stdout[] = $log->getMessage();
                        $message_type = 'alert';
                    break;
                    case 'errormessage':
                        if (is_null($isError)) {
                            $isError = true;
                        }
                        $stdout[] = $log->getMessage();
                        if ($isError) {
                            $message_type = 'error';
                        }
                    break;
                    default:
                        /* ignore unknown logs */
                    break;
                }
            } else {
                assert('false; /* unexpected result: Log entry is not a subclass of Report */');
            }
        }/* end foreach */
        unset($log);
        $this->setVar('STDOUT', $stdout);
    }

    /**
     * flush events to log
     *
     * @access  private
     * @ignore
     */
    function _writeLog()
    {
        /* 0. initialize vars */
        $event_logging = (bool) $this->getVar('PROFILE.LOGGING');

        /* 1. event logging */
        if ($event_logging === true) {

            $log_use_mail = (bool) $this->getVar('PROFILE.LOG.USE_MAIL');
            $log_length = (int) $this->getVar('PROFILE.LOG_LENGTH');
            if ($log_length < 1) {
                $log_length = 50;
            }
            $db = $this->connect('log');
            $table = $db->structure->getTables();
            assert('is_array($table) && count($table) === 1 && isset($table[0]);');
            $table = $table[0];
            $log_changed = false;
            $nr_of_logs = count($this->log);

            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($log); /* cannot redeclare variable $log */');
            foreach ($this->log as $i => $log)
            {
                if (strtolower(get_class($log)) == "log") {
                    $previousLog = $db->get("${table}.?.${table}_id");
                    $previousLog = $db->get("${table}.${previousLog}");
                    /* do not create duplicate entries */
                    if ($log->logEquals($previousLog, "${table}_") === false) {
                        $newLog = $log->getLog("${table}_");
                        if (is_array($newLog) && !empty($newLog)) {
                            if ($db->insert("${table}.*", $newLog) === false) {
                                trigger_error("Unable to insert log. Insert on table '${table}' failed.", E_USER_NOTICE);
                                return null;
                            } else {
                                $log_changed = true;
                            }
                        } else {
                            /* not a valid log entry */
                        }
                    } /* end if */
                    array_splice($this->log, $i, 1);
                } /* end if */
            } /* end foreach */
            unset($i, $log);

            if ($log_changed === true) {
                if ($db->write() === false) {
                    trigger_error("Unable to write log. Commit on table '${table}' failed.", E_USER_WARNING);
                    return null;
                } elseif ($log_use_mail) {

                    if ($db->length($table) > $log_length) {
                        $log_mail = $this->getVar('PROFILE.LOG.MAIL');
                        if (!empty($log_mail)) {
                            for ($i = 0; $i <= ($log_length / 200); $i++)
                            {
                                    $mail = new FormMailer();
                                    $mail->content = $db->get($table, array(), array('LOG_ID'), $i * 200, 200);
                                    $mail->subject = 'JOURNAL';
                                    $mail->send($log_mail);
                                    unset($mail);
                            }
                            unset($i);
                        }
                    }
                    foreach ($db->get("${table}.*.${table}_id") as $id)
                    {
                        if (is_scalar($id) && !empty($id)) {
                            if ($db->remove("${table}.${id}") === false) {
                                trigger_error("Unable to remove entry from table '${table}'.", E_USER_WARNING);
                                return null;
                            }
                        }
                    } /* end for */
                    unset($id);
                    $db->write();

                } else {
                    $list_of_ids = $db->get("${table}.*.${table}_id", "", "", $log_length, $db->length($table) + $nr_of_logs, true);
                    if (is_array($list_of_ids)) {
                        foreach ($list_of_ids as $id)
                        {
                            if (is_scalar($id) && !empty($id)) {
                                if ($db->remove("${table}.${id}") === false) {
                                    trigger_error("Unable to remove entry from table '${table}'.", E_USER_WARNING);
                                    return null;
                                }
                            }
                        } /* end for */
                        $db->write();
                    } else {
                        $db->reset();
                    }
                }
            } else {
                /* intentionally left blank */
            } /* end if */

        } else {
            /* intentionally left blank */
        } /* end if */
    }

}
// Copyright © 2015  Thomas Meyer  http://www.yanaframework.net
?>
